#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include <time.h>

#define name "matrix.dat"

extern int solve(double*,int,double*);
extern void local_solve(int);

void nev(double *m, int size, double *x, double *z)
{
 int i,j;
 double y;
 *z=0;
  for (i=0;i<size;i++)
   {
    y=-m[i*(size+1)+size];
     for (j=0;j<size;j++)
      y+=m[i*(size+1)+j]*x[j];
    (*z)+=fabs(y);  
   }
}   

void write_x(double *x,int size)
{
 int i;
  for (i=0;i<size;i++)
      printf("%lf  ",x[i]);
 printf("\n");      
}


int load_matrix_file(double *m, int size)
{
 FILE *f;
 int i,j;
  if (!(f=fopen(name,"r"))) return -1;
   for (i=0; i<size; i++)
    for (j=0; j<=size; j++)
     if (1!=fscanf(f,"%lf",m+i*(size+1)+j)) return -2 ;
  fclose(f);
 return 0;
}      

int load_matrix_func(double *m, int size)
{
 int i,j;
   for (i=0; i<size; i++)
    for (j=0; j<size; j++)
/*     m[i*(size+1)+j]=1./(i+j+1);*/
     m[i*(size+1)+j]=abs(i-j);
  for (i=0;i<size;i++)
  {
   m[i*(size+1)+size]=0;
    for (j=0; j<size; j+=2)
     m[i*(size+1)+size]+=m[i*(size+1)+j];
  }    
 return 0;
}      

int main(int argc, char * argv[])
{
 int my_rank;
 double dt;
 int t_;
 double *m;
 int way=0;
 double *x;
 double y;
 int size,t;
 
 MPI_Init(&argc, &argv);
 MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
 
  if (my_rank==0)
   {
    if (argc<3) 
    {
     printf("Usage: %s size data_sourse (1-file, 2 - function)\n",argv[0]);
     MPI_Abort(MPI_COMM_WORLD,15);
    } 
    size=atoi(argv[1]);
    way=atoi(argv[2]);      
      if ((way!=1)&&(way!=2))
      MPI_Abort(MPI_COMM_WORLD,14);
   
       if(!(m=(double*)malloc(size*(size+1)*sizeof(double)))) return 12; 
       if(!(x=(double*)malloc((size+1)*sizeof(double)))) return 12; 
     switch (way)
     {
       case 1:
        {if (load_matrix_file(m,size)<0) return -3; break;}
       case 2:
        {if (load_matrix_func(m,size)<0) return -3; break;}
      }
  
   dt=MPI_Wtime();
     if (solve(m,size,x))
      {
       printf("Virodilis`\n");
       return 1;
      } 
    dt=MPI_Wtime()-dt;
    printf("Wtime = %lf\n",dt);
    printf("\nOtvet\n");
    write_x(x,size);
    printf("Povtornij vvod matrix\n");
     switch (way)
     {
      case 1:
      {if (load_matrix_file(m,size)<0) return -3; break;}
      case 2:
      {if (load_matrix_func(m,size)<0) return -3; break;}
     }
    nev(m,size,x,&y);
    printf("Nevazka=%le\n",y);
     if (way==2)
     {
      y=0;
       for (t=0;t<size;t++)
        y+= (((t+1)%2)-x[t])>0?(((t+1)%2)-x[t]):-(((t+1)%2)-x[t]);
      printf("Oshibka=%le\n",y);
      }
    free(x);
    free(m);
    }
    else
    {
     local_solve(my_rank);
    } 
    MPI_Finalize();      
    return 0; 
}
