#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
/*#include <time.h>*/
#define E 0
#define LOCAL(size,p,rank) ((size+1)/p+((size+1)%p>rank?1:0))

void Mprint(double *m, int lines, int cols)
{
 int i,j;
  printf("----------------\n"); 
  for (i=0;i<lines;i++)
   {
    for(j=0;j<cols;j++)
     printf("%lf  ",m[i*cols+j]);
    printf("\n");
   }     
  printf("----------------\n"); 
}

int gauss(double *m, int size, double *x)
{
 int i,j;
  for (i=size-1;i>=0;i--)
   {
    x[i]=m[i*(size+1)+size];
     for (j=i+1;j<size;j++)
      x[i]-=x[j]*m[i*(size+1)+j];
     if (fabs(m[i*(size+2)])<=E) return 10;
    x[i]/=m[i*(size+2)];
   }      
   return 0;
}


void calculate(double *m, int size, int p, int rank)
{
 int col,localcol,line,i;
 double a,b,r,ycol,yline;
  for (col=0;col<size-1;col++)
  {
   localcol=(col-rank)/p+((col-rank)%p>0?1:0);
   for (line=col+1;line<size;line++)
   {

    if (col%p==rank)
     {
       r=sqrt(m[col*LOCAL(size,p,rank)+localcol]*
	    m[col*LOCAL(size,p,rank)+localcol]+
            m[line*LOCAL(size,p,rank)+localcol]*
	    m[line*LOCAL(size,p,rank)+localcol]);
	if (r<=E)
	 {
	  b=0; a=1; 
	 }
	else
	 {      
          a=m[col*LOCAL(size,p,rank)+localcol]/r;/*cos*/
          b=-m[line*LOCAL(size,p,rank)+localcol]/r;/*sin*/
	 } 
     }  
     
    MPI_Bcast(&a,1,MPI_DOUBLE,col%p,MPI_COMM_WORLD); 
    MPI_Bcast(&b,1,MPI_DOUBLE,col%p,MPI_COMM_WORLD); 
     for (i=localcol;i<LOCAL(size,p,rank);i++)
      {
       ycol=m[col*LOCAL(size,p,rank)+i]*a-
       m[line*LOCAL(size,p,rank)+i]*b;
       yline=m[col*LOCAL(size,p,rank)+i]*b+
       m[line*LOCAL(size,p,rank)+i]*a;
       m[col*LOCAL(size,p,rank)+i]=ycol;
       m[line*LOCAL(size,p,rank)+i]=yline;
      }
  }       
 } 
}


int solve(double *m,int size, double *x)
{
 int p;
 int i,j,l;
 double *localm;
 MPI_Status status;
 int t,g;
 t=clock();
 MPI_Comm_size(MPI_COMM_WORLD,&p);
 MPI_Bcast(&size,1,MPI_INT,0,MPI_COMM_WORLD);
 localm=(double*)malloc(size*LOCAL(size,p,0)*sizeof(double));
  if (!localm) { MPI_Abort(MPI_COMM_WORLD,10);}
  for (i=p-1;i>=0;i--)
  {
   for (l=0;l<size;l++)
    for (j=i;j<=size;j+=p)
    {
       localm[l*LOCAL(size,p,i)+(j-i)/p]=m[l*(size+1)+j];
     }  
   if (i!=0)
    MPI_Send(localm,size*LOCAL(size,p,i),MPI_DOUBLE,i,0,MPI_COMM_WORLD);   
   }  
  
  calculate(localm,size,p,0);
  for (i=0;i<p;i++)
  {
   if (i!=0)
    MPI_Recv(localm,size*LOCAL(size,p,i),MPI_DOUBLE,i,0,MPI_COMM_WORLD,&status);   
   for (l=0;l<size;l++)
    for (j=i;j<=size;j+=p)
    {
       m[l*(size+1)+j]=localm[l*LOCAL(size,p,i)+(j-i)/p];
     }  
   }  
             
  free(localm); 
  g= gauss(m,size,x);
  t=clock()-t;
  printf("Node = 0, clock=%d\n",t); 
  return g;
}

void local_solve(int rank)
{
 int size,p;
 double * m;
 int t;
 MPI_Status status;
 t=clock();
 MPI_Comm_size(MPI_COMM_WORLD,&p);
 MPI_Bcast(&size,1,MPI_INT,0,MPI_COMM_WORLD);
 m=(double*)malloc(size*LOCAL(size,p,rank)*sizeof(double));
  if (!m) MPI_Abort(MPI_COMM_WORLD,10);
 MPI_Recv(m,size*LOCAL(size,p,rank),MPI_DOUBLE,0,0,MPI_COMM_WORLD,&status);
 calculate(m,size,p,rank);
 MPI_Send(m,size*LOCAL(size,p,rank),MPI_DOUBLE,0,0,MPI_COMM_WORLD);
 free(m);
 t=clock()-t;
 printf("Clock from %d = %d\n",rank,t);
}

 