#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#include "solve.h"

extern void sync(int);
extern long int get_time(void);
extern long int get_full_time(void);

static long int total_time=0;
static pthread_mutex_t _mutex=PTHREAD_MUTEX_INITIALIZER;

typedef struct ARGS
{
 double *m;
 int size;
 int my_rank;
 int total;
} ARGS;

int p_getTsti(double * m,int size,int st,int i,double * a,double * b,int stolb)
{
double r;
size++;
     r=sqrt(m[st*size+stolb]*m[st*size+stolb]+m[i*size+stolb]*m[i*size+stolb]);
      if (r-fabs(m[st*size+stolb])<=e) return 1; //viroshdenie
     *a=m[st*size+stolb]/r;//cos
     *b=-m[i*size+stolb]/r;//sin
return 0;
}

void p_multT(double * m,int size,int st,int i,double a,double b,int stolb)
{
int k,l,p;
double yi,yj;
      for (k=stolb;k<=size;k++)
       {
        l=st*(size+1);
	p=i*(size+1);
        yi=m[l+k]*a-m[p+k]*b;
        yj=m[l+k]*b+m[p+k]*a;
        m[l+k]=yi;
        m[p+k]=yj;
       }    
}

int do_colom(double *m, int size, int st,int num)
{
 int i,tmp;
 double a,b;
  sync(num+1);
  
  tmp=size-st;
  tmp/=num;
  if (tmp)
   for(i=1;i<num;i++)
    {
     if (p_getTsti(m,size,st,st+i*tmp,&a,&b,st)==0)
      p_multT(m,size,st,st+i*tmp,a,b,st);
    }
  sync(num+1);    
 return 0;  
}

void write_matrix(double *m,int size)
{
 int i,j;
   for (i=0; i<size; i++)
   {
     for (j=0; j<=size; j++)
      printf("%lf  ",m[i*(size+1)+j]);
    printf("\n");
   }     
}

int gauss(double *m, int size, double *x)
{
 int i,j;
  for (i=size-1;i>=0;i--)
   {
    x[i]=m[i*(size+1)+size];
     for (j=i+1;j<size;j++)
      x[i]-=x[j]*m[i*(size+1)+j];
     if (fabs(m[i*(size+2)])<=e) return 10;
    x[i]/=m[i*(size+2)];
   }      
   return 0;
}

void * thread_fun(void * pa)
{
 double *m;
 double a,b;
 int size;
 int i;
 int my_rank,total;
 int st=0,first,last;
 long int t;
 t=get_time();
 m=((ARGS*)pa)->m;
 size=((ARGS*)pa)->size;
 my_rank=((ARGS*)pa)->my_rank;
 total=((ARGS*)pa)->total;
  while (st<size-1)
   {
    first=size-st;
    first/=total;
    last=first;
    first*=my_rank;
    first+=st;
     if (total==my_rank+1)
      last=size;
     else 
      last+=first;
     for (i=first+1;i<last;i++)
      {
       if (p_getTsti(m,size,first,i,&a,&b,st)==0)
        p_multT(m,size,first,i,a,b,st);
//       printf("rank=%d st=%d first=%d last=%d a=%le b=%le:\n",my_rank,st,first,last,a,b);	
//       write_matrix(m,size);
      }	
    sync(total+1);       
    st++;
    sync(total+1);       
   }
 t=get_time()-t;
// printf("Thread %d time =%ld\n",my_rank,t);
 pthread_mutex_lock(&_mutex);     
  total_time+=t;
 pthread_mutex_unlock(&_mutex); 
 return 0;  
} 

int start_threads(double *m,int size,int num, pthread_t * threads,ARGS * args)
{
 int i;
  for (i=0;i<num;i++)
   {
    args[i].m=m;
    args[i].my_rank=i;
    args[i].size=size;
    args[i].total=num;
    if (pthread_create(threads+i,0,thread_fun,args+i)) return -2;
   }
 return 0; 
}


int solve(double *m, int size, double *x, long int * t_th, long int * t_main, int number)
{
 pthread_t * threads;
 ARGS * args;

 int i;
 long int t;
 long int t_full;
 t_full=get_full_time();
 t=get_time();
 
  if(!(threads=(pthread_t*)malloc(number*sizeof(pthread_t)))) return -1;
  if(!(args=(ARGS*)malloc(number*sizeof(ARGS)))) return -1;
 if (start_threads(m,size,number,threads,args)) return 10;
  for(i=0;i<size-1;i++)
   if (do_colom(m,size,i,number)) return 1;
  if (!m[(size-1)*(size+1)+size-1]) return 1;
// printf("Matrix:\n");
// write_matrix(m,size); 
 t=get_time()-t;
 if (gauss(m,size,x)) return 10;
// printf("Main time =%ld\n",t);
 pthread_mutex_lock(&_mutex);     
  total_time+=t;
 pthread_mutex_unlock(&_mutex); 
  t_full=get_full_time()-t_full;

 *t_th=total_time;
 *t_main=t_full;
 return 0; 
   
}

void nev(double *m, int size, double *x, double *z)
{
 int i,j;
 double y;
 *z=0;
  for (i=0;i<size;i++)
   {
    y=-m[i*(size+1)+size];
     for (j=0;j<size;j++)
      y+=m[i*(size+1)+j]*x[j];
    (*z)+=y<0? -y:y;  
   }
}   
