#ifdef TSYSTEM
// grace.hh required for parallel execution
#ifdef __cplusplus
#include "grace/grace.hh"
#endif

#ifndef __TGCC__
#define TCALL(texpr,tinfo) { CallInfo call_info_default tinfo; texpr; }
#else
/* In TCC T starter is a compound statement, but in TGCC it is an
   inline function, so the above call_info_default scope masking won't
   work.  The runtime resetting of call_info default used below is in
   no way elegant, please suggest something better. */
#define TCALL(texpr,tinfo) \
 do { \
   CallInfo temp = call_info_default; \
   call_info_default = CallInfo tinfo; \
   texpr; \
   call_info_default = temp; \
 } while (0)
#endif /* __TGCC__ */

#define talloc(size) (shared_allocator->allocate(size))
#define tfree(ptr) (shared_allocator->deallocate(ptr))
#define tnew(type) new T_Val<type>
#define tnewarray(type,n) new T_Val<type>[n]
#define tdelete(ptr) ptr.tdelete()
#define tdeletearray(ptr) delete[] ptr
// Under TGCC, T_Val methods are not accessible to the user.
// Re-implemented as inline function, see ../src/gr_tval.hh
//#define tvalidate(val) ((val).validate())
#define abort grace::abort
#define tptrwait(x) (x).wait()
#define atexit grace::atexit
#define tsize (comm_layer->get_size())
// Under TGCC, T_Val methods are not accessible to the user.
// Re-implemented as inline function, see ../src/gr_tval.hh
//#define tdrop(a) (a).drop()
#define stop_workers Worker::stop_workers
#define cont_workers Worker::cont_workers
#define nworkers     cpu_number
#else // TSYSTEM
// Dummy stuff for sequential TC-program execution
#define tfun
#define tval
#define tout
#define tptr *
#define tmemo
#define tmain main
#define TCALL(texpr,tinfo) texpr
#define talloc malloc
#define tfree  free
#define tnew(type) ((type*)malloc(sizeof(type)))
#define tnewarray(type,n) ((type*)malloc(n*sizeof(type)))
#define tdelete(ptr) free(ptr)
#define tdeletearray(ptr) free(ptr)
#define tvalidate(val) (val)
#define tprint printf
#define terrprint(x...) fprintf(stderr, x)
#define tptrwait(x) do{}while(0)
#define finish() exit(0)
#define tsize 1
#define tdrop(a)
#ifdef DEBUG
#define dmesg  printf
#else
#define dmesg(x...)
#endif
#define vmesg  printf
#endif // TSYSTEM
