//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/compiler.cc,v $
// $Revision: 1.16 $
// $Date: 2001/08/14 13:41:46 $

/* Sufficient TC -> C transformer */

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include "node.h"

/* Something very, very traditional ;-) */

char *dprintf (const char *fmt, ...) {
    static char tmp[64*1024];
    va_list args;
    size_t  len;

    va_start (args, fmt);
    vsprintf (tmp, fmt, args); 
    va_end (args);
    len = ::strlen(tmp);
    assert (len < sizeof (tmp));
    return strdup(tmp);
}

/* main node tree transformer */

void node::compile() {

  if (is_tc_keyword() && !(find_up(TAG_PARLIST)))
    text = "";

  /* Now handle T-Function definition & body */
  if ((attrib & ATTR_TFUN) && tag == TAG_SYMDEF) {
    static bool ttypes_included = false;
    if (!ttypes_included) {
      node *fundef = find_up(TAG_FUNDEF);
      if (fundef) {
	/* Find first essensial token: we need insert it immediately before tfun */
	node *leaf = fundef;
	while (!leaf->childs.empty()) leaf = *(leaf->childs.begin());
	/* Insert `include' */
	extern char *ttypes_fname;
	leaf->cl = dprintf("%s\n\n#include \"%s\"\n\n", 
			   leaf->cl, ttypes_fname);
	ttypes_included = true;
      }
    }

    /* Now we should rename our T-function */
    text = dprintf("__tfun_%s", text);

  }


  if (tag == TAG_SYMREF && (attrib & ATTR_TFUNARG)) {
    /* First of all, refer it as struct member */
    text = dprintf("_%s->%s", (attrib & ATTR_TOUT) ? "ress" : "args", text);
    text = dprintf (((attrib & ATTR_TOUT) || (attrib & ATTR_TPTR)) ? 
       "(%s)" : "(%s.get_value())", text);
       //"(&(%s.get_value()))" : "(*&(%s.get_value()))", text);
  }

  if (tag == TAG_SYMDEF && (attrib & ATTR_TVAL) && !(attrib & ATTR_TFUNARG)) {
    node *decl = find_up(TAG_DECL);
    if (!decl) decl = find_up(TAG_FUNDEF);
    if (decl == NULL) {
      error("Unable locate declaration for non-tfun-arg tval definition `%s'",
	    text);
    }
    node *decl1 = decl->find_first_leaf();
    // FIXME: hack
    if (!strstr(decl1->cl,"T_Val<")) {
      node *const1 = decl1->find_first(TAG_CONST);
      char *const_str = "";
      if (const1) {
	const1->text = "/*const*/";
	const_str    = " const ";
      }
      decl1->cl = dprintf("%s%sT_Val<", decl1->cl, const_str);
      node *symdef = decl->find_first(TAG_SYMDEF);
      symdef->cl = dprintf(" >%s", symdef->cl);
    }
  }

  if (tag == TAG_SYMDEF && (attrib & ATTR_TPTR) && !(attrib & ATTR_TFUNARG)) {
    node *decl = find_up(TAG_DECL);
    if (!decl) decl = find_up(TAG_FUNDEF);
    if (decl == NULL) {
      error("Unable locate declaration for non-tfun-arg tptr definition `%s'",
	    text);
    }
    node *decl1 = decl->find_first_leaf();
    // FXIME: hack
    if (!strstr(decl1->cl,"T_Ptr<")) {
      node *const1 = decl1->find_first(TAG_CONST);
      char *const_str = "";
      if (const1) {
	const1->text = "/*const*/";
	const_str    = " const ";
      }
      
      decl1->cl = dprintf("%s%sT_Ptr<", decl1->cl, const_str);
      node *symdef = decl->find_first(TAG_SYMDEF);
      symdef->cl = dprintf(" >%s", symdef->cl);
    }
  }

  /* Now we are going to handle T-variables _carefully_ */

  if ((attrib & ATTR_TVAL) && tag == TAG_SYMREF) {
    if (!(attrib & ATTR_LVALUE)) {
      node *func = find_up(TAG_FUNCALL);
      bool inside_speccall = func && func->find_first(TAG_SYMREF) &&
	 !strcmp(func->find_first(TAG_SYMREF)->text,"tdrop");
      if (inside_speccall) 
	;
      else {
	node *unop = find_up(TAG_UNOP);
	node *amp_node = NULL;
	if (unop) {
	  list<node*>::iterator it = unop->childs.begin();
	  amp_node = *it;
          if (!(*(*(it++))->text == '&' && *(it) == this)) amp_node = NULL;
	}
	if (amp_node) {
          amp_node->text=""; // remove '&'
	  text = dprintf("%s.get_tptr()", text); 
	} else {
	  text = dprintf((attrib & ATTR_TREADY) ? 
			 "%s.get_value(/*R*/)" : "%s.get_value()",
			 text);
	}
      }
    }
  }

  /* General recursion */
  if (!childs.empty()) { 
    list<node*>::iterator it = childs.begin();
    while (it != childs.end()) { (*(it))->compile(); it++; } 
  }
}

void node::profile_codegen() {
  if (tag == TAG_SCOPE) {
    node *fst = find_first_attrib(ATTR_STATEMENT);
    if (fst) fst->cl = dprintf(" GR_PROFILE_FUNCTION; %s", fst->cl);
    profile_statements();
  } else {
    /* General recursion */
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (it != childs.end()) { (*(it))->profile_codegen(); it++; } 
    }
  }
}

void node::profile_statements() {
  if ((attrib & ATTR_STATEMENT) && !find_first_inside_attrib(ATTR_STATEMENT)) {
      extern bool do_profile_statements();
      if (do_profile_statements()) {
        cl = dprintf("%s GR_PROFILE_STATEMENT(", cl);
        cr = dprintf("); %s", cr);
      }
  } 
  /* General recursion */
  if (!childs.empty()) { 
    list<node*>::iterator it = childs.begin();
    while (it != childs.end()) { (*(it))->profile_statements(); it++; } 
  }
}

