//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/gentypes.cc,v $
// $Revision: 1.21 $
// $Date: 2002/01/31 17:22:17 $

/* Additional types & macros generation */

#include "node.h"

static char* tfunname;
struct tfunparinfo {
  char type[4096];
  node *symdef;
} tfunpars[32];
static int tfunparnum;

static FILE *ttypes = NULL;
#define tprintf(x...) do { \
   extern char *ttypes_fname; \
   if (!ttypes) ttypes = fopen(ttypes_fname,"w"); \
   fprintf(ttypes,x); \
} while(0)

void node::gen_types_parlist_param(char **s, bool &stop) {
  if (attrib & ATTR_TFUNARG) {
    stop = true; /* End of our walk */
  } else {
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (!stop && it != childs.end()) { 
	(*(it))->gen_types_parlist_param(s,stop); 
	it++;
      } 
    } else if (!is_tc_keyword()) {
      *s += sprintf(*s, "%s ", text);
    }
  }
}

void node::gen_types_parlist() {
  if (tag == TAG_PARDECL) {

    node *param = find_first(TAG_SYMDEF);
    if (!param)
      error("Unable get parameter name for tfun `%s'", tfunname);

    /* Propogate ATTR_TFUNARG */
    param->attrib |= ATTR_TFUNARG;
    if (!param->refs.empty()) { 
      list<node*>::iterator it = param->refs.begin();
      while (it != param->refs.end()) { 
	(*(it))->attrib |= ATTR_TFUNARG;
	it++; 
      } 
    }
    
    /* Fill out tfunparinfo structure */
    tfunpars[tfunparnum].symdef = param;
    char *s = tfunpars[tfunparnum].type;
    bool stop = false;    
    gen_types_parlist_param(&s,stop);

    // Strip last space
    *--s = 0;
    //fprintf (stderr, "tfunpar.type='%s'\n", tfunpars[tfunparnum].type);
    // Strip last asterisk for TOUT parameter
    if ((tfunpars[tfunparnum].symdef->attrib & ATTR_TOUT))
      if (*--s == '*') *s = 0;
      else error("Can't find '*' at end of tout `%s' type (tfun `%s')", 
		 tfunpars[tfunparnum].symdef->text, tfunname);
    // Iterate 
    tfunparnum++;

  } else if (!childs.empty()) { 
    list<node*>::iterator it = childs.begin();
    while (it != childs.end()) { 
      (*(it))->gen_types_parlist(); 
      it++; 
    } 
  }
  
}

void node::gen_types() {
  if ((attrib & ATTR_TFUN) && tag == TAG_SYMDEF) {
    tprintf("\n/* T-wrappers for tfun `%s' */\n", text);
    tfunname = text;

    /* FIXME: return type may contain PARLIST/PARDECL ! */
    node *up = find_up(TAG_FUNDEF);
    if (!up)
      error("Can't find function definition node for `%s'\n", tfunname);
    node *pars = up->find_first(TAG_PARLIST);
    if (!pars) 
      pars = up->find_first(TAG_PARDECL);
    if (!pars)
      error("Can't find parameter list for tfun `%s'", tfunname);

    tfunparnum = 0; 
    pars->gen_types_parlist();

    /* Generate args & ress structures */
    int i;

    tprintf("struct __%s_args {\n", tfunname);
    for (i = 0; i < tfunparnum; i++) 
      if (!(tfunpars[i].symdef->attrib & ATTR_TOUT)) {

	//FIXME !!!: VERY dirty hack!!!
	char *const1 = strstr(tfunpars[i].type,"const");
	if (const1) memcpy(const1,"/*c*/",5);
	const char *const_str = tfunpars[i].symdef->attrib & ATTR_CONST ? 
	  "" : ""; //FIXME!!!!!!!!!!!!!!!!!!!!!

	if (!(tfunpars[i].symdef->attrib & ATTR_TPTR)) {
	  tprintf("\t%sT_Val<%s> %s;\n",
		  const_str, tfunpars[i].type, tfunpars[i].symdef->text);
	} else {
	  tprintf("\t%sT_Ptr<%s> %s;\n",
		  const_str, tfunpars[i].type, tfunpars[i].symdef->text);
	}
      }
    tprintf("};\n");

    tprintf("struct __%s_ress {\n", tfunname);
    for (i = 0; i < tfunparnum; i++) 
      if (tfunpars[i].symdef->attrib & ATTR_TOUT)
	tprintf("\tT_Ptr<%s> %s;\n", tfunpars[i].type,tfunpars[i].symdef->text);
    tprintf("};\n");

    /* Some additional things if we are a memo-function */
    if (attrib & ATTR_TMEMO) {
      /* Our _args contains only ready tval's and globalized tptr's */
      /* So, we can just compute hash and check local memo table */
      tprintf("extern void __tfun_%s(__%s_args const *, __%s_ress *);\n",
	      tfunname, tfunname, tfunname);
      tprintf("inline void ptfun_%s(__%s_args const &_args, __%s_ress& _ress) {\n",
	      tfunname, tfunname, tfunname);
      for (i = 0; i < tfunparnum; i++) {
	if (tfunpars[i].symdef->attrib & ATTR_TOUT) {
	  /* Alocate persistent intermediate tval's for results */
	  tprintf("\t _ress.%s = tnew(%s);\n", 
		  tfunpars[i].symdef->text, tfunpars[i].type);
	  /* ... and bind it to current (original) continuation */
	  tprintf("\t (*(_ress.%s)).bind(current->get_continuation());\n", 
		  tfunpars[i].symdef->text);
	  tprintf("\t _ress.%s.globalize();\n",
		  tfunpars[i].symdef->text);
	}
      }

      tprintf("\t MemoMessage<__%s_args,__%s_ress> *msg = new MemoMessage<__%s_args,__%s_ress>(_args,&ptfun_%s,_ress);\n",
	      tfunname, tfunname, tfunname, tfunname, tfunname);
      tprintf("\t comm_layer->broadcast(msg);\n");

      tprintf("\t __tfun_%s(&_args,&_ress);\n", tfunname);
      tprintf("}\n");

      tprintf("DEFINE_MEMO_FUNCTION(__%s_ress,ptfun_%s,__%s_args)\n\n",
	      tfunname, tfunname, tfunname);

      /* Additional trampoline will only call _ptfun and set result */
      tprintf("inline void memo_tfun_%s(__%s_args const *_args, __%s_ress *_ress) {\n",
	      tfunname, tfunname, tfunname);    
      tprintf("\t __%s_ress *ress = &memo_ptfun_%s(*_args);\n",
	      tfunname, tfunname);
      for (i = 0; i < tfunparnum; i++) {
	if (tfunpars[i].symdef->attrib & ATTR_TOUT) {
	    tprintf("\t *_ress->%s = *ress->%s;\n", 
		    tfunpars[i].symdef->text, tfunpars[i].symdef->text);
	}
      }
      tprintf("}\n");
    }

    /* Generate macro for tfunction call */
    tprintf("#define %s(", tfunname);
    for (i = 0; i < tfunparnum; i++) {
      tprintf(i == 0 ? "_%s" : ",_%s", tfunpars[i].symdef->text);
    }
    tprintf(") { \\\n"
	   "\tContinuation<__%s_args, __%s_ress>* __cont = \\\n"
	   "\tnew Continuation<__%s_args, __%s_ress>( \\\n"
	   "\t  %s_tfun_%s, new ComputationStateFactory<FastProcess>, \\\n"
           "\t  call_info_default \\\n"
	   "\t); \\\n", 
	    tfunname, tfunname, 
	    tfunname, tfunname, 
	    (attrib & ATTR_TMEMO) ? "memo" : "_", tfunname);
    for (i = 0; i < tfunparnum; i++) {
      if (!(tfunpars[i].symdef->attrib & ATTR_TOUT)) {
	tprintf("\t__cont->args.%s = (_%s); \\\n", 
		tfunpars[i].symdef->text, tfunpars[i].symdef->text);
      } else {
	tprintf("\t(*(_%s)).bind(__cont); \\\n", 
		tfunpars[i].symdef->text);
	tprintf("\t__cont->ress.%s = (_%s); \\\n", 
	       tfunpars[i].symdef->text, tfunpars[i].symdef->text);
      }
    }
    tprintf("\t__cont->try_fixed(); \\\n");
    tprintf("}\n\n");
    

    /* Now we just comment out old parlist and set new */
    pars->cl = "/* ";
    pars->cr = dprintf(" */\n\t__%s_args const* _args, __%s_ress* _ress",
		       tfunname, tfunname);

  } else if (!childs.empty()) { 
    list<node*>::iterator it = childs.begin();
    while (it != childs.end()) { 
      (*(it))->gen_types(); 
      it++; 
    } 
  }
}
