/*
//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/lexer.l,v $
// $Revision: 1.10 $
// $Date: 2001/08/23 20:24:14 $
*/

D                       [0-9]
L                       [a-zA-Z_]
H                       [a-fA-F0-9]
E                       [Ee][+-]?{D}+
FS                      (f|F|l|L)
IS                      (u|U|l|L)*

%{
#include <stdio.h>
#include "node.h"
#include ".y.tab.h"

/* Add yytext yo comment container */
void add_to_cc();

/* Update source file line information */
extern int src_line;
void src_file_info();

/* Skip "C" comment, add it to comment container */
void comment();

/* Check that name is a typename */
int check_type();

/*#define YY_DECL int yylex1 YY_PROTO(( void )) */
/* Makes (w/ different count) token from 
   comment container (w/ cleanup) and yytext */ 
void count_token();

#define CRET(x) {count_token(); return(x);}

%}

%%
"/*T"                   { CRET(TCOMMENT_BEGIN); }
"T*/"                   { CRET(TCOMMENT_END); }
"/*"                    { add_to_cc(); comment(); }
^"#".*\n                { add_to_cc(); src_file_info(); }

"auto"                  { CRET(AUTO); }
"break"                 { CRET(BREAK); }
"case"                  { CRET(CASE); }
"char"                  { CRET(CHAR); }
"const"                 { CRET(CONST); }
"continue"              { CRET(CONTINUE); }
"default"               { CRET(DEFAULT); }
"do"                    { CRET(DO); }
"double"                { CRET(DOUBLE); }
"else"                  { CRET(ELSE); }
"enum"                  { CRET(ENUM); }
"extern"                { CRET(EXTERN); }
"float"                 { CRET(FLOAT); }
"for"                   { CRET(FOR); }
"goto"                  { CRET(GOTO); }
"if"                    { CRET(IF); }
"int"                   { CRET(INT); }
"inline"                { CRET(INLINE); }
"__complex__"           { CRET(COMPLEX); }
"long"                  { CRET(LONG); }
"register"              { CRET(REGISTER); }
"return"                { CRET(RETURN); }
"short"                 { CRET(SHORT); }
"signed"                { CRET(SIGNED); }
"sizeof"                { CRET(SIZEOF); }
"static"                { CRET(STATIC); }
"struct"                { CRET(STRUCT); }
"switch"                { CRET(SWITCH); }
"typedef"               { CRET(TYPEDEF); }
"typeof"                { CRET(TYPEOF); }
"union"                 { CRET(UNION); }
"unsigned"              { CRET(UNSIGNED); }
"void"                  { CRET(VOID); }
"volatile"              { CRET(VOLATILE); }
"tval"                  { CRET(TVAL); }
"tfun"                  { CRET(TFUN); }
"tout"                  { CRET(TOUT); }
"tptr"                  { CRET(TPTR); }
"tnew"                  { CRET(TNEW); }
"tnewarray"             { CRET(TNEWARRAY); }
"tmemo"                 { CRET(TMEMO); }
"while"                 { CRET(WHILE); }
"__label__"             { CRET(LABEL); }

{L}({L}|{D})*           { CRET(check_type()); }

0[xX]{H}+{IS}?          { CRET(CONSTANT); }
0[xX]{H}+{IS}?          { CRET(CONSTANT); }
0{D}+{IS}?              { CRET(CONSTANT); }
0{D}+{IS}?              { CRET(CONSTANT); }
{D}+{IS}?               { CRET(CONSTANT); }
{D}+{IS}?               { CRET(CONSTANT); }
'(\\.|[^\\'])+'         { CRET(CONSTANT); }

{D}+{E}{FS}?            { CRET(CONSTANT); }
{D}*"."{D}+({E})?{FS}?  { CRET(CONSTANT); }
{D}+"."{D}*({E})?{FS}?  { CRET(CONSTANT); }

\"(\\.|[^\\"])*\"       { CRET(STRING_LITERAL); }

">>="                   { CRET(RIGHT_ASSIGN); }
"<<="                   { CRET(LEFT_ASSIGN); }
"+="                    { CRET(ADD_ASSIGN); }
"-="                    { CRET(SUB_ASSIGN); }
"*="                    { CRET(MUL_ASSIGN); }
"/="                    { CRET(DIV_ASSIGN); }
"%="                    { CRET(MOD_ASSIGN); }
"&="                    { CRET(AND_ASSIGN); }
"^="                    { CRET(XOR_ASSIGN); }
"|="                    { CRET(OR_ASSIGN); }
">>"                    { CRET(RIGHT_OP); }
"<<"                    { CRET(LEFT_OP); }
"++"                    { CRET(INC_OP); }
"--"                    { CRET(DEC_OP); }
"->"                    { CRET(PTR_OP); }
"&&"                    { CRET(AND_OP); }
"||"                    { CRET(OR_OP); }
"<="                    { CRET(LE_OP); }
">="                    { CRET(GE_OP); }
"=="                    { CRET(EQ_OP); }
"!="                    { CRET(NE_OP); }
";"                     { CRET(';'); 
}
"{"                     { CRET('{'); }
"}"                     { CRET('}'); }
","                     { CRET(','); }
":"                     { CRET(':'); }
"="                     { CRET('='); }
"("                     { CRET('('); }
")"                     { CRET(')'); }
"["                     { CRET('['); }
"]"                     { CRET(']'); }
"."                     { CRET('.'); }
"&"                     { CRET('&'); }
"!"                     { CRET('!'); }
"~"                     { CRET('~'); }
"-"                     { CRET('-'); }
"+"                     { CRET('+'); }
"*"                     { CRET('*'); }
"/"                     { CRET('/'); }
"%"                     { CRET('%'); }
"<"                     { CRET('<'); }
">"                     { CRET('>'); }
"^"                     { CRET('^'); }
"|"                     { CRET('|'); }
"?"                     { CRET('?'); }

[ \t\v\n\f]             { add_to_cc(); if (*yytext == '\n') src_line++; }
.                       { add_to_cc(); /*ignore bad characters*/ }

%%

int yywrap()
{
   return(1);
}


/* flex input routine */
static int input();

/* Comment container */
#define MAX_CMNT (256*1024)
char cc[MAX_CMNT];
int  cpos = 0;

/* Input line/column info */
#define MAX_LINE (64*1024)
char line_text[MAX_LINE];
int  line_text_pos = 0;
int  column = 0;
int  line = 1;
char src_file[MAX_LINE] = "<unknown>";
int  src_line = 1; 


/* Add yytext yo comment container */
void add_to_cc() {
   int len = strlen(yytext);
   memcpy(cc+cpos,yytext,len);
   cpos += len;
}

/* Update source file line information */
void src_file_info() {
   char file[MAX_LINE]; int line;
   memset(file,0,sizeof(file));
   if (2 == sscanf(yytext, "# %d \"%s\"", &line, file)) {
      strncpy(src_file, file, strlen(file)-1); // Suppress extra '"'
      src_line = line;
   }
}

/* Skip "C" comment body & end, add ones to comment container */
void comment()
{
  char *term = "*/";
  char *termp = term;
  while(*termp) {
    int c = yyinput();
    if (c == 0 || c == EOF) return;
    cc[cpos++] = c;
    termp = c == *termp ? termp+1 : term;
  }
}

/* Carefully count line, column and printable string */
void count (char *text)
{
  char *p;
  int len = strlen(text);
  for (p = text; p < text+len; p++) {
    if (*p == '\n') {
      line++;
      column = 0;
      line_text_pos = 0;
    } else {
      line_text[line_text_pos++] = *p;
      if (*p == '\t') {
	column += 8 - (column % 8);
      } else {
	column++;
      }
    }
  }
   // To update our typename database ;-)
   extern void typextr(char*);
   typextr(text);
}


/* Makes (w/ different count) token from  
   comment container (w/ cleanup) and yytext */ 

void count_token() 
{
   yylval = NODE(TAG_NI);
   if (cpos != 0) { 
     cc[cpos] = 0;
     count(cc);
     yylval->cl = strdup(cc);
     cpos = 0;
   }
   count(yytext);
}


/* We need it to understand ANSI "C" _at all_ ! */
hash_map<const char*, char*, hash<const char*>, eqstr> typenames;

int check_type()
{
// It is important to copy `yytext'
   return typenames[strdup(yytext)] ? TYPE_NAME : IDENTIFIER;
}
/*
int yylex()
{
 int x = yylex1();
 printf("%d ", x);
 return x;
}

*/
