//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/node.h,v $
// $Revision: 1.18 $
// $Date: 2001/08/23 20:24:14 $

/* Syntax Tree node definition and operations */

//#include <stl.h>
#include <list>
#if __GNUC__ >= 3
#include <ext/hash_map> // Hmm, SGI extension, however!
#else
#include <hash_map.h> // Hmm, SGI extension, however!
#endif
#include <cstdio>

using namespace std;

/* printf to dynamic memory */
char *dprintf (const char *fmt, ...);

/* Error reporiting routine */
#define error(x...) { fprintf(stderr,"Error: "); \
                      fprintf(stderr,x); \
                      fprintf(stderr,". Exiting\n"); \
                      exit(1); }


/* node tags */

#define TAG_NI           0   /* Nothing Interesting */
#define TAG_TVAL         1   /* TVAL token */
#define TAG_TFUN         2   /* TFUN token */
#define TAG_SYMDEF       3   /* Symbol to be defined */
#define TAG_SYMREF       4   /* Symbol to be referenced */
#define TAG_SCOPE        5   /* A chance to fork symbol scope */
#define TAG_FUNDEF       6   /* Completed function definition */
#define TAG_PARLIST      7   /* Parameter definition list */
#define TAG_PARDECL      8   /* Parameter definition */
#define TAG_TOUT         9   /* TOUT token */
#define TAG_DECL         10  /* Declaration */
#define TAG_FUNCALL      11  /* Function call */
#define TAG_FPAR         12  /* Function parameter list */
#define TAG_TPTR         13  /* T-pointer */
#define TAG_UNOP         14  /* Unary operator */
#define TAG_CONST        15  /* C `const' */
#define TAG_SYMPREDEF    16  /* Predefinition inside T-comment */
#define TAG_TMEMO        17  /* Memoized function */

/* node attributes */

#define ATTR_TVAL     (1<<0)  /* It marks T-variable */
#define ATTR_TFUN     (1<<1)  /* It marks T-function */
#define ATTR_TREADY   (1<<2)  /* It marks ready T-variable */
#define ATTR_LVALUE   (1<<3)  /* It marks variable is LVALUE */
#define ATTR_DFLOWA   (1<<4)  /* It marks dataflow re-area */
#define ATTR_DENTRY   (1<<5)  /* It marks possible entries */
#define ATTR_TOUT     (1<<6)  /* It marks T-out variables */
#define ATTR_TFUNARG  (1<<7)  /* It marks T-function agruments */
#define ATTR_TPTR     (1<<8)  /* It marks T-pointer */
#define ATTR_CONST    (1<<9)  /* It marks `const' */
#define ATTR_STATEMENT (1<<10) /* It marks C statement (operator) */
#define ATTR_TMEMO    (1<<11)  /* It marks memo-function */
#define ATTR_MASK ((unsigned long)((1<<12) - 1))

/* No more than 31 */

/* All we need is node */

class scope;

class node {
 public:

  int tag;
  unsigned long attrib;
  char *cl, *text, *cr;

  node *parent;
  list <node*> childs; /* Node Tree childs */

  node *next_ident_arg;      /* bounded to function-calls */

  node *def;           /* in this node symbol "text" is defined */
  list <node*> refs;   /* in these nodes symbol "text" is referenced to*/

  node (int tag, char *cl, char *text, char *cr) {
    this->tag = tag;
    this->attrib = 0;

    this->cl = cl;
    this->text = text;
    this->cr = cr;

    parent = def = next_ident_arg = NULL;
  }

  node *attach(node *n1 = NULL, node *n2 = NULL, 
	       node *n3 = NULL, node *n4 = NULL,
	       node *n5 = NULL, node *n6 = NULL,
	       node *n7 = NULL, node *n8 = NULL,
	       node *n9 = NULL, node *n10 = NULL) {
    if (n1) { childs.push_back(n1); n1->parent = this; }
    if (n2) { childs.push_back(n2); n2->parent = this; }
    if (n3) { childs.push_back(n3); n3->parent = this; }
    if (n4) { childs.push_back(n4); n4->parent = this; }
    if (n5) { childs.push_back(n5); n5->parent = this; }
    if (n6) { childs.push_back(n6); n6->parent = this; }
    if (n7) { childs.push_back(n7); n7->parent = this; }
    if (n8) { childs.push_back(n8); n8->parent = this; }
    if (n9) { childs.push_back(n9); n9->parent = this; }
    if (n10){ childs.push_back(n10); n10->parent = this; }
    return this;
  }

  void dump(FILE *fo) {
    fprintf(fo,"%s",cl);
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (it != childs.end()) { (*(it))->dump(fo); it++; } 
    } else {
      fprintf(fo,"%s",text); 
    }
    fprintf(fo,"%s",cr); 
  }

  void debugdump(FILE *fo, int offset = 0) {
    char *tagstring;
    char atrstring[100];

    switch(tag){
       case TAG_NI: tagstring = "TAG_NI";
            break;
       case TAG_TVAL: tagstring = "TAG_TVAL";
            break;
       case TAG_TFUN: tagstring = "TAG_TFUN";
            break;
       case TAG_SYMDEF: tagstring = "TAG_SYMDEF";
            break;
       case TAG_SYMREF: tagstring = "TAG_SYMREF";
            break;
       case TAG_SCOPE: tagstring = "TAG_SCOPE";
            break;
       case TAG_FUNDEF: tagstring = "TAG_FUNDEF";
            break;
       case TAG_PARLIST: tagstring = "TAG_PARLIST";
            break;
       case TAG_PARDECL: tagstring = "TAG_PARDECL";
            break;
       case TAG_TOUT: tagstring = "TAG_TOUT";
            break;
       case TAG_DECL: tagstring = "TAG_DECL";
            break;
       case TAG_FUNCALL: tagstring = "TAG_FUNCALL";
            break;
       case TAG_FPAR: tagstring = "TAG_FPAR";
            break;
       case TAG_TPTR: tagstring = "TAG_TPTR";
            break;
       case TAG_UNOP: tagstring = "TAG_UNOP";
            break;
       case TAG_CONST: tagstring = "TAG_CONST";
            break;
       case TAG_SYMPREDEF: tagstring = "TAG_SYMPREDEF";
            break;
    default:  fprintf(stderr, "ERR in function debugdump: unknown tag!\n");
              exit (-1);   
    }
    if(attrib){
       if (attrib & (~ATTR_MASK)) {
         fprintf(stderr, "Bad attrib value %lx!!!\n", attrib);
         exit(-1);
       }
       strcpy(atrstring, "ATTR: ");
       if(attrib & ATTR_TVAL) strcat(atrstring, "TVAL, ");
       if(attrib & ATTR_TFUN) strcat(atrstring, "TFUN, ");
       if(attrib & ATTR_TREADY) strcat(atrstring, "TREADY, ");
       if(attrib & ATTR_LVALUE) strcat(atrstring, "LVALUE, ");
       if(attrib & ATTR_DFLOWA) strcat(atrstring, "DFLOWA, ");
       if(attrib & ATTR_DENTRY) strcat(atrstring, "DENTRY, ");
       if(attrib & ATTR_TOUT) strcat(atrstring, "TOUT, ");
       if(attrib & ATTR_TFUNARG) strcat(atrstring, "TFUNARG, ");
       if(attrib & ATTR_TPTR) strcat(atrstring, "TPTR, ");
       if(attrib & ATTR_CONST) strcat(atrstring, "CONST, ");
       char *p = strrchr(atrstring, ',');
       if (p)
         *p = '\0';
    }
    else atrstring[0] = '\0';

    fprintf(fo,"%*sNODE %p: %s %s cl='%s' text='%s' cr='%s'\n",
	    offset, "", this, tagstring, atrstring, cl, text, cr);

    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (it != childs.end()) { (*(it))->debugdump(fo, offset+3); it++; } 
    }
  }

  void dump(char **s) {
    *s += sprintf(*s,"%s",cl);
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (it != childs.end()) { (*(it))->dump(s); it++; } 
    } else {
      *s += sprintf(*s,"%s",text); 
    }
    *s += sprintf(*s,"%s",cr); 
  }

  node* find_up(int tag) {
    node *n = this;
    while(n->tag != tag) {
      if (n->parent) n = n->parent;
      else return NULL;
    }
    return n;
  }

  node* find_first(int tag) {
    node *n = this->tag == tag ? this : NULL;
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (!n && it != childs.end()) { 
	n = (*(it))->find_first(tag); 
	it++; 
      } 
    }
    return n;
  }

  node* find_first_attrib(unsigned long attrib) {
    node *n = (this->attrib & attrib) ? this : NULL;
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (!n && it != childs.end()) { 
	n = (*(it))->find_first_attrib(attrib); 
	it++; 
      } 
    }
    return n;
  }

  node* find_first_inside(int tag) {
    node *n = NULL;
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (!n && it != childs.end()) { 
	n = (*(it))->find_first(tag); 
	it++; 
      } 
    }
    return n;
  }

  node* find_first_inside_attrib(unsigned long attrib) {
    node *n = NULL;
    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (!n && it != childs.end()) { 
	n = (*(it))->find_first_attrib(attrib); 
	it++; 
      } 
    }
    return n;
  }

  node* find_first_leaf() {
    return childs.empty() ? this : (*(childs.begin()))->find_first_leaf();
  }

  /* clone breaks symbol connection ! */
  node* clone() {
    node *n = new node(tag,strdup(cl),strdup(text),strdup(cr));
    n->tag = tag;
    n->attrib = attrib;

    if (!childs.empty()) { 
      list<node*>::iterator it = childs.begin();
      while (it != childs.end()) { 
	node *cn = (*it)->clone();
	cn->parent = n;
	n->childs.push_back(cn);
	it++; 
      } 
    }
    return n;
  }
  
  bool is_tc_keyword() {
    return 
      (tag == TAG_TVAL && !::strcmp(text,"tval")) ||
      (tag == TAG_TFUN && !::strcmp(text,"tfun")) ||
      (tag == TAG_TPTR && !::strcmp(text,"tptr")) ||
      (tag == TAG_TOUT && !::strcmp(text,"tout")) ||
      (tag == TAG_TMEMO && !::strcmp(text,"tmemo"));
  }

  void calc_attrib(scope *sc, long &attrib); /* attribute.cc */

  void gen_types();                           /* gentypes.cc */
  void gen_types_parlist();                  /* gentypes.cc */
  void gen_types_parlist_param(char **s, bool &stop); /* gentypes.cc */

  void strictness();                         /* dataflow.cc */
  void strictness_set_tready(node *dflowa);  /* dataflow.cc */

  void make_c_function();                    /* compiler.cc */
  void clean_c_function();                   /* compiler.cc */
  void compile();                            /* compiler.cc */

  void profile_codegen();                    /* compiler.cc */
  void profile_statements();                 /* compiler.cc */

  static node *root;
};


/* Symbol Scope */

struct eqstr {
  bool operator()(const char* s1, const char* s2) const {
    return ::strcmp(s1,s2) == 0;
  }
};


class scope {
public:
  scope *parent;

  scope(scope *parent) {
    this->parent = parent;
  }

  hash_map<const char*, node*, hash<const char*>, eqstr> syms;

  void define(node* n) {
    if (syms[n->text])
      fprintf(stderr, "FIXME: multiply definition for symbol `%s'\n", n->text);
    syms[n->text] = n;
  }
   
  node* lookup(char *s) { 
    if (!syms[s]) 
      return parent ? parent->lookup(s) : NULL;
    else
      return syms[s];
  }

};


/* Definitions for lexer/parser */

#define NODE(tag,nx...)  ((new node(tag,"",strdup(yytext),""))->attach(nx))
#define NODE_OP(tag,op)  ((new node(tag,"",op,"")))

#define YYSTYPE node*

extern YYSTYPE yylval;

extern hash_map<const char*, char*, hash<const char*>, eqstr> typenames;

/* */






