/*
//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/parser.y,v $
// $Revision: 1.17 $
// $Date: 2001/10/18 05:45:07 $
*/

%{ /* -*-Fundamental-*- */
/* Full ANSI "C" grammar w/ extensions */

#include "node.h"
extern int yylex(void);
extern char yytext[];
int yyerror(char *errmsg);

#define YYERROR_VERBOSE
%}

%token IDENTIFIER CONSTANT STRING_LITERAL SIZEOF
%token PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP EQ_OP NE_OP
%token AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN ADD_ASSIGN
%token SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token XOR_ASSIGN OR_ASSIGN TYPE_NAME TYPEOF LABEL

%token TYPEDEF EXTERN STATIC AUTO REGISTER
%token CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE CONST VOLATILE VOID 
%token TVAL TFUN TOUT TPTR TNEW TNEWARRAY TCOMMENT_BEGIN TCOMMENT_END TMEMO
%token INLINE COMPLEX
%token STRUCT UNION ENUM ELIPSIS RANGE

%token CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE BREAK RETURN

%start file
%%

string_literal_list
	: string_literal_list STRING_LITERAL { $$ = NODE(TAG_NI,$1,$2); }
	| STRING_LITERAL

primary_expr
        : identifier     {$$ = $1; $$->tag = TAG_SYMREF;} 
        | CONSTANT       
        | string_literal_list 
        | '(' expr ')'   {$$ = NODE(TAG_NI,$1,$2,$3);}
	| '(' compound_statement ')'   {$$ = NODE(TAG_NI,$1,$2,$3);}   
        ;

postfix_expr
        : primary_expr                       
        | postfix_expr '[' expr ']'   {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        | postfix_expr '(' ')'        {$$ = NODE(TAG_NI,$1,$2,$3);}
        | postfix_expr '(' argument_expr_list ')' {
		$$ = NODE(TAG_FUNCALL,$1,$2,$3,$4);
		$$->next_ident_arg = 
			$3->tag==TAG_SYMREF ? $3 : $3->next_ident_arg;
	}
        | postfix_expr '.' identifier {$$ = NODE(TAG_NI,$1,$2,$3);}
        | postfix_expr PTR_OP identifier  {$$ = NODE(TAG_NI,$1,$2,$3);}
        | postfix_expr INC_OP         {$$ = NODE(TAG_NI,$1,$2);}
        | postfix_expr DEC_OP         {$$ = NODE(TAG_NI,$1,$2);} 
        ;

argument_expr_list
        : assignment_expr
        | argument_expr_list ',' assignment_expr {
        	$$ = NODE(TAG_NI,$1,$2,$3);
		$3->next_ident_arg = 
			$1->tag==TAG_SYMREF ? $1 : $1->next_ident_arg;
		$$->next_ident_arg = 
			$3->tag==TAG_SYMREF ? $3 : $3->next_ident_arg;
        }
        ;

unary_expr
        : postfix_expr              
        | INC_OP unary_expr         {$$ = NODE(TAG_NI,$1,$2);}
        | DEC_OP unary_expr         {$$ = NODE(TAG_NI,$1,$2);}
        | unary_operator cast_expr  {$$ = NODE(TAG_UNOP,$1,$2);}
        | SIZEOF unary_expr         {$$ = NODE(TAG_NI,$1,$2);}
        | SIZEOF '(' type_name ')'  {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
	| TNEW '(' type_name ')'    {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
	| AND_OP identifier         {$$ = NODE(TAG_NI,$1,$2);}
	| TNEWARRAY '(' type_name ',' expr ')'    {
			$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6);
	}
        ;

unary_operator
        : '&'  
        | '*' 
        | '+'  
        | '-'  
        | '~' 
        | '!' 
        ;

cast_expr
        : unary_expr
        | '(' type_name ')' cast_expr    {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        ;

multiplicative_expr
        : cast_expr
        | multiplicative_expr '*' cast_expr   {$$ = NODE(TAG_NI,$1,$2,$3);} 
        | multiplicative_expr '/' cast_expr   {$$ = NODE(TAG_NI,$1,$2,$3);}
        | multiplicative_expr '%' cast_expr   {$$ = NODE(TAG_NI,$1,$2,$3);} 
        ;

additive_expr
        : multiplicative_expr
        | additive_expr '+' multiplicative_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        | additive_expr '-' multiplicative_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

shift_expr
        : additive_expr
        | shift_expr LEFT_OP additive_expr    {$$ = NODE(TAG_NI,$1,$2,$3);} 
        | shift_expr RIGHT_OP additive_expr   {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

relational_expr
        : shift_expr
        | relational_expr '<' shift_expr      {$$ = NODE(TAG_NI,$1,$2,$3);} 
        | relational_expr '>' shift_expr      {$$ = NODE(TAG_NI,$1,$2,$3);}
        | relational_expr LE_OP shift_expr    {$$ = NODE(TAG_NI,$1,$2,$3);}
        | relational_expr GE_OP shift_expr    {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

equality_expr
        : relational_expr
        | equality_expr EQ_OP relational_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        | equality_expr NE_OP relational_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

and_expr
        : equality_expr
        | and_expr '&' equality_expr    {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

exclusive_or_expr
        : and_expr
        | exclusive_or_expr '^' and_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

inclusive_or_expr
        : exclusive_or_expr
        | inclusive_or_expr '|' exclusive_or_expr  {
		$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

logical_and_expr
        : inclusive_or_expr
        | logical_and_expr AND_OP inclusive_or_expr {
		$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

logical_or_expr
        : logical_and_expr
        | logical_or_expr OR_OP logical_and_expr  {
		$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

conditional_expr
        : logical_or_expr
        | logical_or_expr '?' logical_or_expr ':' conditional_expr {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5); 
		$3->attrib |= ATTR_DFLOWA; $5->attrib |= ATTR_DFLOWA; 
	}
        ;

assignment_expr
        : conditional_expr
        | cast_expr assignment_operator assignment_expr {
	  $$ = NODE(TAG_NI,$1,$2,$3); $1->attrib |= ATTR_LVALUE; 
       	  if($1->tag == TAG_SYMREF && strcmp($2->text,"=")) { 
            /* Desugaring "C" tricks */
            node *rvalue = $1->clone(); rvalue->attrib &= ~ATTR_LVALUE; 
	    $3->cl = dprintf("%s(",$3->cl); 
	    $3->cr = dprintf(")%s",$3->cr); 
            char *op = strdup($2->text); $2->text = "="; *strstr(op,"=") = 0; 
            $$ = NODE(TAG_NI,$1,$2,rvalue,NODE_OP(TAG_NI,op),$3); 
	  }
          } 
        ;

assignment_operator
        : '='           
        | MUL_ASSIGN /* Oh, my God! Ok, I will desugar it for T-system ;-) */
        | DIV_ASSIGN
        | MOD_ASSIGN
        | ADD_ASSIGN
        | SUB_ASSIGN
        | LEFT_ASSIGN
        | RIGHT_ASSIGN
        | AND_ASSIGN
        | XOR_ASSIGN
        | OR_ASSIGN
        ;

expr
        : assignment_expr
        | expr ',' assignment_expr {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

constant_expr
        : conditional_expr
        ;

declaration
        : declaration_specifiers ';' {$$ = NODE(TAG_DECL,$1,$2);}
        | declaration_specifiers init_declarator_list ';' {
		$$ = NODE(TAG_DECL,$1,$2,$3);}
        ;



/*nested_function
	: declarator compound_statement {$$ = NODE(TAG_FUNDEF,$1,$2);
		$$->attrib |= ATTR_DFLOWA; }
	| declarator declaration_list compound_statement {    
		$$ = NODE(TAG_FUNDEF,$1,$2,$3); $$->attrib |= ATTR_DFLOWA; }
        ; */

declaration_specifiers
        : storage_class_specifier
        | storage_class_specifier declaration_specifiers {
		$$ = NODE(TAG_NI,$1,$2);
	}
        | type_specifier 
        | type_specifier declaration_specifiers  {$$ = NODE(TAG_NI,$1,$2);}
        ;

init_declarator_list
        : init_declarator
        | init_declarator_list ',' init_declarator  {
		$$ = NODE(TAG_NI,$1,$2,$3);
        }
        ;

init_declarator
        : declarator
        | declarator '=' initializer   {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

storage_class_specifier
        : TYPEDEF  
        | EXTERN
        | STATIC
        | AUTO
        | REGISTER
        ;
 
type_specifier
        : CHAR
        | SHORT
        | INT
        | LONG
        | SIGNED
        | UNSIGNED
        | FLOAT
        | DOUBLE
        | CONST                      { $$ = $1; $$->tag = TAG_CONST; }
        | VOLATILE
        | VOID
	| TVAL                       { $$ = $1; $$->tag = TAG_TVAL; }
	| TFUN                       { $$ = $1; $$->tag = TAG_TFUN; }
	| TOUT                       { $$ = $1; $$->tag = TAG_TOUT; }
	| TPTR                       { $$ = $1; $$->tag = TAG_TPTR; }
	| TMEMO                      { $$ = $1; $$->tag = TAG_TMEMO; }
	| INLINE
	| COMPLEX
        | struct_or_union_specifier
        | enum_specifier
        | TYPE_NAME
	| typeof_specifier
        ;

struct_or_union_specifier
        : struct_or_union identifier '{' struct_declaration_list '}'  {
		$$ = NODE(TAG_SCOPE,$1,$2,$3,$4,$5);
        }
        | struct_or_union '{' struct_declaration_list '}' {
		$$ = NODE(TAG_SCOPE,$1,$2,$3,$4);
        }
        | struct_or_union identifier {$$ = NODE(TAG_NI,$1,$2);}
        ;

struct_or_union
        : STRUCT
        | UNION
        ;

struct_declaration_list
        : struct_declaration    { $$ = NODE(TAG_NI,$1); }
        | struct_declaration_list struct_declaration {$$ = NODE(TAG_NI,$1,$2);}
        ;

struct_declaration
        : type_specifier_list struct_declarator_list ';' {
		$$ = NODE(TAG_DECL,$1,$2,$3);
        }
        ;

struct_declarator_list
        : struct_declarator
        | struct_declarator_list ',' struct_declarator {
		$$ = NODE(TAG_NI,$1,$2,$3);
        }
        ;

struct_declarator
        : declarator         
        | ':' constant_expr  {$$ = NODE(TAG_NI,$1,$2);}
        | declarator ':' constant_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

enum_specifier
        : ENUM '{' enumerator_list '}'  {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        | ENUM identifier '{' enumerator_list '}' {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5);
        }
        | ENUM identifier {$$ = NODE(TAG_NI,$1,$2);}
        ;

enumerator_list
        : enumerator
        | enumerator_list ',' enumerator  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

enumerator
        : identifier
        | identifier '=' constant_expr  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

typeof_specifier
	: TYPEOF '(' expr ')'  { $$ = NODE(TAG_NI,$1,$2,$3,$4);}
	| TYPEOF '(' type_name ')' { $$ = NODE(TAG_NI,$1,$2,$3,$4);}
	;

declarator
        : declarator2           {$$ = NODE(TAG_NI,$1);}
        | pointer declarator2   {$$ = NODE(TAG_NI,$1,$2);}
        ;

declarator2
        : identifier            {$$ = $1; $$->tag = TAG_SYMDEF; }
        | '(' declarator ')'    {$$ = NODE(TAG_NI,$1,$2,$3);}
        | declarator2 '[' ']'   {$$ = NODE(TAG_NI,$1,$2,$3);}
        | declarator2 '[' constant_expr ']'  {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        | declarator2 '(' ')'   {$$ = NODE(TAG_NI,$1,$2,$3);}
        | declarator2 '(' parameter_type_list ')'  {
		$$ = NODE(TAG_NI,$1,$2,$3,$4);
        }
        | declarator2 '(' parameter_identifier_list ')' {
                $$ = NODE(TAG_NI,$1,$2,$3,$4);
        }
        ;

pointer
        : '*'
        | '*' type_specifier_list   {$$ = NODE(TAG_NI,$1,$2);}
        | '*' pointer               {$$ = NODE(TAG_NI,$1,$2);}
        | '*' type_specifier_list pointer  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

type_specifier_list
        : type_specifier                     {$$ = NODE(TAG_NI,$1);}
        | type_specifier_list type_specifier {$$ = NODE(TAG_NI,$1,$2);} 
        ;

parameter_identifier_list
        : identifier_list
        | identifier_list ',' ELIPSIS  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

identifier_list
        : identifier                      {$$ = $1; $$->tag = TAG_SYMDEF; }
        | identifier_list ',' identifier  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

parameter_type_list
        : parameter_list              {$$ = NODE(TAG_FPAR,$1);}         
        | parameter_list ',' ELIPSIS  {$$ = NODE(TAG_FPAR,$1,$2,$3);}
        ;

parameter_list 
        : parameter_declaration       
        | parameter_list ',' parameter_declaration  {
		$$ = NODE(TAG_PARLIST,$1,$2,$3);
        }
        ;

parameter_declaration
        : type_specifier_list declarator {$$ = NODE(TAG_PARDECL,$1,$2);}
        | type_name                      {$$ = NODE(TAG_PARDECL,$1);}
        ;

type_name
        : type_specifier_list
        | type_specifier_list abstract_declarator {$$ = NODE(TAG_NI,$1,$2);}
        ;

abstract_declarator
        : pointer
        | abstract_declarator2
        | pointer abstract_declarator2  {$$ = NODE(TAG_NI,$1,$2);}
        ;

abstract_declarator2
        : '(' abstract_declarator ')'  {$$ = NODE(TAG_NI,$1,$2,$3);}
        | '[' ']'                      {$$ = NODE(TAG_NI,$1,$2);}
        | '[' constant_expr ']'        {$$ = NODE(TAG_NI,$1,$2,$3);}
        | abstract_declarator2 '[' ']' {$$ = NODE(TAG_NI,$1,$2,$3);}
        | abstract_declarator2 '[' constant_expr ']' {
		$$ = NODE(TAG_NI,$1,$2,$3,$4);
        }
        | '(' ')'                      {$$ = NODE(TAG_NI,$1,$2);}
        | '(' parameter_type_list ')'  {$$ = NODE(TAG_NI,$1,$2,$3);}
        | abstract_declarator2 '(' ')' {$$ = NODE(TAG_NI,$1,$2,$3);}
        | abstract_declarator2 '(' parameter_type_list ')' {
		$$ = NODE(TAG_NI,$1,$2,$3,$4);
        }
        ;

initializer
        : assignment_expr
        | '{' initializer_list '}'      {$$ = NODE(TAG_NI,$1,$2,$3);}
        | '{' initializer_list ',' '}'  {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        ;

initializer_list
        : initializer
        | initializer_list ',' initializer  {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

statement
        : labeled_statement    { 
		$$ = $1; 
                $$->attrib |= ATTR_STATEMENT|ATTR_DENTRY|ATTR_DFLOWA; 	
	}
        | compound_statement   {$$ = $1; $$->attrib |= ATTR_STATEMENT;}
        | expression_statement {$$ = $1; $$->attrib |= ATTR_STATEMENT};
        | selection_statement  {$$ = $1; $$->attrib |= ATTR_STATEMENT;}
        | iteration_statement  {$$ = $1; $$->attrib |= ATTR_STATEMENT;}
        | jump_statement       {$$ = $1; $$->attrib |= ATTR_STATEMENT;}
        ;

labeled_statement
        : identifier ':' statement         {$$ = NODE(TAG_NI,$1,$2,$3);}
        | CASE constant_expr ':' statement {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        | DEFAULT ':' statement            {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

compound_statement     
        : '{' '}'                     {$$ = NODE(TAG_SCOPE,$1,$2);} 
        | '{' statement_list '}'      {$$ = NODE(TAG_SCOPE,$1,$2,$3);}
        | '{' declaration_list '}'    {$$ = NODE(TAG_SCOPE,$1,$2,$3);}
	| '{' label_decls declaration_list '}'  {$$ = NODE(TAG_SCOPE,$1,$2,$3,$4);}
        | '{' declaration_list statement_list '}' {
		$$ = NODE(TAG_SCOPE,$1,$2,$3,$4); }
	| '{' label_decls declaration_list statement_list '}' {
		$$ = NODE(TAG_SCOPE,$1,$2,$3,$4,$5); }
        ;

declaration_list
        : declaration
        | declaration_list declaration {$$ = NODE(TAG_NI,$1,$2);}
        ;

statement_list
        : statement                 
        | statement_list statement  {$$ = NODE(TAG_NI,$1,$2);}
        ;

label_decls
	: label_decl
	| label_decls label_decl {$$ = NODE(TAG_NI,$1,$2);}
	;

label_decl
	: LABEL label_declarators ';' {$$ = NODE(TAG_NI,$1,$2,$3);}
	;

label_declarators
	: identifier
	| label_declarators identifier {$$ = NODE(TAG_NI,$1,$2);}
	;

expression_statement
        : ';' 
        | expr ';'                  {$$ = NODE(TAG_NI,$1,$2);}
        ;

selection_statement
        : IF '(' expr ')' statement  {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5); 
		$5->attrib |= ATTR_DFLOWA; 
        }
        | IF '(' expr ')' statement ELSE statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7); 
		$5->attrib |= ATTR_DFLOWA; $7->attrib |= ATTR_DFLOWA; 
	}
        | SWITCH '(' expr ')' statement  {$$ = NODE(TAG_NI,$1,$2,$3,$4,$5);}
        ;

iteration_statement
        : WHILE '(' expr ')' statement  {$$ = NODE(TAG_NI,$1,$2,$3,$4,$5);}
        | DO statement WHILE '(' expr ')' ';' {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7);
	}
        | FOR '(' ';' ';' ')' statement  {$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6);}
        | FOR '(' ';' ';' expr ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7);
        }
        | FOR '(' ';' expr ';' ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7);
	}
        | FOR '(' ';' expr ';' expr ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7,$8);
	}
        | FOR '(' expr ';' ';' ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7);
	}
        | FOR '(' expr ';' ';' expr ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7,$8);
	}
        | FOR '(' expr ';' expr ';' ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7,$8);
	}
        | FOR '(' expr ';' expr ';' expr ')' statement {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5,$6,$7,$8,$9);
	}
        ;

jump_statement
        : GOTO identifier ';' {$$ = NODE(TAG_NI,$1,$2,$3);}
        | GOTO '*' expr ';'   {$$ = NODE(TAG_NI,$1,$2,$3,$4);}
        | CONTINUE ';'        {$$ = NODE(TAG_NI,$1,$2);}
        | BREAK ';'           {$$ = NODE(TAG_NI,$1,$2);}
        | RETURN ';'          {$$ = NODE(TAG_NI,$1,$2);}
        | RETURN expr ';'     {$$ = NODE(TAG_NI,$1,$2,$3);}
        ;

file
        : external_definition       { node::root = $$ = $1; } 
        | file external_definition  { node::root = $$ = NODE(TAG_NI,$1,$2);} 
        ;

external_definition
        : function_definition 
        | declaration
	| tattributes 
        ;

tattributes
	: TCOMMENT_BEGIN type_specifier tattr_list ';' TCOMMENT_END {
		$$ = NODE(TAG_NI,$1,$2,$3,$4,$5); 
          }
	;

tattr_list
	: IDENTIFIER  { $$->tag = TAG_SYMPREDEF; }
	| tattr_list ',' IDENTIFIER  { 
		$3->tag = TAG_SYMPREDEF;
		$$ = NODE(TAG_NI,$1,$2,$3); 
	}
	;

function_definition
        : declarator function_body  {
		$$ = NODE(TAG_FUNDEF,$1,$2); $$->attrib |= ATTR_DFLOWA; 
	}
        | declaration_specifiers declarator function_body {
		$$ = NODE(TAG_FUNDEF,$1,$2,$3); $$->attrib |= ATTR_DFLOWA; 
	}
        ;

function_body
        : compound_statement
        | declaration_list compound_statement {$$ = NODE(TAG_NI,$1,$2);}
        ;

identifier
        : IDENTIFIER 
        ;

%%

#include <stdio.h>

extern int  column, line;
extern int  src_line;
extern char src_file[];
extern int  line_text_pos;
extern char line_text[];

int yyerror(char *errmsg)
{
	int bugcol = column - strlen(yytext);
	if (bugcol < 0) bugcol = column;

	fprintf(stderr, "%s:%d:%d:\n - %s:\n", 
		src_file, src_line, bugcol, errmsg);

	fprintf(stderr, "\n%.*s\n", line_text_pos, line_text);

	int i; for (i=0; i<bugcol; i++) fputc(' ', stderr);
	fprintf(stderr, "^\n");

	fprintf(stderr, "(input: line:%d, column:%d)\n", line, bugcol);
	exit(1);
	return 0;
}



