//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/tc2c.cc,v $
// $Revision: 1.9 $
// $Date: 2001/08/14 13:41:46 $

//#include <stl.h>
#if __GNUC__ >= 3
#include <ext/hash_map> // Hmm, SGI extension, however!
#else
#include <hash_map.h> // Hmm, SGI extension, however!
#endif
#include <cstdio>
#include <cstdlib>

#include "node.h"

node *node::root = NULL;
scope *topscope = new scope(NULL); /* Top-level bindings */

char *ttypes_fname;

bool do_profile() {
  static int prof = -1;
  return (prof < 0) ? (prof = getenv("GR_PROFILE") ? 1:0) : prof;
}

bool do_profile_statements() {
  static int prof = -1;
  return (prof < 0) ? (prof = getenv("GR_PROFILE_STATEMENTS") ? 1:0) : prof;
}

int main (int argc, char *argv[]) {
  // Check argument presence
  if (argc != 3)
    error("usage: %s <ttypes_fname> <typextr_in>\n", argv[0]);

  // Store filename for generated types
  ttypes_fname = argv[1];

  // Extract type information from 'full' CCP output
  FILE *typextr_in = fopen(argv[2],"r");
  if (!typextr_in)
    error("%s: unable open <typextr_in> file `%s'", argv[0], argv[2]);
  extern void typextr(FILE *in);
  typextr(typextr_in);
  fclose(typextr_in);

  // Parse 
  extern int yyparse(void);
  yyparse();

  // Generate preample
  printf("/* This code generated from TC source.  DO NOT EDIT ! */\n"
	 "/* For convensions and C++ interfaces see file `tc.h' */\n");
  printf("#ifndef TSYSTEM\n"
	 "#define TSYSTEM\n"
	 "#endif\n");

  // Do passes
  long attrib = 0;
  if (!node::root) error("There is no root node ");
  node::root->calc_attrib(topscope, attrib);
  node::root->gen_types();
  node::root->strictness();
  node::root->compile();
  if (do_profile()) node::root->profile_codegen();

  // Dump converted tree
  node::root->dump(stdout);
   
#if 1 
  // debug dump of tree 
  char *dbgdumpfile = new char[strlen (argv[2]) + 30];
  strcpy (dbgdumpfile, argv[2]);
  char *p = strchr (dbgdumpfile, '.');
  if (!p)
    p = strchr (dbgdumpfile, '\0');
  strcpy(p, ".dbgdump");
//  FILE *fp = fopen(dbgdumpfile, "w+");
//  node::root->debugdump(fp);
//  fclose(fp);
#endif 

  printf("\n");
  return 0;
}
