//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/langs/tcc/typextra.cc,v $
// $Revision: 1.6 $
// $Date: 2001/12/18 12:39:09 $

// User-defined type extractor

#include <assert.h>
#include <ctype.h>
#include "node.h"

void typextr(char *in) {
  /* Token lookup state */
  static char *typedef_token = "typedef";
  static char *tokenp = typedef_token;
  /* Bracket level state */
  static int level = 0;
  /* Typedef name state */
  static char name[4096];
  static char *namep = name;
  static bool name_closed = true;
  /* Previous name inside typedef body */
  static char *prev_name = "";
  /* Finite state machine loop */
  int  c;
  while ((c = *in++)) {
    if (*tokenp) { 
      // We are still looking for typedef     
      if (c == *tokenp) {
	tokenp++;
      } else {
	tokenp = typedef_token;
      }
    } else { 
      switch(c) {
      case '{': 
	// Bracket open
	prev_name = ""; // Reset "previous name"
	//fprintf(stderr,"RESET\n");
	level++;
	break;
      case '}':
	// Bracket closed
	level--;
	break;
      default_char:
      default:
	if (level == 0) {
	  if (isalnum(c) || c == '_') {
	    if (name_closed) {
	      namep = name;
	      name_closed = false;
	    } 
	    *namep++ = c;
	    assert((unsigned)(namep-name) < sizeof(name));
	  } else {
	    *namep++ = 0;
	    char *copy_name = strdup(name);
	    // Intern only non-keywords and non-struct-or-union names
	    if (strcmp(copy_name,"struct") && strcmp(copy_name,"union") &&
		strcmp(prev_name,"struct") && strcmp(prev_name,"union")) {
	      if (!typenames[copy_name]) {
		typenames[copy_name] = copy_name; 
		//else
		//fprintf(stderr,"exception: %s (prev=%s)\n", copy_name, prev_name);
		// FIXME: why {1,6} are types ?!
		//fprintf(stderr, "typenames[%s] = %s\n", copy_name, typenames[copy_name]);
	      }
	    }
	    name_closed = true;
	    prev_name = copy_name;
	    // We are forcing to make all lparts as types
	  }
	}
	break;
	// May be a terminator
      case ';':
      case '=':	
      case '[':	
	if (level == 0) {
	  // End of typedef construction;
	  // Reset state
	  tokenp = typedef_token;
	  prev_name = "";
	  // Now handle a name
	  goto default_char;
	}
	break;
      }
    }
  }
}

void typextr(FILE *in) 
{
  char buf[65536];
  bool stop=false;
  while(!stop) {
    stop = fread(buf,1,sizeof(buf),in) != sizeof(buf);
    typextr(buf);
  }
}

