//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_comm_layer.hh,v $
// $Revision: 1.11 $
// $Date: 2001/06/19 07:57:26 $

#ifndef __gr_comm_layer_hh__
#define __gr_comm_layer_hh__

#include "gr_message.hh"
#include "gr_shared_allocator.hh"

namespace grace
{

//------------------------------------------------------------------------------
//! Communication layer interface
//------------------------------------------------------------------------------

class CommLayer
{

public:
  //
  //! Constructor
  inline CommLayer () {}
  //
  //! Destructor
  inline virtual ~CommLayer () {}
  //
  //! Initialization function
  virtual void init (int& _argc, char**& _argv) = 0 ;
  //
  //! Initialization check
  virtual bool is_initialized () const = 0 ;
  //
  //! Returns a number of nodes in communication space
  inline virtual unsigned get_size () const = 0 ;
  //
  //! Returns an ID of our communication space
  inline virtual CompSpaceID get_id () const = 0 ;
  //
  //! Returns an ID of main communication space
  inline virtual CompSpaceID get_main_id () const = 0 ;
  //
  //! Send a message to another communication space
  inline virtual int send (Message const* _msg, CompSpaceID _dst) const = 0 ;
  //
  //! Send a message to all communication spaces
  inline virtual int broadcast (Message const* _msg) const = 0 ;
  //
  //! Check for a message arrived
  inline virtual int poll () const = 0 ;
  //
  //! Signal a communication layer to close all communications and exit a loop
  inline virtual void finish () const = 0 ;
  //
  //! A loop for message processing (may be empty)
  inline virtual void loop () const = 0 ;
  //
  //! Overloaded operator new
  OVERLOAD_NEW
  //
  //! Overloaded operator delete
  OVERLOAD_DELETE

};

//
//! Pointer to communication layer object
extern CommLayer* comm_layer ;

extern "C" {
//
//! Static method creating specific communication layer (should be defined
//  in appropriate communication layer implementation
void gr_comm_layer_create () ;
//
//! Static method destroying specific communication layer (should be defined
//  in appropriate communication layer implementation
void gr_comm_layer_destroy () ;
//
//! Static method running communication loop
void gr_comm_layer_loop () ;
}

}

#endif // __gr_comm_layer_hh__
