//
// Copyright (C) 2000 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_common.hh,v $
// $Revision: 1.25 $
// $Date: 2002/01/25 12:37:29 $

#ifndef __gr_common_hh__
#define __gr_common_hh__

#include "gr_options.hh"
#include "gr_tprint.hh"
#include "pxx_common.hh"
#include "pxx_spinlock.hh"
#include <setjmp.h>

#if DEBUG
#define dmesg(a...) { \
  if (dmesg_lock) synchronize(pxx::Spinlock, *dmesg_lock); \
  if(comm_layer) \
    tmesg("%u: ", comm_layer->get_id()); \
  tmesg(a); \
}
#define dmethod dmesg("%p: %s\n", this, __PRETTY_FUNCTION__)
#else
#define dmesg(a...)
#define dmethod
#endif // DEBUG

#if VERBOSE
#define vmesg(a...) if (options.verbose) { \
  if (dmesg_lock) synchronize(pxx::Spinlock, *dmesg_lock); \
  if(comm_layer && comm_layer->is_initialized()) \
    tmesg("[%u] ", comm_layer->get_id()); \
  tmesg(a); \
}
#else
#define vmesg(a...)
#endif // VERBOSE

#define emesg(a...) { \
  if (dmesg_lock) synchronize(pxx::Spinlock, *dmesg_lock); \
  if(comm_layer) \
    tmesg("[%u] ", comm_layer->get_id()); \
  tmesg(a); \
}

#define twarn(a...) { \
  if (dmesg_lock) synchronize(pxx::Spinlock, *dmesg_lock); \
  if(comm_layer) \
    tmesg("[%u] ", comm_layer->get_id()); \
  tmesg("*** GRACE warning: "); \
  tmesg(a); \
}

#if defined (__i386__)
#define PEIP __jmpbuf[JB_PC]
#define PESP __jmpbuf[JB_SP]
#define PEBP __jmpbuf[JB_BP]
#elif defined (__alpha__)
#define PEIP __jmpbuf[JB_PC]
#define PESP __jmpbuf[JB_SP]
#define PEBP __jmpbuf[JB_FP]
#endif

#if PARANOIA

#define OVERLOAD_NEW \
  void* operator new (size_t _size) \
  { \
    extern SharedAllocator* shared_allocator ; \
    void* p = shared_allocator->allocate(_size); \
    if (p == null) FATAL("Memory allocation failed"); \
    return p; \
  } \
\
  void* operator new[] (size_t _size) \
  { \
    extern SharedAllocator* shared_allocator ; \
    void* p = shared_allocator->allocate(_size); \
    if (p == null) FATAL("Memory allocation failed"); \
    return p; \
  } \
\
  void* operator new (size_t, void* _ptr) \
  { \
    return _ptr; \
  }\

#else

#define OVERLOAD_NEW \
  void* operator new (size_t _size) \
  { \
    extern SharedAllocator* shared_allocator ; \
    return shared_allocator->allocate(_size); \
  } \
\
  void* operator new[] (size_t _size) \
  { \
    extern SharedAllocator* shared_allocator ; \
    return shared_allocator->allocate(_size); \
  } \
\
  void* operator new (size_t, void* _ptr) \
  { \
    return _ptr; \
  }

#endif // PARANOIA

#define OVERLOAD_DELETE \
  inline void operator delete (void* _ptr) \
  { \
    extern SharedAllocator* shared_allocator ; \
    shared_allocator->deallocate(_ptr); \
  } \
\
  inline void operator delete[] (void* _ptr) \
  { \
    extern SharedAllocator* shared_allocator ; \
    shared_allocator->deallocate(_ptr); \
  }

//
// FIXME: should be actual memory barrier
#define BARRIER
extern "C"
void __libc_longjmp (sigjmp_buf env, int val) __attribute__ ((noreturn)) ;
extern "C" ssize_t __libc_write(int fd, const void *buf, size_t count) ;

namespace grace
{

extern pxx::Spinlock* global_lock ;
extern pxx::Spinlock* dmesg_lock ;

}

#endif // __gr_common_hh__
