//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_comp_resource.hh,v $
// $Revision: 1.21 $
// $Date: 2002/01/29 17:05:59 $

#ifndef __gr_comp_resource_hh__
#define __gr_comp_resource_hh__

#include "gr_message.hh"
#include "gr_core.hh"
#include "gr_call_info.hh"
#include "pxx_sync.hh"


namespace grace
{

#define GR_LOAD_LOW 2
#define GR_LOAD_HIGH 4
#define MAX_NODE 512 // FIXME
#define CPU_MFLOPS 667

//! todo implement this and resource waiting
class ResourceRequest {
};

/* A [sub]cluster power integrable value */

class ComputationResource : public Message {

protected:

  int free_mflops; /* Megaflops free at this time */
  bool valid;    /* Statistic up-to-date flag */

  //! Get index for any given ComputationResource
  unsigned get_index() ;

  static ComputationResource *root() ;

  ComputationResource *up() ;

  ComputationResource *left() ;

  ComputationResource *right() ;

  bool leafp() ;

  void merge() ;

  void recalc() ;

  int weight(ResourceRequest*) ;

  void invalidate() ;

  void handle() ;

  ComputationResource *find_best_rec(ResourceRequest *rq, unsigned _depth) ;

  static ComputationResource *find_best_free(ResourceRequest *rq) ;

  static void indent(int n) ;

  void show_res_tree(int level = 1) ;

  Message* serialize()   { return this; }
  Message* deserialize() { return this; }
  

public:

  ComputationResource() ;

  static void notify_one_worker_busy(int busy) ;

  //! Returns cluster node number either -1 
  //! Should be called after prenatal is taken from queue
  static int allocate_node_for_external_computing(CallInfo const&) ;

};


struct ComputationResourceData {

  ComputationResource resources[2*MAX_NODE-1];
  pxx::Spinlock comp_res_lock;
  int local_free_mflops;
  unsigned nodes_computed ;

  ComputationResourceData() :
    local_free_mflops(0),
    nodes_computed(0)
  {}

  OVERLOAD_NEW
  OVERLOAD_DELETE

};



}

#endif // __gr_comp_resource_hh__
