//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_computation_state.hh,v $
// $Revision: 1.14 $
// $Date: 2002/01/29 17:05:59 $

#ifndef __gr_computation_state_hh__
#define __gr_computation_state_hh__

#include "gr_common.hh"
#include "gr_shared_allocator.hh"
#include "gr_computation.hh"

#include <stdlib.h>

namespace grace
{

//
// forward reference to node class
class Node ;

template <ComputationType> class ComputationState ;

class GenericComputationState
{

public:

  //
  //! Constructor
  inline GenericComputationState () ;
  //
  //! Destructor
  inline virtual ~GenericComputationState () ;
  //
  //! Save state when process is switched off
  inline virtual void save () = 0 ;
  //
  //! Restore state when process is switched on
  inline virtual void restore () = 0 ;
  //
  //! Compute a node
  inline virtual void compute (Node* _node) = 0 ;
  //
  //! Wait for further computation
  inline virtual void wait () = 0;
  //
  //! Overloaded operator new
  OVERLOAD_NEW
  //
  //! Overloaded operator delete
  OVERLOAD_DELETE

};

template <ComputationType> class ComputationStateFactory ;

//
//! A factory interface for creating different computation states

class GenericComputationStateFactory
{

  //
  //! Computation type
  enum ComputationType type ;

public:

  //
  //! Constructor
  inline GenericComputationStateFactory (ComputationType _type) ;
  //
  //! Destructor
  inline virtual ~GenericComputationStateFactory () {}
  //
  //! Virtual method creating a computation state
  inline virtual GenericComputationState* create (Node* _node) = 0 ;
  //
  //!
  inline ComputationType get_type () const ;
  //
  //! Overloaded operator new
  OVERLOAD_NEW
  //
  //! Overloaded operator delete
  OVERLOAD_DELETE

};

}

#endif // __gr_computation_state_hh__

