//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_control_message.hh,v $
// $Revision: 1.12 $
// $Date: 2002/02/13 15:39:44 $

#ifndef __gr_control_message_hh__
#define __gr_control_message_hh__

#include "gr_message.hh"
#include "gr_comm_layer.hh"

#include <string.h>

namespace grace
{

enum ControlMessageType {
  GR_MSG_EXIT,
  GR_MSG_ABORT,
  GR_MSG_PRINT
};

template <ControlMessageType type_t>
class ControlMessage;

template <>
class ControlMessage<GR_MSG_EXIT> : public Message
{

public:

  virtual Message* serialize () { return this; }
  virtual Message* deserialize () { return this; }

  virtual void handle () {
    vmesg("%u: Got ControlMessage<GR_MSG_EXIT>\n", comm_layer->get_id());
    comm_layer->finish();
    delete this;
  }

  ControlMessage() : Message(sizeof(self)) {}

};

template <>
class ControlMessage<GR_MSG_ABORT> : public Message
{

public:

  virtual Message* serialize () { return this; }
  virtual Message* deserialize () { return this; }

  virtual void handle () {
    extern void grace::abort();
    vmesg("%u: Got ControlMessage<GR_MSG_ABORT>\n", comm_layer->get_id());
    abort();
    delete this;
  }

  ControlMessage() : Message(sizeof(self)) {}

};

template <>
class ControlMessage<GR_MSG_PRINT> : public Message
{
private:
  int fd ;
  char buf[1] ;

public:

  virtual Message* serialize () { return this; }
  virtual Message* deserialize () { return this; }

  virtual void handle () {
    fprintf(fd == 1 ? stderr : stdout , "%s", buf);
    //delete this;
  }

  ControlMessage(int _fd, char const* _msg, size_t _size) :
    Message(_size + sizeof(*this)-1, GR_MSG_SYSTEM),
    fd (_fd)
  {
    memcpy(buf, _msg, _size);
  }

};

}

#endif //__gr_control_message_hh__
