//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_core.hh,v $
// $Revision: 1.40 $
// $Date: 2002/02/04 17:29:58 $

#ifndef __gr_core_hh__
#define __gr_core_hh__

#include "gr_common.hh"
#include "gr_shared_allocator.hh"
#include "gr_block_allocator.hh"
#include "gr_shared_chunk_allocator.hh"
#include "gr_continuation.hh"
#include "gr_worker.hh"
#include "gr_node.hh"
#include "gr_comm_layer.hh"
#include "gr_shared_pipe.hh"

#include "pxx_ring.hh"

#include <new>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <signal.h>

namespace grace
{

int atexit (void (*_f)(void)) ;

static void* const shared_heap_start = (void*)0x60000000 ;
static size_t const shared_heap_size = 1U << 29 ;
static void* const block_heap_chunk_mem_start = (void*)0x50000000 ;
static size_t const block_heap_chunk_mem_size = 1U << 26;
static char const* const block_heap_chunk_mem_filename =
  "/tmp/grace_block_heap_chunk_mem.mem" ;
static void* const block_heap_start = (void*)0x80000000 ;
static size_t const block_heap_size = 1U << 29 ;
extern char initial_allocator_buf[];
//extern pxx::HeapAllocator* initial_allocator ;
extern SharedAllocator* shared_allocator ;
extern SharedChunkAllocator* ring_node_allocator ;
extern SharedChunkAllocator* continuation_allocator ;
extern SharedChunkAllocator* node_allocator ;
extern BlockAllocator* block_allocator ;
extern unsigned cpu_number ;
extern Worker* workers ;

static void* const heap = (void*)0xA0000000 ;
static size_t const max_heap_size = 1U << 28 ;
extern size_t heap_size ;
static void* const stack = (void*)0xB0000000 ;
static size_t const max_stack_size = 1U << 27 ;
extern size_t stack_size ;
extern Node* current ;
extern pxx::Spinlock current_lock ;
extern Worker* current_worker ;
class ComputationResourceData ;
extern ComputationResourceData *comp_res_data ;
extern pid_t* comm_pid ;
extern pid_t* system_pid ;
extern pid_t* main_pid ;
extern Node* main_node ;
extern SharedPipe* sys_proc_pipe ;

extern pxx::Ring<ContinuationBase>* prenatal_queue ;
extern pxx::Ring<ContinuationBase>* upload_queue ;
extern pxx::Ring<ContinuationBase>* local_queue ;
extern pxx::Ring<Node>* calculator ;
extern pxx::Ring<Node>* upload_calculator ;
extern pxx::Ring<Node>* collector ;

extern int status ;

extern void terminate (int _error_code) ;
extern void abort () ;
extern void finish () ;

struct Core
{

  //
  //! Number of arguments
  int& argc ;
  //
  //! Array of arguments
  char**& argv ;
  //
  //! Working directory name
  char* work_dir ;
  //
  //! A pointer to main memory allocator
  SharedAllocator* shared_allocator ;
  //
  //! A one-page file for mapping shared_allocator structure
  char const* shared_allocator_buf_filename ;
  //
  //! A file descriptor for shared_allocator_buf_filename
  int shared_allocator_buf_fd ;
  //
  //! A file for mapping a shared heap
  char const* shared_heap_filename ;
  //
  //! A file descriptor for shared heap
  int shared_heap_fd ;
  //
  //! A file for mapping a block memory for stacks allocation
  char const* block_heap_filename ;
  //
  //! A file descriptor for block memory
  int block_heap_fd ;
  int block_heap_chunk_mem_fd ;

  Core (int& _argc, char**& _argv) ;

  ~Core () ;

  void run () ;

};

#ifdef MEM_DEBUG

struct MemCounter
{
  unsigned new_count ;
  unsigned delete_count ;
  char const* label ;
  inline MemCounter (char const* _label) ;
  inline ~MemCounter () ;
  inline MemCounter& operator++ () ;
  inline MemCounter& operator-- () ;
  inline MemCounter& operator++ (int) ;
  inline MemCounter& operator-- (int) ;
  OVERLOAD_NEW
  OVERLOAD_DELETE
};

inline MemCounter::MemCounter (char const* _label) :
  new_count(0),
  delete_count(0),
  label (_label)
{}
inline MemCounter::~MemCounter ()
{
  vmesg("%s: %u allocations\n", label, new_count);
  vmesg("%s: %u deallocations\n", label, delete_count);
}
inline MemCounter& MemCounter::operator++ ()
{
  new_count++;
  return self;
}
inline MemCounter& MemCounter::operator-- ()
{
  delete_count++;
  return self;
}
inline MemCounter& MemCounter::operator++ (int)
{
  new_count++;
  return self;
}
inline MemCounter& MemCounter::operator-- (int)
{
  delete_count++;
  return self;
}

extern MemCounter* continuation_mem_count ;
extern MemCounter* node_mem_count ;

#define COUNT_NEW(a) (*a ## _mem_count)++
#define COUNT_DELETE(a) (*a ## _mem_count)--

#else

#define COUNT_NEW(a)
#define COUNT_DELETE(a)

#endif // MEM_DEBUG

}

#endif // __gr_core_hh__
