#include "gr_core.hh"
#include "gr_worker.ih"
#include "gr_node.ih"
#include "gr_computation_state.ih"
#include "gr_fast_process.ih"
#include "gr_continuation.ih"
#include "gr_tval.ih"
#include "gr_tres.ih"

#include <cstdio>
#include <cstdlib>
#include <math.h>

//---------------------------------------------------------------------------

namespace grace
{

typedef unsigned long long  u64;
typedef long long s64;
typedef double d64;

struct pack_prm {
  u64 rec_count;
  u64 start;
  u64 size;
};

struct pack_res {
  u64 sum[10];
  d64 Sx;
  d64 Sy;
};

struct __f__args
{
  T_Val<pack_prm> parm ;
};

struct __f__ress
{
  T_Val<pack_res> res ;

public:

  __f__ress (ContinuationBase* _cont) :
    res (_cont)
};


#define  FALSE 0
#define  TRUE 1

#define CONST_1 ((u64)1<<46)

#ifdef USE_MACROS

#define max(a,b)  ((a) > (b) ? (a) : (b))
#define lfabs(a)  ((a) < 0 ? (a)*(d64)(-1) : a)
#define get_ab_mod_2_46(a,b)  ((((a) % CONST_1)*((b) % CONST_1)) % CONST_1)

#else

d64 lfabs(d64 a){
  return a < 0 ? -a : a ;
}

u64 get_ab_mod_2_46(u64  a, u64  b ){
  return ((a % CONST_1)*(b % CONST_1)) % CONST_1;
}

#endif

u64 fast_raise(u64 a, u64 x){
  u64 res=1;
  while( x > 0 ){
    if ( x & (u64)1) res=get_ab_mod_2_46(res,a);
    a=get_ab_mod_2_46(a,a);
    x >>= 1;
  }
  return res;
}

void make_gist(u64 k,u64 packet_len, u64 a, u64 s, pack_res *res){
u64 count=0;
u64 tmp1,tmp2,x0=fast_raise(a,(u64)2*k);
  tmp2=get_ab_mod_2_46(x0,s); //x_0
  for(; count < (u64)packet_len; count++){
    d64 x,y,t;
    tmp1=get_ab_mod_2_46(a,tmp2);          //x_1
    tmp2=get_ab_mod_2_46(a,tmp1);          //y_1
    x=(d64)((s64)(tmp1<<1) - (s64)CONST_1)/(d64)CONST_1;
    y=(d64)((s64)(tmp2<<1) - (s64)CONST_1)/(d64)CONST_1;
    t=x*x+y*y;
    if(t<=1.0){
      t=sqrt(-2.0*log(t)/t);
      x*=t;
      y*=t;
      res->sum[(int)max(lfabs(x),lfabs(y))]++;
      res->Sx+=x;
      res->Sy+=y;
    }
  }
}

void iterator (__f__args const* _args, __f__ress* _ress)
{

  int i;

  pack_prm const& parm = _args->parm;
  pack_res res;

  if(parm.rec_count == 0 ){
    memset(&res,0,sizeof(pack_res));
    make_gist(parm.start,parm.size,(u64)pow(5,13),271828183,&res);
  } else {
    pack_prm pt1,pt2;
    T_Val<pack_res> res1,res2;
    pt1.rec_count=pt2.rec_count=parm.rec_count-1;
    pt1.start=parm.start;
    pt2.start=parm.start+parm.size/2;
    pt1.size=pt2.size=parm.size/2;

    Continuation<__f__args, __f__ress>* cont1 =
      new Continuation<__f__args, __f__ress>(
        iterator, new ComputationStateFactory<FastProcess>
      );
    cont1->argsp->parm = pt1;
    res1 = cont1->ressp->res;
    prenatal_queue->put_front((GenericContinuation*)cont1);

    Continuation<__f__args, __f__ress>* cont2 =
      new Continuation<__f__args, __f__ress>(
        iterator, new ComputationStateFactory<FastProcess>
      );
    cont2->argsp->parm = pt2;
    res2 = cont2->ressp->res;
    prenatal_queue->put_front((GenericContinuation*)cont2);

    pack_res const& r1 = res1;
    pack_res const& r2 = res2;

    for(i=0;i<10;i++) res.sum[i] = r1.sum[i] + r2.sum[i];
    res.Sx = r1.Sx + r2.Sx;
    res.Sy = r1.Sy + r2.Sy;
  }
  _ress->res = res;
}

}

using namespace grace;

int tmain (int argc, char* argv[])
{
  int i;
  T_Val<pack_res> res;
  pack_prm prm = { (u64)10, (u64)0, (u64)1<<24 };

  printf("v20 sizeof(u64)=%d sizeof(d64)=%d\nProblem size 0x%Lx\n",
         sizeof(u64),sizeof(d64),((pack_prm const&)prm).size);

  Continuation<__f__args, __f__ress>* cont =
    new Continuation<__f__args, __f__ress>(
      iterator, new ComputationStateFactory<FastProcess>
    );
  cont->argsp->parm = prm;
  res = cont->ressp->res;
  prenatal_queue->put_front((GenericContinuation*)cont);

  for(i=0;i<10;i++){
    printf("Q[%d]=%Lu\n",i,((pack_res const&)res).sum[i]);
  }
  printf("Sx=%le\nSy=%le\n",
         ((pack_res const&)res).Sx,((pack_res const&)res).Sy);
  return 0;
}
