//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_fast_process.hh,v $
// $Revision: 1.21 $
// $Date: 2002/01/25 12:37:29 $

#ifndef __gr_fast_process_hh__
#define __gr_fast_process_hh__

#include "gr_common.hh"
#include "gr_block_allocator.hh"
#include "gr_computation_state.hh"
#include "gr_core.hh"

namespace grace
{

extern size_t const fast_process_default_heap_size ;

//
//! Computation state for fast process

template <>
class ComputationState<FastProcess> :
  public GenericComputationState
{

private:

  //
  //! Handler of process heap
  BlockAllocator::BlockTreeItem* heap ;
  //
  //! Handler of process stack
  BlockAllocator::BlockTreeItem* stack ;

public:

  //
  //! Constructor
  inline ComputationState (
    Node* _node,
    size_t _heapsize = fast_process_default_heap_size,
    size_t _stksize = options.stack_size * pxx::page_size
  ) ;
  //
  //! Destructor
  inline ~ComputationState () ;
  //
  //! Save state
  inline void save () {}
  //
  //! Restore state
  inline void restore () {}
  //
  //! Compute a node
  void compute (Node* _node) ;
  //
  //! Wait for further computation
  void wait () ;

private:

  //
  //! Get stack address
  inline void* get_stk () ;
  //
  //! A trampoline to compute function on a switched stack
  static void trampoline (Node* _node) ;

};

//
//! A computation state factory for fast process

template <>
class ComputationStateFactory<FastProcess> :
  public GenericComputationStateFactory
{

private:

  //
  //! Heap size of a process
  size_t heap_size ;
  //
  //! Stack size of a process
  size_t stack_size ;

public:

  //
  //! Constructor
  ComputationStateFactory (
    size_t _heapsize = fast_process_default_heap_size,
    size_t _stksize = options.stack_size * pxx::page_size
  ) :
    GenericComputationStateFactory (FastProcess),
    heap_size (_heapsize),
    stack_size (_stksize)
  {}
  //
  //! Destructor
  ~ComputationStateFactory () {}
  //
  //! Create a computation state
  inline ComputationState<FastProcess>* create (Node* _node) ;

};

}

#endif // __gr_fast_process_hh__
