//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_main_process.hh,v $
// $Revision: 1.8 $
// $Date: 2001/01/23 01:26:49 $

#ifndef __gr_main_process_hh__
#define __gr_main_process_hh__

#include "gr_common.hh"
#include "gr_block_allocator.hh"
#include "gr_computation_state.hh"
#include "gr_shared_pipe.hh"

namespace grace
{

//
//! Computation state for main process

template <>
class ComputationState<MainProcess> :
  public GenericComputationState
{

private:

  //
  //! Handler of process heap
  BlockAllocator::BlockTreeItem* heap ;
  //
  //! Handler of process stack
  BlockAllocator::BlockTreeItem* stack ;
  //
  //! A PID of main process
  pid_t pid ;
  //
  //! Pipe file descriptors for signalling pipe
  SharedPipe* sig_pipe;

public:

  //
  //! Constructor
  inline ComputationState () ;
  //
  //! Destructor
  inline ~ComputationState () ;
  //
  //! Compute a node
  inline void compute (Node*) ;
  //
  //! Wait while node will be ready for further computation
  inline void wait () ;
  //
  //! Save computation state
  inline void save () {}
  //
  //! Restore computation state
  inline void restore () {}
  //
  //! Get stack address
  inline void* get_stk () ;

};

}

#endif // __gr_main_process_hh__
