//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_message.hh,v $
// $Revision: 1.13 $
// $Date: 2002/01/29 17:05:59 $

#ifndef __gr_message_hh__
#define __gr_message_hh__

#include "gr_common.hh"
#include "gr_shared_allocator.hh"
#include "gr_computation_space.hh"

namespace grace
{

enum MessageTag {
  GR_MSG_GENERIC = 0,
  GR_MSG_SYSTEM,
  GR_MSG_CONTINUATION,
  GR_MSG_AUTODELETE
};

class Message
{

friend class CommLayer ;

private:
  //
  //! Message size
  size_t size ;
  //
  //! Source computation space
  mutable CompSpaceID src ;

  //! Destination computation space
  mutable CompSpaceID dst ;

public:
  //
  //! A message tag
  MessageTag tag ;
  //
  //! Constructor
  inline Message (size_t _size, MessageTag _tag = GR_MSG_GENERIC) ;
  //
  //! Destructor
  inline virtual ~Message () {}
  //
  //! A method which serializes an object to a message
  inline virtual Message* serialize () ;
  //
  //! A method which deserializes a message to a real object
  inline virtual Message* deserialize () ;
  //
  //! Handle an active message
  inline virtual void handle () ;
  //
  //! Get message size
  inline size_t get_size () const;
  //
  //! Get message source
  inline CompSpaceID get_src () const;
  //
  //! Get message destination
  inline CompSpaceID get_dst () const;
  //
  //! Set message source
  inline void set_src (CompSpaceID _id) const;
  //
  //! Set message destination
  inline void set_dst (CompSpaceID _id) const;
  //
  //! Overloaded operator new
  OVERLOAD_NEW
  //
  //! Overloaded operator delete
  OVERLOAD_DELETE

};

}

#endif // __gr_message_hh__
