//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_mpi_comm_layer.hh,v $
// $Revision: 1.17 $
// $Date: 2002/01/25 12:37:29 $

#ifndef __gr_mpi_comm_layer_hh__
#define __gr_mpi_comm_layer_hh__

#include "gr_comm_layer.hh"
#include "gr_shared_pipe.hh"
#include "pxx_spinlock.hh"
#include "pxx_ring.hh"

//
//! this is a hack to prevent LAM from referring to C++ library
#ifndef MPIPP_H
#define MPIPP_H
#endif

#include <mpi.h>

namespace grace
{

//------------------------------------------------------------------------------
//! MPI-based communication layer
//------------------------------------------------------------------------------

class MPI_CommLayer :
  public CommLayer
{

private:
  //
  //! Signalling pipe
  SharedPipe* sig_pipe ;
  //
  //! \todo Use semaphore instead of lock/count pair
  //! A spinlock for signalling pipe
  pxx::Spinlock sig_lock ;
  //
  //! A number of items in signalling pipe
  mutable volatile int sig_count ;
  //
  //! Incoming message queue
  mutable pxx::Ring<Message> in_queue ;
  //
  //! Outgoing message queue
  mutable pxx::Ring<Message> out_queue ;
  //
  //! Current rank
  int rank ;
  //
  //! A size of communication space
  unsigned size ;
  //
  //! A buffer for system messages
  static char sys_buf[4096] ;
  //
  //!
  bool initialized ;

public:
  //
  //! Constructor
  MPI_CommLayer () ;
  //
  //! Destructor
  ~MPI_CommLayer () ;
  //
  //! Initialization function
  void init (int& _argc, char**& _argv) ;
  //
  //! Initialization check
  bool is_initialized () const ;
  //
  //! Returns a number of nodes in communication space
  inline unsigned get_size () const ;
  //
  //! Returns an ID of our communication space
  inline CompSpaceID get_id () const ;
  //
  //! Returns an ID of main communication space
  inline CompSpaceID get_main_id () const ;
  //
  //! Send a message to another communication space
  inline int send (Message const*  _msg, CompSpaceID _dst) const ;
  //
  //! Send a message to all communication spaces
  inline int broadcast (Message const*  _msg) const ;
  //
  //! Check for a message arrived
  inline int poll () const ;
  //
  //! Signal a communication layer to close all communications and exit a loop
  void finish () const ;
  //
  //! Receive incoming messages
  void recv (int _tag) const ;
  //
  //! A loop for message processing (may be empty)
  void loop () const ;

};

}

#endif // __gr_mpi_comm_layer_hh__
