//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_node.hh,v $
// $Revision: 1.21 $
// $Date: 2001/08/24 10:25:50 $

#ifndef __gr_node_hh__
#define __gr_node_hh__

#include "pxx_ring.hh"
#include "gr_continuation.hh"
#include "gr_profile.hh"

#include <setjmp.h>

namespace grace
{

class GenericComputationState ;

//! A computational node

class Node :
  public CallInfo
{

template <typename val_t> friend class T_Val ;
template <typename val_t> friend class T_Ptr ;
template <typename val_t> friend class ComputationState ;

private:

  //
  //! Back reference to continuation
  ContinuationBase* cont ;
  //
  //! Pointer to the state of T-process
  GenericComputationState* cstate ;
  //
  //! setjmp() buffer
  jmp_buf jbuf ;
  //
  //! A list of producers
  pxx::Ring<Node> in ;
  //
  //! A list of consumers
  pxx::Ring<Node> out ;
  //
  //! A flag indicating if a node is computed
  bool computed ;

public:

  //
  //! Default constructor (for main process pseudo node)
  inline Node () ;
  //
  //! Constructor
  inline Node (ContinuationBase* _cont) ;
  //
  //! Destructor
  inline ~Node () ;
  //
  //! Overloaded operator new
  inline void* operator new (size_t) ;
  //
  //! Overloaded operator delete
  inline void operator delete (void* _ptr) ;
  //
  //! Compute a node
  inline void compute () ;
  //
  //! Wait for a signal from another node
  inline void wait () ;
  //
  //! Check if a node has consumers
  inline bool has_consumer (Node*) ;
  //
  //! Tell a node to forget about consumer
  inline void forget (Node*) ;
  //
  //! Set a flag indicating that a node is computed
  inline void finish_computation () ;
  //
  //! Check if a node is computed
  inline bool is_computed () ;
  //
  //! Get a continuation referring to a node
  inline ContinuationBase* get_continuation () ;
  //
  //! Get computation state attached to a node
  inline GenericComputationState* get_cstate () const ;
  //
  //! Delete computation state attached to a node
  inline void delete_cstate () ;
  //
  //! Get node's setjmp() buffer
  inline jmp_buf& get_jmp_buf () ;
  //
  //! Signals to dependent nodes
  inline void signal () ;

  //! profile information
  ProfBlock* prof_block ;

};

}

#endif // __gr_node_hh__
