//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_shared_allocator.hh,v $
// $Revision: 1.5 $
// $Date: 2001/01/23 01:26:49 $

#ifndef __gr_shared_allocator_hh__
#define __gr_shared_allocator_hh__

#include "pxx_heap_allocator.hh"
#include "pxx_spinlock.hh"
#include "pxx_sync.hh"

namespace grace
{

class SharedAllocator :
  public pxx::HeapAllocator,
  public pxx::Spinlock
{

public :

  SharedAllocator (
    void* const _start,
    size_t const _heap_size,
    int _fd = 0,
    int _need_remap = false
  ) :
    pxx::HeapAllocator (_start, _heap_size, _fd, _need_remap)
  {}

  ~SharedAllocator ()
  {}

  inline void* allocate (size_t _size)
  {
    synchronize(Spinlock, self);
    return pxx::HeapAllocator::allocate (_size);
  }

  void deallocate (void* _ptr)
  {
    synchronize(Spinlock, self);
    return pxx::HeapAllocator::deallocate(_ptr);
  }

  void* reallocate (void* _ptr, size_t _size)
  {
    synchronize(Spinlock, self);
    return pxx::HeapAllocator::reallocate(_ptr, _size);
  }

};

}

#endif // __gr_shared_allocator_hh__
