//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_tarray.hh,v $
// $Revision: 1.2 $
// $Date: 2001/01/25 01:43:16 $

#ifndef __gr_tarray_hh__
#define __gr_tarray_hh__

#include "gr_tval.hh"

namespace grace
{

//------------------------------------------------------------------------------
// Dynamic array of t-values
//------------------------------------------------------------------------------

template <typename val_t>
class T_Array
{

private:
  //
  //! Pointer to data
  T_Val<val_t>* ptr ;
  //
  //! Size of array
  size_t size ;
  //
  //! Primary flag
  bool primary ;

public:
  //
  //! Default constructor
  inline T_Array () ;
  //
  //! Constructor
  inline T_Array (size_t _size) ;
  //
  //! Copy constructor
  inline T_Array (T_Array const& _array) ;
  //
  //! Destructor
  inline ~T_Array () ;
  //
  //! Assignment operator
  inline T_Array& operator = (T_Array const& _array) ;
  //
  //! Indexing operator
  inline T_Val<val_t>& operator [] (size_t _index) ;
  //
  //! Indexing operator
  inline T_Val<val_t> const& operator [] (size_t _index) const ;
  //
  //! Subarray operator
  inline T_Array<val_t> operator () (size_t _start, size_t _length) ;
  //
  //! Subarray operator
  inline T_Array<val_t> const operator () (
    size_t _start, size_t _length
  ) const ;

};

}

#endif // __gr_tarray_hh__
