//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_tgcc.hh,v $
// $Revision: 1.1 $
// $Date: 2002/01/31 17:22:17 $

#ifndef __gr_tgcc_hh__
#define __gr_tgcc_hh__

#include "gr_core.hh"

// This file contains helper functions used by tgcc.

namespace grace
{

// tgcc generates:
//
// void
// __ptfun_F (__F_args &args, __F_ress& ress)
// {
//   ___ptfun_a (ress.res1);
//   ___ptfun_a (ress.res2);
//   ...
//   ___ptfun_b (args, &__ptfun_F, &__tfun_F, ress);
// }
//
// ... which used the helper functions below:

template <typename RESTYPE>
static inline void
___ptfun_a (T_Out<RESTYPE> &res)
{
  res = new T_Val<RESTYPE>;
  (*res).bind (current->get_continuation());
  res.globalize ();
}

template <typename ARGSTYPE, typename RESSTYPE>
static inline void
___ptfun_b (ARGSTYPE const &args,
	    RESSTYPE &ress,
	    void (*ptfunp) (ARGSTYPE const &, RESSTYPE &),
	    void (*tfunp) (ARGSTYPE const *, RESSTYPE *))
{
  MemoMessage<ARGSTYPE,RESSTYPE> *msg
    = new MemoMessage<ARGSTYPE,RESSTYPE> (args, ptfunp, ress);
  comm_layer->broadcast (msg);
  tfunp (&args, &ress);
}


// tgcc generates:
//
// void
// __memo_tfun_F (__F_args const *args, __F_ress *ress)
// {
//   __F_ress *memo_ress = ___memo_tfun_a (args, ress, ptfun_F);
//   ___memo_tfun_b (ress->res1, memo_ress->res1);
//   ___memo_tfun_b (ress->res2, memo_ress->res2);
//   ...
// }
// 
// ... which used the helper functions below:

template <typename ARGSTYPE, typename RESSTYPE>
RESSTYPE *
___memo_tfun_a (ARGSTYPE const *args, RESSTYPE *ress,
	     void (*ptfunp) (ARGSTYPE const &, RESSTYPE &))
{
  MemoClosure<ARGSTYPE,RESSTYPE> closure(*args, ptfunp);
  return &MemoTableEntry<ARGSTYPE,RESSTYPE>::get_value(closure);
}

template <typename RESTYPE>
static inline void
___memo_tfun_b (RESTYPE &user, RESTYPE &memo)
{
  // copy memoized results from temporary locations to the places
  // which the user reserved for T-function results
  *user = *memo;
}

}

#endif // __gr_tgcc_hh__
