//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_block_tree.hh,v $
// $Revision: 1.6 $
// $Date: 2001/01/23 01:26:49 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_block_tree_hh__
#define __pxx_block_tree_hh__

#include "pxx_chunk_allocator.hh"

#include <new>

namespace pxx
{

class BlockTree :
  protected ChunkAllocator
{

private:

  struct Item
  {
    void* mask ;

    Item (void* _mask) :
      mask (_mask)
    {}

    void for_all (void (*_func)(void*, size_t)) ;

  };

  struct Node :
    Item
  {
    Item* left ;
    Item* right ;

    Node (void* _mask, Item* _l, Item* _r) :
      Item (_mask),
      left (_l),
      right (_r)
    {}

  };

  struct Leaf :
    Item
  {
    void* key ;
    size_t value ;

    Leaf (void* _key, size_t _value) :
      Item (0),
      key (_key),
      value (_value)
    {}

  };

  Item* root ;

  size_t size ;

public:

  BlockTree (Allocator& _allocator) :
    ChunkAllocator (_allocator,  max(sizeof(Leaf), sizeof(Node))),
    root (0),
    size (0)
  {}

  ~BlockTree ()
  {}

  int insert (void* _key, size_t _value) ;
  int remove (void* _key, size_t& _value = zref(size_t)) ;
  int get (void* _key, size_t& _value) ;
  int set (void* _key, size_t _value) ;
  void for_all (void (*_func)(void*, size_t));

private:

  inline is_leaf (Item* _item)
  {
    return _item->mask == 0;
  }

  inline is_node (Item* _item)
  {
    return _item->mask != 0;
  }

  inline Leaf* new_leaf (void* _key, size_t _value)
  {
    void* p = allocate ();
    return new(p) Leaf(_key, _value);
  }

  inline void destroy_item (Item* _item)
  {
    if (is_node(_item)) {
      destroy_item(((Node*)_item)->left);
      destroy_item(((Node*)_item)->right);
    }
    deallocate(_item);
  }

};

}

#endif // __pxx_block_tree_hh___
