//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_chunk_allocator.hh,v $
// $Revision: 1.8 $
// $Date: 2001/01/23 01:26:49 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_chunk_allocator_hh__
#define __pxx_chunk_allocator_hh__

#include "pxx_i_chunk_allocator.hh"
#include "pxx_i_allocator.hh"

namespace pxx
{

class ChunkAllocator :
  public I_ChunkAllocator
{

private:

  I_Allocator& block_allocator ;

  void* block_list ;

  void* free_list ;

  size_t chunk_size ;

  size_t block_size ;

public:

  inline ChunkAllocator (I_Allocator& _allocator, size_t const _chunksize) :
    block_allocator (_allocator),
    block_list (0),
    free_list (0),
    chunk_size (get_chunk_size(_chunksize)),
    block_size (
      size_align(256 * chunk_size, page_size) - sizeof(uintptr_t)
    )
  {}

  inline ~ChunkAllocator ()
  {
    void* p = block_list;
    while (p != 0) {
      void* q = p;
      p = *(void**)p;
      block_allocator.deallocate(q);
    }
  }

  inline void* allocate ()
  {
    if (free_list != 0) {
      void* p = free_list;
      free_list = *(void**)free_list;
      return p;
    } else {
      void* p = block_allocator.allocate(block_size);
      init_block(p);
      return ChunkAllocator::allocate();
    }
  }

  inline void deallocate (void* _ptr)
  {
    *(void**)_ptr = free_list;
    free_list = _ptr;
  }

private:

  static inline size_t get_chunk_size (size_t const _size)
  {
    return size_align(_size, sizeof(void*));
  }

  inline void init_block (void* _ptr)
  {
    *(void**)_ptr = block_list;
    block_list = _ptr;
    void* p = _ptr;
    void* q = (char*)p + block_size - chunk_size;
    while (p <= q) {
      p = *(void**)p = (char*)p + chunk_size;
    }
    p = (char*)p - chunk_size;
    *(void**)p = free_list;
    free_list = _ptr;
  }

};

}

#endif // __pxx_chunk_allocator_hh__
