//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_common.hh,v $
// $Revision: 1.16 $
// $Date: 2001/03/12 18:37:50 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_common_hh__
#define __pxx_common_hh__

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

namespace pxx
{

#define __STDC_CONSTANT_MACROS

#ifndef DEBUG
#define DEBUG 0
#endif

#ifndef PARANOIA
#define PARANOIA 0
#endif

#if UINTPTR_MAX == UINT32_MAX
#define UINTPTR_C(x) UINT32_C(x)
#elif UINTPTR_MAX == UINT64_MAX
#define UINTPTR_C(x) UINT64_C(x)
#else
#error "Unsupported pointer size"
#endif

#define null (0)

#define self (*this)

typedef void* Ptr ;

#define zref(_type) (*((_type*)(0)))

#define FATAL \
  fprintf( \
    stderr,"Fatal error in %s\nat %s:%d\n", \
    __PRETTY_FUNCTION__, __FILE__, __LINE__ \
  ), \
  pxx::fatal

#define TRY(a) \
  try { \
    a; \
  } catch (pxx::SysError const& err) { \
    FATAL("System error code = %d", err.get_code()); \
  }

extern void (*terminate) (int _error_code) ;
extern void fatal (char* _fmt, ...) ;

extern size_t page_size ;
extern size_t orders [256] ;
extern size_t rounds [sizeof(size_t) * 8] ;

static inline
size_t get_order (size_t _size)
{
  if (_size > 0) _size--;
#if 1
  size_t b;
  int n = -8;
  do {
    b = _size & size_t(0xFF);
    _size >>= 8;
    n += 8;
  } while (_size != 0);
  return orders [b] + n + 1;
#else
  size_t n = 0;
  if (_size & 0xFFFF0000) { _size &= 0xFFFF0000; n += 16; }
  if (_size & 0xFF00FF00) { _size &= 0xFF00FF00; n += 8; }
  if (_size & 0xF0F0F0F0) { _size &= 0xF0F0F0F0; n += 4; }
  if (_size & 0xCCCCCCCC) { _size &= 0xCCCCCCCC; n += 2; }
  if (_size & 0xAAAAAAAA) n += 1;
  return n + 1;
#endif
}

static inline
bool aligned (void* _ptr, uintptr_t _align)
{
  return ((_align - 1) & uintptr_t(_ptr)) == 0;
}

static inline
void* ptr_align (void* _ptr, uintptr_t _align)
{
  return (void*)(((uintptr_t)_ptr) & (~(_align - 1)));
}

static inline
size_t size_align (size_t _size, size_t _align)
{
  _align--;
  return (_size + _align) & (~(_align));
}

inline void* ptr_and (void* _p1, void* _p2)
{
  return (void*)((uintptr_t)_p1 & (uintptr_t)_p2);
}

inline void* ptr_or (void* _p1, void* _p2)
{
  return (void*)((uintptr_t)_p1 | (uintptr_t)_p2);
}

inline void* ptr_xor (void* _p1, void* _p2)
{
  return (void*)((uintptr_t)_p1 ^ (uintptr_t)_p2);
}

inline void* ptr_diff (void* _p1, void* _p2)
{
  uintptr_t diff = (uintptr_t)_p1 ^ (uintptr_t)_p2;
  return (void*)(((diff ^ (diff - 1)) >> 1) + 1);
}

template <typename typeT>
inline typeT* ptr_add (typeT* _ptr, size_t _value)
{
  return (typeT*)((char*)_ptr + _value);
}

template <typename typeT>
inline typeT* ptr_sub (typeT* _ptr, size_t _value)
{
  return (typeT*)((char*)_ptr - _value);
}

template <typename typeT1, typename typeT2>
inline ssize_t sub_ptr (typeT1* _ptr1, typeT2* _ptr2)
{
  return ((char*)_ptr1 - (char*)_ptr2);
}

#define pxx_max(a,b) ((a) > (b) ? (a) : (b))

}

#endif // __pxx_common_hh__
