//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_heap.cc,v $
// $Revision: 1.8 $
// $Date: 2001/03/30 13:40:44 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#include "pxx_heap.hh"

#include <sys/mman.h>
#include <unistd.h>

namespace pxx
{

Heap::Heap (
  void* const _start,
  size_t const _max_size,
  size_t _init_size = 0,
  int _fd = 0,
  bool _need_remap = true
) :
  start (_start),
  max_size (_max_size),
  max_order (get_order(max_size)),
  current_size (_init_size == 0 ? page_size : _init_size),
  current_order (get_order(current_size)),
  fd (_fd),
  need_remap (_need_remap)
{
  if (rounds[max_order] != max_size)
    FATAL("heap size is not a power of 2");
  if (!aligned(_start, page_size))
    FATAL("Heap start address is bad aligned");
  if (current_size > max_size)
    FATAL("Initial heap size greater than maximal size");
  if (need_remap) {
    void* res = mmap(
      start, current_size,
      PROT_READ | PROT_WRITE,
      (fd == 0 ? MAP_PRIVATE | MAP_ANON : MAP_SHARED) | MAP_FIXED,
      fd, 0
    );
    if (res == MAP_FAILED)
      FATAL("mmap() failed, errno = %d", errno);
  }
  if (fd != 0) {
    int res = ftruncate(fd, current_size);
    if (res != 0)
      FATAL("ftruncate() failed, errno = %d", errno);
  }
}

Heap::~Heap ()
{
  if (need_remap) {
    int res = munmap(start, current_size);
    if (res != 0)
      FATAL("munmap() failed, errno = %d", errno);
  }
}

void Heap::expand (unsigned _order = 1)
{
  size_t new_size = current_size << _order;
  if ((new_size > max_size) || (new_size == 0))
    FATAL("Heap exhausted");
  if (need_remap) {
    void* res = mremap(start, current_size, new_size, 0);
    if (res == MAP_FAILED)
      FATAL("mremap() failed, errno = %d", errno);
    if (res != start)
      FATAL("mremap() changes a heap start address");
  }
  current_order += _order;
  current_size = new_size;
  if (fd != 0) {
    int res = ftruncate(fd, current_size);
    if (res != 0)
      FATAL("ftruncate() failed, errno = %d", errno);
  }
}

}
