//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_heap_allocator.hh,v $
// $Revision: 1.16 $
// $Date: 2001/01/23 01:26:49 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_heap_allocator_hh__
#define __pxx_heap_allocator_hh__

#include "pxx_heap.hh"
#include "pxx_i_allocator.hh"

namespace pxx {

//
// heap allocator class
class HeapAllocator :
  public I_Allocator,
  protected Heap
{

private:
  //
  // memory block header structure
  struct BlockInfo
  {
    uintptr_t size ;            // size of block
    BlockInfo* prev ;           // pointer to previous free block
    BlockInfo* next ;           // pointer to next free block

    BlockInfo (BlockInfo* _prev, BlockInfo* _next, size_t _size = 0) :
      size (_size),
      prev (_prev),
      next (_next)
    {}

    inline void remove ()
    {
      prev->next = next;
      next->prev = prev;
    }

    inline uintptr_t get_size () const
    {
      return size & ~((uintptr_t)(0x01));
    }

    inline void set_size (uintptr_t _size)
    {
      size = _size | 0x01;
    }

  };

  struct FreeBlockList :
    public BlockInfo
  {

    FreeBlockList () :
      BlockInfo (this, this)
    {}

    void insert (BlockInfo* _block, size_t _size)
    {
      BlockInfo* last = prev;
      _block->next = this;
      _block->prev = last;
      last->next = _block;
      this->prev = _block;
      _block->size = _size;
    }

  };

  size_t min_free_size ;
  FreeBlockList lists [sizeof(size_t) * 8] ;

public:
  //
  // constructor
  HeapAllocator (
    void* const _start,
    size_t const _heap_size,
    int _fd = 0,
    bool _need_remap = true
  ) ;
  //
  // destructor
  ~HeapAllocator () ;
  //
  // allocate memory block of _size bytes
  void* allocate (size_t _size) ;
  //
  // free memory block at _ptr
  void deallocate (void* _ptr) ;
  //
  // resize memory block at _ptr to _size bytes
  void* reallocate (void* _ptr, size_t _size) ;
  //
  //
  inline static size_t get_size (void* _ptr)
  {
    return ((BlockInfo*)ptr_sub(_ptr, sizeof(uintptr_t)))->get_size();
  }

#if DEBUG
  //
  // dump memory map
  void memory_dump () ;
#endif // DEBUG

private:

  void* _reallocate (void* _ptr, size_t _size, size_t _oldsize) ;

  inline void* user_ptr (void* _ptr) const
  {
    return ptr_add(_ptr, sizeof(uintptr_t));
  }

  inline BlockInfo* true_ptr (void* _ptr) const
  {
    return (BlockInfo*)ptr_sub(_ptr, sizeof(uintptr_t));
  }

};

}

#endif // __pxx_heap_allocator_hh__
