//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_slist.hh,v $
// $Revision: 1.3 $
// $Date: 2001/01/23 01:26:49 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_slist_hh__
#define __pxx_slist_hh__

#include "pxx_common.hh"
#include "pxx_i_chunk_allocator.hh"

namespace pxx
{

template <typename type_t>
class SList
{

friend class Iterator ;

private:

  class Node
  {

  protected:

    Node* next ;
    type_t* value ;

    inline Node (type_t* _value = null, Node* _next = null) :
      value (_value),
      next (_next)
    {}

    inline ~Node ()
    {}

  };

  I_ChunkAllocator& allocator;
  Node* first ;
  Node* last ;

public:

  inline SList (I_ChunkAllocator& _allocator) :
    allocator (_allocator),
    first (null),
    last (null)
  {}

  inline ~SList ()
  {
    while (first != null) {
      Node* p = first->next;
      allocator.deallocate(first);
      first = p;
    }
  }

  inline void put (type_t* _value)
  {
    Node* p =
      (Node*)allocator.allocate();
    #if PARANOIA
    if (p == 0)
      FATAL("Unable to allocate new slist node");
    #endif
    new(p) Node<type_t>(_value);
    if (first == null) first = last = p;
    else last = last->next = p;
  }

  class Iterator
  {

  private:

    Node* current ;

  public:

    inline Iterator (SList& _list) :
      current (_list.first)
    {}

    operator bool ()
    {
      return (current != null);
    }

    Iterator& operator ++ ()
    {
      if (current != null) current = current->next;
      return self;
    }

    type_t* operator -> ()
    {
      return ptr->item;
    }

  };

};

}

#endif // __pxx_slist_hh__
