#include <stdlib.h>
#include "method1.h"

/******************************************************************************
*	Classes
******************************************************************************/

struct strPixel16 // Bit fields to get access to RGB chanals if biBitCount==16
{
 unsigned int b:5;
 unsigned int g:5;
 unsigned int r:5;
};

struct strPixel24 // Bit fields to get access to RGB chanals if biBitCount==24
{
 unsigned char b;
 unsigned char g;
 unsigned char r;
};

struct strPixel32 // Bit fields to get access to RGB chanals if biBitCount==32
{
 unsigned char b;
 unsigned char g;
 unsigned char r;
 unsigned char zero;
};

struct POSITION
{
  int x;
  int y;
  float	Corr_value;
};	// Struct for the search direction coding

struct VECTOR
{  int x;
   int y;
};  // Struct for various vector coding

struct RGB_CORR
{	double r;
	double b;
	double g;
};

/******************************************************************************
*	Prototypes
******************************************************************************/

struct POSITION TA_GetNewPosition_DS(struct POSITION prev_pos, int _bDiagn, int _nNStep, int _nVectDrctNormal,FrameInfo* pCurFrame, RGB1* tmpl, RGB1* img);
void TA_GetCorrValue_DS(struct POSITION* _pos,FrameInfo* pCurFrame, RGB1* tmpl, RGB1* img);
int TA_CutImage_DS(FrameInfo* pCurFrame,struct POSITION _pos, RGB1* img);
float TA_CompareTemp2Image_02_DS(int _MAX_PIXEL_VALUE, RGB1* tmpl, RGB1* img);

/******************************************************************************
*	Implementation
******************************************************************************/

void TA_GetCorrValue_DS(struct POSITION* _pos, FrameInfo* pCurFrame, RGB1* tmpl, RGB1* img)
{// The function compare the template with the image in the _pos position  and compute the
 //	correlation coefficient

	if (TA_CutImage_DS(pCurFrame, *_pos, img)) // The template size _nTemplSize*_nTemplSize
	{
		int _MAX_PIXEL_VALUE;
		if (pCurFrame->m_nBitsPixel==16)
			_pos->Corr_value=TA_CompareTemp2Image_02_DS(32, tmpl, img);
		else
			_pos->Corr_value=TA_CompareTemp2Image_02_DS(256, tmpl, img);
	}

}

float TA_CompareTemp2Image_02_DS(int _MAX_PIXEL_VALUE, RGB1* tmpl, RGB1* img)
{
	struct RGB_CORR Corr;	// The sity-block metric
	int i,j,t;
	struct RGB_CORR S1,S2,S3;

	Corr.r =0;Corr.g =0;Corr.b =0;
	S1.r =0;S1.g =0;S1.b =0;
	S2.r =0;S2.g =0;S2.b =0;
	S3.r =0;S3.g =0;S3.b =0;

	t=_nTemplSize1*_nTemplSize1;
	
	for (i=0;i<t;i=i+1)
	{
		 S1.r= S1.r+ abs(tmpl[i].r - img[i].r);
		 S1.g= S1.g+ abs(tmpl[i].g - img[i].g);
		 S1.b= S1.b+ abs(tmpl[i].b - img[i].b);
			 			
	}

	
		Corr.r = 1- S1.r/(t*_MAX_PIXEL_VALUE);
		Corr.g = 1- S1.g/(t*_MAX_PIXEL_VALUE);
		Corr.b = 1- S1.b/(t*_MAX_PIXEL_VALUE);
		
	return (float)((float)(Corr.r+Corr.g+Corr.b)/3.); // return template to image correlation coefficient
	//return (float)(sqrt(Corr.r*Corr.r+Corr.g*Corr.g +Corr.b*Corr.b));

}

struct POSITION TA_GetNewPosition_DS(struct POSITION prev_pos, int _bDiagn, int _nNStep, int _nVectDrctNormal, FrameInfo* pCurFrame, RGB1* tmpl, RGB1* img)
{
// The recursive function finds the best position of aim template correlation in the new frame.
//
//
	struct	POSITION New_pos[8], Best_pos;
	
	int t=0,i=0;
	int best_pos_number =-1;
	int x_=0,y_=0;

	struct	VECTOR _v8ass[8]; // The vectors defind 8 nearest neigbors of the current element

	float _fNewDirectionTresh = 0.97;
	int _SEARCH_STEP_LIMIT = 15;
	float Corr_value = prev_pos.Corr_value;

	_v8ass[0].x= 1;_v8ass[0].y= 0;//
	_v8ass[1].x= 1;_v8ass[1].y=-1;//			(-1,-1) ( 0,-1) ( 1,-1)
	_v8ass[2].x= 0;_v8ass[2].y=-1;//			(-1, 0) ( 0, 0) ( 1, 0)
	_v8ass[3].x=-1;_v8ass[3].y=-1;//			(-1, 1) ( 0, 1) ( 1, 1)
	_v8ass[4].x=-1;_v8ass[4].y= 0;//
	_v8ass[5].x=-1;_v8ass[5].y= 1;//					3  2  1
	_v8ass[6].x= 0;_v8ass[6].y= 1;//					4  x  0
	_v8ass[7].x= 1;_v8ass[7].y= 1;//					5  6  7
	
	for (i=0;i<8;i=i+1)			//Initialization may be not obligatory
	{	New_pos[i].x=0;
		New_pos[i].y=0;
		New_pos[i].Corr_value=0;
	}
		
	Best_pos.x=prev_pos.x;
	Best_pos.y=prev_pos.y;
	
	// Corr_value - the maximal value is the best

	if (_nNStep==0) // If it is just first frame after target was caught.
	{
		TA_GetCorrValue_DS(&Best_pos,pCurFrame, tmpl, img); //-------------------------------------------//
		
		for (i=0;i<8;i=i+1)
		{	New_pos[i].x=prev_pos.x+_v8ass[i].x;
			New_pos[i].y=prev_pos.y+_v8ass[i].y;
			
			TA_GetCorrValue_DS(&New_pos[i],pCurFrame, tmpl, img); //-------------------------------------------//

			if (New_pos[i].Corr_value > Best_pos.Corr_value)
			{	Best_pos.x=New_pos[i].x;
				Best_pos.y=New_pos[i].y;
				Best_pos.Corr_value = New_pos[i].Corr_value;
				best_pos_number =i;
			}				
		}

		if (Best_pos.x == prev_pos.x && Best_pos.y == prev_pos.y)
		{
			Corr_value = Best_pos.Corr_value;
			if (Corr_value<_fNewDirectionTresh)
			{
				_nVectDrctNormal = _nVectDrctNormal+4;
				if (_nVectDrctNormal>7) _nVectDrctNormal=_nVectDrctNormal-8;
				_nNStep = _nNStep+1;
				Best_pos =TA_GetNewPosition_DS(Best_pos, _bDiagn, _nNStep, _nVectDrctNormal,pCurFrame,tmpl,img);
			}

			if (Best_pos.Corr_value <prev_pos.Corr_value)
			return prev_pos;
			else
			return Best_pos;
		}
		else
		{	
			// To define the grow direction			
			x_ = Best_pos.x - prev_pos.x;
			y_ = Best_pos.y - prev_pos.y;
			
			for (i=0;i<8;i=i+1)
			{
				if (_v8ass[i].x==x_ &&_v8ass[i].y==y_)
					_nVectDrctNormal =i;
			}
			
			
			// To detect if the new position is the corner of the search area rectangle or not
			t= (int)(best_pos_number/2); t= t*2;

			if (best_pos_number != t) _bDiagn=1;
			else _bDiagn=0;

			_nNStep = _nNStep+1;

			//Recursion
			Best_pos =TA_GetNewPosition_DS(Best_pos, _bDiagn, _nNStep, _nVectDrctNormal,pCurFrame,tmpl,img);
			Corr_value = Best_pos.Corr_value;
			return Best_pos;
		}			
	}
	else
	{	if (_nNStep>_SEARCH_STEP_LIMIT)
			return prev_pos;
		
		Best_pos.Corr_value=prev_pos.Corr_value;
					
		// To define the relative coordinates
		if (_bDiagn) // if the best position ia a diagonal direction
		{
			//int x_=0,y_=0;
			x_=_v8ass[_nVectDrctNormal].x;
			y_=_v8ass[_nVectDrctNormal].y;
	
			New_pos[0].x=x_+x_;	New_pos[0].y=y_-y_;	//	(x-x,y+y)	(x,y+y)	(x+x,y+y)
			New_pos[1].x=x_+x_;	New_pos[1].y=y_;	//				(x,y)	(x+x,y	)
			New_pos[2].x=x_+x_;	New_pos[2].y=y_+y_;	//						(x+x,y-y)
			New_pos[3].x=x_;	New_pos[3].y=y_+y_;	//			4	  3		2	
			New_pos[4].x=x_-x_;	New_pos[4].y=y_+y_;	//				(x,y)	1
													//						0				
		}
		else
		{	
			x_=_v8ass[_nVectDrctNormal].x;
			y_=_v8ass[_nVectDrctNormal].y;
			
			if(x_==0)
			{
				New_pos[0].x=-2*y_;	 New_pos[0].y=y_;	//			(x-2y,y+y)	
				New_pos[1].x= - y_;  New_pos[1].y=y_;	//			(x-y,y+y)	
				New_pos[2].x=0;		 New_pos[2].y=y_;	//			(x,y+y)	
				New_pos[3].x=y_ ;	 New_pos[3].y=y_;	//			(x+y,y+y)		
				New_pos[4].x=2*y_;	 New_pos[4].y=y_;	//			(x+2y,y+y)
			}
			else
			{
				New_pos[0].y=-2*x_;	New_pos[0].x=x_;	//			(y+y,x-2y)	
				New_pos[1].y=-  x_;	New_pos[1].x=x_;	//			(y+y,x-y)	
				New_pos[2].y=0;		New_pos[2].x=x_;	//			(y+y,x)	
				New_pos[3].y=x_ ;	New_pos[3].x=x_;	//			(y+y,x+y)		
				New_pos[4].y=2*x_;	New_pos[4].x=x_;	//			(y+y,x+2y)
			}
		}

		for (i=0;i<5;i=i+1)
		{	// To define the exact coordinates
			New_pos[i].x=prev_pos.x+New_pos[i].x;
			New_pos[i].y=prev_pos.y+New_pos[i].y;

			TA_GetCorrValue_DS(&New_pos[i],pCurFrame, tmpl, img); //-------------------------------------------//

			if (New_pos[i].Corr_value > Best_pos.Corr_value)
			{	Best_pos.x=New_pos[i].x;
				Best_pos.y=New_pos[i].y;
				Best_pos.Corr_value = New_pos[i].Corr_value;
				best_pos_number =i;
			}				
		}

		if (Best_pos.x == prev_pos.x && Best_pos.y == prev_pos.y)
		{	
			Corr_value = Best_pos.Corr_value;
			if (Corr_value<_fNewDirectionTresh)
			{
				_nVectDrctNormal = _nVectDrctNormal+4;
				if (_nVectDrctNormal>7) _nVectDrctNormal=_nVectDrctNormal-8;
				_nNStep = _nNStep+1;
				Best_pos =TA_GetNewPosition_DS(Best_pos, _bDiagn, _nNStep, _nVectDrctNormal,pCurFrame,tmpl,img);
			}

			if (Best_pos.Corr_value <prev_pos.Corr_value)
			return prev_pos;
			else
			return Best_pos;
		}
		else
		{	
			// To define the grow direction			
			x_ = Best_pos.x - prev_pos.x;
			y_ = Best_pos.y - prev_pos.y;
			
			if (_bDiagn)
			{
				if (!(_v8ass[_nVectDrctNormal].x==x_ &&_v8ass[_nVectDrctNormal].y==y_))
				{
						if (x_ ==_v8ass[_nVectDrctNormal].x)
							y_=0;
						else	
							x_=0;
						for (i=0;i<8;i=i+1)
						{
							if (_v8ass[i].x==x_ &&_v8ass[i].y==y_)
							_nVectDrctNormal =i;
						}
				}				
			}
			else
			{
				if (_v8ass[_nVectDrctNormal].y==0) // Then it will be the horizontal direction
				{
					if (y_*y_> 1)
					{
						y_=y_/2;
					}
					else y_=0;					
				}
				else
				{	
					if (x_*x_> 1)
					{
						x_=x_/2;
					}
					else x_=0;
				}
					for (i=0;i<8;i=i+1)
						{
							if (_v8ass[i].x==x_ &&_v8ass[i].y==y_)
							_nVectDrctNormal =i;
						}
			}

			// To detect if the new position is the corner of the search area rectangle or not

			if (_bDiagn)
			{	
				if (best_pos_number !=2) _bDiagn =0;
			}
			else
			{	
				if (best_pos_number ==0) _bDiagn = 1;
				if (best_pos_number ==4) _bDiagn = 1;
			}
			
			_nNStep = _nNStep+1;

			//Recursion
			Best_pos=TA_GetNewPosition_DS(Best_pos, _bDiagn, _nNStep, _nVectDrctNormal, pCurFrame,tmpl,img);
			Corr_value = Best_pos.Corr_value;
			return Best_pos;				
		}
	}
	return Best_pos;
}

int TA_CutImage_DS(FrameInfo* pCurFrame,struct POSITION _pos, RGB1* img)
{	// It should detect the image type 16- 24 - or 32 bits and detect if the template will be out of frames

	int t_x, t_y, i, j, t;
	int _offset;
	char* pBits;
	char* pbiPixel;
	long byteLength;
	int _nBytes_onPixel;

	struct strPixel16* pbiPixel16;

	pBits = (char*)pCurFrame->m_pBits;
	_nBytes_onPixel = (pCurFrame->m_nBitsPixel/8);

	byteLength =  _nBytes_onPixel * pCurFrame->m_nWidth;

	_offset = (int)(_nTemplSize1/2);
	
	t_x =_pos.x - _offset;
	t_y =_pos.y - _offset;

	if (t_x<0 || t_y<0)
		return 0;
	if (t_x+_nTemplSize1>pCurFrame->m_nWidth || t_y+_nTemplSize1>pCurFrame->m_nHeight)
		return 0;
	
	if (pCurFrame->m_nBitsPixel==16)
    {
		t=0;
	 for (j=0;j<_nTemplSize1;j=j+1)
	  for (i=0;i<_nTemplSize1;i=i+1)
	  	{
	   	    pbiPixel16 = (struct strPixel16*)(pBits+(t_y+j)*byteLength+(t_x+i)*_nBytes_onPixel);
			img[t].r = pbiPixel16->r;
			img[t].g = pbiPixel16->g;
			img[t].b = pbiPixel16->b;
			t=t+1;
		}
	}
	if (pCurFrame->m_nBitsPixel==24)
    {
		t=0;
	 for (j=0;j<_nTemplSize1;j=j+1)
	  for (i=0;i<_nTemplSize1;i=i+1)
	  	{
	   	    pbiPixel = (pBits+(t_y+j)*byteLength+(t_x+i)*_nBytes_onPixel);
			img[t].b = *pbiPixel;
			pbiPixel++;
			img[t].g = *pbiPixel;
			pbiPixel++;
			img[t].r = *pbiPixel;
			t=t+1;
		}
	}
	if (pCurFrame->m_nBitsPixel==32)
    {
		t=0;
	 for (j=0;j<_nTemplSize1;j=j+1)
	  for (i=0;i<_nTemplSize1;i=i+1)
	  	{
	   	    pbiPixel = (pBits+(t_y+j)*byteLength+(t_x+i)*_nBytes_onPixel);
			img[t].b = *pbiPixel;
			pbiPixel++;
			img[t].g = *pbiPixel;
			pbiPixel++;
			img[t].r = *pbiPixel;
			t=t+1;
		}
	}
	return 1; // if everything is OK.
}

void method1(FrameInfo frameInfo, InfoMethod1 in, FullSight* fullSight, InfoMethod1* out)
{
	int i;
	struct POSITION _pos;
	float _fNewTemplateTresh = 0.7;

	_pos.x = in.prev_pos.center.x;
	_pos.y = in.prev_pos.center.y;
	_pos.Corr_value = in.Corr_value;

	if(_pos.Corr_value == 0)
        {
	  TA_CutImage_DS(&frameInfo, _pos, in.image);
	  _pos.Corr_value = 1;
	}
	
	_pos = TA_GetNewPosition_DS(_pos, 0, 0, 0, &frameInfo, in.image, out->image);

	if (_pos.Corr_value< _fNewTemplateTresh)
	{	
		TA_CutImage_DS(&frameInfo, _pos, in.image);
	}

	fullSight->sight.center.x =_pos.x;
	fullSight->sight.center.y =_pos.y;
	fullSight->sight.thckns.x = 10;
	fullSight->sight.thckns.y = 10;	
	fullSight->probab = 1;


	out->Corr_value = _pos.Corr_value;
	out->prev_pos.center.x = _pos.x;
	out->prev_pos.center.y = _pos.y;
	for(i=0;i<_nTemplSize1*_nTemplSize1;i=i+1)
		out->image[i] = in.image[i];
	return;
}
//max's comments:
//if you want to modify this method you can change:
//simple changes:
//1 - fNewTemplateTresh - the tresholf for taking new template
//2 - _nTemplSize - size if template to take
//3 - way of calculating of correlation value(but i guess it won't lead to any visible difference)
//4 - _nNewDirectionTresh - the treshold for finishing our search
//5 - _SEARCH_STEP_LIMIT - the worst thing of this method is frames are changing to fast to be processed
//complecated changes:
//6 - the idea of this method is if you're checking some direction(which is good at the begining)
//    you keep checking it. But the object might be at different direction(which is not so good at
//    the begining but quite OK at the end). May be this is the field to think about.
//    Also this method uses recursion but doesn't use the advantage of it such as easy access to
//    the previous step. It's seemed that it's making next step all the time even if you have to
//    step back. This is also field for changes.
