/******************************************************************************
*                     Copyright 2002 UIIP HPS Lab                            *
*                           All rights reserved                              *
*                                                                            *
* The material contained herein may not be reproduced in whole or in part,   *
*        without the prior written consent of HPS Lab                        *
*                                                                            *
******************************************************************************/

/******************************************************************************
*	Includes
******************************************************************************/

#include "method2.h"

/******************************************************************************
*	Forward declarations
******************************************************************************/


/******************************************************************************
*	Defines and constants
******************************************************************************/


/******************************************************************************
*	Typedefs
******************************************************************************/


/******************************************************************************
*	Classes
******************************************************************************/


/******************************************************************************
*	Prototypes
******************************************************************************/
// This function is used when it is necessary to restart object tracing
// under new conditions
eBool CatchObject2(FrameInfo frameInfo,
                 InfoMethod2 InfoInput,
                 InfoMethod2 *pInfoOutput,
                 FullSight *pFullSight);

// This function is used when it is necessary to continue object tracing
// without recalculation of the object description
eBool TraceObject2(FrameInfo frameInfo,
                 InfoMethod2 InfoInput,
                 InfoMethod2 *pInfoOutput,
                 FullSight *pFullSight);

// This function is used for object model building.
// Is based on analysis and processing of input information.
eBool BuildModel_2(FrameInfo frameInfo,
                  InfoMethod2 InfoInput,
                  InfoMethod2 *pInfoOutput,
                  FullSight *pFullSight,
                  SightRect ImPartRect);

// This function is used for comparison the current object model with previous
// frame one.
int CompareModels_2(InfoMethod2 InfoInput,
                     InfoMethod2 *pInfoOutput,
                     FullSight *pFullSight);

/******************************************************************************
*	Implementation
******************************************************************************/
//=============================================================================
//-----------------------------------------------------------------------------
void method2(FrameInfo frameInfo,
             InfoMethod2 InfoInput,
             FullSight *pFullSight,
             InfoMethod2 *pInfoOutput)
{

if(InfoInput.bIsAimTaken==eFalse ||
   InfoInput.sPrevPosition.center.x!=InfoInput.sNewPosition.center.x ||
   InfoInput.sPrevPosition.center.y!=InfoInput.sNewPosition.center.y
  )
   {
   //to restart object tracing
   // verify validity starting object coordinates

   // receive and initialize input data
   if(frameInfo.m_nHeight<frameInfo.m_nWidth)
      InfoInput.iMaxObjSize=frameInfo.m_nHeight/OBJ_SCALE_FACTOR;
      else
      InfoInput.iMaxObjSize=frameInfo.m_nWidth/OBJ_SCALE_FACTOR;
   if(!(InfoInput.iMaxObjSize%2))
      InfoInput.iMaxObjSize--;
   if(InfoInput.iMaxObjSize<MIN_FRAME_SIZE)
      return;
   InfoInput.iMaxObjSpeed=InfoInput.iMaxObjSize/MAX_OBJ_SHIFT;
   //InfoInput.iMaxObjSpeed
   // verify input object coordinates

   if(InfoInput.sNewPosition.center.x<InfoInput.iMaxObjSize/2||
      InfoInput.sNewPosition.center.y<InfoInput.iMaxObjSize/2||
      InfoInput.sNewPosition.center.x>=(frameInfo.m_nWidth-InfoInput.iMaxObjSize/2)||
      InfoInput.sNewPosition.center.y>=(frameInfo.m_nHeight-InfoInput.iMaxObjSize/2))
      {
      pFullSight->sight.center.x=0;
      pFullSight->sight.center.y=0;
      pFullSight->sight.thckns.x=InfoInput.iMaxObjSize/2;
      pFullSight->sight.thckns.y=InfoInput.iMaxObjSize/2;
      pFullSight->probab = 0;
      InfoInput.bIsAimTaken=pInfoOutput->bIsAimTaken=eFalse;
      pInfoOutput->sPrevPosition.center.x=pInfoOutput->sPrevPosition.center.y=0;
      return;
      }


   InfoInput.bIsAimTaken=CatchObject2(frameInfo, InfoInput, pInfoOutput, pFullSight);
   }
   else
   {
   //to continue object tracing
   TraceObject2(frameInfo, InfoInput, pInfoOutput, pFullSight);
   }

return;
}
//=============================================================================
//-----------------------------------------------------------------------------
// This function is used when it is necessary to restart object tracing
// under new conditions - the previous frame is not analyzing.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The current frame is processed to find the "object" that will be traced
eBool CatchObject2(FrameInfo frameInfo,
                 InfoMethod2 InfoInput,
                 InfoMethod2 *pInfoOutput,
                 FullSight *pFullSight)
{
eBool bSuccess=eFalse;

SightRect ImPartRect;   // image part rectangle

ImPartRect.LeftUpCorner.x=InfoInput.sNewPosition.center.x-InfoInput.iMaxObjSize/2;
ImPartRect.LeftUpCorner.y=InfoInput.sNewPosition.center.y-InfoInput.iMaxObjSize/2;
ImPartRect.RightDownCorner.x=InfoInput.sNewPosition.center.x+InfoInput.iMaxObjSize/2;
ImPartRect.RightDownCorner.y=InfoInput.sNewPosition.center.y+InfoInput.iMaxObjSize/2;

//printf("\ninside CatchObject:\n");

bSuccess=BuildModel_2(frameInfo,InfoInput,pInfoOutput,pFullSight,ImPartRect);
if (!bSuccess)
   return bSuccess;

// TEMPORARY!!!
pInfoOutput->sPrevPosition.center.x=InfoInput.sNewPosition.center.x;
pInfoOutput->sPrevPosition.center.y=InfoInput.sNewPosition.center.y;
pInfoOutput->sNewPosition.center.x=InfoInput.sNewPosition.center.x;
pInfoOutput->sNewPosition.center.y=InfoInput.sNewPosition.center.y;
pInfoOutput->sPrevPosition.thckns.x =InfoInput.iMaxObjSize;
pInfoOutput->sPrevPosition.thckns.y =InfoInput.iMaxObjSize;
pInfoOutput->sNewPosition.thckns.x =InfoInput.iMaxObjSize/2;
pInfoOutput->sNewPosition.thckns.y =InfoInput.iMaxObjSize/2;
pInfoOutput->iMaxObjSize=InfoInput.iMaxObjSize;
pInfoOutput->iMaxObjSpeed=InfoInput.iMaxObjSpeed;
pInfoOutput->iProbability=1;
pInfoOutput->bIsAimTaken=eTrue;

pFullSight->sight.center.x=InfoInput.sNewPosition.center.x;
pFullSight->sight.center.y=InfoInput.sNewPosition.center.y;
pFullSight->sight.thckns.x =InfoInput.iMaxObjSize/2;
pFullSight->sight.thckns.y =InfoInput.iMaxObjSize/2;
pFullSight->probab=1;
bSuccess=eTrue;
// TEMPORARY!!!
return bSuccess;
}
//.............................................................................

//=============================================================================
//-----------------------------------------------------------------------------
// This function is used when it is necessary to continue object tracing
// without recalculation of the object description.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The current frame part is processing and built object model is comparing with
// the previous one.
// New coordinates of the object are recalculating.
eBool TraceObject2(FrameInfo frameInfo,
                 InfoMethod2 InfoInput,
                 InfoMethod2 *pInfoOutput,
                 FullSight *pFullSight)
{
// common parameters
eBool bSuccess=eFalse;
SightRect ImPartRect;   // image part rectangle
unsigned long bestCompareResult=0;
unsigned long curCompareResult;
int bestX, bestY;
InfoMethod2 tmpOutput;
int minX, minY, maxX, maxY;

// parameters for square search
int candX[4], candY[4];
int midX, midY;
int cntrX, cntrY;
int distXY, distX, distY;

// parameters for spiral search
long cntrSp=0; // spiral counter
long limSp;  // spiral limit
int xShiftSp;
int yShiftSp;
int xPosSp;
int yPosSp;
int cntrStep=0;
int limStep;
int cntrDir;
int SpDir=0; // spiral direction : 0-right, 1-down, 2-left, 3-up
             // 0: xShiftSp=1;  yShiftSp=0;
             // 1: xShiftSp=0;  yShiftSp=1;
             // 2: xShiftSp=-1; yShiftSp=0;
             // 3: xShiftSp=0;  yShiftSp=-1;


bestX=InfoInput.sNewPosition.center.x;
bestY=InfoInput.sNewPosition.center.y;
// adjust the limits of search frame
minX=InfoInput.sNewPosition.center.x-
     InfoInput.iMaxObjSize/2-
     InfoInput.iMaxObjSpeed;
maxX=InfoInput.sNewPosition.center.x-
     InfoInput.iMaxObjSize/2+
     InfoInput.iMaxObjSpeed;
minY=InfoInput.sNewPosition.center.y-
     InfoInput.iMaxObjSize/2-
     InfoInput.iMaxObjSpeed;
maxY=InfoInput.sNewPosition.center.y-
     InfoInput.iMaxObjSize/2+
     InfoInput.iMaxObjSpeed;
if(minX<0)
   { maxX-=minX; minX=0; }
if(minY<0)
   { maxY-=minY; minY=0; }
if(maxX>(frameInfo.m_nWidth-InfoInput.iMaxObjSize))
   {
   minX+=(frameInfo.m_nWidth-InfoInput.iMaxObjSize-maxX);
   maxX=frameInfo.m_nWidth-InfoInput.iMaxObjSize;
   }
if(maxY>frameInfo.m_nHeight-InfoInput.iMaxObjSize)
   {
   minY+=(frameInfo.m_nHeight-InfoInput.iMaxObjSize-maxY);
   maxY=frameInfo.m_nHeight-InfoInput.iMaxObjSize;
   }

InfoInput.sNewPosition.center.x=minX+InfoInput.iMaxObjSize/2+InfoInput.iMaxObjSpeed;
InfoInput.sNewPosition.center.y=minY+InfoInput.iMaxObjSize/2+InfoInput.iMaxObjSpeed;
bestX=InfoInput.sNewPosition.center.x;
bestY=InfoInput.sNewPosition.center.y;


// /*
// spiral search (traces better, but is paralleling with difficulties)
// determine the limits for spiral search
cntrSp=0;
limSp=(InfoInput.iMaxObjSpeed)*(InfoInput.iMaxObjSpeed)*4;
SpDir=0; xShiftSp=1; yShiftSp=0;
xPosSp=InfoInput.sNewPosition.center.x-InfoInput.iMaxObjSize/2;
yPosSp=InfoInput.sNewPosition.center.y-InfoInput.iMaxObjSize/2;

// processing of central point
bestCompareResult=0;
bestCompareResult=~bestCompareResult;
ImPartRect.LeftUpCorner.x=xPosSp;
ImPartRect.LeftUpCorner.y=yPosSp;
ImPartRect.RightDownCorner.x=ImPartRect.LeftUpCorner.x+InfoInput.iMaxObjSize;
ImPartRect.RightDownCorner.y=ImPartRect.LeftUpCorner.y+InfoInput.iMaxObjSize;

bSuccess=BuildModel_2(frameInfo,InfoInput,&tmpOutput,pFullSight,ImPartRect);
if (!bSuccess)
   return bSuccess;

curCompareResult=CompareModels_2(InfoInput,&tmpOutput,pFullSight);
if (curCompareResult<bestCompareResult)
   {
   bestCompareResult=curCompareResult;
   bestX=xPosSp;
   bestY=yPosSp;
   }
//printf("%3d,%3d   \n",xPosSp,yPosSp);
yPosSp--;
cntrSp++;
cntrStep=0;
limStep=1;
cntrDir=1;
limSp--;
while(limSp>0)
   {
   while(cntrDir)
      {
      // make steps on the current direction
      while(cntrStep<limStep)
         {
         ImPartRect.LeftUpCorner.x=xPosSp;
         ImPartRect.LeftUpCorner.y=yPosSp;
         ImPartRect.RightDownCorner.x=ImPartRect.LeftUpCorner.x+InfoInput.iMaxObjSize;
         ImPartRect.RightDownCorner.y=ImPartRect.LeftUpCorner.y+InfoInput.iMaxObjSize;

         bSuccess=BuildModel_2(frameInfo,InfoInput,&tmpOutput,pFullSight,ImPartRect);
         if (!bSuccess)
            return bSuccess;

         curCompareResult=CompareModels_2(InfoInput,&tmpOutput,pFullSight);
         if (curCompareResult<bestCompareResult)
            {
            bestCompareResult=curCompareResult;
            //printf("%10d  ",bestCompareResult);
            bestX=xPosSp;
            bestY=yPosSp;
            }
         //printf("%3d,%3d   ",xPosSp,yPosSp);
         cntrSp++;
         cntrStep++;
         xPosSp+=xShiftSp;
         yPosSp+=yShiftSp;
         limSp--;
         if(limSp<=0)
            break;
         }
      // change direction
      if(limSp<=0)
         break;
      SpDir++;
      switch(SpDir%4)
         {
         case 0: xShiftSp=1; yShiftSp=0;  break;
         case 1: xShiftSp=0; yShiftSp=1;  break;
         case 2: xShiftSp=-1; yShiftSp=0;  break;
         case 3: xShiftSp=0; yShiftSp=-1; break;
         }
      cntrDir--;
      cntrStep=0;
      }
   if(limSp<=0)
      break;
   cntrDir=2;
   limStep++;
   }

// */

//printf("\n");


// square search (not best tracer, but easy for paralleling)
// determine the limits of square search frame
 /*
// adjust the limits of square search frame
midX=(minX+maxX)/2;
midY=(minY+maxY)/2;
cntrSp=0;

bestCompareResult=0;
bestCompareResult=~bestCompareResult;
for(cntrX=midX; cntrX>=minX; cntrX--)
   for(cntrY=midY; cntrY>=minY; cntrY--)
      {
      ImPartRect.LeftUpCorner.x=cntrX;
      ImPartRect.LeftUpCorner.y=cntrY;
      ImPartRect.RightDownCorner.x=ImPartRect.LeftUpCorner.x+InfoInput.iMaxObjSize;
      ImPartRect.RightDownCorner.y=ImPartRect.LeftUpCorner.y+InfoInput.iMaxObjSize;
      cntrSp++;

      bSuccess=BuildModel_2(frameInfo,InfoInput,&tmpOutput,pFullSight,ImPartRect);
      if (!bSuccess)
         return bSuccess;

      curCompareResult=CompareModels_2(InfoInput,&tmpOutput,pFullSight);
      if (curCompareResult<bestCompareResult)
         {
         bestCompareResult=curCompareResult;
         candX[0]=cntrX;
         candY[0]=cntrY;
         }
      }
bestCompareResult=0;
bestCompareResult=~bestCompareResult;
for(cntrX=midX; cntrX>=minX; cntrX--)
   for(cntrY=midY+1; cntrY<maxY; cntrY++)
      {
      ImPartRect.LeftUpCorner.x=cntrX;
      ImPartRect.LeftUpCorner.y=cntrY;
      ImPartRect.RightDownCorner.x=ImPartRect.LeftUpCorner.x+InfoInput.iMaxObjSize;
      ImPartRect.RightDownCorner.y=ImPartRect.LeftUpCorner.y+InfoInput.iMaxObjSize;
      cntrSp++;

      bSuccess=BuildModel_2(frameInfo,InfoInput,&tmpOutput,pFullSight,ImPartRect);
      if (!bSuccess)
         return bSuccess;

      curCompareResult=CompareModels_2(InfoInput,&tmpOutput,pFullSight);
      if (curCompareResult<bestCompareResult)
         {
         bestCompareResult=curCompareResult;
         candX[2]=cntrX;
         candY[2]=cntrY;
         }
      }
bestCompareResult=0;
bestCompareResult=~bestCompareResult;
for(cntrX=midX+1; cntrX<maxX; cntrX++)
   for(cntrY=midY; cntrY>=minY; cntrY--)
      {
      ImPartRect.LeftUpCorner.x=cntrX;
      ImPartRect.LeftUpCorner.y=cntrY;
      ImPartRect.RightDownCorner.x=ImPartRect.LeftUpCorner.x+InfoInput.iMaxObjSize;
      ImPartRect.RightDownCorner.y=ImPartRect.LeftUpCorner.y+InfoInput.iMaxObjSize;
      cntrSp++;

      bSuccess=BuildModel_2(frameInfo,InfoInput,&tmpOutput,pFullSight,ImPartRect);
      if (!bSuccess)
         return bSuccess;

      curCompareResult=CompareModels_2(InfoInput,&tmpOutput,pFullSight);
      if (curCompareResult<bestCompareResult)
         {
         bestCompareResult=curCompareResult;
         candX[1]=cntrX;
         candY[1]=cntrY;
         }
      }
bestCompareResult=0;
bestCompareResult=~bestCompareResult;
for(cntrX=midX+1; cntrX<maxX; cntrX++)
   for(cntrY=midY+1; cntrY<maxY; cntrY++)
      {
      ImPartRect.LeftUpCorner.x=cntrX;
      ImPartRect.LeftUpCorner.y=cntrY;
      ImPartRect.RightDownCorner.x=ImPartRect.LeftUpCorner.x+InfoInput.iMaxObjSize;
      ImPartRect.RightDownCorner.y=ImPartRect.LeftUpCorner.y+InfoInput.iMaxObjSize;
      cntrSp++;

      bSuccess=BuildModel_2(frameInfo,InfoInput,&tmpOutput,pFullSight,ImPartRect);
      if (!bSuccess)
         return bSuccess;

      curCompareResult=CompareModels_2(InfoInput,&tmpOutput,pFullSight);
      if (curCompareResult<bestCompareResult)
         {
         bestCompareResult=curCompareResult;
         candX[3]=cntrX;
         candY[3]=cntrY;
         }
      }
distX=InfoInput.sNewPosition.center.x-InfoInput.iMaxObjSize/2-candX[0];
if(distX<0)
   distX=-distX;
distY=InfoInput.sNewPosition.center.y-InfoInput.iMaxObjSize/2-candY[0];
if(distY<0)
   distY=-distY;
distXY=distX+distY;
bestX=candX[0];
bestY=candY[0];

distX=InfoInput.sNewPosition.center.x-InfoInput.iMaxObjSize/2-candX[1];
if(distX<0)
   distX=-distX;
distY=InfoInput.sNewPosition.center.y-InfoInput.iMaxObjSize/2-candY[1];
if(distY<0)
   distY=-distY;
if(distXY<(distX+distY))
   {
   distXY=distX+distY;
   bestX=candX[1];
   bestY=candY[1];
   }

distX=InfoInput.sNewPosition.center.x-InfoInput.iMaxObjSize/2-candX[2];
if(distX<0)
   distX=-distX;
distY=InfoInput.sNewPosition.center.y-InfoInput.iMaxObjSize/2-candY[2];
if(distY<0)
   distY=-distY;
if(distXY<(distX+distY))
   {
   distXY=distX+distY;
   bestX=candX[2];
   bestY=candY[2];
   }

distX=InfoInput.sNewPosition.center.x-InfoInput.iMaxObjSize/2-candX[3];
if(distX<0)
   distX=-distX;
distY=InfoInput.sNewPosition.center.y-InfoInput.iMaxObjSize/2-candY[3];
if(distY<0)
   distY=-distY;
if(distXY<(distX+distY))
   {
   distXY=distX+distY;
   bestX=candX[3];
   bestY=candY[3];
   }
 */


//printf("\ninside TraceObject:\n");

// TEMPORARY!!!
pInfoOutput->sPrevPosition.center.x=bestX+InfoInput.iMaxObjSize/2;
pInfoOutput->sPrevPosition.center.y=bestY+InfoInput.iMaxObjSize/2;
pInfoOutput->sPrevPosition.thckns.x =InfoInput.iMaxObjSize;
pInfoOutput->sPrevPosition.thckns.y =InfoInput.iMaxObjSize;
pInfoOutput->iMaxObjSize=InfoInput.iMaxObjSize;
pInfoOutput->sNewPosition.center.x=bestX+InfoInput.iMaxObjSize/2;
pInfoOutput->sNewPosition.center.y=bestY+InfoInput.iMaxObjSize/2;
pInfoOutput->sNewPosition.thckns.x =InfoInput.iMaxObjSize/2;
pInfoOutput->sNewPosition.thckns.y =InfoInput.iMaxObjSize/2;
pInfoOutput->iProbability=1;
pInfoOutput->bIsAimTaken=eTrue;

pFullSight->sight.center.x=bestX+InfoInput.iMaxObjSize/2;
pFullSight->sight.center.y=bestY+InfoInput.iMaxObjSize/2;
pFullSight->sight.thckns.x =InfoInput.iMaxObjSize/2;
pFullSight->sight.thckns.y =InfoInput.iMaxObjSize/2;
pFullSight->probab=1;
bSuccess=eTrue;
// TEMPORARY!!!
return bSuccess;
}
//.............................................................................
//=============================================================================
//-----------------------------------------------------------------------------
//
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
eBool BuildModel_2(FrameInfo frameInfo,
                  InfoMethod2 InfoInput,
                  InfoMethod2 *pInfoOutput,
                  FullSight *pFullSight,
                  SightRect ImPartRect)
{
eBool bSuccess=eFalse;
int cntr;
int cntrX;
int cntrY;
unsigned int cntrB;
unsigned long aveB;
unsigned int cntrG;
unsigned long aveG;
unsigned int cntrR;
unsigned long aveR;
unsigned int cntrL;
unsigned long aveL;
unsigned char *pTMP;
unsigned char *pBitCut;

// find the field of interest, i.e. cut part of the image to be processed
switch(frameInfo.m_nBitsPixel)
   {
   case 8:
      //
      break;
   case 16:
      //
      break;
   case 32:
      //
      {
      /*
      for(cntr=0; cntr<256; cntr++)
         InfoInput.sObjModel.uiHistoR[cntr]=
         InfoInput.sObjModel.uiHistoG[cntr]=
         InfoInput.sObjModel.uiHistoB[cntr]=0;
      for(cntr=0; cntr<256; cntr++)
         InfoInput.sObjModel.uiHistoL[cntr]=0;
      */
      /*
      for(cntr=0; cntr<MAX_FRAME_SIZE; cntr++)
         InfoInput.sObjModel.uiChange_L[cntr]=0;
      */
      /*
      printf("|");
      for(cntrX=0; cntrX<InfoInput.iMaxObjSize; cntrX++)
         printf("--");
      printf("|\n");
      */
      cntrB=0;
      aveB=0;
      cntrG=0;
      aveG=0;
      cntrR=0;
      aveR=0;
      for(cntrY=0; cntrY<InfoInput.iMaxObjSize; cntrY++)
         {
         pTMP=((unsigned char *)frameInfo.m_pBits+
            (cntrY+ImPartRect.LeftUpCorner.y)*frameInfo.m_nWidthBytes+
            ImPartRect.LeftUpCorner.x*4);

         //printf("|");
         for(cntrX=0; cntrX<InfoInput.iMaxObjSize; cntrX++)
            {
            /*
            if(*pTMP<64)
               printf("##");
               else if(*pTMP<111)
                  printf("++");
                  else if(*pTMP<158)
                     printf("--");
                     else if(*pTMP<196)
                        printf(". ");
                        else
                        printf("  ");
            */
            //InfoInput.sObjModel.sObjAveColor.l
            aveB+=*pTMP;
            cntrB++;
            pTMP++;
            aveG+=*pTMP;
            cntrG++;
            pTMP++;
            aveR+=*pTMP;
            cntrR++;
            pTMP++;
            pTMP++;
            }
         //printf("|\n");
         }
      /*
      printf("|");
      for(cntrX=0; cntrX<InfoInput.iMaxObjSize; cntrX++)
         printf("--");
      printf("|\n");
      */
      aveL=aveB+aveG+aveR;
      cntrL=cntrB+cntrG+cntrR;
      aveL/=cntrL;
      aveB/=cntrB;
      aveG/=cntrG;
      aveR/=cntrR;
      pInfoOutput->sObjModel.sObjAveColor.l=aveL;
      pInfoOutput->sObjModel.sObjAveColor.r=aveR;
      pInfoOutput->sObjModel.sObjAveColor.g=aveG;
      pInfoOutput->sObjModel.sObjAveColor.b=aveB;
      }
      break;
   default:
      return eFalse;
   }
//printf("  aveB=%4d  aveG=%4d  aveR=%4d  aveL=%4d  ",aveB,aveG,aveR,aveL);
// TEMPORARY!!!
bSuccess=eTrue;
// TEMPORARY!!!
return bSuccess;
}
//.............................................................................

//=============================================================================
//-----------------------------------------------------------------------------
//
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
int CompareModels_2(InfoMethod2 InfoInput,
                     InfoMethod2 *pInfoOutput,
                     FullSight *pFullSight)
{
int res=1;
int dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.b-(int)pInfoOutput->sObjModel.sObjAveColor.b;
if(dif<0)
   dif=-dif;
res*=dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.g-(int)pInfoOutput->sObjModel.sObjAveColor.g;
if(dif<0)
   dif=-dif;
res*=dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.r-(int)pInfoOutput->sObjModel.sObjAveColor.r;
if(dif<0)
   dif=-dif;
res*=dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.l-(int)pInfoOutput->sObjModel.sObjAveColor.l;
if(dif<0)
   dif=-dif;
res*=dif*2;
//++++++++++++++++
dif=(int)InfoInput.sObjModel.sObjAveColor.b-(int)pInfoOutput->sObjModel.sObjAveColor.b;
if(dif<0)
   dif=-dif;
res+=dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.g-(int)pInfoOutput->sObjModel.sObjAveColor.g;
if(dif<0)
   dif=-dif;
res+=dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.r-(int)pInfoOutput->sObjModel.sObjAveColor.r;
if(dif<0)
   dif=-dif;
res+=dif;

dif=(int)InfoInput.sObjModel.sObjAveColor.l-(int)pInfoOutput->sObjModel.sObjAveColor.l;
if(dif<0)
   dif=-dif;
res+=dif*2;

//printf("res=%5d  ",res);
return res;
}

//.............................................................................

/* end of file */

//=============================================================================
//-----------------------------------------------------------------------------
//
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//


//.............................................................................

