/******************************************************************************
*                     Copyright 2002 UIIP HPS Lab                             *
*                           All rights reserved                               *
*                                                                             *
* The material contained herein may not be reproduced in whole or in part,    *
*        without the prior written consent of HPS Lab                         *
*                                                                             *
******************************************************************************/

/******************************************************************************
*
* Description: Method #2
*           implementation of object tracing algorithm #2
*
*
*           Created and implemented by O!Patsko (c) for HPS Lab
******************************************************************************/

#ifndef I_METHOD2_H
#define I_METHOD2_H

/******************************************************************************
*	Includes
******************************************************************************/

#include "FrameInfo.h"
#include "Sight.h"
#include "types.h"

/******************************************************************************
*	Forward declarations
******************************************************************************/

typedef struct
   {
   ColorRGB sObjAveColor;     // average color of the object
   ColorRGB sObjColorDisp; // color dispersion of the object (color stability)
   TPoint   sObjCenter;    // center of the object
   TPoint   sObjDimensions;// x- & y- dimensions of the object
   TPoint   sObjShift;     // last shift of the object
		 //   eObjType  sObjType;     // object type exerts influence on processing technology
   //unsigned int uiChange_L[MAX_FRAME_SIZE];
   //unsigned int uiHistoR[256];
   //unsigned int uiHistoG[256];
   //unsigned int uiHistoB[256];
   //unsigned int uiHistoL[256];
   } ObjModel2;   // holds the object description information

typedef struct
   {
   int   iMaxObjSize;      // determines the maximal object size in depending of frame size
   int   iMaxObjSpeed;     // determines the object shift between current and previous frames
   int   iProbability;     // shows the robustness of the object detection and tracing
   ObjModel2 sObjModel;    // describes the traced object
   Sight sNewPosition;    // holds coordinates of the new object position
   Sight sPrevPosition;    // holds coordinates of the previous object position
   eBool  bIsAimTaken;     // is "true" if the object was catched
   } InfoMethod2; // holds information that is used for implementation of the method #2


/******************************************************************************
*	Classes
******************************************************************************/


/******************************************************************************
*	Prototypes
******************************************************************************/

void method2(FrameInfo frameInfo,
             InfoMethod2 in,
             FullSight* fullSight,
             InfoMethod2* out);

/******************************************************************************
*	Implementation
******************************************************************************/


#endif /* I_METHOD2_H */

/* end of file */
