/******************************************************************************
*                     Copyright 2002 IEC HPS Lab                              *
*                           All rights reserved                               *
*                                                                             *
* The material contained herein may not be reproduced in whole or in part,    *
*        without the prior written consent of HPS Lab                         *
*                                                                             *
******************************************************************************/

/******************************************************************************
*
* Description: Method #3
*           implementation of object tracing algorithm #3
*
*
*           Created and implemented by O!Patsko (c) for HPS Lab
******************************************************************************/

#ifndef I_METHOD3_H
#define I_METHOD3_H

/******************************************************************************
*	Includes
******************************************************************************/

#include "Sight.h"
#include "FrameInfo.h"
//#include "math.h"
#include "types.h"

/******************************************************************************
*	Forward declarations
******************************************************************************/

typedef struct
   {
   ColorRGB sObjAveColor;     // average color of the object
   ColorRGB sObjColorDisp; // color dispersion of the object (color stability)
   TPoint   sObjCenter;    // center of the object
   TPoint   sObjDimensions;// x- & y- dimensions of the object
   TPoint   sObjShift;     // last shift of the object
   eObjType  sObjType;     // object type exerts influence on processing technology
   //unsigned int uiChange_L[MAX_FRAME_SIZE];
   int uiHistoR[256];
   int uiHistoG[256];
   int uiHistoB[256];
   //int uiHistoL[256];
   } ObjModel3;   // holds the object description information

typedef struct
   {
   int   iMaxObjSize;      // determines the maximal object size in depending of frame size
   int   iMaxObjSpeed;     // determines the object shift between current and previous frames
   int   iProbability;     // shows the robustness of the object detection and tracing
   ObjModel3 sObjModel;    // describes the traced object
   Sight sNewPosition;    // holds coordinates of the new object position
   Sight sPrevPosition;    // holds coordinates of the previous object position
   eBool  bIsAimTaken;     // is "true" if the object was catched
   } InfoMethod3; // holds information that is used for implementation of the method #1


/******************************************************************************
*	Classes
******************************************************************************/


/******************************************************************************
*	Prototypes
******************************************************************************/

void method3(FrameInfo frameInfo,
             InfoMethod3 in,
             FullSight* fullSight,
             InfoMethod3* out);

/******************************************************************************
*	Implementation
******************************************************************************/


#endif /* I_METHOD3_H */

/* end of file */




