#include "tc.h"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#define  CLASS_S
#include "task.h"

#define SIZE      NA
#define PARTS     1400
#define PART_SIZE (SIZE / PARTS)

#define SEED  314159265
#define AMULT 1220703125

// data types

struct vpart
{
    double data[PART_SIZE];
};

struct vector
{
	double data[SIZE];
};

struct spvector
{
	double data[NONZERO + 1];
	int    idx[NONZERO + 1];
};

struct cell
{
	int    addr;
	double data;
};

struct matrix
{
	int         idx[PARTS + 1];
	struct cell data[((NONZERO + 1) * (NONZERO + 1) + 1) * SIZE];
};

// qsort

void _qsort1( struct cell *base,
              int l,
              int r )
{
	do
	{
		int	l1;
		int	r1;
		int	x;
		int	m;

		m = (l + r) / 2;
		x = base[m].addr;

		l1 = l;
		r1 = r;
		
		do
		{
			while ( base[l1].addr < x ) l1++;
			while ( base[r1].addr > x ) r1--;
			if ( l1 <= r1 )
			{
				struct cell	t;
				
				t.addr = base[l1].addr;
				base[l1].addr = base[r1].addr;
				base[r1].addr = t.addr;
	
				t.data = base[l1].data;
				base[l1].data = base[r1].data;
				base[r1].data = t.data;
			
				l1++;
				r1--;
			}
		} while ( l1 <= r1 );
		
		if ( r1-l > r-l1 )
		{
			if ( l1 < r ) _qsort1( base, l1, r );
			r = r1;
		}
		else
		{
			if ( l < r1 ) _qsort1( base, l, r1 );
			l = l1;
		}
		
	} while ( l < r );
}

void _qsort( struct cell *base,
             int nmemb )
{
	_qsort1( base, 0, nmemb - 1 );
}

// create data

void make_0( struct vector *result )
{
	int i;

	for ( i = 0; i < SIZE; i++ )
	{
        result->data[i] = 0;
	}
}


void make_1( struct vector tptr result )
{
	int i;
	
	for ( i = 0; i < SIZE; i++ )
	{
        result->data[i] = 1.0;
	}
}

// matrix generator

#define R23 0.00000011920928955078125
#define R46 0.0000000000000142108547152020037174224853515625
#define T23 8388608.0
#define T46 70368744177664.0

double randlc( double a,
               double *x )
{
	double t1;
	double t2;
	double t3;
	double t4;
	double a1;
	double a2;
	double x1;
	double x2;
	double z;

	t1 = R23 * a;
	a1 = (int)(t1);
	a2 = a - T23 * a1;

	t1 = R23 * (*x);
	x1 = (int)(t1);
	x2 = (*x) - T23 * x1;
	t1 = a1 * x2 + a2 * x1;
	t2 = (int)(R23 * t1);
	z = t1 - T23 * t2;
	t3 = T23 * z + a2 * x2;
	t4 = (int)(R46 * t3);
	*x = t3 - T46 * t4;

	return ( R46 * (*x) );
}

void make_v( double *seed,
             int j,
             struct spvector *spv )
{
	int	i;
	int	mult;
	int	*mark;

	mark = new int[SIZE];

	for ( i = 0; i < SIZE; i++ ) mark[i] = 0;
	
	for ( i = 0; i <= NONZERO; i++ )
	{
		spv->idx[i] = -1;
	}

	mult = 1;
	while ( mult < SIZE ) mult <<= 1;

	for ( i = 0; i < NONZERO; i++ )
	{
		int    ind;
		double val;
		
		do
		{
			val = randlc( AMULT, seed );
			ind = (int)(randlc( AMULT, seed ) * mult);
		} while (!((ind >= 0) && (ind < SIZE) && (!mark[ind])));

		spv->data[i] = val;
		spv->idx[i] = ind;
		mark[ind] = 1;
	}

	for ( i = 0; (i < NONZERO) && (spv->idx[i] != j ); i++);
	spv->data[i] = 0.5;
	spv->idx[i] = j;

	delete[] mark;
}

void make_a( struct matrix tptr result )
{
	int             i;
	int             j;
	int             k;
	int             index;
    double          seed;
	double          zeta;
	double          scale;
	double          ratio;
	double          size;
	struct spvector spv;

	seed = SEED;
	zeta = randlc( AMULT, &seed );
	
    index = 0;

	for ( i = 0; i < SIZE; i++ )
	{
		result->data[index].data = RCOND - SHIFT;
		result->data[index].addr = i * (SIZE + 1);
		index++;
	}

	size = 1;
	ratio = pow( RCOND, 1.0 / SIZE );
	
	for ( k = 0; k < SIZE; k++ )
	{
		make_v( &seed, k, &spv );
		for ( j = 0; j <= NONZERO; j++ )
		{
			if ( spv.idx[j] < 0 ) continue;
			scale = size * spv.data[j];
			for ( i = 0; i <= NONZERO; i++ )
			{
				if ( spv.idx[i] < 0 ) continue;
				result->data[index].data = scale * spv.data[i];
				result->data[index].addr = spv.idx[j] * SIZE + spv.idx[i];
				index++;
			}
		}
		size *= ratio;
	}

	_qsort( result->data, index );

	j = k = 0;
	for ( i = 0; i < PARTS; i++ )
	{
		result->idx[i] = j;
		while ((k < index) && (result->data[k].addr / (SIZE * PART_SIZE) == i))
		{
			result->data[j].data = result->data[k].data;
			result->data[j].addr = result->data[k].addr;
			k++;
			while ((k < index) && (result->data[j].addr == result->data[k].addr))
			{
				result->data[j].data += result->data[k].data;
				k++;
			}
			j++;
		}
	}
	result->idx[i] = j;
    printf( "NZ = %d (%d)\n", j, index );
}

// cg: step 1

tfun void cgs1_part( struct matrix tptr m,
                     int num,
                     struct vector tptr p,
                     tout struct vpart *q,
                     tout double *mpq)
{
	int i;
    int x;
    int y;
    
	for ( i = 0; i < PART_SIZE; i++ )
	{
		q->data[i] = 0;
	}
	
    for ( i = m->idx[num]; i < m->idx[num + 1]; i++ )
	{
        x = m->data[i].addr % SIZE;
		y = (m->data[i].addr / SIZE) - num * PART_SIZE;
		q->data[y] += m->data[i].data * p->data[x];
	}
	
	*mpq = 0;
    for ( i = 0; i < PART_SIZE; i++ )
	{
		*mpq += q->data[i] * p->data[num * PART_SIZE + i];
	}
}

void cgs1( struct matrix tptr a,
           struct vector *x,
           struct vector tptr p,
           struct vector tptr xi,
           double rho,
           struct vector *x1,
           struct vector tptr p1,
           struct vector tptr xi1,
           double *nxi1 )
{
    int          i;
    int          j;
    double       alpha;
	double       beta;
	double       rho0;
	double       rho1;
    double       d;
    double       tptr b2arr;
	struct vpart tptr qarr;

    b2arr = new tval double[PARTS];
	qarr = new tval struct vpart[PARTS];

    for ( i = 0; i < PARTS; i++ )
	{
		cgs1_part( a, i, p, (qarr + i), (b2arr + i) );
	}

    d = 0;
    for ( i = 0; i < PARTS; i++ ) d += b2arr[i];
    
    rho0 = rho;
    alpha = rho0 / d;
    
    double res;
    res = 0;

    for ( j = 0; j < PARTS; j++ )
	{
        double tptr part;

        part = (qarr + j)->data;
			
		for ( i = 0; i < PART_SIZE; i++ )
		{
			int    idx;
			double tmp;
            
            idx = j * PART_SIZE + i;

			tmp = xi1->data[j * PART_SIZE + i] = xi->data[j * PART_SIZE + i] - alpha * part[i];
			x1->data[idx] = x->data[idx] + alpha * p->data[idx];
				
			res += tmp * tmp;
		}
	}

	*nxi1 = rho1 = res;
    
    beta = rho1 / rho0;

	for ( i = 0; i < SIZE; i++ )
	{
		p1->data[i] = xi1->data[i] + beta * p->data[i];
	}
    
	delete[] qarr;
	delete[] b2arr;
}


// cg

void cg( struct matrix tptr a,
         struct vector tptr x0,
         double rho,
         struct vector tptr x,
         double *nxi,
         double *result )
{
    int           k;
	int           i;
    double	      nxi1;
    struct vector *vx0;
	struct vector tptr vxi0;
	struct vector tptr vp0;
    struct vector *vx1;
	struct vector tptr vxi1;
	struct vector tptr vp1;
	struct vector *zero;

	vx0  = new struct vector;
	vxi0 = new tval struct vector;
	vp0  = new tval struct vector;  
	
	zero = new struct vector;
    make_0( zero );

	cgs1( a, zero, x0, x0, rho, vx0, vp0, vxi0, &nxi1 );
    
	delete zero;
	
	k = 0;
    for ( i = 1; i < 25; i++ )
	{
        vx1  = new struct vector;
		vxi1 = new tval struct vector;
		vp1  = new tval struct vector;

		cgs1( a, vx0, vp0, vxi0, nxi1, vx1, vp1, vxi1, &nxi1 );
        
        delete vx0;
        vx0  = vx1;
		delete vxi0;
        vxi0 = vxi1;
		delete vp0;
        vp0  = vp1;
        
		k++;
	}

	*nxi = nxi1;
	
    double t;
    double xz;
    double nz;
    
    xz = nz = 0;
		
	for ( i = 0; i < SIZE; i++ )
    {
        t = vx0->data[i];
		nz += t * t;
	}
    
    nz = sqrt( nz );
		
	for ( i = 0; i < SIZE; i++ )
    {
		x->data[i] = vx0->data[i] / nz;
		xz += vx0->data[i] * x0->data[i];
    }

	*result = 1 / xz;
    
	delete vp0;
	delete vxi0;
	delete vx0;
}

// main

void main_loop( double *res )
{
	int           i;
    struct matrix tptr a;
	struct vector tptr x;
	struct vector tptr z;

	a = new tval struct matrix;
	x = new tval struct vector;
	z = new tval struct vector;

	make_a( a );
	make_1( x );

	for ( i = 0; i < NITER; i++ )
    {
		double	      res1;
		double	      nxi;
        struct vector tptr t;

		cg( a, x, (i)?1.0:SIZE, z, &nxi, &res1 );

		*res = SHIFT + res1;		        
		t = x;
        x = z;
        z = t;
        printf( "%3d %25.19e %19.17f\n", i, sqrt( nxi ), *res );
	}

    delete a;
	delete z;
	delete x;
}

// t-main

tfun int main( int argc,
               char *argv[] )
{
	double res;
    
	main_loop( &res );
    printf( "%25.19e %25.19e\n", res, fabs( res - ZETA ) );
	
	return 0;
}
