/************************************************************
*    perceptron based classifier 
*
*       $Id: nnet.c,v 3.1 2003/06/28 18:54:03 school Exp school $
*            
*       OS : linux
* COMPILER : GNU project C and C++ Compiler (gcc-2.95.3)
*
*************************************************************
*
*  AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://mechanoid.narod.ru
*  e-mail : periodical@ukrpost.net
* 
************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/**********************************************/

#define uint     unsigned int 
#define weight_t double
#define data_t   char

#define T  0
#define LO -1
#define HI +1

// teaching speed
#define ETA 0.001

// max net error
#define MAXERR 0.0001

//max number iteration of teaching
#define MAX_ITER 555

#define IN_SIZE  35
#define OUT_SIZE 10

#include	"patterns.h"

data_t    I[IN_SIZE]; // inputs  
weight_t  S[OUT_SIZE]; // states 
data_t    O[OUT_SIZE]; // outputs 
data_t   IDO[OUT_SIZE]; // ideal outputs 
weight_t  E[OUT_SIZE]; // error vector
weight_t  net_err=0; // error of net

weight_t W[OUT_SIZE][IN_SIZE]; // weight matrix

/**********************************************/

void initWeight();
void printOut();
void printIn();
void step(); // one step of net

data_t   f(weight_t); // activation func 
weight_t s(uint); // state of neuron  
void setPattern(uint); 
void teach(); 

/*****************************/
void initWeight() {
	uint i=0;
	uint j=0;

   srand(4711);
   for(j=0;j<OUT_SIZE;j++)  {   
	  for(i=0;i<IN_SIZE;i++) {	
		 W[j][i]=sin(rand())/2.0;
	  }
   }
}

/**********************************************/
void step() {
	uint i=0;

	for(i=0;i<OUT_SIZE;i++)	{
	   S[i]=s(i); 
	   O[i]=f(S[i]); 
	}
}

/**********************************************/
data_t f(weight_t s) {
   return (s<T)?LO:HI; 
}

/**********************************************/
weight_t s(uint j) {
	weight_t sum=0.0;
   uint     i=0;
	
	for(i=0;i<IN_SIZE;i++) { sum+=I[i]*W[j][i]; }
	
	return sum;
}

/*************************/

#define GPRINT 
#ifndef GPRINT

void printOut() {
   uint i=0;
	
	printf("O => ");	  
   for(i=0;i<OUT_SIZE;i++) { printf("%d ",O[i]); }
	printf("\n");
}

void printIn() {
   uint i=0;
	
	printf("I => ");	  
   for(i=0;i<IN_SIZE;i++) { printf("%d ",I[i]); }
	printf("\n");
}

#else 

void printIn() {
   uint i=0;
   uint j=0;
	
   for(j=0;j<Y;j++) {
	  for(i=0;i<X;i++) {
		 printf("%c",((I[j*X+i]==HI)?MARK:NOMARK));
	  }
	  printf("\n");
   }
}

void printOut() {
   uint j=0;

   for(j=0;j<OUT_SIZE;j++) {
	  if(O[j]==HI) {
		 printf("->%d\n",j);
		 return;
	  }
   }
   printf("unknown\n");
}
#endif

/**********************************************/
void setPattern(uint n) {
   uint j=0;
   uint i=0;
			  
   if(n>(N-1)){ return;}
   
   for(j=0;j<Y;j++) {
	  for(i=0;i<X;i++) {
		 I[i+j*X]=(PATTERNS[n][j][i]==MARK)?HI:LO;
	  }
   }
   
  for(j=0;j<OUT_SIZE;j++){ IDO[j]=TARGET[n][j]; }
}

/**********************************************/
void outError() {
   uint j=0;

   net_err=0.0;
   for(j=0;j<OUT_SIZE;j++) {
	  E[j]=(weight_t)IDO[j]-S[j];
	  net_err+=0.5*(E[j]*E[j]);
   }
}

/**********************************************/
void teach() {
   uint n=0;
   uint m=0;
   uint i=0; 
   uint j=0; 
   uint iter=0; 
   weight_t err=0.0;  
	   
   for(iter=0;iter<MAX_ITER;iter++)  {
	  printf(".");	  
	  ///////////////
	  // calculate error of net on all patterns		  
	  err=0.0;
	  for(n=0;n<N;n++) {
		 setPattern(n);
		 step();
		 outError();
		 err=(net_err>err)?net_err:err; //max error
	  }
	  //
	  ////////////////

	  if(err<=MAXERR) { break; }
	
	  //////////////////////
	  // corect weight
	  for(m=0;m<N*10;m++)  {
		 // random from [0,N-1]
		 n=(uint)(((float)rand()/RAND_MAX)*(N));
		 setPattern(n);
		 step();
		 outError();
		 
		 for(j=0; j < OUT_SIZE; j++) {
			for(i=0 ; i < IN_SIZE; i++) {
				W[j][i]+= ETA*I[i]*E[j]; 
	      }
	    }
	  }
  	  //
	  /////////////////////
   }
}

/**********************************************
 **********************************************/
int main(int argc, char *argv[]) {
	uint i=0;

	printf("init\n");	  
	
   initWeight();
	printf("teach ");	  
	teach();
	printf(" ok\n\n");	  

	for(i=0;i<N;i++) {
		setPattern(i);
		printIn();
		step();
		printOut();
	   printf("\n");	  
	}
	
	return 0;
}
