/************************************************************
*    perceptron based classificator 
*    ( see about neural net on http://mechanoid.narod.ru )
*
*     FILE : nnet.tc
*  CREATED : Thu Jul 10 03:45:52 EDT 2003
*       OS : linux
* COMPILER : T-System (GRACE)  PSI RAS Russia
*
*************************************************************
*
*  AUTHOR : Evgeny S. Borisov
*
*  Glushkov Institute of Cybernetics
*  National Academy of Sciences of Ukraine
* 
*    http://mechanoid.narod.ru
*  e-mail : periodical@ukrpost.net
* 
************************************************************/
#include <txx>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/**********************************************/

#define T  0
#define LO -1
#define HI +1

// teaching speed
#define ETA 0.001

// max net error
#define MAXERR 0.0001

//max number iteration of teaching
#define MAX_ITER 555

#define IN_SIZE  35
#define OUT_SIZE 10

#include	"patterns.h"

char    I[IN_SIZE]; // inputs  
double  S[OUT_SIZE]; // states 
char    O[OUT_SIZE]; // outputs 
char    IDO[OUT_SIZE]; // ideal outputs 
double  E[OUT_SIZE]; // error vector
double  net_err=0; // error of net

double W[OUT_SIZE][IN_SIZE]; // weight matrix

/**********************************************/

void initWeight();
void printOut();
void printIn();
void step(); // one step of net

char   f(double); // activation func 
double s(unsigned int); // state of neuron  
void setPattern(unsigned int); 
void teach(); 

/*****************************/
void initWeight() {
	unsigned int i=0;
	unsigned int j=0;

   srand(4711);
   for(j=0;j<OUT_SIZE;j++)  {   
	  for(i=0;i<IN_SIZE;i++) {	
		 W[j][i]=sin(rand())/2.0;
	  }
   }
}

// T-PARALLEL +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
tfun void out(unsigned int i) {
   S[i]=s(i); 
   O[i]=f(S[i]); 
}

/**********************************************/
void step() {
	unsigned int i=0;

	for(i=0;i<OUT_SIZE;i++)	{ out(i); }
}

/**********************************************/

char f(double s) { return (s<T)?LO:HI; }

/**********************************************/
double s(unsigned int j) {
	double sum=0.0;
   unsigned int     i=0;
	
	for(i=0;i<IN_SIZE;i++) { sum+=I[i]*W[j][i]; }
	
	return sum;
}

/*************************/

void printIn() {
   unsigned int i=0;
   unsigned int j=0;
	
   for(j=0;j<Y;j++) {
	  for(i=0;i<X;i++) {
		 printf("%c",((I[j*X+i]==HI)?MARK:NOMARK));
	  }
	  printf("\n");
   }
}

void printOut() {
   unsigned int j=0;

   for(j=0;j<OUT_SIZE;j++) {
	  if(O[j]==HI) {
		 printf("->%d\n",j);
		 return;
	  }
   }
   printf("unknown\n");
}

/**********************************************/
void setPattern(unsigned int n) {
   unsigned int j=0;
   unsigned int i=0;
			  
   if(n>(N-1)){ return;}
   
   for(j=0;j<Y;j++) {
	  for(i=0;i<X;i++) {
		 I[i+j*X]=(PATTERNS[n][j][i]==MARK)?HI:LO;
	  }
   }
   
  for(j=0;j<OUT_SIZE;j++){ IDO[j]=TARGET[n][j]; }
}

/**********************************************/
void outError() {
   unsigned int j=0;

   net_err=0.0;
   for(j=0;j<OUT_SIZE;j++) {
	  E[j]=(double)IDO[j]-S[j];
	  net_err+=0.5*(E[j]*E[j]);
   }
}

/**********************************************/
void teach() {
   unsigned int n=0;
   unsigned int m=0;
   unsigned int i=0; 
   unsigned int j=0; 
   unsigned int iter=0; 
   double err=0.0;  
	   
   for(iter=0;iter<MAX_ITER;iter++)  {
	  ///////////////
	  // calculate error of net on all patterns		  
	  err=0.0;
	  for(n=0;n<N;n++) {
		 setPattern(n);
		 step();
		 outError();
		 err=(net_err>err)?net_err:err; //max error
	  }
	  //
	  ////////////////

	  if(err<=MAXERR) { break; }
	
	  //////////////////////
	  // corect weight
	  for(m=0;m<N*10;m++)  {
		 // random from [0,N-1]
		 n=(unsigned int)(((float)rand()/RAND_MAX)*(N));
		 setPattern(n);
		 step();
		 outError();
		 
		 for(j=0; j < OUT_SIZE; j++) {
			for(i=0 ; i < IN_SIZE; i++) {
				W[j][i]+= ETA*I[i]*E[j]; 
	      }
	    }
	  }
  	  //
	  /////////////////////
   }
}

/**********************************************/
tfun int main (int argc, char *argv[]) {
	unsigned int i=0;

	printf("init\n");	  
	
   initWeight();
	printf("teach ");	  
	teach();
	printf(" ok\n\n");	  

	for(i=0;i<N;i++) {
		setPattern(i);
		printIn();
		step();
		printOut();
	   printf("\n");	  
	}
	
	return 0;
}
