#include "output.h"
#include "fractal.h"

#define MAX_JOB_SIZE 256

int fragment (struct rect r, struct pixel *pxl, struct params *par)
{
	int x, y;
	
	for (y = 0; y < r.height; y++) {
		for (x = 0; x < r.width; x++) {
			*(pxl++) = render (r.left+x, r.top+y, par);
		}
	}
	return 1;
}

struct out_data {
	struct rect rect;
	struct pixel px[MAX_JOB_SIZE];
};

struct tree {
	tptr struct tree left;
	tptr struct tree right;
	tptr struct out_data od;
};

tfun struct out_data do_out (struct rect r, struct params par)
{
	struct out_data od;
			
	od.rect = r;
	fragment(r, od.px, &par);

	return od;
}

tfun tree h_tree (struct rect r, struct params par)
{
	struct tree res;

	enum dmeth {
		no_div,
		vert_div,
		hor_div,
	} div_method;

//	fprintf(stderr, "[%d] (%d,%d)-(%d,%d)\n", ts::myRank, r.left, r.top, r.width, r.height);
	
	if (r.width*r.height <= MAX_JOB_SIZE) div_method = no_div;
	else div_method = (r.width > r.height)?hor_div:vert_div;
	
	switch (div_method) {
	case hor_div:
		{
			tval tree t1;
			tval tree t2;
			struct rect r1 = r;
			struct rect r2 = r;
			int w2 = r.width/2;
		
			r1.width = w2;
		
			r2.width -= w2;
			r2.left += w2;
		
			t1 = h_tree(r1,par);
			res.left = &t1;
			t2 = h_tree(r2,par);
			res.right = &t2;
			res.od = NULL;
		
			return res;
		}
	case vert_div:
		{
			tval tree t1;
			tval tree t2;
			struct rect r1 = r;
			struct rect r2 = r;
			int h2 = r.height/2;
		
			r1.height = h2;
		
			r2.height -= h2;
			r2.top += h2;
			
			t1 = h_tree(r1,par);
			res.left = &t1;
			t2 = h_tree(r2,par);
			res.right = &t2;
			res.od = NULL;
		
			return res;
		}
	case no_div:
		{
			tval struct out_data pd;

			pd = do_out (r, par);
			
			res.left = NULL;
			res.right = NULL;
			res.od = &pd;
			
			return res;
		}
	default:
		fprintf(stderr, "BUG\n");
	}
	return 0;
}

static struct out_descr desc;

int consumer (struct tree *t)
{
	while (t) {
		struct out_data *p;
		
		p = t->od;
		if (p) {
			out_fragment (&desc, p->rect, p->px);
		}
		
		consumer(t->left);
		t = t->right;
	}
	return 0;
}

tfun int main (int argc , char *argv[])
{
	int sk = 0;
	struct rect r;
	r.left = 0;
	r.top = 0;
	r.width = 512;
	r.height = 512;
	char fname[256] = "fr";

	int frames = 10;
	int i;
	
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i],"-width")) {
			r.width = atoi(argv[++i]);
		} else
		if (!strcmp(argv[i],"-height")) {
			r.height = atoi(argv[++i]);
		} else
		if (!strcmp(argv[i],"-frames")) {
			frames = atoi(argv[++i]);
		} else
		if (!strcmp(argv[i],"-sock")) {
			strncpy(fname, argv[++i], 256);
			sk = 1;
		} else
		if (!strcmp(argv[i],"-output")) {
			strncpy(fname, argv[++i], 256);
			sk = 0;
		} else
		if (!strcmp(argv[i],"--")) {
			i++;
			break;
		} else break;
	}

	if (!sk) out_open (&desc, fname, out_type_tga, r.width, r.height, frames);
	else out_open (&desc, fname, out_type_sock, r.width, r.height, frames);
	for (int frame = 0; frame < frames; frame++) {
		struct params par;
		struct tree t;
		
		make_params(&par, r, ((double)frame)/(frames-1));
		t = h_tree (r,par);
		consumer (&t);
		out_next(&desc);
	}
	
	return 0;
}
