#include <stdio.h>
#include "fractal.h"

#define RADIUS 2.
#define PART_SIZE 80
#define PARTS_IN_SIDE 8

#define POINTS_IN_SIDE (PARTS_IN_SIDE*PART_SIZE)
#define TOTAL_PARTS (PARTS_IN_SIDE*PARTS_IN_SIDE)
#define TOTAL_POINTS (TOTAL_PARTS*PART_SIZE*PART_SIZE)

#define PART_STEP (2*RADIUS/PARTS_IN_SIDE)
#define BEGIN     (-RADIUS)
#define STEP      (PART_STEP/PART_SIZE)

#define MAX_ITERATION    4096

#define RE(x)   (__real__ (x)) // real part
#define IM(x)   (__imag__ (x)) // image part

#define CC(x,y) ((x) + (y)*1i) // complex constant

typedef __complex__ float Complex;

static inline Complex f(Complex _z, Complex _param) {
	static const double t = 0.1;
	Complex rot = CC(cos(t),sin(t));
	Complex z2, z5;
	_z *= rot;
	z2 = _z * _z;
	z5 = z2 * z2 * _z;
	return _z -
		_param * 0.015 * (z5 * z2 + 1./z5 + _z + .1/_z);

	//0.1*_param * CC(cos(5*RE(_z*_z)), sin(3*IM(_z*_z)));
	//1.3*_param * CC(cos(1*RE(_z*_z*_z*_z)), IM(_z*_z));
	//0.2 * _param * CC(RE(_z*_z*_z*_z), IM(_z*_z*_z));
}

static int trace (Complex z0, Complex param)
{
	int iter = 0;
	Complex z = z0;
	do {
		if (iter >= MAX_ITERATION) break;
		z = f(z,param);
		iter++;
	} while (RE(z)*RE(z) + IM(z)*IM(z) < RADIUS*RADIUS);
	return iter;
}

struct pixel render (int x, int y, struct params *par)
{
	struct pixel px;
	int iter;
	
	x -= par->xoffs;
	y -= par->yoffs;
	
	iter = trace (CC(x,y)*(par->scale), CC(par->re,par->im));

#if 1
	{
		px.r = (iter & 0xf00) >> 4;
		px.g = (iter & 0x0f0);
		px.b = (iter & 0x00f) << 4;
	}
#else
	px.r = (iter & 0xf00) >> 4;
	px.g = (iter & 0x0f0);
	px.b = (iter & 0x00f) << 4;
#endif
	
	return px;
}


int make_params (struct params *par, struct rect r, double tm)
{
	par->re = -(1-tm)*(1-tm);
	par->im = 0.0;
	par->scale = 4.0/r.width;
	par->xoffs = r.left+r.width/2;
	par->yoffs = r.top+r.height/2;
}
