//
// This implements allocators as specified in the C++ standard.  
//
// Note that standard-conforming allocators use many language features
// that are not yet widely implemented.  In particular, they rely on
// member templates, partial specialization, partial ordering of function
// templates, the typename keyword, and the use of the template keyword
// to refer to a template member of a dependent type.
//


#ifndef __T_ALLOCATOR_H__
#define __T_ALLOCATOR_H__

#ifndef GRACE
#define t_allocator std::allocator

#else

#define __TA_NOTHROW throw ()

template <class _Tp>
class t_allocator {
public:
  typedef size_t     size_type;
  typedef ptrdiff_t  difference_type;
  typedef _Tp*       pointer;
  typedef const _Tp* const_pointer;
  typedef _Tp&       reference;
  typedef const _Tp& const_reference;
  typedef _Tp        value_type;

  template <class _Tp1> struct rebind {
    typedef t_allocator<_Tp1> other;
  };

  t_allocator() __TA_NOTHROW {}
  t_allocator(const t_allocator&) __TA_NOTHROW {}
  template <class _Tp1> t_allocator(const t_allocator<_Tp1>&) __TA_NOTHROW {}
  ~t_allocator() __TA_NOTHROW {}

  pointer address(reference __x) const { return &__x; }
  const_pointer address(const_reference __x) const { return &__x; }

  // __n is permitted to be 0.  The C++ standard says nothing about what
  // the return value is when __n == 0.
  static _Tp* allocate(size_type __n, const void* = 0) {
    //return __n != 0 ? static_cast<_Tp*>(_Alloc::allocate(__n * sizeof(_Tp))) 
    //                : 0;
    return __n != 0 ? static_cast<_Tp*>(talloc(__n * sizeof(_Tp))) 
                    : 0;
  }

  // __p is not permitted to be a null pointer.
  static void deallocate(pointer __p, size_type __n)
     { tfree(__p); }
    // { _Alloc::deallocate(__p, __n * sizeof(_Tp)); }

  size_type max_size() const __TA_NOTHROW 
    { return size_t(-1) / sizeof(_Tp); }

  void construct(pointer __p, const _Tp& __val) { new(__p) _Tp(__val); }
  void destroy(pointer __p) { __p->~_Tp(); }
};

template<>
class t_allocator<void> {
public:
  typedef size_t      size_type;
  typedef ptrdiff_t   difference_type;
  typedef void*       pointer;
  typedef const void* const_pointer;
  typedef void        value_type;

  template <class _Tp1> struct rebind {
    typedef t_allocator<_Tp1> other;
  };
};


template <class _T1, class _T2>
inline bool operator==(const t_allocator<_T1>&, const t_allocator<_T2>&) 
{
  return true;
}

template <class _T1, class _T2>
inline bool operator!=(const t_allocator<_T1>&, const t_allocator<_T2>&)
{
  return false;
}

#endif /* GRACE */
#endif
