#include <automat.h>

// ----------------------------------------------------------------------------

long add_loc_code(LocDictionary &LDict, string w)
{
  for(LocDictionary::iterator it = LDict.begin(); it != LDict.end(); it ++)
    if(it->word == w) return it->code;

  LocCode LC;
  LC.code = LDict.size() + 1;
  LC.word = w;

  LDict.insert(LDict.end(), LC);

  return LC.code;
}

// ----------------------------------------------------------------------------

Token::Token()
{
}

Token::Token(TokenType t, const char *v):
  type(t), value(v)
{
}

Token &Token::operator= (const Token &t)
{
  type = t.type; value = t.value; return *this;
}

bool Token::operator== (const Token &t) const
{
  return (type == t.type && value == t.value);
}

bool Token::operator!= (const Token &t) const
{
  return !(*this == t);
}

Lex::Lex(const char *str): 
  buffer(str), 
  cursor(0), 
  star(STAR,""), 
  plus(PLUS,""), 
  question(QUESTION,""), 
  verticalbar(VERTICALBAR,""), 
  left(LEFT,""), 
  right(RIGHT,""), 
  end(END,"")
{
}

const Token Lex::current_lexem()
{
  if(cursor >= buffer.length()) 
    return end;
  else {
    switch(buffer[cursor]) {
      case '*':
        return star;
      case '+':
        return plus;
      case '?':
        return question;
      case '|':
        return verticalbar;
      case '(':
        return left;
      case ')':
        return right;
      case '"':
        {
          string Str("");
          char str[2];
          str[1] = '\0';
	  cursor ++;
          while( cursor < buffer.length() ) 
            if(buffer[cursor] == '"')
              break;
            else {	    
              str[0] = buffer[cursor];
	      Str += str;
	      cursor ++;
	    }
          return Token(CHAR,Str.c_str());	  	  
	}
      default:
	char str[2];
	str[0] = buffer[cursor];
	str[1] = '\0';
        return Token(CHAR,str);	    
    }	  
  }	
}

//#ifdef DEBUG
//ostream &operator<< (ostream &s, const Token &t)
//{
//  return s << t.type << " " << t.value; 
//}	
//#endif // end of DEBUG
	
const Token Lex::push_lexem()
{
  Token t = current_lexem(); cursor ++; return t;
}	
	
Parser::Parser(const char *str, GenerateAutomat &ga):
  Lex(str), generate_automat(&ga)
{
}

//
// atom ::= CHAR | '(' regular ')'
//

void Parser::treat_atom(RegExpNFA_ptr result)
{
  Token t = push_lexem();

  if(t == Token(LEFT, "")) {
      treat_regular(result);

      if(push_lexem() != Token(RIGHT, ""))
        throw 1;
  }
  else
  if(t.type == CHAR) {
    // result =  new RegExpNFA("");
    result->addState(1);
    result->addState(2);
    result->addTransition(1,2,t.value.c_str());
    result->addStartState(1);
    result->addFinalState(2);
  }
  else
    throw 1;
}

//
// Atom ::= atom | atom '*' 
//

void Parser::treat_Atom(RegExpNFA_ptr result)
{
  treat_atom(result);

  if(current_lexem() == Token(STAR, "")) {
    push_lexem();
    result->iterate();
  }
}	

//
// sequence ::= Atom | Atom sequence
//
	
void Parser::treat_sequence(RegExpNFA_ptr result)
{
  treat_Atom(result);

  if(current_lexem() == Token(LEFT, "") | current_lexem().type == CHAR) {

    StrNFA r("");
    treat_sequence(&r);
    result->concat(r);
  }	     
}

//
// regular ::= sequence | sequence '|' sequence
//

void Parser::treat_regular(RegExpNFA_ptr result)
{
  treat_sequence(result);

  if(current_lexem() == Token(VERTICALBAR, "")) {
    push_lexem();
    StrNFA r("");
    treat_sequence(&r);
    result->alternate(r);
  }	  
}	

void Parser::start() throw(Parser::Exception)

  // main ::= regular 

{
  StrNFA r("");
  treat_regular(&r);
  generate_automat->set_result(r);
}	

GenerateAutomat::GenerateAutomat(const char* reg_expr):
  error_(false), M_("")
{
  Parser P(reg_expr, *this);

  try {
    P.start();
  }
  catch(const Parser::Exception &) {
    cerr << "syntax error\n";
  }
}


