// ----------------------------------------------------------------------------
//    
// ----------------------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <stdlib.h>

using namespace std;

struct Label {
  string label;
  double min;
  double max;

  Label(string l, double mi, double ma): label(l), min(mi), max(ma) {}
};	

typedef list<Label> Labels;

static double d_get_rand(double minval, double maxval)
{
  return minval + (maxval - minval)*((double) rand()/RAND_MAX);
}	

static long get_rand(long minval, long maxval)
{
  return minval + (long)((maxval - minval)*((double) rand()/RAND_MAX));
}

int main (int argc, char* argv[])
{
  if(argc < 7) {
    cerr <<  "Usage:" 
	 << endl
	 << argv[0] 
	 << " <srandom>"
	 << " <output_file>" 
	 << " <vertex_size>" 
	 << " <min_vertex_degree>" 
	 << " <max_vertex_degree>" 
	 << " <label>" 
	 << " <probability>" 
	 << " ..." 
	 << endl;  
    return -1;	  
  }	  

  cout << "Random graph generation" << endl;

  unsigned int  _srandom = 0;
  string output;
  long _vertex_size;
  long _min_vertex_degree;
  long _max_vertex_degree;
  Labels L;
  double  P = 1.; 
  double _P = 0.;

  for(int i = 1; i < argc; i ++)
    if(i == 1) {
      _srandom = atoi(argv[i]);	  
    }
    else  
    if(i == 2) {
      output = argv[i];
    }
    else
    if(i == 3) {
      _vertex_size = atol(argv[i]);
    }
    else
    if(i == 4) {
      _min_vertex_degree = atol(argv[i]);
    }
    else
    if(i == 5) {
      _max_vertex_degree = atol(argv[i]);
    }
    else {
      string l = argv[i];
      i ++;
      double p = atof(argv[i]);
      p = (P > p) ? p : P;
      P -= p;	      
      _P += p;
      L.insert(L.end(), Label(l, _P - p, _P)); 
    }	    

  cout << "Srandom:           " << _srandom           << endl 
       << "Output file:       " << output             << endl
       << "Vertex size:       " << _vertex_size       << endl
       << "Min vertex degree: " << _min_vertex_degree << endl
       << "Max vertex degree: " << _max_vertex_degree << endl;
  
  srand( _srandom );
  
  for(Labels::const_iterator j = L.begin(); j != L.end(); j ++) {
    cout << "\t" << j->label << " " << j->min << " " << j->max << endl;
  }

  //
  //  
  //
    
  ofstream O(output.c_str());
  
  for(long f = 0; f < _vertex_size; f ++) {
  
    long f_degree = get_rand(_min_vertex_degree, _max_vertex_degree);
	 
    for(long i = 0; i < f_degree; i ++) {
    
      long t = get_rand(0, _vertex_size-1); 	    

      double r = d_get_rand(0., 1.);
     
      for(Labels::iterator j = L.begin(); j != L.end(); j ++)
        if(r >= j->min && r <= j->max) {
		
          O << f << " " << t << " " << j->label << endl;    
	  break;	
        }		
    }	    
  }	  
	  
  O.close();
  
  return 0;	
}	
