#include <automat.h>

int main (int argc, char* argv[])
{
  //
  //   	
  // 

  if(argc != 5) {
    cerr << "Usage:" 
	 << endl
	 << argv[0] 
	 << " <db_file_in> <q_file_in> <db_file_out> <q_file_out>" 
	 << endl;  
    return -1; 
  }	  

  string db_in = argv[1], 
          q_in = argv[2], 
	db_out = argv[3], 
         q_out = argv[4];

  //
  //   
  //

  Graph DB;
	 
  LocDictionary LDict;

  ifstream DI( db_in.c_str() );

  while( !DI.eof() ) {
    long f, 
         t;
    string l;

    DI >> f >> t >> l;
     
    if( !DI.eof() ) {
      DB.addVertex(f, 0L);
      DB.addVertex(t, 0L);
      DB.addEdge(f, t, add_loc_code(LDict, l));
    }
  }	  
  
  DI.close();
 
  //
  //   
  //

  ofstream DO( db_out.c_str() );
 
  DO << DB.getEdges().size() << endl;

  Graph::Edges::const_iterator i;
  for(i = DB.getEdges().begin(); i != DB.getEdges().end(); i ++) 
    DO << i->first << " " << i->second.to << " " << i->second.label << endl;

  DO.close();
  
  //
  //  
  //

  Query Q;
  
  ifstream QI( q_in.c_str() );
  
  while( !QI.eof() ) {
    QueryEdge e;
	  
    QI >> e.from >> e.to >> e.label;

    if( !QI.eof() ) 
      Q.insert(Q.end(),  e );   
  }
  
  QI.close();
  
  //
  //  
  //

  ofstream QO( q_out.c_str() );

  QO << Q.size() << endl;

  for(Query::const_iterator i = Q.begin(); i != Q.end(); i ++) {
	  
    QO << i->from 
       << " " 
       << i->to 
       << endl;

    StrNFA A = GenerateAutomat(i->label.c_str()).result();

    // epsilon
    
    QO << "\t" 
       << add_loc_code(LDict, A.epsilon())	    
       << endl; 
    
    // transitions
  
    AttGraph<long,long,string>::Edges E = A.getTransitions();
    
    QO << "\t"
       << E.size()
       << endl;
   
    AttGraph<long,long,string>::Edges::const_iterator j;
    
    for(j = E.begin(); j != E.end(); j ++) {
      QO << "\t\t"
         << j->first
	 << " "
	 << j->second.to
	 << " "
	 << add_loc_code(LDict, j->second.label)
         << endl;	      
    }	    
   
    // start states

    set<long> SS = A.getStartStates();
    
    QO << "\t"
       << SS.size()
       << endl;
   
    for(set<long>::const_iterator j = SS.begin(); j != SS.end(); j ++)
      QO << "\t\t"
	 << (*j)   
	 << endl;     
	 
    // final states
    
    set<long> FS = A.getFinalStates();

    QO << "\t"
       << FS.size()
       << endl;
    
    for(set<long>::const_iterator j = FS.begin(); j != FS.end(); j ++)
      QO << "\t\t"
	 << (*j)   
	 << endl;     
  }
  
  QO.close();
	 
  return 0;
}	
