#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "consts.h"
#include "utils.h"

/*Подпрограмма вычисления перепада давлений и коэффициентов нормальной силы C 
и азимутального момента AMZ.*/
void aer(
//input
double *ga1,	//старые циркуляции на цилиндре
double *gak1,	//старые циркуляции на пластинах
double *ga,	//циркуляция вихрей на цилиндре
double *gak,	//циркуляция вихрей на пластины
//double *xt,
//double *yt,	//контрольные точки на цилиндре
double *xb,
double *yb,	//вихри на цилиндре
double *xtk,
double *ytk,	//контрольные точки на пластине
double *xbk,
double *ybk,	//вихри на пластине
double *x1,
double *y1,	//координаты свободных вихрей
double q,
double *dg,	//циркуляции свободных вихрей
int ir,		//текущий шаг
double *al,	//углы пластин
double aldt,	//угловая скорость
double ta,
double *p,
double *pk,	//вектора перепада давления (для внутреннего использования)
//output
double *c,
double *amz,
double *cx,	//коэффициент сопротивления 
double *cy,	//коэффициент подъемной силы
double *sum31	//суммарное изменение циркуляции
)
{
 double gg1,gas1,wn,wt,wte,ytl;
 int ip,im,ie,ic;//,is;
 double vort[FOUR], vortsum;
 double apnew,apold,dga,ron;
 int ig;

#ifdef VERB
 int i,j,k;
  for (i=0;i<FOUR;i++)
   for (j=0;j<NK;j++)
   {
    printf("gak1[%d,%d]=%lf\n",i,j,gak1[K(i,j)]);
    printf("gak[%d,%d]=%lf\n",i,j,gak[K(i,j)]);
    printf("xtk[%d,%d]=%lf\n",i,j,xtk[K(i,j)]);
    printf("ytk[%d,%d]=%lf\n",i,j,ytk[K(i,j)]);
    printf("xbk[%d,%d]=%lf\n",i,j,xbk[K(i,j)]);
    printf("ybk[%d,%d]=%lf\n",i,j,ybk[K(i,j)]);
   }
 for (i=0;i<N;i++)
 {
  printf("ga[%d]=%lf\n",i,ga[i]);
  printf("xb[%d]=%lf\n",i,xb[i]);
  printf("yb[%d]=%lf\n",i,yb[i]);
 }

 printf("q=%lf\nir=%d\nta=%lf\n",q,ir,ta);

 for (i=0;i<FOUR;i++)
  printf("al[%d]=%lf\n",i,al[i]);
 for (i=0;i<FOUR;i++)
  for (j=0;j<NC;j++)
   for (k=0;k<=ir;k++)
    {
     printf("x1[%d,%d,%d]=%lf\n",i,j,k,x1[F(i,j,k)]);
     printf("y1[%d,%d,%d]=%lf\n",i,j,k,y1[F(i,j,k)]);
     printf("dg[%d,%d,%d]=%lf\n",i,j,k,dg[F(i,j,k)]);
    }

#endif

 for (ip=0;ip<FOUR;ip++)//1
 {
//обнуление коэффициентов
  c[ip]=0;
  amz[ip]=0;
//цикл по всем контрольным точкам на пластине
 for (ie=0;ie<NK;ie++)//10
  {
   /*Вычисление изменения циркуляции по замкнутому контуру, охватывающему
   все свободные вихри, сошедшие с кромки пластины, и 
   все присоединенные вихри на пластине с номерами,
   большеми, чем IE */
   gg1=0;
    for (ic=0;ic<NC;ic++)//11
     {
	gg1-=dg[F(ip,ic,ir)];//Другая индексация
     }//11
/*
   gl1=0;
    for (ic=0;ic<NC;ic++)//20
     for (is=0;is<=ir;is++) 
	gl1-=dg[F(ip,ic,is)];//20
*/
   for (im=ie;im<NK-1;im++)//30
    {
     gg1+=-gak[K(ip,im)]+gak1[K(ip,im)];
//     gl1+=-gak[K(ip,im)];
    }//30
/*Вычисление вектора скорости потока в контрольной точке с номером ie
на пластине */
  ospeed(xtk[K(ip,ie)],ytk[K(ip,ie)],
    xb,yb,xbk,ybk,x1,y1,ga,gak,dg,ta,ir,q,
    &wn,&wt);
//Вычисление средней циркуляции в той же контрольной точке
  if (0==ie)
   gas1=-gak[K(ip,0)];
  else
  if (NK-1==ie)
   gas1=-gak[K(ip,NK-2)];
  else
//  if ((ie>0)&&(ie<NK-1))
   gas1=-(gak[K(ip,ie)]+gak[K(ip,ie-1)])/2;
  wte=wn*cos(al[ip])+wt*sin(al[ip]);
/*Перепад давлений в конторльной точке с номером ie на пластине*/
  pk[K(ip,ie)]=-1./DT*(wte*gas1*DT*N/(2*M_PI*R)-gg1);
   if (0==ie) vort[ip]=-gg1;
 }//10
/*Вычисление коэффициентов нормальной силы C и азимутального момента AMZ путем
интегрирования перепада давлений на пластине*/
 for (ie=0;ie<NK-1;ie++) //110
 {
  c[ip]+=pk[K(ip,ie)]/2*DY;
  ytl=sqrt(xtk[K(ip,ie)]*xtk[K(ip,ie)]+ytk[K(ip,ie)]*ytk[K(ip,ie)]);
  amz[ip]+=pk[K(ip,ie)]/2.*DY*ytl;
 }//110
 c[ip]+=pk[K(ip,NK-1)]/4.*DY;
 ytl=sqrt(xtk[K(ip,NK-1)]*xtk[K(ip,NK-1)]+ytk[K(ip,NK-1)]*ytk[K(ip,NK-1)]);
 amz[ip]+=pk[K(ip,NK-1)]/4.*DY*ytl;
 pk[K(ip,NK-1)]=0;
 c[ip]*=2;
 amz[ip]*=-2;
 }//1
//Вычисление давлений на поверхности цилиндра
vortsum=0;
 for (ie=0;ie<N;ie++)//3
  {
   if (0!=ie) 
    for (ip=0;ip<FOUR;ip++)//5
     {
      apnew=-xb[ie]*sin(al[ip])+yb[ie]*cos(al[ip]);
      apold=-xb[ie-1]*sin(al[ip])+yb[ie-1]*cos(al[ip]);
       if ((apnew>0)&&(apold<=0))
        vortsum+=vort[ip];
     }//5
   dga=0;
    for (ig=1;ig<=ie;ig++)//7
     dga+=ga[ig]-ga1[ig];//7
   p[ie]=ga[ie]*aldt*R/HS-(vortsum+dga)/DT;
   p[ie]+=-0.5*(ga[ie]/HS)*(ga[ie]/HS);
  }//3
//Вычисление коэффициентов сопротивления и подъемной силы
*cx=0;
*cy=0;
 for (ip=0;ip<FOUR;ip++)//13
  {
   (*cx)+=c[ip]*sin(al[ip]);
   (*cy)-=c[ip]*cos(al[ip]);
  }//13
 for (ie=0;ie<N;ie++)//17
  {
   ron=sqrt(xb[ie]*xb[ie]+yb[ie]*yb[ie]);
   (*cx)-=p[ie]*HS*xb[ie]/ron;
   (*cy)-=p[ie]*HS*yb[ie]/ron;
  }//17
//Вычисление суммарного изменения циркуляции
 *sum31=0;
  for (ig=0;ig<N;ig++)//19
   (*sum31)+=ga[ig]-ga1[ig];//19
  for (ip=0;ip<FOUR;ip++)//31
   (*sum31)+=vort[ip];//31
}//aer

void ospeed(
//input
double xx, double yy,	//координаты точки
double *xb, double *yb, //вихри на цилиндре
double *xbk, double *ybk,	//вихри на пластинах
double *x1, double *y1,	//свободные вихри
double *ga, double *gak, double *dg,	//циркуляции на цилиндре, пластине и свободные
double ta,	//текущее время
int ir,
double q,	//регуляризирующий источник
//output -- компоненты скорости
double *w1, double *w2)
{
 int ki,i;
 int k,ic;
 double wx,wy,ro2;

 *w1=0;
 *w2=0;
  for (ki=0;ki<FOUR;ki++)//1
   for (i=0;i<NK-1;i++)//2
    {
     speed(xbk[K(ki,i)],ybk[K(ki,i)],xx,yy,&wx,&wy);
     (*w1)+=wx*gak[K(ki,i)];
     (*w2)+=wy*gak[K(ki,i)];
    }//1 2
  for (k=0;k<N;k++)//10
   {
    speed(xb[k],yb[k],xx,yy,&wx,&wy);
    (*w1)+=wx*ga[k];
    (*w2)+=wy*ga[k];
   }//10
  for (k=0;k<FOUR;k++)
   for (ic=0;ic<NC;ic++)
    for (i=0;i<=ir;i++)//20
    {
#ifdef SPEEDF
     if (ir==i)
      speed(x1[F(k,ic,i)],y1[F(k,ic,i)],xx,yy,&wx,&wy);
     else
      speedf(x1[F(k,ic,i)],y1[F(k,ic,i)],xx,yy,&wx,&wy);
#endif
#ifndef SPEEDF
      speed(x1[F(k,ic,i)],y1[F(k,ic,i)],xx,yy,&wx,&wy);
#endif
     (*w1)+=wx*dg[F(k,ic,i)];
     (*w2)+=wy*dg[F(k,ic,i)];
    }//20
  ro2=xx*xx+yy*yy;
  (*w1)+=q*xx/(2*M_PI*ro2);
  (*w2)+=q*yy/(2*M_PI*ro2);
  (*w1)+=VEL(ta);
}//ospeed

//Скорость, индуцированная вихрем в точке
void speed(
double xb, double yb,	//вихрь
double xt, double yt,	//точка
double *wx, double *wy	//скорость
)
{
 double xm,ym,wxm,wym;
 double ro2=(xb-xt)*(xb-xt)+(yb-yt)*(yb-yt);
 double ro=sqrt(ro2);

  if (ro>EPS)
   speed2(xb,yb,xt,yt,wx,wy);
  else
  {
    if (ro>ZERO)
    {
     xm=xb+EPS/ro*(xt-xb);
     ym=yb+EPS/ro*(yt-yb);
     speed2(xb,yb,xm,ym,&wxm,&wym);
     *wx=ro/EPS*wxm;
     *wy=ro/EPS*wym;
    }
   else
    {
     *wx=*wy=0;
    }
  }
}//speed

//speedf == speed mod speed2f=speed2

//Скорость без зоны дискретности
void speed2(
double xb, double yb,	//вихрь
double xt, double yt,	//точка
double *wx, double *wy)	//скорость
{
 double ro2=(xt-xb)*(xt-xb)+(yt-yb)*(yt-yb);
 *wx=-(yt-yb)/(2*M_PI*ro2);
 *wy=(xt-xb)/(2*M_PI*ro2);
}//speed2

//Ошибка решения системы
double m_err(double *ans, double *matr, double *b, int size)
{
 int i,j;
 double err=0,tmp;
  for (i=0;i<size;i++)
   {
    tmp=0;
     for (j=0;j<size;j++)
      tmp+=ans[j]*matr[i+size*j];
    err+=fabs(tmp-b[i]);
   }
 return err;
}
