      			
		      /* fun.cpp */

#include "cav.h"
#include "timer.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

extern double *ka,*kb,*kc,*u;
extern unsigned N,K;
extern double ag[],x1[];
extern double sn,cn,m2pi,mpi_2;
extern int nL;

struct InverseMas {
#define MAX_N 1000000
	double* data;
	InverseMas() { 
		int i;
		if (!(data=(double*)malloc(MAX_N*sizeof(double)))) {
			printf("No memory\n");
			exit(1);
		}
		for (i=0;i<N;i++)
			data[i]=1/(double)(2*i);
	}
	~InverseMas() {
		free(data);
	}
} massiv;
                       /*  7  */
                    /* nrnz.cpp  */
static Timer tt("fun");

void nrnz(unsigned j, double x, double *nr, double *nz)
{
	double a,d;

//	tt.startT();
	if(j<K)  
		d=dr(j); 
	else 
		d=dz(j);
	a=atan((3.0*ka(j)*x*x+2.0*kb(j)*x+kc(j))/d);
	if(j<K) {
		*nr=-sin(a); *nz=cos(a);
	}
	else {
	   	*nr=-cos(a); *nz=sin(a);
	}
//	tt.stopT();
	     
}		    
	
			/*  8  */
		     /* kdL.cpp  */

double kdL(unsigned j, double x)
{
	double v,xL,p,d;
        int L;
//	tt.startT(1);
	if(j<K) 
		d=dr(j)*dr(j);
	else
	        d=dz(j)*dz(j);  
	for(L=0,p=0.0; L<nL; L++) {
		v=x1[L]*x;
	  	xL=3.0*v*v*ka(j)+2.0*v*kb(j)+kc(j);
 	        p+=ag[L]*sqrt(d+xL*xL);
	}
//	tt.stopT(1);
	return (x*p);
}

			/*  9  */
		     /*  ekm.cpp  */


double ekm(double m, double *em)
{
	double a,m1,k,p,j1,*p_inv;
	unsigned i,n;
	signed long j;
	
	p_inv=massiv.data;
	m1=sqrt(1.0-m);
	p=(1.0-m1)/(1.0+m1); 
	tt.startT();
	a=2.0-20.0/log10(p); 
	tt.stopT();
	n=(unsigned)a;
	a=0.25*p*p;
	*em=k=1.0+a;
	for(i=2; i<n; i++) {
		j=2*(signed long)i;// j1=(j-1)*(j-1);
		a*=(p*(*p_inv))*(p*(*p_inv));
  	        *em+=a;
		a*=j-1;
		a*=j-1;
		k+=a;
		p_inv++;
        }
	*em*=mpi_2/(1.0+p);
	return (k*mpi_2*(1.0+p));
}
	
	



			/*  10  */
         	      /*  om.cpp  */

void om(double *a, double *b, unsigned n)
{
 	unsigned L,j,i,k;
	double r,r1;
//	tt.startT(3);
	for(L=0; L<n; L++) {
	        r=0.0;
    		for(i=L; i<n; i++)
		        if(fabs(a(i,L)) > fabs(r)) { 
				r=a(i,L); 
				k=i; 
			}
		if(k !=L) {
			for(i=L; i<n; i++)  {
				r1=a(k,i)/r; 
				a(k,i)=a(L,i); 
				a(L,i)=r1;
			}
			r1=b(k)/r; 
			b(k)=b(L); 
			b(L)=r1;
		} else {
		        for(i=L; i<n; i++)  
				a(k,i)/=r;
		    	b(k)/=r;
		}
		for(i=0; i<n; i++)
			if(i!=L) {
				r=a(i,L);
				if(L<n-1) 
				for(j=L+1; j<n; j++) 
					a(i,j)-=a(L,j)*r;
				b(i)-=b(L)*r;
			}
	} 
//	tt.stopT(3);
}			
                   /*  11  */
		  /* fCx.cpp */

double fCx(double sgm)
{
	unsigned j;
        int L;
        double v,p,c,cx;

//	tt.startT(4);
	for(j=0, cx=0.0; j<N; j++) {
		for(L=0, p=0.0; L<nL; L++) {
			c=kdL(j,x1[L]);
			v=u(j)+(u(j+1)-u(j))*c/ds(j);
		        p+=(1.0-v*v)*(r(j)+dr(j)*x1[L])*ag[L];
	        }
	        cx+=p*dr(j);
	}
//	tt.stopT(4);
	return(2.0*cx+sgm);
}

  
	            /* 12 */
	           /* Gf1.cpp */
double Gf1(double x)
{
  	double a,b2,b3,b4,b6,b9,b11;

//	tt.startT(5);
	b2=x*x; 
	b3=x*b2; 
	b6=b3*b3; 
	b9=b3*b6;
        b4=b2*b2;
	b11=b2*b9;	
           
        a=-0.00000000304686671180843449547807046477
	  +0.000000318397571383981404777458363569*2.0*x    
	  -0.0000146462882836631446197630847*3.0*b2
          +0.0003734803512334101878039586604*4.0*b3
          -0.0059756856197345630048633385674*5.0*b4
          +0.0647365942137910992193528344808*6.0*b4*x
          -0.4993965839349599082635790088520*7.0*b6
          +2.8403180711300844782491056128400*8.0*b6*x
          -12.202848009299622202848009299600*9.0*b6*b2
          +40.26939843068875326939843068880*10.0*b9
          -103.1695331695331695331695331700*11.0*b9*x
          +206.3390663390663390663390663390*12.0*b11
	  -322.3284823284823284823284823280*13.0*b11*x
          +391.3988713988713988713988713990*14.0*b11*b2
          -365.3056133056133056133056133060*15.0*b11*b3
          +256.8555093555093555093555093560*16.0*b11*b4
          -131.5384615384615384615384615380*17.0*b11*b4*x
          +46.28205128205128205128205128210*18.0*b11*b6
	  -190.0*b9*b9
	  +20.0*b9*b9*x; 

//	tt.stopT(5);
	return a;
}

	              /* 13 */
	            /* ck.cpp */

double ck(double x)
{
        double a;
	a=23.0*29.0*31.0*132.0*Gf1(x);
	a*=35.0*37.0*39.0;
	return(1.0/(a*a*x*(1.0-x)));
}
