#ifndef __TIMER_H__
#define __TIMER_H__

//#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CPU_FREQUENCY 1009021000
#define UINT32_MAX 4294967295U

class KTimer {
	
		class String {
				char* data;
			public:
				String(char* string=NULL)
				{
					data=NULL;
					initString(string);
				}
				void initString(char* string=NULL)
				{
					int length;
					if (data!=NULL)
						delete[] data;
					length = (string==NULL) ? strlen("default string")+1 : strlen(string)+1;
					data = new char[length]; 
					(string==NULL) ? memcpy(data,"default string",length) : memcpy(data,string,length);
				}
				char* operator () () const {
					return data;
				}
				~String()
				{
					delete[] data;
				}
		};
		typedef struct _TimeData {
			unsigned long long sec,tmp_sec;
			unsigned long long step_no;
			String description;
		} TimeData;

#define TIME_SIZE 10
	
		struct CpuInfo {
			unsigned long long frequency;
		};

		CpuInfo cpuInfo;
		TimeData data[TIME_SIZE];
		KTimer* parent;
		String timer_desc;
		int showResult;
		
		void addTimeData(int counter,unsigned long long value);
		void getCpuInfo();
		
	public:
		inline KTimer(char* name=NULL,KTimer* parent=NULL,int showInfo=0);
		
		inline void setKTimer(int counter=0,char* desc=NULL) ;
		inline void startT(int counter=0) ;
		inline void stopT(int counter=0) ;
		inline double wTime();
		inline double getTimeData(int counter=0);
		
		inline void showData(int counter=0) const ;
		inline void showAllData() const;
		inline void addDataToParent(int counter);
		inline ~KTimer();
};

inline void KTimer::addTimeData(int counter,unsigned long long value)
{
	data[counter].sec+=value;
}

inline void KTimer::getCpuInfo()
{
	FILE* f;
	char buf[256];
	
	f=fopen("/proc/cpuinfo","r");
	if (!f) {
		if (showResult==1)
			printf("Can not open file /proc/cpuinfo. Assuming Cpu Frequency is %d\n",CPU_FREQUENCY);
		cpuInfo.frequency=CPU_FREQUENCY;
	}
	while (fscanf(f,"%s",buf)==1) {
		if (strcmp(buf,"MHz")==0) {
			fscanf(f,"%s",buf);
			fscanf(f,"%s",buf);
			cpuInfo.frequency=(unsigned long long)(atof(buf)*1000000);
			if (showResult==1)
				printf("Found cpu. Frequency %lld\n",cpuInfo.frequency);
			fclose(f);
			return;
		}
	}
	if (showResult==1)
		printf("Can not find any appropriate info\n");
	cpuInfo.frequency=CPU_FREQUENCY;
	return;
}

inline KTimer::KTimer(char* name,KTimer* pnt,int _showResult)
{
	for (int i=0;i<TIME_SIZE;i++) {
		data[i].sec=data[i].tmp_sec=data[i].step_no=0;
		data[i].description.initString();
	}
	parent=pnt;
	timer_desc.initString(name);
	showResult=_showResult;

	if (parent==NULL)
		getCpuInfo();
	else
		cpuInfo=parent->cpuInfo;
//	printf("Arch i686 found. Assuming CPU frequency %uhz. Using RDTSC.\n",CPU_FREQUENCY);
	
}
inline void KTimer::setKTimer(int counter,char* desc)
{
	data[counter].sec=data[counter].tmp_sec=data[counter].step_no=0;
	data[counter].description.initString(desc);
}
inline void KTimer::startT(int counter)
{
	asm volatile (  "rdtsc\n\t"
	    : "=A" (data[counter].tmp_sec) );
	
}
inline void KTimer::stopT(int counter)
{
	long long tmp;

	asm volatile (  "rdtsc\n\t"
	    : "=A" (tmp) );

	data[counter].sec+=tmp-data[counter].tmp_sec;
	data[counter].step_no++;
//	printf("%lld\n",data[counter].sec);
}
inline double KTimer::wTime()
{
	long long tmp;
	asm volatile ( "rdtsc\n\t"
	    : "=A" (tmp) );
	return tmp/(((double)cpuInfo.frequency));
}
inline  void KTimer::showData(int counter) const
{
	long long tmp;

	tmp=data[counter].sec;
	tmp-=((long long)11)*data[counter].step_no;
	tmp=(long long)(tmp/(((double)cpuInfo.frequency)/1000000));
	if (showResult==1)
		printf("description: \"%s\" \n nomer=%d Total time: %u sec %u usec step_no=%lld \n",data[counter].description(),counter,(unsigned)(tmp/1000000),(unsigned)(tmp%1000000),data[counter].step_no);
}
inline void KTimer::showAllData() const
{
	for (int i=0;i<TIME_SIZE;i++)
		showData(i);
	printf("\n");
}

inline double KTimer::getTimeData(int counter)
{
	long long tmp;
	tmp=data[counter].sec;
	tmp-=((long long)11)*data[counter].step_no;
	tmp=(long long)(tmp/(((double)cpuInfo.frequency)/1000000));
	
	return ((double)tmp)/1000000;
}
inline void KTimer::addDataToParent(int counter)
{
	parent->addTimeData(counter,data[counter].sec);
}
inline KTimer::~KTimer()
{
	int i;
	printf("Results of Timer \"%s\" are presented below\n",timer_desc());
	if (parent==NULL) {
		if (showResult==1)	
		showAllData();
	}
	else
		for (i=0;i<TIME_SIZE;i++) 
			parent->addTimeData(i,data[i].sec);
}


class KTimerInitializer {
	public:
		KTimerInitializer(KTimer* timer,int counter,char* desc)
		{
			timer->setKTimer(counter,desc);
		}
};
#endif

