#ifndef __app___
#define __app___

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "WWAppTdef.h"
#include "WWAppTimp.h"
#include "body.h"
#include "windfield.h"
#include "consts.h"



class Application: public WWApplicationT<Body,WindField>
    {
	private:
	//file pointer for the file in which we shall save al on
        //each step of program execution
        FILE *fp_al;
#ifdef VISION
	FILE *fifo;
#endif		
	int SetWorkTime();

	int SavePrepare();
	int SaveStepResults(Body*,WindField*,double);
	int SaveResults(Body*,WindField*);

    };

int Application::SetWorkTime()
    {
	time_start=TIME_START;
	time_end=TIME_END;
	time_step=TIME_STEP;
    
    	return 0;
    }


// This function is for preparing files for saving per step information
// In order not to open & close files more then one time!!!...

int Application::SavePrepare()
    {
      mkdir("results",0700);
	if ((fp_al=fopen("results/al","w+"))==NULL)
	    {
	    fprintf(stderr,"\nCann`t open file al :(!!! Terminating...\n");
	    Abort(11);
	    }

#ifdef VISION
	if ((fifo=fopen("results/fifo","w+"))==NULL)
	    {
	    fprintf(stderr,"\nCann`t open file fifo :(!!! Terminating...\n");
	    Abort(11);
	    }
#endif	    


	return 0;
    }

//---------this function saves results per step of the program

int Application::SaveStepResults(Body *body,WindField *wf,double time)
    {
#ifdef VISION    
	int i;
#endif	
	printf("\nStep=%lf\n",time);
	
	fprintf(fp_al,"time=%lf  al=%5.25lf  aldt=%5.25lf\n",time,body->al,body->aldt);
#ifdef VISION
	fprintf(fifo,"begin\n");
 	  for (i=0;i<body->__NP+body->__SOURCE;i++)
           fprintf(fifo,"%d\n%le\n%le\n",2,body->awnd[i].x,body->awnd[i].y);
          for (i=0;i<wf->n;i++)
           fprintf(fifo,"%d\n%le\n%le\n",wf->fwnd[i].circ>0?1:0,wf->fwnd[i].x,wf->fwnd[i].y);
        fprintf(fifo,"end\n");
#endif	
	
	return 0;
    }

//----------This function saves the results at the end of execution
int Application::SaveResults(Body *body,WindField *wf)
    {
	FILE *fp;
	int i;
	
	if ((fp=fopen("results/awnd","w+"))==NULL)
	    {
	    fprintf(stderr,"\nCann`t open file awnd :(!!! Terminating...\n");
	    Abort(10);
	    }
	
	fprintf(fp,"      x          y        circ\n");
	
	for (i=0;i<NP;i++)
	    fprintf(fp,"%8.4lf   %8.4lf   %8.4lf\n",
			body->awnd[i].x,body->awnd[i].y,body->awnd[i].circ);
			
	fclose(fp);
	
	if ((fp=fopen("results/fwnd","w+"))==NULL)
	    {
	    fprintf(stderr,"\nCann`t open file fwnd :(!!! Terminating...\n");
	    Abort(12);
	    }
	
	fprintf(fp,"      x          y        circ\n");
	
	for (i=0;i<wf->n;i++)
	    fprintf(fp,"%8.4lf   %8.4lf   %8.4lf\n",
			wf->fwnd[i].x,wf->fwnd[i].y,wf->fwnd[i].circ);
	
	fclose(fp);
	
	
	
	fclose(fp_al);
#ifdef VISION
	fclose(fifo);
#endif	
	
	return 0;
    }

#endif
