//The Body class
#ifndef __body___
#define __body___

#include <math.h>

#include "WWBodyTdef.h"
#include "WWBodyTimp.h"
#include "windfield.h"
#include "consts.h"

/* 
 *   ""
 * The definition of the "Body" - class
 */
class Body: public WWBodyT<SOURCE,    //the body with a source
		NP,	      //the number of control points
		NUMS,	      //the number of points where new free winds
			      //				come from
		WindField     //the WindField class 
		>
{
 public:
 static const int __SOURCE=SOURCE;
 static const int __NP=NP;
 virtual int UserInit();
 virtual int CountCenter(double time);
 virtual int CountPoints();
 virtual int CountWinds();
 virtual int CountNorms();
};		
/*
 * 㭪 CountCenter    業ࠫ쭮
 * 窨 ⥪ ⥫  ᨬ  ६.
 * 室 ࠬ time - ⥪饥 ६.
 *     㦭 ᢮ 室 祭
 * ᫥騬 ६:
 * center.x -   ᮢ न 業ࠫ쭮 窨
 * center.y -   ४ न 業ࠫ쭮 窨
 * centerspeed.x - ࢠ न  ᪮ 業ࠫ쭮 窨
 * centerspeed.y -  न  ᪮ 業ࠫ쭮 窨
 */
#define V -1		

int Body::CountCenter(double time)
    {
	 center.x=V*time;
    	 center.y=0;
	
	 centerspeed.x=V;
    	 centerspeed.y=0;	
    
	 return 0;
    }

/*
 * 㭪 UserInit() 樠 砫쭮  ⥫
 *  樠樨 ⥫ 室  ᫥騥 稭
 * ६ al ᢮ 祭 砫쭮 㣫  ⥫ 
 * ⭮⥫쭮 ன , aldt - 砫쭠 㣫 ᪮ ⥫.
 * inds[] - 祭 ᮢ  뢠 ३  饬 ᨢ 
 * ᮥ ३. 
 * inertion - 祭 樨
 * rho - ⭮ ।
 */

int Body::UserInit()
    {
	al=M_PI/4;
	aldt=0;
	aldtold=0;

	inertion=100;
	rho=1;

	inds[0]=0;
	inds[1]=20;
	inds[2]=40;
	inds[3]=60;
/*	
	inds[4]=10;
	inds[5]=30;
	inds[6]=50;
	inds[7]=70;
*/
	return 0;
    }
/*
 * 㭪 CountWinds  न ᮥ ३  ⥫
 *  ᨬ  㣫 al.  न  뢠 
 * ᨢ awnd[].
 * 室 ⠪ 뢠 ,  ⢮ ᮥ ३
 *   ⨯ ࠭ ⥫.
 * ᫨ ⥫  筨,  ⢮ ३  NP+1
 */
//#define NW 	NP+1
#define NW	NP

int Body::CountWinds()
    {
	int i;
	double cos_al=cos(al);
	double sin_al=sin(al);

	for (i=0;i<20;i++)
	    {
		awnd[i].x=cos_al-sin_al*(-1.+i*0.1)+center.x;
    		awnd[i].y=sin_al+cos_al*(-1.+i*0.1)+center.y;	    
	    }
	    
	for (i=20;i<40;i++)
	    {	
		awnd[i].x=cos_al*(1-(i-20)*0.1)-sin_al+center.x;
		awnd[i].y=sin_al*(1-(i-20)*0.1)+cos_al+center.y;
	    }
	    
	for (i=40;i<60;i++)
	    {
		awnd[i].x=-cos_al-sin_al*(1-(i-40)*0.1)+center.x;
		awnd[i].y=-sin_al+cos_al*(1-(i-40)*0.1)+center.y;
	    }
	    
	for (i=60;i<80;i++)
	    {
		awnd[i].x=(-1+(i-60)*0.1)*cos_al+sin_al+center.x;
		awnd[i].y=(-1+(i-60)*0.1)*sin_al-cos_al+center.y;
	    }
	
	return 0;
    }

/* 㭪 CountPoints()  न ஫ 祪  ⥫
 *  ᨬ  㣫  ⥫ al  ࠬ஢ 業ࠫ쭮 窨 ⥫,
 * ⢮ ஫ 祪 ࠢ NP
 * न 室   ᨢ  cpnt[],
 * cpnt[].x - ࢠ न,
 * cpnt[].y -  न
 */

int Body::CountPoints()
    {
	int i;
	double sin_al=sin(al);
	double cos_al=cos(al);
	
	for (i=0;i<20;i++)
	    {
		cpnt[i].x=cos_al-sin_al*(-1+0.05+i*0.1)+center.x;
    		cpnt[i].y=sin_al+cos_al*(-1+0.05+i*0.1)+center.y;	    
	    }
	    
	for (i=20;i<40;i++)
	    {	
		cpnt[i].x=cos_al*(1-0.05-(i-20)*0.1)-sin_al+center.x;
		cpnt[i].y=sin_al*(1-0.05-(i-20)*0.1)+cos_al+center.y;
	    }
	    
	for (i=40;i<60;i++)
	    {
		cpnt[i].x=-cos_al-sin_al*(1-0.05-(i-40)*0.1)+center.x;
		cpnt[i].y=-sin_al+cos_al*(1-0.05-(i-40)*0.1)+center.y;
	    }
	    
	for (i=60;i<80;i++)
	    {
		cpnt[i].x=(-1+0.05+(i-60)*0.1)*cos_al+sin_al+center.x;
		cpnt[i].y=(-1+0.05+(i-60)*0.1)*sin_al-cos_al+center.y;
	    }
	
	return 0;
    
    }
/*
 * 㭪 CountNorms() 室   ஢ ଠ  ⥫
 *  ஫ 窠  ᨬ  㣫  ⥫ al
 * ࠪ⨪ ஢ 뢠  ᨢ norm[]
 */

int Body::CountNorms()
    {
	int i;
	double sin_al=sin(al);
	double cos_al=cos(al);
	for (i=0;i<20;i++)
	    {
		norm[i].x=cos_al;
    		norm[i].y=sin_al;	    
	    }
	    
	for (i=20;i<40;i++)
	    {	
		norm[i].x=-sin_al;
		norm[i].y=cos_al;
	    }
	    
	for (i=40;i<60;i++)
	    {
		norm[i].x=-cos_al;
		norm[i].y=-sin_al;
	    }
	    
	for (i=60;i<80;i++)
	    {
		norm[i].x=sin_al;
		norm[i].y=-cos_al;
	    }
	
	return 0;
    }
    
#endif
