#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "geometry.h"

#define BUFSIZE		256

static char *_fgets (char *s, int size, char **buf)
{
	char *s1 = s;
	while (size && ((*s = *((*buf)++)) != 0)) {
		if (*s++ == '\n')
			break;
		size--;
	}
	if (size)
		*s = 0;
	return (s == s1) ? 0 : s1;
}

static size_t fsize(int fd)
{
	int res;
	struct stat st;

	res=fstat(fd,&st);
	if(res!=-1)res=st.st_size;
	return res;
}

#define tmesg(x...) fprintf(stderr,x)

static inline int read_vector (struct vector *v, char *src, char**end)
{
	v->x = strtod (src, end);
	v->y = strtod (*end, end);
	v->z = strtod (*end, end);
	return 0;
}

int read_scene (struct scene *sc, const char *fname)
{
	int	source_scene;
	size_t	fsz;
	char	*inbuf;
	char	*bufp;
	char	cp_buff[BUFSIZE];
	char	*end_ptr;

	int	nx, ny;
	
	struct vector from, at, updir, v_tmp1, v_tmp2;
	struct texture *texture;
	struct solid *tmps;
	int poly_n_points;
	
	int s_cnt = 0;
	int p_cnt = 0;
	int l_cnt = 0;
	int v_cnt = 0;
	
	int cur_tex = -1;

	
	
	if ((source_scene = open (fname, O_RDONLY)) < 0) {
		error ("Can't open scene");
		return 1;
	}
	
	scene_init (sc);

	fsz = fsize (source_scene);
	tmesg ("malloc(%d)\n", fsz);
	inbuf = (char *) malloc (fsz + 1);
	tmesg ("after malloc\n");
	if ((!inbuf) || (read (source_scene, inbuf, fsz) != fsz)) {
		error ("Can't read scene");
	}
	inbuf[fsz] = 0;
	bufp = inbuf;

	while (_fgets (cp_buff, BUFSIZE, &bufp) != NULL) {
		if (l_cnt >= MAX_LIGHTS) {
			error("Out of scene size (too many point lights)");
		}

		switch (cp_buff[0]) {
		case 'b': read_vector(&(sc->scs.background), cp_buff + 1, &end_ptr);
			break;
		case 'v':
			if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
				error ("Can't read.");
			}
			read_vector (&from, strpbrk (cp_buff, " "), &end_ptr);
			if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
				error ("Can't read.");
			}
			read_vector (&at, strpbrk (cp_buff, " "), &end_ptr);
			if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
				error ("Can't read.");
			}
			read_vector (&updir, strpbrk (cp_buff, " "), &end_ptr);
			set_camera (&(sc->scs), &from, &at, &updir);
			
			if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
				error ("Can't read.");
			}
			if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
				error ("Can't read.");
			}
			if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
				error ("Can't read.");
			}
			nx = (int) strtod (strpbrk (cp_buff, " "), &end_ptr);
			ny = (int) strtod (end_ptr, &end_ptr);
			v_cnt++;
			break;

		case 'f':
			cur_tex = alloc_texture (sc, TEXTURE_SIZE);
			if (cur_tex < 0) {
				error ("Out of scene size (texture allocation)");
			}
			texture = GET_TEXTURE (sc, cur_tex);
			read_vector (&(texture->color), cp_buff + 1, &end_ptr);
			
			texture->kd = strtod (end_ptr, &end_ptr);
			texture->ks = strtod (end_ptr, &end_ptr);
			texture->p = (int) strtod (end_ptr, &end_ptr);
			texture->kt = strtod (end_ptr, &end_ptr);
			texture->ka = 0.3;	// What the fucking const?
			set_med (&(texture->med), strtod (end_ptr, &end_ptr), 0);
			break;

		case 's': read_vector (&v_tmp1, cp_buff + 1, &end_ptr);
		
			tmps = alloc_solid(sc,SPHERE_SIZE);
			if (!tmps) {
				error ("Out of scene size\n");
			}
			
			sphere_set (tmps, &v_tmp1, strtod (end_ptr,&end_ptr));
			
			tmps->texture = cur_tex;
			
			s_cnt++;
			break;

		case 'p':

			poly_n_points = (int) strtod (cp_buff + 1, &end_ptr);
			tmps = alloc_solid(sc,POLY_SIZE(poly_n_points));
			if (!tmps) {
				error ("Out of scene size\n");
			}
			
			for (v_cnt = 0; v_cnt < poly_n_points; v_cnt++) {
				if (_fgets (cp_buff, BUFSIZE, &bufp) == NULL) {
					error ("Can't read.");
				}
				read_vector (&(tmps->obj[0].poly.vertex[v_cnt].point),
						cp_buff, &end_ptr);
			}
			poly_set (tmps,v_cnt);
			
			tmps->texture = cur_tex;
			
			p_cnt++;

			break;

		case 'l': read_vector (&v_tmp1, cp_buff + 1, &end_ptr);
			setv_fk (&v_tmp2, 1, 1, 1);
			set_plight (&sc->light[sc->lights_count],
				    &v_tmp1, &v_tmp2, 2);
			sc->lights_count++;
			l_cnt++;
			break;

		default:
			v_cnt=0;
			while(cp_buff[v_cnt]>=' ')
				v_cnt++;
			cp_buff[v_cnt]=0;
			tmesg("`%s'\n",cp_buff);
			error ("Unsupported input data");
		}
	}

	free (inbuf);
	close (source_scene);
	
	setv_f (&sc->scs.ambient, 1.0);
	set_med (&sc->scs.air, 1, 0);
	set_med (&sc->scs.glass, 1.5, 0);
	
	tmesg ("\nLight sources: Point lights .. %d\n", l_cnt);
	tmesg ("Solids:        Spheres ....... %d\n", s_cnt);
	tmesg ("               Polygons ...... %d\n\n", p_cnt);

}
