#include <malloc.h>
#include <fcntl.h>

#include <unistd.h>

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "tracer.h"
#include "render.h"

#include "output.h"
#include "rt.h"

#define DEFAULT_HALF_WIDTH 0.4
#define DEFAULT_HALF_HEIGHT 0.4

/* #define GREEN_DEBUG 1 */

double rnd (void)
{
	return (rand () / (double) RAND_MAX);
}

static char fname[256];
int parse_cmdline (int cnt, int argc, char **argv)
{
	if (cnt < argc) {
		strncpy(fname, argv[cnt], 256);
	} else {
		strncpy(fname, "input.nff", 256);
	}
	fprintf(stderr, "Input file: '%s'\n", fname);
	return 0;
}

int make_params (struct params *par, struct rect r, double tm)
{
	par->f_ulx	= -DEFAULT_HALF_WIDTH;
	par->f_stepx	= 2.0 * DEFAULT_HALF_WIDTH / (r.width);
	par->f_uly	= -DEFAULT_HALF_HEIGHT;
	par->f_stepy	= 2.0 * DEFAULT_HALF_HEIGHT / (r.height);

	read_scene (&par->sc, fname);
	
	par->sc.scs.threshold		= 1e-5 /*thresh*/;
	par->sc.scs.geomthreshold	= 1e-5 /*gthresh*/;
	par->sc.scs.maxlevel		= 5 /*depth*/;
	par->sc.scs.nx			= r.width;
	par->sc.scs.ny			= r.height;
}

struct pixel render (int x, int y, struct params *par)
{
	double		tmp;
	struct ray	r;
	struct vector	color;
	struct pixel	res;
	const struct scene *s = &par->sc;

	camera (&s->scs, (par->f_ulx) + x*(par->f_stepx),
			 (par->f_uly) + y*(par->f_stepy), &r);
	trace (s, s->scs.air, 1.0, &r, &color, 0);
			
	tmp = (color.x * 255);
	res.r = (tmp >= 255) ? 255 : (tmp <= 0) ? 0 : (unsigned char) tmp;
			
	tmp = (color.y * 255);
	res.g = (tmp >= 255) ? 255 : (tmp <= 0) ? 0 : (unsigned char) tmp;
			
	tmp = (color.z * 255);
	res.b = (tmp >= 255) ? 255 : (tmp <= 0) ? 0 : (unsigned char) tmp;

	return res;
}
