#ifndef __TIMAGE_H__
#define __TIMAGE_H__

#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>

typedef unsigned char byte;	/* byte type */
typedef unsigned long Pixel;	/* what X thinks a pixel is */

/*
 * header information for TARGA image file format (TGA),
 * mode 2, no mapping
 */

struct tga_header {
	byte magic[12];
	/* {0,0,2,0,0,0,0,0,0,0,0,0}; */
	short int nx, ny;
	byte magic_plus[2];
	/* {0x18,0}; */
};

#define SIZE_TARGA_HEADER	(sizeof(struct tga_header))

#ifdef _IN_WRITER_
static byte tga_magic[] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
static byte tga_magic_plus[] = { 0x18, 0 };
#endif

#define IMGSIZE_WIDTH_MAX 720
#define IMGSIZE_HEGHT_MAX 540

#define BYTE_PER_TGAPIXEL 3

#define LENGTH_SOCKNAME_MAX (128 - sizeof (unsigned short))
				/* ^:sa_family */

typedef union {
	struct sockaddr saa[1];	/* == { PF_UNIX, ... } */
	struct {
		unsigned short sa_family;
		char ca_name[LENGTH_SOCKNAME_MAX + /* For ASCIIZ ZERO */ 1];
	} sa_forname;
} trtunixsockaddr_t;

 /* TRT -> TRTIMAGE command structure */
typedef struct tga_cmd {
	unsigned short
	 ush_tgsign, ush_cmd;
	union {
		struct {
			int i_width, i_height;
		} dim;
		struct {
			int i_firstline, i_nlines;
		} fragspec;
	} data;
} tgcmd_t;

/* TRT -> TRTIMAGE command signature */
#define TGCMD_SIGN 0x7991

/* TRT -> TRTIMAGE command types */
enum { TGCMD_START, TGCMD_FRAG, TGCMD_END };

#define WIDTH_TGCMD(tgcmd) (tgcmd.data.dim.i_width)
#define HEIGHT_TGCMD(tgcmd) (tgcmd.data.dim.i_height)
#define FLINE_TGCMD(tgcmd) (tgcmd.data.fragspec.i_firstline)
#define NLINES_TGCMD(tgcmd) (tgcmd.data.fragspec.i_nlines)

#endif				/* __TIMAGE_H__ */
