#include <math.h>

#include "tracer.h"

void set_med (struct medium *m, double r, double b)
{
	m->n_refr = r;
	m->beta = b;
}

const struct solid *s_scene_isect (const struct scene *s, const struct ray *r, double *t)
{
	const struct solid	*closest_obj = NULL;
	const struct solid	*currobj;
	double			closest_dist = _INFINITY;
	double			dist;
	int			i;

	// FIXME: this is _very_ slow
	currobj = FIRST_SOLID (s);
	for (i = 0; i < s->solids_count; i++) {
		if (vmt[currobj->type].isect (&(currobj->obj[0]), r, &dist)) {
			if (dist < closest_dist) {
				closest_dist = dist;
				closest_obj = currobj;
			}
		}
		currobj = next_solid (s, currobj);
	}
	*t = closest_dist;
	return closest_obj;
}

void s_scene_shadebackground (const struct scene *s, const struct ray *r,
		struct vector *color)
{
	*color = s->scs.background;
}

void set_camera (struct sc_set *scsp, const struct vector *from,
		const struct vector *at, const struct vector *updir)
{
	struct vector cross;

	scsp->eye = *from;
	sub_v_v (&(scsp->eyedir), *at, *from);
	normalyze (&(scsp->eyedir), scsp->eyedir);
	vmult_v_v (&cross, scsp->eyedir, *updir);
	normalyze (&(scsp->vx), cross);
	vmult_v_v (&cross, scsp->vx, scsp->eyedir);
	normalyze (&(scsp->vy), cross);
}

/* CAMERA
 *
 *   Create a first hit ray (goes from the eye to the scene)
 */
void camera (const struct sc_set *pscp, const double x,
		const double y, struct ray *r)
{
	double f_rdir_norm;

	r->org = pscp->eye;

	/*  r -> dir =  *(normalyze (*add_v_v (*mult_v_f (eyedir, 10),
	 *  *add_v_v (*mult_v_f (vx, x),
	 *  *mult_v_f (vy, y))
	 *  )));
	 */
	(r->dir).x = pscp->eyedir.x * 1 + pscp->vx.x * x + pscp->vy.x * y;
	(r->dir).y = pscp->eyedir.y * 1 + pscp->vx.y * x + pscp->vy.y * y;
	(r->dir).z = pscp->eyedir.z * 1 + pscp->vx.z * x + pscp->vy.z * y;
	f_rdir_norm = norma_v (r->dir);
	(r->dir).x /= f_rdir_norm;
	(r->dir).y /= f_rdir_norm;
	(r->dir).z /= f_rdir_norm;
}

double trace (const struct scene *s, const struct medium curmed,
		const double weight, const struct ray *r,
		struct vector *color, const int level)
{
	const struct solid	*obj;
	double			dist = _INFINITY;
	double			t;
	struct vector		p;

	if ((obj = s_scene_isect (s, r, &t)) != NULL) {
		/* p = * isect_point (r, t); */
		p.x = r->org.x + r->dir.x * t;
		p.y = r->org.y + r->dir.y * t;
		p.z = r->org.z + r->dir.z * t;
		dist = t;
		shade (s, curmed, weight, &p, &(r->dir), obj,
				color, level + 1);
	} else {
		s_scene_shadebackground (s, r, color);
	}
	return dist;
}

void shade (const struct scene *s, const struct medium curmed,
		const double weight, const struct vector *p,
		const struct vector *view, const struct solid *obj,
		struct vector *color, const int level)
{
	struct texture	texture;
	struct ray	r;
	struct vector	tn;
	struct vector	tcol;
	struct vector	l;
	double		dist;
	double		sh;
	struct vector	h;
	double		ln, vn;
	int		entering_obj = 1;
	int		i;
	double		eta, ci, ctsq;
	double		rweight;
	double		tweight;
	double		f_tmp;

	texture = *GET_TEXTURE(s, obj->texture);
	vmt[obj->type].norm (&(obj->obj[0]), p, &tn);
	//(*(obj->get_tex)) (&s->scs, &obj->g, p, &texture);
	/*  if (( vn = dot_v_v (view, texture.n)) > 0) { ** */
	if ((vn = dot_v_v (*view, tn)) >= 0) {
		tn.x = -(tn.x);
		tn.y = -(tn.y);
		tn.z = -(tn.z);
		vn = -vn;
		entering_obj = 0;
	}
	r.org = *p;
	/* *color = *mult_v_f (*mult_v_v (ambient, texture.color), texture.ka); */
	color->x = s->scs.ambient.x * texture.color.x * texture.ka;
	color->y = s->scs.ambient.y * texture.color.y * texture.ka;
	color->z = s->scs.ambient.z * texture.color.z * texture.ka;
	for (i = 0; i < (s->lights_count); i++) {
		const struct plight *lsp = &s->light[i];
		sh = plight_shadow (s, lsp, p, &l);
		if (sh > s->scs.threshold) {
			ln = dot_v_v (l, tn);
			if (ln > s->scs.threshold) {
				if (texture.kd > s->scs.threshold) {
					/*
					 *color = *add_v_v (*color, *mult_v_v (scene -> light[i] -> color,
					 *mult_v_f (texture.color, texture.kd*sh*ln)));
					 */
					color->x +=
					    (lsp->color).x * texture.color.x *
					    (texture.kd * sh * ln);
					color->y +=
					    (lsp->color).y * texture.color.y *
					    (texture.kd * sh * ln);
					color->z +=
					    (lsp->color).z * texture.color.z *
					    (texture.kd * sh * ln);
				}
				if (texture.ks > s->scs.threshold) {
					/*
					   h = * normalyze (*sub_v_v (l, view));
					   *color = *add_v_v (*color, *mult_v_f (scene -> light[i] -> color,
					   (texture.ks*sh*pow(dot_v_v (texture.n, h),
					   texture.p))));
					 */
					h.x = l.x - view->x;
					h.y = l.y - view->y;
					h.z = l.z - view->z;
					f_tmp =
					    sqrt (h.x * h.x + h.y * h.y +
						  h.z * h.z);
					h.x /= f_tmp;
					h.y /= f_tmp;
					h.z /= f_tmp;
					color->x += (lsp->color).x *
						(texture.ks * sh *
						pow (dot_v_v (tn, h),
						texture.p));
					color->y += (lsp->color).y *
						(texture.ks * sh *
						pow (dot_v_v (tn,h),
						texture.p));
					color->z += (lsp->color).z *
						(texture.ks * sh *
						pow (dot_v_v (tn,h),
						texture.p));
				}
			}
		}
	}
	rweight = weight * texture.ks;
	tweight = weight * texture.kt;

	if (rweight > s->scs.threshold && level < s->scs.maxlevel) {
		/* r.dir = *sub_v_v (view, * mult_v_f (texture.n, (2*vn))); */
		r.dir.x = view->x - tn.x * 2 * vn;
		r.dir.y = view->y - tn.y * 2 * vn;
		r.dir.z = view->z - tn.z * 2 * vn;
		dist = trace (s, curmed, rweight, &r, &tcol, level);
		/* *color = *add_v_v (*color, *mult_v_f (tcol, texture.ks)); */
		color->x += tcol.x * texture.ks;
		color->y += tcol.y * texture.ks;
		color->z += tcol.z * texture.ks;
	}
	
	if (tweight > s->scs.threshold && level < s->scs.maxlevel) {
		eta = curmed.n_refr / (entering_obj ? texture.med.n_refr :
				       s->scs.air.n_refr);
		ci = -vn;
		ctsq = 1 + eta * eta * (ci * ci - 1);

		if (ctsq > s->scs.threshold) {
			/*
			   r.dir = *add_v_v (*mult_v_f (view, eta),
			   *mult_v_f (texture.n, (eta*ci-sqrt(ctsq))));
			 */
			r.dir.x = view->x * eta
				+ tn.x * (eta * ci - sqrt (ctsq));
			r.dir.y = view->y * eta 
				+ tn.y * (eta * ci - sqrt (ctsq));
			r.dir.z = view->z * eta
				+ tn.z * (eta * ci - sqrt (ctsq));
			if (entering_obj) {
				dist = trace (s, texture.med, tweight, &r,
						&tcol, level);
			} else {
				dist = trace (s, s->scs.air, tweight, &r,
						&tcol, level);
			}
			/* *color += *add_v_v (*color, *mult_v_f (tcol, texture.kt)); */
			color->x += tcol.x * texture.kt;
			color->y += tcol.y * texture.kt;
			color->z += tcol.z * texture.kt;
		}
	}
}

