#ifndef __VECTOR__
#define __VECTOR__

#ifndef TC
#include <math.h>
#include <stdlib.h>
#else
#ifdef __cplusplus
extern "C"
#endif
double sqrt (double);
#endif

struct vector {
	double x, y, z;
};

struct ray {
	struct vector org;	/* ray begining */
	struct vector dir;	/* direction (must be normalized) */
};

static inline void setv_f (struct vector *v, double f)
{
	v->x = f;
	v->y = f;
	v->z = f;
}

static inline void setv_v (struct vector *v, struct vector u)
{
	v->x = u.x;
	v->y = u.y;
	v->z = u.z;
}

static inline void setv_fk (struct vector *v, double vx, double vy, double vz)
{
	v->x = vx;
	v->y = vy;
	v->z = vz;
}

static inline void add_v_v (struct vector *v_tmp, struct vector v,
			    struct vector u)
{
	v_tmp->x = v.x + u.x;
	v_tmp->y = v.y + u.y;
	v_tmp->z = v.z + u.z;
}

static inline void sub_v_v (struct vector *v_tmp, struct vector v,
			    struct vector u)
{
	v_tmp->x = v.x - u.x;
	v_tmp->y = v.y - u.y;
	v_tmp->z = v.z - u.z;
}

static inline void invert_v (struct vector *v_tmp, struct vector v)
{
	v_tmp->x = -v.x;
	v_tmp->y = -v.y;
	v_tmp->z = -v.z;
}

static inline void mult_v_f (struct vector *v_tmp, struct vector v, double f)
{
	v_tmp->x = v.x * f;
	v_tmp->y = v.y * f;
	v_tmp->z = v.z * f;
}

static inline void mult_v_v (struct vector *v_tmp, struct vector v,
			     struct vector u)
{
	v_tmp->x = v.x * u.x;
	v_tmp->y = v.y * u.y;
	v_tmp->z = v.z * u.z;
}

static inline void div_v_f (struct vector *v_tmp, struct vector v, double f)
{
	v_tmp->x = v.x / f;
	v_tmp->y = v.y / f;
	v_tmp->z = v.z / f;
}

static inline double dot_v_v (struct vector v, struct vector u)
{
	return (v.x * u.x + v.y * u.y + v.z * u.z);
}

static inline void vmult_v_v (struct vector *cross, struct vector v,
			      struct vector u)
{
	cross->x = v.y * u.z - v.z * u.y;
	cross->y = v.z * u.x - v.x * u.z;
	cross->z = v.x * u.y - v.y * u.x;
}

static inline double norma_v (struct vector v)
{
	return sqrt (v.x * v.x + v.y * v.y + v.z * v.z);
}

static inline void setr_v_v (struct ray * r, struct vector v, struct vector u)
{
	setv_v (&(r->org), v);
	setv_v (&(r->dir), u);
}

static inline void isect_point (struct vector *v_tmp, const struct ray r, double t)
{
	struct vector tmp;
	mult_v_f (&tmp, r.dir, t);
	add_v_v (v_tmp, r.org, tmp);
}

 
static inline void normalyze (struct vector * v_tmp, struct vector v)
{
	double norma;
	
	norma = norma_v (v);
	div_v_f (v_tmp, v, norma);
}


static inline void rndvector (struct vector *v_tmp)
{
	setv_fk (v_tmp, rand () - 0.5 * RAND_MAX,
		 rand () - 0.5 * RAND_MAX, rand () - 0.5 * RAND_MAX);
	normalyze (v_tmp, *v_tmp);
}

#endif
