/* mary */

#include <draw.h>

void
cleanup_canvas(GraphCanvas *gcanvas)
{
  GtkWidget *canvas = gcanvas->canvas;
  GList *list;
  gnome_canvas_update_now (GNOME_CANVAS (canvas));
  for(list = gcanvas->item_list; list != NULL; list = list->next) {
      GnomeCanvasItem *item = GNOME_CANVAS_ITEM(list->data);
      gtk_object_destroy(GTK_OBJECT(item));
  }
  if(gcanvas->item_list != NULL) {
    g_list_free(gcanvas->item_list);
    gcanvas->item_list = NULL;
  }
}

void
init_canvas(GraphCanvas *gcanvas, double width, double height, double scale)
{
  gcanvas->culc_width = width;
  gcanvas->culc_height = height;  
  //don't resize canvas
  gcanvas->canvas_width = MAX (gcanvas->canvas_width, gcanvas->culc_width);
  gcanvas->canvas_height = MAX (gcanvas->canvas_height, gcanvas->culc_height);
  gcanvas->item_list = NULL;
  gnome_canvas_set_pixels_per_unit (GNOME_CANVAS (gcanvas->canvas), scale);
  gtk_widget_set_usize (gcanvas->canvas, 
			CANVAS_WIDTH, CANVAS_HEIGHT);

  gnome_canvas_set_scroll_region (GNOME_CANVAS (gcanvas->canvas), 0.0, 0.0, 
				  gcanvas->canvas_width, gcanvas->canvas_height);

  gcanvas->x_offset = (gcanvas->canvas_width - gcanvas->culc_width) / 2;
  gcanvas->y_offset = (gcanvas->canvas_height - gcanvas->culc_height) / 2;
}

double
trans_x_coord(GraphCanvas *gcanvas, double x) { return x + gcanvas->x_offset; }

double
trans_y_coord(GraphCanvas *gcanvas, double y) { return (gcanvas->culc_height - y) + gcanvas->y_offset; }

GnomeCanvasItem *
draw_ellipse(GraphCanvas *gcanvas, double x1, double y1, double x2, double y2, 
	     char *color, char *fillcolor)
{
  GtkWidget *canvas = gcanvas->canvas;
  guint type;
  GnomeCanvasItem *item;
  type = gnome_canvas_ellipse_get_type ();
  item = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (canvas)),
				type,
				"x1", x1,
				"y1", y1,
				"x2", x2,
				"y2", y2,
				"fill_color", fillcolor,
				"outline_color", color,
				"width_units", 1.0,
				NULL);
  gcanvas->item_list = g_list_prepend(gcanvas->item_list, item);
  return item;
}

GnomeCanvasItem *
draw_rect(GraphCanvas *gcanvas, double x1, double y1, double x2, double y2,
	  char *color, char *fillcolor)
{
  GtkWidget *canvas = gcanvas->canvas;
  guint type;
  GnomeCanvasItem *item;
  type = gnome_canvas_rect_get_type ();
  item = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (canvas)),
				type,
				"x1", x1,
				"y1", y1,
				"x2", x2,
				"y2", y2,
				"fill_color", fillcolor,
				"outline_color", color,
				"width_units", 1.0,
				NULL);
  gcanvas->item_list = g_list_prepend(gcanvas->item_list, item);
  return item;
}

pointf*
pointfof(double x, double y)
{
  pointf *rv = g_new0(pointf,1);
  rv->x = x, rv->y = y;
  return rv;
}

void
draw_bezier(GraphCanvas *gcanvas, bezier bz, char *style, char *color)
{
  GtkWidget *canvas = gcanvas->canvas;
  int i, n;
  double t, step = 0.01;
  guint type = gnome_canvas_line_get_type ();
  GnomeCanvasItem *item;
  GList *pointf_list = NULL, *list;
  int num_points = 0;
  GnomeCanvasPoints *points;
  double x_prev = -1, y_prev = -1;
  int line_style = !strcmp(style, "solid") ? GDK_LINE_SOLID : GDK_LINE_ON_OFF_DASH;

  if(bz.sflag) {
    double sx = trans_x_coord(gcanvas, bz.sp.x);
    double sy = trans_y_coord(gcanvas, bz.sp.y);
    pointf_list = g_list_prepend(pointf_list, pointfof(sx, sy)); num_points++;
  }

  for (i = 0; i < bz.size - 3; i += 3) {
    double x0 = trans_x_coord(gcanvas, bz.list[i].x);
    double y0 = trans_y_coord(gcanvas, bz.list[i].y);
    double x1 = trans_x_coord(gcanvas, bz.list[i+1].x);
    double y1 = trans_y_coord(gcanvas, bz.list[i+1].y);
    double x2 = trans_x_coord(gcanvas, bz.list[i+2].x);
    double y2 = trans_y_coord(gcanvas, bz.list[i+2].y);
    double x3 = trans_x_coord(gcanvas, bz.list[i+3].x);
    double y3 = trans_y_coord(gcanvas, bz.list[i+3].y);

    //r(t)=(((1-t)*V0+3*t*V1)*(1-t)+3*t^2*V2)*(1-t)+t^3*V3
    for(n = 0; n <= 100; n++) {
      double x, y;
      t = step*n;
      x = (((1-t)*x0+3*t*x1)*(1-t)+3*pow(t, 2.0)*x2)*(1-t)+pow(t, 3.0)*x3;
      y = (((1-t)*y0+3*t*y1)*(1-t)+3*pow(t, 2.0)*y2)*(1-t)+pow(t, 3.0)*y3;
      pointf_list = g_list_prepend(pointf_list, pointfof(x, y)); num_points++;
    }
  }

  if(bz.eflag) {
    double ex = trans_x_coord(gcanvas, bz.ep.x);
    double ey = trans_y_coord(gcanvas, bz.ep.y);
    pointf_list = g_list_prepend(pointf_list, pointfof(ex, ey)); num_points++;
  }

  pointf_list = g_list_reverse(pointf_list);
  points = gnome_canvas_points_new(num_points);
  for(list = pointf_list, i = 0; list != NULL; list = list->next, i += 2) {
      pointf *p = (pointf*)(list->data);
      points->coords[i] = p->x; 
      points->coords[i+1] = p->y;
      free(p);
  }
  assert(i == 2*num_points);
  if(pointf_list != NULL)
    g_list_free(pointf_list);

  item = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (canvas)),
				type,
				"points", points,
				"fill_color", color,
				"width_units", 1.0,
				"line_style", line_style,
				"first_arrowhead", bz.sflag ? TRUE : FALSE,
				"last_arrowhead", bz.eflag ? TRUE : FALSE,
				"smooth", FALSE,
				"arrow_shape_a", 5.0,
				"arrow_shape_b", 10.0,
				"arrow_shape_c", 3.0,
				NULL);
  gcanvas->item_list = g_list_prepend(gcanvas->item_list, item);
}

GnomeCanvasItem *
draw_text(GraphCanvas *gcanvas, double x, double y, char *str, char *color)
{
  GtkWidget *canvas = gcanvas->canvas;
  guint type;
  GnomeCanvasItem *item;
  type = gnome_canvas_text_get_type ();
  item = gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (canvas)),
				type, 
				"text", str,
				"x", x,
				"y", y,
				"justification", GTK_JUSTIFY_CENTER,
				"font", FONT,
				"fill_color", color,
				NULL);
  gcanvas->item_list = g_list_prepend(gcanvas->item_list, item);
  return item;
}

void
draw_graph(Agraph_t* g, GraphCanvas *canvas)
{
  int i;
  node_t *n;
  edge_t *e;
  bezier bz;
  GnomeCanvasItem *item;

  //GD_drawing(g)->margin.x = GD_drawing(g)->margin.y = DEFAULT_MARGIN;
  /* determine final drawing size and scale to apply. */
  setup_graph(g);
  cleanup_canvas(canvas);
  init_canvas(canvas, (double)GD_bb(g).UR.x, (double)GD_bb(g).UR.y, GD_drawing(g)->scale);
  
  for (n = agfstnode(g); n; n = agnxtnode(g,n)) {
    double x1 = trans_x_coord(canvas, ND_coord_i(n).x) - POINTS(ND_width(n)/2.0);
    double y1 = trans_y_coord(canvas, ND_coord_i(n).y) - POINTS(ND_height(n)/2.0);
    double x2 = trans_x_coord(canvas, ND_coord_i(n).x) + POINTS(ND_width(n)/2.0);
    double y2 = trans_y_coord(canvas, ND_coord_i(n).y) + POINTS(ND_height(n)/2.0);
    char *color = agget(n,"color"); if(!color) color = COLOR;
    char *fillcolor = agget(n,"fillcolor"); 
    char *label = agget(n,"label"); if(!label) label = ND_label(n)->text;
    if(!fillcolor) {
      if(!strcmp(agget(n,"style"), "filled")) 
	fillcolor = color;
      else
	fillcolor = "lightgrey";
    }

    if(strcmp(ND_shape(n)->name, "record") == 0 || strcmp(ND_shape(n)->name, "box") == 0 || 
       strcmp(ND_shape(n)->name, "polygon") == 0)
      item = draw_rect (canvas, x1, y1, x2, y2, color, fillcolor); 
    else
      item = draw_ellipse (canvas, x1, y1, x2, y2, color, fillcolor);
    x1 = trans_x_coord(canvas, ND_coord_i(n).x);
    y1 = trans_y_coord(canvas, ND_coord_i(n).y);

    draw_text (canvas, x1, y1, label, COLOR);
  }

  for (n = agfstnode(g); n; n = agnxtnode(g,n)) {
    for (e = agfstout(g,n); e; e = agnxtout(g,e)) {
      bz = ED_spl(e)->list[0];
      draw_bezier (canvas, bz, LINE_STYLE, COLOR);

      if (ED_label(e)) {
	double x = trans_x_coord(canvas, ED_label(e)->p.x);
	double y = trans_y_coord(canvas, ED_label(e)->p.y);
	draw_text (canvas, x, y, ED_label(e)->text, COLOR);
      }
    }
  }

  //fflush(Output_file);
}

GraphCanvas *
create_canvas()
{
  GraphCanvas *gcanvas = g_new0(GraphCanvas, 1);
  gcanvas->canvas = gnome_canvas_new ();
  gcanvas->canvas_height = CANVAS_HEIGHT;
  gcanvas->canvas_width = CANVAS_WIDTH;
  gtk_widget_set_usize (gcanvas->canvas, CANVAS_WIDTH, CANVAS_HEIGHT);
  gnome_canvas_set_scroll_region (GNOME_CANVAS (gcanvas->canvas), 0.0, 0.0, 
				  gcanvas->canvas_width, gcanvas->canvas_height);
  return gcanvas;
}


/* end of mary */
