/* mary */

#include "graph_ops.h"
#include "draw.h"

void
layout_graph(Agraph_t *g)
{
  dot_cleanup(g);
  dot_layout(g);
}

Agraph_t *
create_graph(char *name)
{
  Agraph_t *g = agopen(name, AGDIGRAPH);
  Agsym_t  *a;

  if (!(a = agfindattr(g->proto->n, "fontsize")))
    a = agnodeattr(g, "fontsize", "");
  agxset(g->proto->n, a->index, FONTSIZE);

  if (!(a = agfindattr(g->proto->n, "fontname")))
    a = agnodeattr(g, "fontname", "");
  agxset(g->proto->n, a->index, FONTNAME);

  if (!(a = agfindattr(g->proto->n, "fontcolor")))
    a = agnodeattr(g, "fontcolor", "");
  agxset(g->proto->n, a->index, FONTCOLOR);
    
  if (!(a = agfindattr(g->proto->n, "fillcolor")))
    a = agnodeattr(g, "fillcolor", "");
  agxset(g->proto->n, a->index, FILLCOLOR);

  if (!(a = agfindattr(g->proto->n, "color")))
    a = agnodeattr(g, "color", "");
  agxset(g->proto->n, a->index, COLOR);
  
  if (!(a = agfindattr(g->proto->n, "style")))
    a = agnodeattr(g, "style", "");
  agxset(g->proto->n, a->index, STYLE);

  if (!(a = agfindattr(g->proto->n, "label")))
    a = agedgeattr(g, "label", "");
  agxset(g->proto->n, a->index, "");

  if (!(a = agfindattr(g->proto->e, "label")))
    a = agedgeattr(g, "label", "");
  agxset(g->proto->e, a->index, "");

  return g;
}

void
set_node_attr(Agraph_t *g, char *name, char *label, char *style, char *fillcolor)
{
  node_t *n;
  for (n = agfstnode(g); n; n = agnxtnode(g,n)) {
    if(!strcmp(ND_label(n)->text, name))
    break;
  }
  if(!n) return;
  ND_label(n) = make_label (strdup (label), 
			    ND_label(n)->fontsize, ND_label(n)->fontname, ND_label(n)->fontcolor,
			    n->graph);
  ND_shape(n)->initfn(n);
  agset(n, "label", label);
  agset(n, "style", style);
  agset(n, "fillcolor", fillcolor);
}

Agnode_t *
insert_node(Agraph_t *g, char *name, int do_layout)
{
  Agnode_t *n;
  n = agnode(g, name);//!!
  agset(n, "label", name);
  if(do_layout)
    layout_graph(g);
  return n;
}

Agedge_t *
insert_edge(Agraph_t *g, char *tail, char *head, char *label, int do_layout)
{
  Agnode_t *t = insert_node(g, tail, FALSE);
  Agnode_t *h = insert_node(g, head, FALSE);
  Agedge_t *e = agedge(g, t, h);
  if(label && strcmp(label,"") != 0)
    agset(e, "label", label);
  if(do_layout)
    layout_graph(g);
  return e;
}

void
delete_node(Agraph_t *g, char *name, int do_layout)
{
  Agnode_t *n = agfindnode(g, name);
  agDELnode(g, n);
  if(do_layout)
    layout_graph(g);
}

void
delete_edge(Agraph_t *g, char *tail, char *head, int do_layout)
{
  Agnode_t *t = agfindnode(g, tail);
  Agnode_t *h = agfindnode(g, head);
  Agedge_t *e = agfindedge(g, t, h);
  agDELedge(g, e);
  if(do_layout)
    layout_graph(g);
}

/* end of mary */
