/* mary */

#include "win.h"
#include "graph_ops.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>

static MainWin *win;

int
file_save(char *fname)
{
  if(!(Output_file = file_select(fname)))
    return FALSE;
  Output_lang = lang_select("ps",1);
  dotneato_write_one(win->g);
  return TRUE;
}

int
file_open(char *fname)
{
  Agraph_t *g;
  if(!freopen(fname,"r",stdin))//by deafault, dot reads graph from stdin
    return FALSE;
  if ((g = next_input_graph())) {
    win->g = g;
    layout_graph(win->g);
    draw_graph(win->g, win->canvas);
    return TRUE;
  }
  return FALSE;
}

/* callback for the OK button on the file selection dialog */
void
file_open_ok_clicked_cb(GtkButton *button, GtkFileSelection *fsel)
{
  char *fname;
  g_return_if_fail(GTK_IS_FILE_SELECTION(fsel));
  /* get the filename */
  fname = gtk_file_selection_get_filename(fsel);
  if(/* if no file was selected */
     !fname ||
     /* or we can't read it */
     !file_open(fname)) {
    GtkWidget *dlg;
    
    /* make a new dialog with the file selection as parent */
    dlg = gnome_error_dialog_parented(_("Can't open file"),
				      GTK_WINDOW(fsel));
    gtk_window_set_modal(GTK_WINDOW(dlg), TRUE);
  } else {
    /* we have read the file without a problem so just
       close the open dialog */
    gtk_widget_destroy(GTK_WIDGET(fsel));
  }
}

/* callback for the OK button on the file selection dialog */
void
file_save_ok_clicked_cb(GtkButton *button, GtkFileSelection *fsel)
{
  char *fname;
  g_return_if_fail(GTK_IS_FILE_SELECTION(fsel));
  /* get the filename */
  fname = gtk_file_selection_get_filename(fsel);
  if(/* if no file was selected */
     !fname ||
     /* or we can't write it */
     !file_save(fname)) {
    GtkWidget *dlg;
    /* make a new dialog with the file selection as parent */
    dlg = gnome_error_dialog_parented(_("Can't save file"),
				      GTK_WINDOW(fsel));
    gtk_window_set_modal(GTK_WINDOW(dlg), TRUE);
  } else {
    if(win->fname)
      g_free(win->fname);
    win->fname = g_strdup(fname);
    gtk_widget_destroy(GTK_WIDGET(fsel));
  }
}

void
help_about_activate_cb(GtkMenuItem *menuitem, gpointer data)
{
  static GtkWidget* dialog = NULL;
  /* a list of authors */
  const char *authors[] = {
    "David Dobkin",
    "John Ellson",
    "Emden Gansner",
    "Eleftherios Koutsofios",
    "Stephen North",
    "Kiem-Phong Vo",
    "Gordon Woodhull",
    NULL
  };
  /* if the dialog is already created, just show and raise it
     (just in case it is obscured or minimized) */
  if(dialog) {
    gdk_window_show(dialog->window);
    gdk_window_raise(dialog->window);
    return;
  }
  /* create the dialog */
  dialog = gnome_about_new("Graphviz", "1.10",//VERSION, 
			   "2003",
			   authors,
			   "Very simple graphical frontend to dot, written by Maria Landina. Please note that program understands only two languages: DOT as input and POSTSCRIPT as output.",
			   NULL);
  /* set the parent of the dialog to be the main application
     window, this will help windowmanagers */
  gnome_dialog_set_parent(GNOME_DIALOG(dialog), GTK_WINDOW(win->window));  
  /* when the dialog gets destroyed, the function gtk_widget_destroyed
     is a utility function from gtk that will NULL the pointer that is
     passed as the data of the signal */
  gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
		     GTK_SIGNAL_FUNC(gtk_widget_destroyed),
		     &dialog);  
  gtk_widget_show(dialog);
}

void
file_open_activate_cb(GtkMenuItem *menuitem, gpointer user_data)
{
  GtkFileSelection *fsel;
  /* make a new gtk file selection */
  fsel = GTK_FILE_SELECTION(gtk_file_selection_new(_("Open")));
  /* if we have a current filename add it to the box */
  //if(win->fname)
  //gtk_file_selection_set_filename(fsel, win->fname);    
  /* Connect the signals for Ok and Cancel */
  gtk_signal_connect(GTK_OBJECT(fsel->ok_button), "clicked",
		     GTK_SIGNAL_FUNC(file_open_ok_clicked_cb), fsel);
  /* connect gtk_widget_destroy to the cancel button, so that
     we just kill the dialog */
  gtk_signal_connect_object
    (GTK_OBJECT(fsel->cancel_button), "clicked",
     GTK_SIGNAL_FUNC(gtk_widget_destroy), 
     GTK_OBJECT(fsel));
  /* position this dialog where the mouse is, this is normaly
     set according to preferences for gnome dialogs, but there
     is yet no file selection dialog in gnome */
  gtk_window_position(GTK_WINDOW(fsel), GTK_WIN_POS_MOUSE);
  /* set the parent of the dialog to be the main application
     window, this will help windowmanagers */
  gtk_window_set_transient_for(GTK_WINDOW(fsel),GTK_WINDOW(win->window));
  /* show the dialog */
  gtk_widget_show(GTK_WIDGET(fsel));
}

void
file_save_as_activate_cb(GtkMenuItem *menuitem, gpointer data)
{
  GtkFileSelection *fsel;
  /* make a new gtk file selection */
  fsel = GTK_FILE_SELECTION(gtk_file_selection_new(_("Save As...")));
  /* if we have a current filename add it to the box */
  if(win->fname)
    gtk_file_selection_set_filename(fsel, win->fname);
  /* Connect the signals for Ok and Cancel */
  gtk_signal_connect (GTK_OBJECT (fsel->ok_button), "clicked",
		      GTK_SIGNAL_FUNC (file_save_ok_clicked_cb), fsel);
  /* connect gtk_widget_destroy to the cancel button, so that
     we just kill the dialog */
  gtk_signal_connect_object
    (GTK_OBJECT (fsel->cancel_button), "clicked",
     GTK_SIGNAL_FUNC (gtk_widget_destroy), 
     GTK_OBJECT(fsel));
  /* position this dialog where the mouse is, this is normaly
     set according to preferences for gnome dialogs, but there
     is yet no file selection dialog in gnome */
  gtk_window_position (GTK_WINDOW (fsel), GTK_WIN_POS_MOUSE);
  /* set the parent of the dialog to be the main application
     window, this will help windowmanagers */
  gtk_window_set_transient_for(GTK_WINDOW(fsel), GTK_WINDOW(win->window));
  /* show the dialog */
  gtk_widget_show (GTK_WIDGET(fsel));
}

void
file_save_activate_cb (GtkMenuItem *menuitem, gpointer data)
{
  /* if no default filename is set, call on_save_as_activate */
  if(!(win->fname)) {
    file_save_as_activate_cb(NULL, NULL);
  }/* try writing the file now */
  else if(!file_save(win->fname)) {
    /* tell the user what's wrong according to his preferences,
       it can be on the status bar or a dialog */
    GtkWidget *dlg;    
    /* make a new dialog with the file selection as parent */
    dlg = gnome_error_dialog_parented(_("Can't save file"),
				      GTK_WINDOW(win->window));
    gtk_window_set_modal(GTK_WINDOW(dlg), TRUE);
  }
}

static GnomeUIInfo file_menu[] =
{
  GNOMEUIINFO_MENU_OPEN_ITEM (file_open_activate_cb, NULL),
  GNOMEUIINFO_MENU_SAVE_ITEM (file_save_activate_cb, NULL),
  GNOMEUIINFO_MENU_SAVE_AS_ITEM (file_save_as_activate_cb, NULL),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_MENU_EXIT_ITEM (gtk_main_quit, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo help_menu[] =
{
  GNOMEUIINFO_MENU_ABOUT_ITEM (help_about_activate_cb, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo main_menu[] =
{
  GNOMEUIINFO_MENU_FILE_TREE(file_menu),
  GNOMEUIINFO_MENU_HELP_TREE(help_menu),
  GNOMEUIINFO_END
};

static GnomeUIInfo menu_toolbar [] = {
  GNOMEUIINFO_ITEM_STOCK(NULL, _("Open"),
			 GTK_SIGNAL_FUNC (file_open_activate_cb), 
			 GNOME_STOCK_PIXMAP_OPEN),
  GNOMEUIINFO_ITEM_STOCK(NULL, _("Save"),
			 GTK_SIGNAL_FUNC (file_save_activate_cb), 
			 GNOME_STOCK_PIXMAP_SAVE),
  GNOMEUIINFO_ITEM_STOCK(NULL, _("Exit"),
			 GTK_SIGNAL_FUNC(gtk_main_quit),
			 GNOME_STOCK_PIXMAP_EXIT),
  GNOMEUIINFO_END
};

char *
replace_delim_symbols(char* str)
{
  int len = strlen(str);
  if(str[len-1]==',' || str[len-1]==';') {
    str[len-1]='\0';
    len--;
  }
  if(str[0]=='\"') {
    assert(str[len-1]=='\"');
    str[len-1]='\0';
    str++;
  }
  return str;
}

int
read_graph()
{
  char s[256];  
  char tmp1[64], tmp2[64], tmp3[64];
  char tail[64], head1[64], name[64], label1[64], style1[64], fillcolor1[64];
  char *head, *label, *style, *fillcolor;

  fgets(s, 255, stdin);

  sscanf(s, "%s %s %s", tmp1, tmp2, tmp3);
  if(strcmp("callGraph", tmp1) != 0)
    return TRUE;

  if(!strcmp("callgraph", tmp3))//begin of callgraph description
  {
    return TRUE;
  }

  else if(!strcmp("};",tmp2))//end of callgraph description
  {
    return FALSE;//must involve timer remove
    //gtk_timeout_remove(win->idle_id);
  }

  else if(!strcmp("->", tmp3))//edge
  {
    sscanf(s, "callGraph\t\t%s -> %s", tail, head1);
    head = replace_delim_symbols(head1);

    insert_edge(win->g, tail, head, "", TRUE);
    draw_graph(win->g, win->canvas);
    gtk_main_iteration_do(FALSE);
    return TRUE;
  }
  else if(!strcmp("[",tmp3))//node
  {
    sscanf(s,"callGraph\t\t%s [ label = %s style = %s fillcolor = %s ];",name,label1,style1,fillcolor1);
    label = replace_delim_symbols(label1);
    style = replace_delim_symbols(style1);
    fillcolor = replace_delim_symbols(fillcolor1);

    set_node_attr(win->g, name, label, style, fillcolor);
    draw_graph(win->g, win->canvas);
    gtk_main_iteration_do(FALSE);
    return TRUE;
  }
  
  return TRUE;
}

void
destroy_main_window(GtkWidget *widget, gpointer user_data)
{
  gtk_widget_destroy(win->window);
  gtk_main_quit();
}

MainWin*
create_main_win()
{
  GtkWidget *scrolledwindow;
  Agraph_t *g;
 
  win = g_new0(MainWin,1);
  win->window = gnome_app_new("graphviz", "Grapviz");
  
  scrolledwindow = gtk_scrolled_window_new (NULL, NULL);  gtk_widget_show (scrolledwindow);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow), 
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  gnome_app_set_contents(GNOME_APP(win->window), scrolledwindow);

  win->canvas = create_canvas();
  gtk_container_add(GTK_CONTAINER (scrolledwindow), win->canvas->canvas);
 
  /* add the menus to the main window */
  gnome_app_create_menus(GNOME_APP(win->window), main_menu);
  /* add a toolbar */
  gnome_app_create_toolbar(GNOME_APP(win->window), menu_toolbar);

  gtk_widget_show_all(win->window);

  win->fname = NULL;
  win->g = create_graph("callgraph");
  win->idle_id = gtk_timeout_add(TIMEOUT, (GtkFunction) read_graph, NULL);
  
  gtk_signal_connect(GTK_OBJECT(win->window), "destroy",
		     GTK_SIGNAL_FUNC(destroy_main_window),
		     &(win->window));
  return win;
}


/* end of mary */
