#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>        
#include <assert.h>  
#include <fcntl.h>                                                                      
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h> 

void ltdbInit(const char *progfile) {
  { // check file exists :-)
    struct stat buf;
    if (stat(progfile,&buf)) {
      fprintf(stderr, "ltdb: unable find program file (trying '%s')\n", progfile); 
      exit(1);
    }
  }
  int attpid = getpid();
  if (fork()) {     // this process ...
    while (true) {} // ... should wait for debugger...
  } else { 
    int gdbpid;
    int toDbg[2],fromDbg[2];
    pipe(toDbg); pipe(fromDbg); 
    if ((gdbpid = fork())) { // comm process

      int console=1; //open("/dev/tty",O_RDWR);
      assert(console >= 0);
      dup2(console,0);
      dup2(console,1);
      dup2(console,2);

      int maxfd = fromDbg[0];
      while(true) { 
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(0,&fds);
	FD_SET(fromDbg[0],&fds);
      
	struct timeval tv;
	tv.tv_sec =  0; 
	tv.tv_usec = 100000;
      
	if (!select(maxfd+1, &fds, NULL, NULL, &tv)) { 
	  int status;
	  if (gdbpid == waitpid(gdbpid, &status, WNOHANG)) {
	    if (WIFEXITED(status))   { fprintf(stderr,"Debugger exited normally\n");  break; } 
	    if (WIFSIGNALED(status)) { fprintf(stderr,"Debugger crashed\n"); break; } 
	    fprintf(stderr,"Unrecognized debugger finish\n"); break;
	  }
	}
	char c;
	if (FD_ISSET(0,&fds)) {
	  read(0,&c,1);
	  write(toDbg[1],&c,1);
	} else if (FD_ISSET(fromDbg[0],&fds)) {
	  read(fromDbg[0],&c,1);
	  write(1,&c,1);
	}
      }
      kill(attpid,SIGKILL); // kill attached process
    } else { // gdb process
      dup2(toDbg[0],0);
      dup2(fromDbg[1],1);
      dup2(fromDbg[1],2);
      char attpidstr[0x10];
      snprintf(attpidstr,sizeof(attpidstr),"%d",attpid);
      fprintf(stderr, "\n *** Attaching GDB to pid %s...\n\n", attpidstr);
      execlp("gdb", "gdb", progfile, attpidstr, NULL);
      fprintf(stderr, "execlp for gdb failed\n");
    }
  }
  exit(1); // both processes
} 

extern "C" {
void  registerFeature(const char *name, const char *comment, void* (*entry)(int, void *arg));
}

class LTDBInit {
public:
  static void* entry(int cmd, void *progfile) { 
    switch (cmd) {
    case 0: 
      return NULL; // Initialization
    default:
      ltdbInit((const char*)progfile);  // avoid start at init stage
      return NULL; 
    }
  }
  LTDBInit() {
    registerFeature("ltdb", "Lightweight TDB support", entry);
  }
} LTDBInit_inistance;

