#include "prof.h"
#include <signal.h>
#include <pthread.h>

static int prof_flag = PROF_UNDEF;
static char prof_server[256];
static int prof_port = 1025;
static int prof_output = PROF_SCREEN;
static pthread_t prof_thread;

/********************************************************************/

static void prof_deinit()
{
  pthread_exit(0);
}

/********************************************************************/

static void *prof_init(void *arg)
{
  struct sigaction sigact;
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = &prof_deinit;
  sigaction(SIGTERM, &sigact, NULL);
  prof_com_server(prof_server, prof_port,  prof_output);
  return 0;
}

/********************************************************************/

static void prof_start()
{
  pthread_create(&prof_thread, 0, &prof_init, 0);
}

/********************************************************************/

static void prof_stop()
{
  prof_com_client(prof_server, prof_port, "close");
  pthread_kill(prof_thread, SIGTERM);
  pthread_join(prof_thread, 0);
}

/********************************************************************/

void prof(int cmd, void *arg)
{
  char hostname[256];
  gethostname(hostname, sizeof(hostname));
  local_hostname(hostname);

  if (cmd > 1 && prof_flag == PROF_UNDEF) {
    prof_flag = PROF_NONE;
    char **argv = (char **) arg;
    
    int i;
    for (i = 1; i < cmd - 1; i ++)
      if (!strcmp(argv[i], "-prof-server")) {
        strcpy(prof_server, argv[i + 1]);
        prof_flag = PROF_ACTIVE;
        i ++;
      } else if (!strcmp(argv[i], "-prof-port")) {
        int port = atoi(argv[i + 1]);
        char s[8];
        sprintf(s, "%d", port);
        s[8] = '\0';
        if (!strcmp(s, argv[i + 1]))
          prof_port = port;
        i ++;
      } else if (!strcmp(argv[i], "-prof-output")) {
        if (!strcmp(argv[i + 1], "window"))
          prof_output = PROF_WINDOW;
        else if (!strcmp(argv[i + 1], "both"))
          prof_output = PROF_BOTH;
        i ++;
      }

    if (prof_flag == PROF_ACTIVE && !strcmp(hostname, prof_server))
      prof_start();
  } else if (!cmd && prof_flag == PROF_ACTIVE && !strcmp(hostname, prof_server)) {
    sleep(2);
    prof_stop();
  } else if (cmd == 1 && prof_flag == PROF_ACTIVE)
    prof_get_stat(prof_server, prof_port, (char *) arg);
}
