#include "prof.h"
#include "prof_hosts.h"
#include <fcntl.h>

/********************************************************************/

static int prof_proc_stat(pid_t pid, double *utime, double *stime, long *rss)
{
  char stat_file[32];
  snprintf(stat_file, sizeof(stat_file), "/proc/%d/stat", pid);
  int stat_fd = open(stat_file, O_RDONLY);
  if (stat_fd == -1)
    return errno;

  char stat[1024];
  size_t stat_len = read(stat_fd, stat, sizeof(stat));
  if (stat_len == -1) {
    close(stat_fd);
    return errno;
  }
  close(stat_fd);
  stat[stat_len] = '\0';

  unsigned long _utime, _stime;
  sscanf(stat, "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %lu %lu %*d %*d %*d %*d %*d %*d %*u %*u %ld",
          &_utime, &_stime, rss);
  *utime = _utime / 100.;
  *stime = _stime / 100.;
  *rss = *rss * getpagesize() / 1024;
  return 0;
}

/********************************************************************/

int prof_get_stat(char *server, int port, char *funcname)
{
  pid_t pid = getpid();
  double utime, stime;
  long rss;
  int err = prof_proc_stat(pid, &utime, &stime, &rss);
  if (err)
    return err;

  char nodename[256];
  if (gethostname(nodename, sizeof(nodename)) == -1)
    return errno;
  local_hostname(nodename);
  
  char message[1024];
  snprintf(message, sizeof(message), "%s %d %s %f %f %ld", nodename, pid, funcname, utime, stime, rss);
  return prof_com_client(server, port, message);
}

/********************************************************************/

void prof_print_stat(char *message)
{
  char nodename[256], funcname[256];
  pid_t pid;
  double utime, stime;
  unsigned long rss;
  sscanf(message, "%s %d %s %lf %lf %lu", nodename, &pid, funcname, &utime, &stime, &rss);
  printf("%s   (%d) %s:   utime(sec)=%.6f   stime(sec)=%.6f   rss(Kb)=%lu\n", nodename, pid, funcname, utime, stime, rss);
}

/********************************************************************/

void local_hostname(char *hostname)
{
  char *p = strchr(hostname, '.');
  if (p)
    *p = '\0';
  int i;
  for (i = 0; strcmp(hosts[i], ""); i += 2)
    if (!strcmp(hosts[i], hostname)) {
      strcpy(hostname, hosts[i + 1]);
      break;
    }
}
