#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "xprof.h"
#include "xprof_hosts.h"

struct view_menu_col {
  gchar title[32];
  gboolean visible;
};

struct view_menu_col view_menu_cols[] = {{"/View/node", TRUE}, {"/View/pid", TRUE}, {"/View/function", TRUE}, {"/View/user time", TRUE}, {"/View/system time", TRUE}, {"/View/resident set size", TRUE}};

/********************************************************************/

void xprof_show_or_hide_column(GtkWidget *menu_item, gpointer data)
{
  gchar *path = gtk_item_factory_path_from_widget(menu_item);
  int column;
  for (column = 0; column < 6 && !strstr(path, view_menu_cols[column].title); column ++)
    ;
  if (column < 6) {
    view_menu_cols[column].visible = !view_menu_cols[column].visible;
    gtk_clist_set_column_visibility(GTK_CLIST(env.stat_lst), column, view_menu_cols[column].visible);
  }
}

/********************************************************************/

void xprof_file_save(GtkWidget *widget, gpointer file_dlg)
{
  gchar *file_name = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_dlg));
  if (file_name != NULL) {
    FILE *fp = fopen(file_name, "w+");
    if (fp != NULL) {
      gchar *text[6];
      int i, j;
      for (i = 0; i < env.rows; i ++) {
        for (j = 0; j < 6; j ++)
          gtk_clist_get_text(GTK_CLIST(env.stat_lst), i, j, &(text[j]));
        fprintf(fp, "%s   (%s) %s:   utime(sec)=%s   stime(sec)=%s   rss(Kb)=%s\n", text[0], text[1], text[2], text[3], text[4], text[5]);
      }
      fclose(fp);
      gtk_widget_destroy(GTK_WIDGET(file_dlg));
    }
  }
}

/********************************************************************/

void xprof_file_cancel(GtkWidget *widget, gpointer file_dlg)
{
  gtk_widget_destroy(GTK_WIDGET(file_dlg));
}

/********************************************************************/

void xprof_save_stat(GtkWidget *widget, gpointer data)
{
  GtkWidget *file_dlg = gtk_file_selection_new("Save stat into file");
  gtk_window_set_modal(GTK_WINDOW(file_dlg), TRUE);
  gtk_signal_connect(GTK_OBJECT(file_dlg), "destroy", GTK_SIGNAL_FUNC(xprof_file_cancel), file_dlg);
  GtkWidget *ok_btn = GTK_FILE_SELECTION(file_dlg)->ok_button;
  gtk_signal_connect(GTK_OBJECT(ok_btn), "clicked", GTK_SIGNAL_FUNC(xprof_file_save), file_dlg);
  GtkWidget *cancel_btn = GTK_FILE_SELECTION(file_dlg)->cancel_button;
  gtk_signal_connect(GTK_OBJECT(cancel_btn), "clicked", GTK_SIGNAL_FUNC(xprof_file_cancel), file_dlg);
  gtk_widget_show(file_dlg);  
}

/********************************************************************/

void xprof_about_close(GtkWidget *widget, gpointer about_dlg)
{
  gtk_widget_destroy(GTK_WIDGET(about_dlg));
}

/********************************************************************/

void xprof_about(GtkWidget *widget, gpointer data)
{
  GtkWidget *about_dlg = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(about_dlg), "About openTS Profiler");
  gtk_window_set_default_size(GTK_WINDOW(about_dlg), 300, 100);
  gtk_window_set_modal(GTK_WINDOW(about_dlg), TRUE);
  gtk_signal_connect(GTK_OBJECT(about_dlg), "destroy", GTK_SIGNAL_FUNC(xprof_about_close), about_dlg);

  GtkWidget *label1 = gtk_label_new("openTS Profiler");
  GtkWidget *label2 = gtk_label_new("version 0.1");

  GtkWidget *close_btn = gtk_button_new_with_label("Close");
  gtk_signal_connect(GTK_OBJECT(close_btn), "clicked", GTK_SIGNAL_FUNC(xprof_about_close), about_dlg);

  GtkWidget *about_box = GTK_DIALOG(about_dlg)->vbox;
  gtk_box_pack_start(GTK_BOX(about_box), label1, FALSE, TRUE, 10);
  gtk_box_pack_start(GTK_BOX(about_box), label2, FALSE, TRUE, 5);
  GtkWidget *action_box = GTK_DIALOG(about_dlg)->action_area;
  gtk_box_pack_start(GTK_BOX(action_box), close_btn, FALSE, FALSE, 0);

  gtk_widget_show_all(about_dlg);  
}

/********************************************************************/

gint xprof_click_column(GtkWidget *widget, gint column, gpointer data)
{
  env.sort_col = column;
  env.sort_type = (env.sort_type == GTK_SORT_ASCENDING) ? GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
  xprof_sort_stat();
  return 0;
}

/********************************************************************/

gint xprof_close(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit();
  return 0;
}

/********************************************************************/

void xprof_deinit()
{
  pthread_exit(NULL);
}
  
/********************************************************************/

void *xprof_init(void *arg)
{
  struct sigaction sigact;
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = &xprof_deinit;
  sigaction(SIGTERM, &sigact, NULL);
  xprof_com_server();
  return NULL;
}

/********************************************************************/

int xprof_start(pthread_t *xprof_thread)
{
  struct sigaction sigact;
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = &xprof_show_stat;
  sigaction(SIGUSR1, &sigact, NULL);
 
  if (pthread_create(xprof_thread, NULL, &xprof_init, NULL) == -1)
    return errno;
  return 0;
}

/********************************************************************/

void xprof_stop(pthread_t xprof_thread)
{
  xprof_com_client("close");
  pthread_kill(xprof_thread, SIGTERM);
  pthread_join(xprof_thread, NULL);
}

/********************************************************************/

static void local_hostname(char *hostname)
{
  char *p = strchr(hostname, '.');
  if (p)
    *p = '\0';
  int i;
  for (i = 0; strcmp(hosts[i], ""); i += 2)
    if (!strcmp(hosts[i], hostname)) {
      strcpy(hostname, hosts[i + 1]);
      break;
    }
}

/********************************************************************/

int main(int argc, char *argv[])
{
  gtk_init(&argc, &argv);

  if (gethostname(env.server, sizeof(env.server)) == -1)
    return errno;
  local_hostname(env.server);
  env.port = 1025;  
  int i;
  for (i = 1; i < argc - 1; i ++)
    if (!strcmp(argv[i], "-prof-port")) {
      int port = atoi(argv[i + 1]);
      char s[8];
      sprintf(s, "%d", port);
      s[8] = '\0';
      if (!strcmp(s, argv[i + 1])) {
        env.port = port;
        break;
      }
    }  
  env.port += 1;
  env.parent = getpid();

  GtkWidget *main_wnd = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(main_wnd), "openTS Profiler");
  gtk_window_set_default_size(GTK_WINDOW(main_wnd), 800, 600);
  gtk_signal_connect(GTK_OBJECT(main_wnd), "destroy", GTK_SIGNAL_FUNC(xprof_close), NULL);

  GtkItemFactoryEntry menu_items[] = {
    {"/_File", NULL, NULL, 0, "<Branch>"},
    {"/File/save _As ...", "<control>S", xprof_save_stat, 0, NULL},
    {"/File/sep", NULL, NULL, 0, "<Separator>"},
    {"/File/_Quit", "<control>Q", gtk_main_quit, 0, NULL},
    {"/_View", NULL, NULL, 0, "<Branch>"},
    {view_menu_cols[0].title, NULL, NULL, 0, "<CheckItem>"},
    {view_menu_cols[1].title, NULL, NULL, 0, "<CheckItem>"},
    {view_menu_cols[2].title, NULL, NULL, 0, "<CheckItem>"},
    {view_menu_cols[3].title, NULL, NULL, 0, "<CheckItem>"},
    {view_menu_cols[4].title, NULL, NULL, 0, "<CheckItem>"},
    {view_menu_cols[5].title, NULL, NULL, 0, "<CheckItem>"},
    {"/_Help", NULL, NULL, 0, "<LastBranch>"},
    {"/Help/_About", NULL, xprof_about, 0, NULL}
  };
  GtkAccelGroup *menu_accel = gtk_accel_group_new();
  GtkItemFactory *menu_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", menu_accel);
  gtk_item_factory_create_items(menu_factory, (gint) (sizeof(menu_items) / sizeof(menu_items[0])), menu_items, NULL);
  GtkWidget *view_item;
  gint item_num;
  for (item_num = 0; item_num < 6; item_num ++) {
    view_item = gtk_item_factory_get_widget(menu_factory, view_menu_cols[item_num].title);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(view_item), TRUE);
    gtk_signal_connect(GTK_OBJECT(view_item), "toggled", GTK_SIGNAL_FUNC(xprof_show_or_hide_column), NULL);
  }
  gtk_window_add_accel_group(GTK_WINDOW(main_wnd), menu_accel);
  GtkWidget *menu_bar = gtk_item_factory_get_widget(menu_factory, "<main>");
  
  GtkWidget *scrol_wnd = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrol_wnd), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  gchar *titles[] = {"node", "pid", "function", "user time (sec)", "system time (sec)", "resident set size (Kb)"};
  env.stat_lst = gtk_clist_new_with_titles(6, titles);
  gtk_clist_set_column_width(GTK_CLIST(env.stat_lst), 0, 80);
  gtk_clist_set_column_width(GTK_CLIST(env.stat_lst), 1, 50);
  gtk_clist_set_column_width(GTK_CLIST(env.stat_lst), 2, 250);
  gtk_clist_set_column_width(GTK_CLIST(env.stat_lst), 3, 100);
  gtk_clist_set_column_width(GTK_CLIST(env.stat_lst), 4, 100);
  gtk_clist_set_column_width(GTK_CLIST(env.stat_lst), 5, 120);
  env.sort_col = 3;
  env.sort_type = GTK_SORT_DESCENDING;
  env.rows = 0;
  env.row_data[0] = '\0';
  gtk_signal_connect(GTK_OBJECT(env.stat_lst), "click_column", GTK_SIGNAL_FUNC(xprof_click_column), NULL);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrol_wnd), env.stat_lst);

  GtkWidget *stat_box = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(stat_box), 20);
  gtk_box_pack_start(GTK_BOX(stat_box), scrol_wnd, TRUE, TRUE, 0);

  GtkWidget *main_box = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_box), menu_bar, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_box), stat_box, TRUE, TRUE, 0);
  
  gtk_container_add(GTK_CONTAINER(main_wnd), main_box);
  gtk_widget_show_all(main_wnd);

  pthread_t xprof_thread;
  xprof_start(&xprof_thread);
 
  gtk_main();

  xprof_stop(xprof_thread);
  return 0;
}
