#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <dmpi.h>

enum {
  TRACE_UNDEF = -1,
  TRACE_NONE,
  TRACE_RECORD,
  TRACE_PLAY
};

static int trace_flag = TRACE_UNDEF;
static char filename[256];
static int fd;

/********************************************************************/

static void *my_trace_record(int size, void *msg)
{
  write(fd, &size, sizeof(int));
  if (msg) {
    fprintf(stderr, "record size=%d\n", size);
    write(fd, msg, size);
    return msg;
  }
  return 0;
}

/********************************************************************/

static void *my_trace_play(int cmd)
{
  int size;
  if (read(fd, &size, sizeof(int)) > 0) {
    if (!cmd) // 0 - read count, 1 - read message
      return (void *) size;
    
    fprintf(stderr, "play size=%d\n", size);
    if (size) {
      void *msg = malloc(size);
      read(fd, msg, size);
      return msg;
    }
  } else
    trace_flag = TRACE_NONE;
  return 0;
}

/********************************************************************/

void *my_trace(int cmd, void *arg)
{
  if (trace_flag == TRACE_UNDEF) {
    trace_flag = TRACE_NONE;
    char **argv = (char **) arg;

    int i;
    for (i = 1; i < cmd - 1 && trace_flag != TRACE_RECORD && trace_flag != TRACE_PLAY; i ++) {
      if (!strcmp(argv[i], "-trace-record"))
        trace_flag = TRACE_RECORD;
      else if (!strcmp(argv[i], "-trace-play"))
        trace_flag = TRACE_PLAY;
    }
    
    if (trace_flag == TRACE_NONE)
      return 0;
    
    if (trace_flag == TRACE_RECORD || trace_flag == TRACE_PLAY) {
      int rank;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      sprintf(filename, "%s.%d", argv[i], rank);
    }
 
    if (trace_flag == TRACE_RECORD)
      fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY, 0644);
    else if (trace_flag == TRACE_PLAY) {
      if (access(filename, R_OK))
        trace_flag = TRACE_NONE;
      else
        fd = open(filename, O_RDONLY, 0644);
    }    
  } else if (trace_flag == TRACE_RECORD)
    return(my_trace_record(cmd, arg));
  else if (trace_flag == TRACE_PLAY)
    return(my_trace_play(cmd));
  return (void *) trace_flag;
}
