/*
 * Dynamic MPI implementation
 * Somewhere in Spring 2002 by var@skif.botik.ru
 * Extended by madmax@botik.ru, A.Inyukhin, E.Stepanov & others
 */

#ifndef DMPI_H
#define DMPI_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <netdb.h>

struct mpi_portal_status { 
  int  MPI_SOURCE;
  int  MPI_TAG;
  int  MPI_ERROR;
// Should be enough for most MPI implementations
  unsigned char real_status[64];
};

struct mpi_portal_request {
 // Should be enough for most MPI implementations
  unsigned char real_request[64];
};

#define _MPI_Comm     int
#define _MPI_Datatype int
#define _MPI_Status   struct mpi_portal_status
#define _MPI_Request  struct mpi_portal_request

#define _MPI_SUCCESS        0
#define _MPI_ERR_OTHER      (-1)
#define _MPI_ANY_TAG        (-1)
#define _MPI_ANY_SOURCE     (-1)
#define _MPI_STATUS_IGNORE  ((_MPI_Status*)(-1))
#define _MPI_COMM_WORLD     ((_MPI_Comm)(-2))
#define _MPI_COMM_SELF      ((_MPI_Comm)(-1))
#define _MPI_COMM_NULL      ((_MPI_Comm)(0))
#define _MPI_BYTE           (1)

struct mpi_portal_t {
  int version;
  const char *revision;
  const char *name;
  const char *build_info;
  int (*_MPI_Init)(int *pargc, char ***pargv);
  int (*_MPI_Error_string)(int errcode, char *msg, int *plen);
  int (*_MPI_Barrier)(_MPI_Comm comm);
  int (*_MPI_Comm_rank)(_MPI_Comm comm, int *rank);
  int (*_MPI_Comm_size)(_MPI_Comm comm, int *psize);
  int (*_MPI_Finalize)(void);
  int (*_MPI_Get_count)(_MPI_Status *stat, _MPI_Datatype dtype, int *count);
  int (*_MPI_Iprobe)(int src,int tag,_MPI_Comm comm,int *flag,
		     _MPI_Status *stat);
  int (*_MPI_Recv)(void *buf, int count, _MPI_Datatype dtype,
		  int src, int tag, _MPI_Comm comm, _MPI_Status *stat);
  int (*_MPI_Send)(void *buf, int count, _MPI_Datatype dtype, int dest,
		  int tag, _MPI_Comm comm);
  int (*_MPI_Isend)(void *buf, int count, _MPI_Datatype dtype,int dest, 
                  int tag, _MPI_Comm comm, _MPI_Request *req);
  int (*_MPI_Test)(_MPI_Request *req, int *flag, _MPI_Status *stat);
  int (*_MPI_Wait)(_MPI_Request *req, _MPI_Status *stat);
  double (*_MPI_Wtime)(void);
};

#define MAXSOPATH 1024
extern char dmpi_so[MAXSOPATH];  
extern int  dmpi_meta;

// Something for UniMPI implementation

#ifndef DMPI_MALLOC
#define DMPI_MALLOC malloc
#define DMPI_FREE   free
#endif

// Redirection for user

extern struct mpi_portal_t *mpi_portal;

#ifndef DMPI_IMPLEMENTATION
//#define MPI_Init         (*mpi_portal->_MPI_Init)
#define MPI_Error_string (*mpi_portal->_MPI_Error_string)
#define MPI_Barrier      (*mpi_portal->_MPI_Barrier)
#define MPI_Comm_rank    (*mpi_portal->_MPI_Comm_rank)
#define MPI_Comm_size    (*mpi_portal->_MPI_Comm_size)
#define MPI_Finalize     (*mpi_portal->_MPI_Finalize)
#define MPI_Get_count    (*mpi_portal->_MPI_Get_count)
#define MPI_Iprobe       (*mpi_portal->_MPI_Iprobe)
#define MPI_Recv         (*mpi_portal->_MPI_Recv)
#define MPI_Isend        (*mpi_portal->_MPI_Isend)
#define MPI_Test         (*mpi_portal->_MPI_Test)
#define MPI_Wait         (*mpi_portal->_MPI_Wait)
#define MPI_Send         (*mpi_portal->_MPI_Send)
#define MPI_Wtime        (*mpi_portal->_MPI_Wtime)
#define MPI_COMM_WORLD   _MPI_COMM_WORLD
#define MPI_COMM_SELF    _MPI_COMM_SELF
#define MPI_COMM_NULL    _MPI_COMM_NULL
#define MPI_SUCCESS      _MPI_SUCCESS
#define MPI_ERR_OTHER    _MPI_ERR_OTHER
#define MPI_BYTE         _MPI_BYTE
#define MPI_ANY_SOURCE   _MPI_ANY_SOURCE
#define MPI_ANY_TAG      _MPI_ANY_TAG
#define MPI_Comm         _MPI_Comm
#define MPI_Datatype     _MPI_Datatype
#define MPI_Status       _MPI_Status
#define MPI_Request      _MPI_Request
#define MPI_STATUS_IGNORE _MPI_STATUS_IGNORE

#ifdef __cplusplus
extern "C" {
#endif
	extern int MPI_Init (int *a, char ***b); // DMPI init
//	extern int MPI_Comm_rank (_MPI_Comm comm, int *rank);
#ifdef __cplusplus
}
#endif

#endif /* DMPI_IMPLEMENTATION */
#endif /* DMPI_H */
