#include <trt>
typedef void* (*entry_t)(int cmd, void *arg);
extern "C" {
	void registerFeature(const char *name, const char *comment, entry_t entry);
	entry_t featureEntry(const char *name);
}
using namespace ts;
namespace gs {

class EstimatorData {
	int n;
	int *a, *b;
	long bw;
public:
	EstimatorData(long bw_=100) {
		bw = bw_;
		n = realsuperSize*(realsuperSize-1)/2;
		a = new int[n];
		b = new int[n];
		int k = 0;
		for (int i=0;i<realsuperSize;i++)
		for (int j=i+1;j<realsuperSize;j++) {
			a[k] = i;
			b[k++] = j;
		}
	}
	~EstimatorData() {
		delete[] a;
		delete[] b;
	}
	long nchannels() {
		return n;
	}
	long bandwidth(long ch) {
		return bw;//FIXME
	}
	bool in_path(long ch, long node1, long node2) {
		if ((a[ch]==node1)&&(b[ch]==node2))
			return true;
		if ((a[ch]==node2)&&(b[ch]==node1))
			return true;
		return false;
	}
};

EstimatorData estimator;

class Resource {
public:
  double proc_usage;
  double net_usage[SUPERSIZE];
  long ntasks;
  
  Resource() : proc_usage(0), ntasks(0) {
    memset(net_usage, 0, SUPERSIZE*sizeof(double));
  }
  Resource(double a) : proc_usage(a), ntasks(0) {
	for (int i=0;i<SUPERSIZE;i++)
		net_usage[i] = a;
  }
  
  bool operator== (Resource& r) { //v! "resource at %p", &r
    if (proc_usage!=r.proc_usage)
      return false;
    if (ntasks!=r.ntasks)
      return false;
    for (int i=0;i<SUPERSIZE;++i)
      if (net_usage[i]!=r.net_usage[i])
        return false;
    return true;
  }
  Resource& operator= (const Resource& r) {
    proc_usage = r.proc_usage;
    ntasks = r.ntasks;
    for (int i=0;i<SUPERSIZE;++i)
      net_usage[i] = r.net_usage[i];
    return *this;
  }
};

class TFunResource : public Resource {
public:
  TFunResource() : Resource(1.) {};
  TFunResource(double a) : Resource(a) {};
};


class MacroScheduler : public ts::MacroScheduler {
public:

  inline SCell<controlHandler>& myResourceCell(int rank) { return tb[4*rank+3]; }
  inline Resource& myResourceAt(int rank) { 
    if (rank != myRank) myResourceCell(rank).read(1);
    MControlData<Resource> *d = (MControlData<Resource> *)&myResourceCell(rank).data();
    if (!d) myResourceCell(rank) = d = new MControlData<Resource>();
    d->vid = "gs_resource";
    return (*d)(); 
  }
  inline static TFunResource* task_resource(ts::Task* t) {
    return (TFunResource *)((char *)t + t->size - sizeof(TFunResource));
  }
  
/*
  inline SCell<controlHandler>& tfunCell() { return sharedCell(2); }
  inline TFunData& tfunData() {
    MControlData<TFunData> *d = (MControlData<TFunData> *)&tfunCell().data();
    if (!d) {
      tfunCell() = d = new MControlData<TFunData>();
      d->vid = "tfundata"; // FIXME: what's that for ?
    }
    return (*d)();
  }
*/

  inline bool mypublish(Resource& res) { //v! "resource at %p", &res
    if (!(myResourceAt(myRank) == res)) {
//      uiprintf("[%d] publishing new Res: proc_usage = %lf\n", myRank, res.proc_usage);
      myResourceAt(myRank) = res;
      myResourceCell(myRank).write(1);
      return true;
    } else {
      return false;
    }
  }
  
  Resource *myRes;
  long *chdata;
  double tmdata[SUPERSIZE];
  static const int mflops_default=1000;
  int mflops[SUPERSIZE];
  MacroScheduler() { 
    uiprintf("gs::MacroScheduler()\n");
    myRes=new Resource; //FIXME: my!
    myResourceAt(myRank) = *myRes ; 
    
    for (int i=0;i<SUPERSIZE;i++)
      mflops[i] = mflops_default; // real mflops not known at this point

    chdata = new long[estimator.nchannels()];
  } 
  virtual ~MacroScheduler() {}

  void update_resource() {
    *myRes = Resource();
    for (int i=0; i<MAXTASKPRIO; i++) {
      Task *curr = static_cast<Task*>(ptq.tasks + i);
      while ((curr = static_cast<Task*>(curr->getFront())) && (curr!=ptq.tasks+i)) {
	TFunResource* tfres = task_resource(curr);
	myRes->ntasks++;
	myRes->proc_usage += tfres->proc_usage;
      }
    }
    myRes->proc_usage /= mflops[myRank]; 
    mypublish(*myRes);
  }
  inline double dispersion(double ts, double ts2) {
    return realsuperSize*ts2 - ts*ts;
  }
  inline void delta(TFunResource* res, long r1, long r2, double& d1, double& d2) {
    // moving res from r1 to r2
    d1 = - (double)((double)res->proc_usage / (double)(mflops[r1]));
    d2 = + (double)((double)res->proc_usage / (double)(mflops[r2]))/* + .005*/;
  }
  int make_decision(Task *tsk) {
    double ts=0; // sum(Ti)
    double ts2=0;// sum(Ti*Ti)
    for (int i=0;i<realsuperSize;i++) {
      double pu = myResourceAt(i).proc_usage;
      ts += pu;
      ts2 += pu*pu;
      tmdata[i] = pu;
    }
/*    printf("tmdata: ");
    for (int i=0;i<realsuperSize;i++) 
      printf("%lf ", tmdata[i]);
    printf("\n");  */
    double DT = dispersion(ts, ts2);
    double DT_best = DT;
    int rank_best = -1;
    TFunResource* res = task_resource(tsk);
    double d1, d2;
    double DT_cur;
    for (int i=0;i<realsuperSize;i++)
      if (i!=myRank) {

	delta(res, myRank, i, d1, d2);
	double dts = d1+d2;
	double dts2 = 2*(tmdata[myRank]*d1 + tmdata[i]*d2) + d1*d1 + d2*d2;
	DT_cur = DT + realsuperSize*dts2 - 2*ts*dts - dts*dts;
	if (DT_cur < DT_best) {
	  DT_best = DT_cur;
	  rank_best = i;
	}
      }
    // debug
/*    uiprintf("^^^^ [%d] pu:%lf, to:%d ; \n", myRank, res->proc_usage, rank_best);
    for (int i=0;i<realsuperSize;i++)
      uiprintf("%d:%d; %lf \n", myRank, i, tmdata[i]);
    uiprintf("\n");    */
    return rank_best; // _never_ returns myRank
  }
  virtual void exchangeByResources() {
    ts::MacroScheduler::exchangeByResources();
/*    uiprintf("mflops: ");
    for (int i=0;i<realsuperSize;i++) {
      mflops[i] = tb.resourceAt(i).mflops;
      uiprintf("%d ", mflops[i]);
    }
    uiprintf("\n"); */
  }
  int check_dst(Task *t, int dst) {
    int atRank = (*t)().atRank;
    if (atRank != -1)
      return atRank;
    return dst;
  }
  virtual bool visit() {
    if (!tb.resourceAt(ts::myRank).mflops)
      return true;
    update_resource();
    if (ptq.count > 0)  { //v! "I am task exporter(%d tasks) - export most havy task FIXME: level", ptq.count
      Task* t = ptq.pop(true);
      TFunResource* res = task_resource(t);
      assert(t != NULL);
      int dst = make_decision(t);
      //uiprintf("dst=%d\n", dst);
      dst = tb.routeTask(t, dst);
      if (dst >= 0) {
	if (dst!=myRank)
	  myResourceAt(dst).proc_usage += res->proc_usage / mflops[dst];
	return true;
      } else {
	ptq.put(*t, true);
	return false;
      }
    } else {              //v I am free and waiting for tasks
	return false;
    }
  }
  virtual bool work() {
    return ts::MacroScheduler::work();
  }
/*  virtual bool visit() {
    long long tm = time();
    bool res = __visit();
    tm = time() - tm;
    total_time += tm;
    return res;
  }
*/
  
};

class GSInit {
public:
  static void* entry(int cmd, void *parm) {
    long long id;
    int size;
    double cpuUsage = ts::tc.cpuUsage;
    switch(cmd) {
      case 0: // init
/*	ts::Task::get_extra = get_extra;
	ts::Task::after_ctor = after_ctor;
	ts::MacroScheduler::New = MacroScheduler_New; */
	return NULL;
/*	case 1: // new TaskBoard
	{ using namespace ts; uiprintf("Initializing grid scheduler\n"); }
	return new gs::MacroScheduler;
	case 2: // returns amount of extra space needed for each task
	  return (void *)sizeof(TFunResource);
	case 3: // (test) initialize allocated space
//	  id = ((ts::Task *)parm)->taskId;
//	  size = ((ts::Task *)parm)->size;
//	  printf("taskId = %lld\n", id);
	  cpuUsage = ts::tc.cpuUsage;
//	  uiprintf("new TFun, cpuUsage=%lf\n", cpuUsage);
	  new(MacroScheduler::task_resource((ts::Task*)parm)) TFunResource(cpuUsage);
	  return NULL; 
	case 4: // Task::onFinish

	  return NULL;  */
      default:
	assert(1);
	return NULL;
      }
  }
  static int get_extra(void) {
    return sizeof(TFunResource);
  }
  static void after_ctor(ts::Task* t) {
    double cpuUsage = ts::tc.cpuUsage;
    new(MacroScheduler::task_resource(t)) TFunResource(cpuUsage);
  }
  static ts::MacroScheduler* MacroScheduler_New(void) {
    { using namespace ts; uiprintf("Initializing grid scheduler\n"); }
    return (ts::MacroScheduler*)(new gs::MacroScheduler);
  }
  GSInit() {
    registerFeature("gs", "Grid-oriented scheduler", entry);
    ts::Task::get_extra = get_extra;
    ts::Task::after_ctor = after_ctor;
    ts::MacroScheduler::New = MacroScheduler_New;
  }
} GSInit_instance;

}
