//+
//  Open T-Superstructure for C++
//
//  See article: "New words for old song"
//  S.M.Abramov,A.I.Adamovich,V.A.Roganov
//
//  MoTOR = Mobile Threads, Mobile Objects & References[DRC]
//-

// Two transport handlers: control & data 

extern TransportHandler controlHandler, dataHandler;
#ifdef IMPL
TransportHandler controlHandler("controlHandler", CTRL_SERVICE_PRIO);
TransportHandler    dataHandler("dataHandler",    DATA_SERVICE_PRIO);
class InitControlHandler {
public:
  static void inPlaceCtor(Cell *c) { new (c) SCell<controlHandler>(); }
  InitControlHandler() { controlHandler.inPlaceCtor = inPlaceCtor; }
} initControlHandler; // Data handler initialized in `tss' module
#endif


// Mobile data w/ COW logic

template <typename type, TransportHandler &handler>
class MDataBase : public SCell<handler>::Data {
public:
  type val;
  inline MDataBase() {
    flags = ctFlags;
    //v! "new MData, flags = %d", flags
  }
  inline MDataBase(const MDataBase& m) : val(m.val) {  
    //v! "new MData %p by copy constructor from %p", this, &m
    assert(m.size == size);
    flags = m.flags & ~SDATAINTRANS;
  }
  inline MDataBase<type,handler>& operator= (const MDataBase<type,handler>& m) { 
    //v! "new MData %p by operator= from %p", this, &m
    assert(m.size == size);
    val = m.val; //v!  COPY DATA ! DONE
    flags = m.flags & ~SDATAINTRANS;
    return *this; 
  }
  inline type& operator ()() { return val; }
  virtual void onZero () {
    if (!(flags & SDATATRANS))  // FIXME: not here!!!
      delete this;
    else
      hfree(this); //FIXME: free == super::free ? 
  } 
};

template <typename type>
class MData        : public MDataBase<type,dataHandler> {
public:
  virtual bool isReady() { return true; }

  //FIXME: why we need it ???
  virtual SCell<dataHandler>::Data* clone () {
    SCell<dataHandler>::Data* result = 
      (SCell<dataHandler>::Data*) new (MData<type>::halloc(size)) MData<type>(*this);
    int extraSize = size - sizeof(MData<type>);
    assert(extraSize >= 0);
    if (extraSize > 0) 
      memcpy((char*)result + sizeof(MData<type>),(char*)this + sizeof(MData<type>),extraSize);
    return result;
  }
  // for extraSize needs
  void* operator new (size_t size, void *mem) { return mem; }

};
template <typename type>
class MControlData : public MDataBase<type,controlHandler> {};


// DRC

#define MAXLONGBIT 31

#define DRCSIG 17
class DRCObj {
public:
  int  lcnt;
  long long taken, we;
  unsigned noDRCDelete : 1;
  unsigned takeRqSent : 1;
  short sig;
  inline DRCObj() : lcnt(0), taken(0), we(0), noDRCDelete(0), takeRqSent(0), sig(DRCSIG) {}
  inline bool drcObjOk() { return sig == DRCSIG; }

  inline void putWEE(char &wee) {
    assert(drcObjOk());
    assert(wee > 0);  //v! "adding &wee=%p 1<<wee=%d to &we=%p, we=%d\n", &wee, 1<<wee, &we, (int)we
    we +=  1 << wee;     
    wee = -1;
    lcnt++;
  }

  static inline char we2wee (long long we) {
    assert(we >= 2);
    if (we > (1UL << MAXLONGBIT)) return MAXLONGBIT; 
    if (we > (1UL << 16)) return 16; 
    if (we > (1UL <<  8)) return 8; 
    if (we > (1UL <<  4)) return 4; 
    if (we > (1UL <<  2)) return 2; 
    return 1;
  }
  inline char getWEE() {
    assert(drcObjOk());
    if (we < 4) { takeWE(); assert(we > 4); }
    char wee = we2wee(we);
    we -= 1 << wee;
    assert(wee > 0 && we >= 0);
    return wee;
  }

  inline void take(long long &_we) {
    assert(drcObjOk());
    taken -= _we;
    _we = 1UL << MAXLONGBIT;
    taken += _we;
  }
  inline void retrieve(long long &_we) {
    assert(drcObjOk());
    taken -= _we;
    assert (taken >= 0);
    _we = 0;
    //if (taken == we && !noDRCDelete) delete this;
    if (taken == we && lcnt == 0 && !noDRCDelete) delete this; // WHY??? lcnt == 0
  }

  inline void onLZero() { 
    retrieveWE();
  }

  // Abstract communicators & destructor
  virtual void takeWE()     = 0;
  virtual void retrieveWE() = 0; 
  virtual ~DRCObj() { sig = 0; }

};

class DRCRef {
public:
  char wee; // lighten<0, null=0, real>0
  inline void lighten() { if (wee > 0) getObj().putWEE(wee);  }
  inline void ctor(DRCObj* obj = NULL) { wee = obj ? obj->getWEE() : 0; }
  char getWEE() { 
    lighten();
    return wee ? getObj().getWEE() : 0;
  }
  inline void dtor() { // Avoid Grace new(this)-style here!!
    //v! "DRCRef::dtor for %p", this
    lighten(); 
    assert(!wee || getObj().lcnt > 0);
    if (wee && !--getObj().lcnt) getObj().onLZero();
  }

  // Abstract factory-function
  virtual DRCObj& getObj() = 0;
};


// Mobile reference

#define MREF_GLUE     1
#define MREF_MAGNETIC 2 // apply recursive w/ level!

template <typename type> 
class MObj;

template <typename type>
class MRef : protected DRCRef {
public:
  int cellOffset;
  long long seqNo;
  int pflags;

  inline MRef(MObj<type>* mo = NULL) { ctor(mo); }

  inline MRef(const MRef<type>& mo) {
    //v! here
    ctor(((MRef<type> &)mo).getMObjPtr());
  }

  inline MObj<type>* operator= (MObj<type>* mo) { 
    //v! here
    MRef<type> holder(getMObjPtr());
    if (getMObjPtr() != mo) { dtor(); ctor(mo); }
    return mo;
  }

  inline MObj<type>* operator= (MRef<type>& mo) { 
    //v! here
    return *this = mo.getMObjPtr(); 
  }

  inline operator type* ()  { 
    //v "cellOffset = %d, mref = %p, mobj = %p", cellOffset, this, &**this
    return cellOffset ? &(**this)()() : NULL;
  }
 
  inline MObj<type>* getMObjPtr() { return cellOffset ? (MObj<type> *)&MObj<type>::cellAt(cellOffset) : NULL; }

  // Implemented after MObj definition; Downloads data!!
  inline MObj<type>& operator* ();

  // DRC dereference
  virtual DRCObj& getObj();

  virtual ~MRef() { dtor(); };

  void ctor(MObj<type>* mo);
  void dtor();
};

struct Resource {
  // Positive - available free, Negative - task
  int  prio;
  int  mflops; 
  int  mbytes;
  int  bogomips;
  int  wantedTaskMFlops; // For requests only!

  Resource() : prio(0),mflops(0),mbytes(0),bogomips(0),wantedTaskMFlops(0) {}
  bool operator== (Resource& r) { //v! "resource at %p", &r
    return prio==r.prio && mflops==r.mflops && bogomips==r.bogomips && 
      mbytes==r.mbytes  && wantedTaskMFlops==r.wantedTaskMFlops;
  }
  Resource& operator= (const Resource& r) {
    prio   = r.prio; mflops = r.mflops; bogomips = r.bogomips; 
    mbytes = r.mbytes; wantedTaskMFlops = r.wantedTaskMFlops;
    return *this;
  }
  Resource& operator += (const Resource &r) {
    prio   += r.prio; mflops += r.mflops; bogomips += r.bogomips; 
    mbytes += r.mbytes; 
    return *this;
  }
  const char* viz() {
    static char buf[0x100];
    snprintf(buf, sizeof(buf)-1, "%'.1fGf[%dBM],%'.2fGB", 
	     mflops/1000., bogomips, mbytes/1000.);  
    return buf;
  }
};

class FreeResource : public Resource {
public:
  static double doFlops (long long n) {
    float x=0.,x1=0.,x2=0.,x3=0.,x4=0.; 
    while ((n-=16) > 0) x = x + x1 * x2 + x3 * x4; //FIXME: invalid!
    return x;
  }
  static long usecFlops(int n) {
    double startTime = wTime(); doFlops(n); double endTime = wTime();
    return (long)(1000000.*(endTime - startTime));
  }
  FreeResource() {
    prio = tc.nodePrio;

    long long flop = 10000000;          // About 0.01 sec
    int flopTime = usecFlops(flop); 

    flop *= 100000; flop /= flopTime;   // We want to compute ~ 0.1 sec
    MPI_Barrier(MPI_COMM_WORLD);        // Wait all...
    mflops = flop/usecFlops(flop);      // ...and do it together!
   
    //uiprintf("FreeResource() : mflops = %d\n", mflops);
    
    const char *bogoStr = fileInfo("/proc/cpuinfo","bogomips",": ","\n");
    assert(bogoStr);
    bogomips = atoi(bogoStr);

    mbytes = ((long long)sysconf(_SC_PHYS_PAGES)*(long long)sysconf(_SC_PAGESIZE))/(1024*1024);
  }
};

int getCurrentMFlops();

#define MAXTASKPRIO 3
class TaskResource : public Resource {
public:
  int atRank;
  TaskResource() : atRank(tc.atRank) {
    assert(atRank >= -1 && atRank < realsuperSize); 
    prio = tc.prio;
    assert(prio >= 0 && prio < MAXTASKPRIO);

    mflops = getCurrentMFlops()+1; // *Very* simple way to aproximate mflops
    if (mflops >= 0) mflops =  -1;

    mbytes = tc.mbytes;
  };
};

class TaskCtxt {
  static long lastTaskClassId;
  static const int max_name_len = 30;
  char name_[max_name_len+1];
public:
  long taskClassId;
  TaskCtxt(char *name) : taskClassId(++lastTaskClassId) {
    strncpy(name_, name, max_name_len);
    name_[max_name_len] = '\0';
  }
  char *name() {
    return name_;
  }
};
#ifdef IMPL
long TaskCtxt::lastTaskClassId = 0;
#endif
class Task : public MControlData<TaskResource>, public Ring, public TaskStatistic {
  static long lowNextTaskId;
  static inline long long newTaskId() { return (long long)++lowNextTaskId + (((long long)myRank) << 32); }
public:
  long long taskId;
  bool packed;

  static inline Task* getCurrent() {
    // FIXME: check whether this is real Task, or just TaskStatistic
    TaskStatistic* ts = TaskStatistic::getCurrent();
    Task* t = (Task*)(ts);
//    assert( (void*)t == (void*)ts );
    return t;
  }
  // getRealCurrent: returns current, if it is a real Task, not statistic Task
  static inline Task* getRealCurrent() {
    Task* t = getCurrent();
    if ((t) && (t->taskId >= 0))
      return t;
    else
      return NULL;
  }
  static inline void setCurrent(Task* t) {TaskStatistic::setCurrent(t);}

  Task *parent;
  Header<ThreadCache>* hth;

  char label[0x40];

  static int (*get_extra)(void);
  static void (*after_ctor)(Task*);
  
  static int get_extra_default(void) {return 0;}
  static void after_ctor_default(Task*) {return;}
  
  virtual TaskCtxt& context(){
    assert(0 && "ERROR: Task::context() called for pseudo-task\n");
    return *(TaskCtxt*)0;
  } // overloaded in TFun

  void setLabel(const char *format, ...) {
    va_list args;
    va_start (args,format);
    vsnprintf(label,sizeof(label)-1,format,args);
    label[sizeof(label)-1] = 0;
    va_end (args);
  }

  inline Task(long long taskId_ = 0) : taskId((taskId_) ? taskId_ : newTaskId()),
      packed(false), parent(getRealCurrent()) {
    label[0] = 0;
    setCurrent(this); //v! "Constructing new task [%p]", this
  }

  inline void stopAndContParent() { 
    setCurrent(parent); //v! "Task [%p] constructed and awaiting for activation; back to %p", this, parent
    parent = NULL; // sanity
  }

  inline void activate() {
    // unpack all TOuts if task is from another node
    if (packed) unpackState();
    hth = new Header<ThreadCache>; // Allocate superthread
    hth->tsk = this;               // Load our coordinates
    hth->wth->contTasks();         // Enter superqueue
  }

  // network 
  virtual void packState()   { packed = true; };
  virtual void unpackState() {};

  virtual void onRecv();

  virtual void work() {
    assert(0 && "ERROR: Task::work() called for pseudo-task\n");
  }

  virtual ~Task() {}

};


extern Task rootTask;
extern Task schedTask;
extern Task mpiTask;
#ifdef IMPL
Task rootTask(-1);
Task schedTask(-2);
Task mpiTask(-3);
#endif

//FIXME: reslg
class TaskPrioQueue {
public:
  Ring tasks[MAXTASKPRIO];
  int count;
  TaskPrioQueue() : count(0) {}
  void put(Task& t, bool fromExport) { 
    fromExport ? tasks[t().prio].pushBack(t) : tasks[t().prio].pushFront(t); 
    count++; 
  }
  Task *get(bool forExport) {
    int i;
    for (i=0; i<MAXTASKPRIO; i++) {
      if (!tasks[i].ringEmpty()) {
	return (Task*)(forExport ? tasks[i].getBack() : tasks[i].getFront());
      }
    }
    return NULL;
  }
  Task* pop(bool forExport) {
    Task *result = get(forExport);
    if (result) { count--; result->remove(); }
    return result;
  }
  void clear() {
    //v! "%d tasks in ptq when destroying queue", count
    for (int i = 0; i < MAXTASKPRIO; i++) {
      while (!tasks[i].ringEmpty()) {
        delete (Task *)tasks[i].getFront();
      }
    }
  }
};

externNoImpl TaskPrioQueue tq;
externNoImpl TaskPrioQueue ptq; //Diffs from tq - it is "prenatal" tasks 

#ifdef IMPL
int getCurrentMFlops() {
  return !Task::getRealCurrent() ? tc.mflops : (*(Task::getRealCurrent()))().mflops;
}

int (*Task::get_extra)(void) = Task::get_extra_default;
void (*Task::after_ctor)(Task*) = Task::after_ctor_default;
long  Task::lowNextTaskId = 0;

void ThreadCache::newTask(ThreadCache *me) {
  Task::setCurrent(me->tsk); 
  assert(Task::getCurrent()); 
  Task::getCurrent()->work(); //v! "Starting task %s", me->tsk->viz() ! "done task %s", me->tsk->viz() 
  delete me->tsk->hth; //v! free superthread
  me->tsk->remove();
  delete me->tsk;
  me->tsk = NULL; //sanity
  Task::setCurrent(NULL); // Amen!

  me->th->initProc((ThrH::proc_t)newTask, me);
  me->wth->stopHereAndRunOthers(); 
}

void Task::onRecv() {
  Ring::ct(); // Reconstruct ring!
  assert(packed); // if we receive task it must be already packed!
  assert((*this)().mflops < 0);
  ptq.put(*this,(*this)().atRank < 0); // Place fixed rank for execution, normal - for export 
}
#endif//IMPL

class MThread : public Task {
public:
  MThread();
  bool isRunning;

  // FIXME: vita is unimplemented
  // MObj<int>* vita; //FIXME: dt seqNo
  // int vitaSeqNo;
  bool finishCondition();

  virtual void start() {
    if (finishCondition()) throw new FinalizeException();
  }
  virtual void stop() {
  }

  void run() {
   if (!isRunning) {  //v! "Running MThread [%p]", this
     ptq.put(*this,false);
     isRunning = true;
    }
  }

  virtual void work() = 0;

  virtual ~MThread() { //v! "Destroying MThread [%p]", this
  }
};

// Mobile object

template <typename type>
class MObj : public SCell<dataHandler>, public ThrH, public DRCObj {
public:
  // FIXME: vita is unimplemented
  // MRef<int> vita; // Hint for TSS: set it for cold values _only_
  long long owner;
  int extraSize;
  inline MObj() : extraSize(tc.extraSize) { 
    lock();  //v! "New MObj %p, owner = %p\n", this, owner
  }

  // Assignment to MData 
  inline MData<type>* operator= (MData<type> *data) {
    return (MData<type>*) (*(SCell<dataHandler>*)this = data);
  }

  inline void waiting() { 
    assert(Task::getRealCurrent());
    Task *saved_current = Task::getCurrent();
    Task::setCurrent(NULL); // sanity
    stopHereAndRunOthers(); //v! "task %p stopped at MObj[%p]", Task::getCurrent(), this ! 
    Task::setCurrent(saved_current); //v! "continue task %s at MObj[%p]", Task::getCurrent()->viz(), this
  }

  inline MData<type>& operator ()() { 
    assert(cellIsOk());
    //FIXME: is it needed NOW ?
    //ARef alive(this); //v "Just to hold this tricky fox, cnt = %d", cnt

    ///FIXME: check we are master or not, Should wait for a proper seqNo if i'm a slave

    while (owner && owner != Task::getCurrent()->taskId) waiting();
      //FIXME: onRead() to force lazy function to run & dynamic priority adjust

    if (&data()==NULL) { 
      assert(owner == Task::getCurrent()->taskId);
      *this = new (MData<type>::halloc(sizeof(MData<type>)+extraSize)) MData<type>; 
    }

    //COW
    assert(data().cnt >= 1);
    if (data().cnt > 1) {
      //v! "BEFORE COW: %d", *(int*)&((MData<type>&)data()).val
      MData<type> *oldData = (MData<type> *) &data();
      *this = new (MData<type>::halloc(sizeof(MData<type>)+extraSize)) MData<type> (*oldData);
      MData<type> *newData = (MData<type> *) &data();
      //v! " AFTER COW: %d", *(int*)&((MData<type>&)data()).val
      assert(extraSize >= 0);
      if (extraSize > 0) {
	// CopyExtent 
	memcpy((char*)newData + sizeof(MData<type>),(char*)oldData + sizeof(MData<type>),extraSize);
      }
    }
    assert(data().cnt == 1);

    return *(MData<type>*)&data();
  }

  inline void unlock() {     
    assert (!Task::getCurrent() || owner == Task::getCurrent()->taskId);
    owner = 0; 
    contTasks(); 
  }

  inline void lock() { if (Task::getCurrent()) owner = Task::getCurrent()->taskId; }

  void onWrite() {
    //FIXME: do not send hot data!
    write(getSeqNo());//FIXME
  }

  void chunkSetLg(int) {
    // FIXME
  }

  virtual ~MObj() {
    //v! here
  }


  // DRC-related methods

  class DRCData : public SCell<dataHandler>::Data {
  public:
    long long we;
    DRCData (int _cellOffset, long long _seqNo, OnRecvMethod _method, long long _we, const char *vid) :
      SCell<dataHandler>::Data(_cellOffset, _seqNo, _method, vid), we(_we) {}
  };

  void onTake(SCell<dataHandler>::Data &_d) {
    DRCData& d = (DRCData&)_d;
    take(d.we);
    // Change method and rank (to avoid new Data creation)
    d.vid = "takeReply";
    d.method = (Data::OnRecvMethod)&ts::MObj<int>::onTakeReply;
    int srcRank = d.srcRank; //FIXME: why we need it ???
    d.srcRank = myRank;
    // Send taken data back
    dataHandler.asyncSend(&d, srcRank); //v! send take reply message
  }

  void onTakeReply(SCell<dataHandler>::Data &_d) {
    DRCData& d = (DRCData&)_d;
    we += d.we;
    takeRqSent = 0;
    contTasks();
  }

  void onRetrieve(SCell<dataHandler>::Data &_d) {
    DRCData& d = (DRCData&)_d;
    assert(getSeqNo() >= d.seqNo);
    retrieve(d.we);
  }

  virtual void takeWE() {
    if (master() == myRank) {
      take(we);
    } else { //v! send take message
      while (we < (1ll << 16)) {
        if (!takeRqSent) {
          dataHandler.asyncSend(new DRCData(offset(), getSeqNo(), (Data::OnRecvMethod)&ts::MObj<int>::onTake, we, "takeWE"), master());
          we = 0; //FIXME:
          takeRqSent = 1; 
	}
        waiting();
      }
    }
  }  

  virtual void retrieveWE() {
     //v! "cell %p", this 
    if (master() == myRank) {
      retrieve(we);
    } else { //v! send retrieve message
      dataHandler.asyncSend(new DRCData(offset(), getSeqNo(), (Data::OnRecvMethod)&ts::MObj<int>::onRetrieve, we, "retrieve"), master());
      we = 0;
      onSlvRetrieve();
    }
  }

  virtual void onSlvRetrieve() {}

  //FIXME: make it abstract
private:
  void* operator new (size_t)   {return NULL;}
  void* operator new[] (size_t) {return NULL;}
};

#ifdef IMPL
extern MObj<int>* newVita();

// FIXME: vita is unimplemented
MThread::MThread() : isRunning(false) /*, vita(newVita()) */ { /* vitaSeqNo = vita->getSeqNo(); */ } 
#endif //IMPL

inline bool MThread::finishCondition()          { return false; /* vitaSeqNo!= vita->getSeqNo(); */ } 

template <typename type>
inline MObj<type>& MRef<type>::operator* () { //FIXME: avoid it to take address by &* - it download data !
  MObj<type> *mo = getMObjPtr();
  if (mo) {
    assert(seqNo >= mo->getSeqNo());
    while (seqNo > mo->getSeqNo()) {
      assert(mo->master() != myRank);
      mo->read(seqNo); 
      mo->waiting();
    }
  }
  return *mo;
}

template <typename type>
inline void MRef<type>::ctor(MObj<type>* mo) {
  cellOffset = mo ? mo->offset() : 0;
  DRCRef::ctor(mo ? mo : NULL);
  seqNo = mo ? mo->getSeqNo() : 0;
  pflags = tc.pflags;//FIXME: pflags - short name
}

template <typename type>
inline void MRef<type>::dtor () { 
  DRCRef::dtor(); 
  cellOffset = 0;
  seqNo = 0;
}

template <typename type>
DRCObj& MRef<type>::getObj() { // Implemented _after_ TObj definition for proper casting
  MObj<type> *mo = getMObjPtr(); 
  assert(mo && mo->drcObjOk() && mo->cellIsOk());
  return (DRCObj &)*mo;
} 

// TaskboardHierarchy-based cooperative macroscheduler 


// 3,2 = Resources + Statistic + Ready, Heartbeat + Relax
#define CSEG_CELLS (4)
#define CSEG_SHARED_CELLS (3+20)
class TaskBoard : public SSegment<SCell<controlHandler>,controlHandler,CSEG_CELLS,CSEG_SHARED_CELLS> {
public:
  inline SCell<controlHandler>& resourceCell(int rank) { return (*this)[CSEG_CELLS*rank]; }
  inline Resource& resourceAt(int rank) { 
    if (rank != myRank) resourceCell(rank).read(1);
    MControlData<Resource> *d = (MControlData<Resource> *)&resourceCell(rank).data();
    if (!d) resourceCell(rank) = d = new MControlData<Resource>();
    d->vid = "resource";
    return (*d)(); 
  }

  ///FIXME: use template function here!

  inline SCell<controlHandler>& statisticCell(int rank) { return (*this)[CSEG_CELLS*rank+1]; }
  inline Statistic& statisticAt(int rank) { 
    if (rank != myRank) statisticCell(rank).read(1);
    MControlData<Statistic> *d = (MControlData<Statistic> *)&statisticCell(rank).data();
    if (!d) statisticCell(rank) = d = new MControlData<Statistic>();
    d->vid = "statistic";
    return (*d)(); 
  }

  // FIXME: use template or macro to avoid copying
  inline SCell<controlHandler>& readyCell(int rank) { return (*this)[CSEG_CELLS*rank+2]; }
  inline char& readyAt(int rank) { 
    if (rank != myRank) readyCell(rank).read(1);
    MControlData<char> *d = (MControlData<char> *)&readyCell(rank).data();
    if (!d) {
      readyCell(rank) = d = new MControlData<char>();
      d->vid = "ready";
      (*d)() = 0;
    }
    return (*d)(); 
  }
 
  
  SCell<controlHandler>& heartbeatCell() { return sharedCell(0); } 
  int& heartbeatData() { 
    MControlData<int> *d = (MControlData<int> *)&heartbeatCell().data();
    if (!d) { 
      heartbeatCell() = d = new MControlData<int>(); 
      d->vid = "heartbeat";
      (*d)() = 0; 
    }
    return (*d)(); 
  }

  SCell<controlHandler>& relaxCell()     { return sharedCell(1); }
  char& relaxData()     { 
    MControlData<char> *d = (MControlData<char> *)&relaxCell().data();
    if (!d) {
      relaxCell() = d = new MControlData<char>();
      d->vid = "relax";
      (*d)() = 0;
    }
    return (*d)(); 
  }

  FreeResource *myFreeRes;
  Resource *zeroRes;
  TaskBoard() : myFreeRes(new FreeResource), zeroRes(new Resource) { 
    // init my statistic
    myStat = &statisticAt(myRank);
    // init my resource
    resourceAt(myRank) = *myFreeRes;
    resourceCell(myRank).write(1);
  } 

  void deleteSuperCells() {
    for (int j=0;j<CSEG_SHARED_CELLS;j++)
      sharedCell(j) = 0;
    for (int i=0; i<realsuperSize; i++)
      for (int j=0; j<CSEG_CELLS;j++)
	(*this)[CSEG_CELLS*i+j] = 0;
  }

  virtual ~TaskBoard(){
    deleteSuperCells();	  
    delete myFreeRes;
    delete zeroRes;
  }

  inline bool publish(Resource& res) { //v! "resource at %p", &res
    if (!(resourceAt(myRank) == res)) {
      resourceAt(myRank) = res;
      resourceCell(myRank).write(1);
      return true;
    } else {
      return false;
    }
  }
  int mostHavyTaskMFlops() { //Somebody (!=me) placed any task here ?
    int i;
    int minMFlops = 0;
    for (i=0; i<realsuperSize; i++) if (i!=myRank && resourceAt(i).mflops < minMFlops) minMFlops = resourceAt(i).mflops;
    return minMFlops;
  }
  int mostFreeRank(int taskMFlops) {
    int maxMFlops = 0, rankFound = -1, i = myRank; // Starting from neighbours
    for (i=(i+1)%realsuperSize; i != myRank; i=(i+1)%realsuperSize) {
      Resource& res = resourceAt(i);
      //v! "look at rank %d resource %d" , i, res.mflops
      if (taskMFlops <= res.wantedTaskMFlops && res.mflops > maxMFlops) {  
	rankFound = i; maxMFlops = res.mflops; //v! "Found rank %d with resource mflops=%d",i,maxMFlops 
      }
    }
    return rankFound;
  }
  // routeTask: returns rank to which task has been send, myRank if it was activated locally, -1 otherwise
  int routeTask(Task *t, int tryToRunWhere) {
    int atRank = (*t)().atRank;
    if (atRank == -1) atRank = tryToRunWhere;
    if (atRank == -1) return -1;

    if (atRank == myRank) { 
      tq.put(*t, false);
      t->activate();
      statisticCell(myRank).write(1);
      myStat->tasksActivated++;
    } else {
      if (!t->packed) t->packState();
      controlHandler.asyncSend(t,atRank); //FIXME: free task memory!
      myStat->tasksExported++;
      if ((*t)().atRank < 0) resourceAt(atRank).mflops = 0; // Speculative action
    }
    //if (atRank!=myRank)
    //  uiprintf("routeTask: %d -> %d\n", myRank, atRank);
    return atRank;
  }

};


//FIXME: it goes to other segment (bad -> oindex < 0); and it is'nt know why it _same_
//externNoImpl TaskBoard MetaBoard;
class MacroScheduler : public Service {

  class Heartbeat : public Prof {

    int lastHeartbeat;
    MacroScheduler* ms; 
 
  public:
     
    Heartbeat(MacroScheduler *_instance):lastHeartbeat(0),ms(_instance) { start(); };

    void check() {
      stop();

      if (tick2Sec(total) > tc.heartbeatTimeout) {
	errprintf("Heartbeat timeout at rank %d \n", myRank);
	exitServices(1);
	goto out;
      }
      if (ms->tb.heartbeatData() < 0) {
	exitServices(0);  // fail every 4 bln seconds
	goto out;
      }

      if (myRank == mainRank) {
	
	if (tick2Sec(total) > tc.heartbeatInterval) {
	  ms->tb.heartbeatData() = lastHeartbeat++; 
	  ms->tb.heartbeatCell().write(0);
	  total = 0; 
	}

      } else {

	// heartbeat listening
	if (lastHeartbeat != ms->tb.heartbeatData() ) {
	  lastHeartbeat = ms->tb.heartbeatData();
	  total = 0;
	}

      }
    out:
      start();
    }  

  };
 
  Heartbeat hb; 
public:
  TaskBoard tb;

  static MacroScheduler* (*New)(void);
  static MacroScheduler* New_default() {
    return new MacroScheduler;
  };

  MacroScheduler(): Service( __FUNCTION__,SCDL_SERVICE_PRIO ),hb(this) {
    //v! MacroScheduler - ready
  }
  virtual ~MacroScheduler() {}
  // Main function: interact with taskboard (FIXME: use onWrite)
  // Returns true if successful, false - if not

  // *VERY* classic T-system {0,1,>1} approach 

  virtual bool visit() {
    if (ptq.count > 1)  { //v! "I am task exporter(%d tasks) - export most havy task", ptq.count
      Task* t = ptq.pop(true);
      assert(t != NULL);
      if (tb.routeTask(t,tb.mostFreeRank((*t)().mflops)) >= 0) { 
	return true; 
      } else {
	ptq.put(*t, true);
	return tb.publish((*t)()); 
      }
    } else if (ptq.count == 1) {
      return tb.publish(*tb.zeroRes);
    } else {              //v I am free and waiting for tasks
      int mflops = tb.mostHavyTaskMFlops();
      if (mflops < 0) {   // Yes, somebody has a task for me!
	tb.myFreeRes->wantedTaskMFlops = mflops;
	return tb.publish(*tb.myFreeRes);
      } else { // Be silent (to avoid job-free manifestation)
	return false;
      }
    }
  }

  bool runLocalTasks() {
    Task *t;
    if ((ThrH::delayed.lnk == &ThrH::delayed) && (t = ptq.pop(false))) { 
      tb.routeTask(t,myRank);
    }
    return ThrH::lastRunOthers();
    //FIXME: do not call Iprobe so often! 
  }

  // FIXME: add set FIRST statistic!

  void setStatistic()  { //FIXME: set LAST statistic!
    myStat->alive = false;
    myStat->total_time = tick2Sec(TaskStatistic::total_time + schedTask.total + rootTask.total);
    myStat->tasks_time = tick2Sec(TaskStatistic::total_time);
/*    myStat->idle_time = idle_time.total;*/
    myStat->sched_time = tick2Sec(schedTask.total);
    myStat->visit_count = schedTask.event_count;
    myStat->mpi_time = tick2Sec(mpiTask.total);
    tb.statisticCell(myRank).write(1); //v! Relax stage: setStatistic done
  }  

  // FIXME: add 'phase' inside statistic:
  // 1 - ready for work
  // 2 - ready for exit

  bool mainStatisticReady() { 
    int i;
    for (i = 0; i < realsuperSize; i++) if (tb.statisticAt(i).alive) return false;
    return true;
  }
  void printStatistic() {
    if (realsuperSize == 1) uiprintf("Tasks activated/exported/Msgs sent:\n  ");
    else                    uiprintf("Tasks activated/exported/Msgs sent at rank[0..%d]:\n  ", realsuperSize-1);
    for (int i = 0; i < realsuperSize; i++) 
      uiprintf("%Ld/%Ld/%Ld  ", 
	       tb.statisticAt(i).tasksActivated, tb.statisticAt(i).tasksExported, tb.statisticAt(i).msgsSent);
    uiprintf("\n");

    if (realsuperSize == 1) uiprintf("Data msgs/size:\n  ");
    else                    uiprintf("Data msgs/size at rank[0..%d]:\n  ", realsuperSize-1);
/*    for (int i = 0; i < realsuperSize; i++) 
      uiprintf("%d/%d  ", 
	       tb.statisticAt(i).msg_count, tb.statisticAt(i).msg_size);
    uiprintf("\n");  */
    uiprintf("Taskboard visits: ");
    for (int i=0; i<realsuperSize; i++)
      uiprintf("%d ", tb.statisticAt(i).visit_count);
    uiprintf("\n");
/*    uiprintf("Idle time: ");
    for (int i=0; i<realsuperSize; i++)
      uiprintf("%lf ", tb.statisticAt(i).idle_time);
    uiprintf("\n");   */
    uiprintf("Scheduler time: ");
    for (int i=0; i<realsuperSize; i++)
      uiprintf("%lf ", tb.statisticAt(i).sched_time);
    uiprintf("\n");
    uiprintf("Total time: ");
    for (int i=0; i<realsuperSize; i++)
      uiprintf("%lf ", tb.statisticAt(i).total_time);
    uiprintf("\n");
    uiprintf("Tasks time: ");
    for (int i=0; i<realsuperSize; i++)
      uiprintf("%lf ", tb.statisticAt(i).tasks_time);
    uiprintf("\n");  
    uiprintf("MPI time: ");
    for (int i=0; i<realsuperSize; i++)
      uiprintf("%lf ", tb.statisticAt(i).mpi_time);
    uiprintf("\n");
  }

  virtual void exchangeByResources() {
    int i, realRess;
    do { 
      Service::service[CTRL_SERVICE_PRIO]->work(); // Exchange by resource data
      for (realRess = 0, i = 0; i < realsuperSize; i++) realRess += !!tb.resourceAt(i).mflops;
    } while (realRess < realsuperSize);
    //v! "received all resources"
    tb.readyAt(myRank) = 1;
    tb.readyCell(myRank).write(1);
    // make sure that all nodes receive right resources (mflops) before
    // they will be modified by sending tasks
    // otherwise race condition may occur
    //v! "waiting for all ready"
    do { 
      Service::service[CTRL_SERVICE_PRIO]->work(); // Exchange by ready values
      for (realRess = 0, i = 0; i < realsuperSize; i++) realRess += !!tb.readyAt(i);
    } while (realRess < realsuperSize);
    //v! "all ready, going on"
  }

  void printClusterResources() {
    Resource *r;
    if (realsuperSize > 1) {
      uiprintf("(");
      int i;
      Resource total;
      for (i = 0; i < realsuperSize; i++) {
	r = &tb.resourceAt(i);
	assert(r->mflops > 0);
	uiprintf("%s%s", i?"+":"", r->viz());
	total += *r;
      }
      uiprintf(") ~= %s\n", total.viz());
    } else {
      uiprintf("%s\n", tb.resourceAt(myRank).viz());
    }
  }

  virtual bool work() {


    hb.check();
    static bool relaxed = false; 
    
    //v! Detect and handle relax condition
    if (tb.relaxData() && !relaxed) {
      setStatistic(); //v! Got Relax
      // We already have result, don't run any tasks more
      ptq.clear();
      relaxed = true;
    }
    
    if (myRank == mainRank) {
      if (relaxed) {
	if (mainStatisticReady()) {
	  tb.heartbeatData() = -1; 
	  tb.heartbeatCell().write(0); //v! Finalize cluster
	}
	return false;
      } else { // !relaxed
	bool mainResultReady();
	if (mainResultReady()) {
	  tb.relaxData() = 1; //v! Relax...
	  tb.relaxCell().write(0); 
	  return false;
	}
      }
    }

    static bool last_result = false;
    

    bool result;
    {
      Task::Holder __taskHolder(&schedTask);
    //v Visit taskboard
      result = visit();
      TaskStatistic::getCurrent()->event_count++;
      result = result || runLocalTasks();
    }
    
    last_result = result;
    return result;
  }
};
#ifdef IMPL
MacroScheduler* (*MacroScheduler::New)(void) = MacroScheduler::New_default;
Feature macroSchedFeature("macroScheduler", "History-sensitive cooperative macro-scheduler");
#endif

//class VizData : public MControlData {
//public:
//  virtual void onRecv() {
//     "%s: Here!", __PRETTY_FUNCTION__
//  }
//};
