//+
//  Open T-Superstructure for C++
//
//  See article: "New words for old song"
//  S.M.Abramov,A.I.Adamovich,V.A.Roganov
//-

// Supermemory & superthreads for supercomputers

double wTime();

class MemCache {
public:
  void pre_ct() {}
#define MEM_CACHE_SIZE 0x10000
  static MemCache *mem_cache[MEM_CACHE_SIZE];
  static inline MemCache **pcache(size_t size) {
    return size<MEM_CACHE_SIZE ? &mem_cache[size] : NULL;
  }
  static inline size_t& totalUsed() { return total_mem_used; }
private:
  static size_t total_mem_used;
};

  //FIXME:
#ifdef IMPL
MemCache *MemCache::mem_cache[MEM_CACHE_SIZE];
size_t MemCache::total_mem_used = 0;

class MemCacheChecker {
public:
  ~MemCacheChecker() {
    if (MemCache::totalUsed()) {
      uiprintf("[%d] MemCache: %ld bytes is in use\n", myRank, MemCache::totalUsed());
    }
   }
} memCacheChecker;

#endif//IMPL

struct thr_attrib_t {
  int prio;
  int stack;
};

#ifdef IMPL
thr_attrib_t rt_thr_attrib = {
  0, 16
};
thr_attrib_t *thr_attrib = &rt_thr_attrib;
#else//IMPL
extern thr_attrib_t rt_thr_attrib, *thr_attrib;
#endif//IMPL

#define TATTRIB_MIN_PRIO (-1)
#define TATTRIB_MAX_PRIO  (1)
#define TATTRIB_MAX_STACK 256

// if we are on IA32, enable IA32 fast threads
#if defined(__i386__) 
#define IA32_ACCEL
#define ACCEL_THREADS
// if we are on x86-64, enable x86-64 fast threads
#elif defined(__x86_64__)
#define X86_64_ACCEL
#define ACCEL_THREADS
#endif

#ifndef ACCEL_THREADS
#define PTH_THREADS
#endif

struct ThrH {
  typedef void (*proc_t)(void*);	// function to call when returning to this thread
  void *stk; proc_t proc; void *arg;	// first; order!
  void *stkMin;				// FIXME: stack grow direction
  void *initStk;			// base stack address
  ThrH *lnk;
#ifdef PTH_THREADS
  pth_t tid;
#endif
  double readyTime;
#if defined(ACCEL_THREADS)
  ThrH(): stk(NULL),proc(NULL),arg(NULL),stkMin((void*)-1),lnk(this),readyTime(0) {};
#else
  ThrH(): proc(NULL),arg(NULL),lnk(this),tid((pth_t)NULL),readyTime(0) {};
#endif

#if 0//FIXME: vptr shifts savepoint
  virtual void start(){}
  virtual void stop(){}
#endif
  static ThrH last,delayed,sleeping;

  static ThrH main_thread;		// thread corresponding to entry main() function
  static ThrH *active_thread;		// currently active thread

  // Thr API
  void initProc(proc_t, void *arg);
  void contTasks(bool delay = true);
  static void saveRestore(ThrH* save, ThrH* restore, void* _unused = NULL);
  static ThrH* self();
  static void yield(ThrH *to);

  //FIXME: please remove it - or saveRestore in not universal ???
  static void spawn(ThrH* target, proc_t func, void *arg);

  void stopHereAndRunOthers();
  static void sleep(double sec);
  static bool lastRunOthers(); // returns `true' when something done
};

#ifdef IMPL

// global thread variables
ThrH  ThrH::main_thread;
ThrH* ThrH::active_thread = 0;
int *context;

// returns thread we are currently in
ThrH* ThrH::self()
{
   // no magic with stack, simply return our variable
   return active_thread;
}

void ThrH::yield(ThrH *to)
{
   // do switch context from current thread
   saveRestore(active_thread, to);
}

#ifdef ACCEL_THREADS

void ThrH::spawn(ThrH* child, proc_t func, void *arg) {
   //v! "spawning (%p) -> %p with func %p and arg %p", active_thread, child, func, arg
   // set child stack pointer
   child->stk = child->initStk;
   // setup function to run
   // FIXME: we are greatly rely on fact that given function (really ThreadCache::cycle()) never returns
   child->proc = func;
   child->arg = arg;
   // switch to child
   saveRestore(active_thread, child);
}

#endif // ACCEL_THREADS

#ifdef IA32_ACCEL

void ThrH::saveRestore(ThrH* save, ThrH* restore, void* tmp) {
  // FIXME: assert(save == &last || last.stk); // Check we are running
  // we should save ourself to active thread or to not-initialized thread only
  assert((save->stk == 0) || (save->stk == active_thread->stk));
  // if we have a new procedure to run, restore stack to initial
  if (restore->proc != 0) {
	  restore->stk = restore->initStk;
  }
  if (save != restore) {
    //v! "%p(esp=%p) -> %p(esp=%p,stack=%p,sleft=%d)", save, save->stk, restore, restore->stk, restore->stkMin, (char*)restore->stk-(char*)restore->stkMin 
  }
  active_thread = restore;
  // barrier before ...
  //save->stop();
  asm("pusha; movl 4(%0),%2; test %2,%2; jne 1f;\n"
      "movl %%esp,(%0); 1: movl (%1),%%esp;\n" 
      "movl 4(%1),%2; test %2,%2; je 2f;\n"
      "push 8(%1); movl $0,4(%1); call *%2;\n"
      "2: popa;\n" 
      :/* no out */: "r"(save),"r"(restore),"r"(tmp) : "memory");
  //FIXME: stack grow direction
  assert ((unsigned)save->stk > (unsigned)save->stkMin);
  //restore->start();
}

Feature IA32Accel ("LTHR", "Light-weight threads accelerated for Intel/IA32 architecture");

#else // IA32_ACCEL
#ifdef X86_64_ACCEL

void ThrH::saveRestore(ThrH* save, ThrH* restore, void* tmp) {
  // FIXME: assert(save == &last || last.stk); // Check we are running
  // we should save ourself to active thread or to not-initialized thread only
  assert((save->stk == 0) || (save->stk == active_thread->stk));
  // if we have a new procedure to run, restore stack to initial
  if (restore->proc != 0) {
	  restore->stk = restore->initStk;
  }
  if (save != restore) {
    //v! "%p(esp=%p) -> %p(esp=%p,stack=%p,sleft=%d)", save, save->stk, restore, restore->stk, restore->stkMin, (char*)restore->stk-(char*)restore->stkMin 
  }
  active_thread = restore;
  // barrier before ...
  //save->stop();
  asm("push %%r12; push %%r13; push %%r14;\n"
      "push %%r15; push %%rbx; push %%rbp;\n"
      "movq 8(%0),%2; test %2,%2; jne 1f;\n"
      "movq %%rsp,(%0); 1: movq (%1),%%rsp;\n" 
      "movq 8(%1),%2; test %2,%2; je 2f;\n"
      "movq 16(%1),%%rdi; movq $0,8(%1); call *%2;\n"
      "2: pop %%rbp; pop %%rbx ; pop %%r15;\n"
      "pop %%r14; pop %%r13; pop %%r12;\n" 
      :/* no out */: "r"(save),"r"(restore),"r"(tmp) : "memory", "rdi");
  //FIXME: stack grow direction
  assert ((unsigned long)save->stk > (unsigned long)save->stkMin);
  //restore->start();
}

Feature X86_64_Accel ("LTHR", "Light-weight threads accelerated for AMD/x86-64 architecture");

#endif // X86_64_ACCEL

#ifdef PTH_THREADS
#warning "No hardware-specific accelerated threads available, using generic (slow)"

void ThrH::spawn(ThrH* child, proc_t func, void *arg) {
   //v! "spawning (%p) -> %p with func %p and arg %p", active_thread, child, func, arg

   // create new pth thread
   pth_attr_t attr = pth_attr_new();
   pth_attr_set(attr, PTH_ATTR_STACK_SIZE, (unsigned long)child->initStk - (unsigned long)child->stkMin);
   pth_attr_set(attr, PTH_ATTR_STACK_ADDR, child->stkMin);
   child->tid = pth_spawn(attr, (void*(*)(void*))func, arg);
   pth_yield(child->tid);
}

void ThrH::saveRestore(ThrH* save, ThrH* restore, void* tmp) {
  //if (save != restore) { 
  // assert(save == &last || last.tid); // Check we are running
  if (save->tid == (pth_t)NULL) { save->tid = pth_self(); } //v! "tid %p -> tid %p", save->tid, restore->tid
  pth_yield(restore->tid); 
  //}
}

Feature pthFeature ("pth", "GNU Posix-compatible threads", true, NULL);

#endif // PTH_THREADS

#endif // IMPL

ThrH ThrH::last, ThrH::delayed, ThrH::sleeping; // last->current->others...

void ThrH::contTasks(bool delay) {
  ThrH *head = delay ? &delayed : &last;
  while (lnk != this) {
    ThrH* lnk1     = lnk->lnk;
    lnk->lnk       = head->lnk->lnk;
    head->lnk->lnk = lnk;  //v! "Thr %p continued", head->lnk
    lnk            = lnk1;
  }
}

void ThrH::initProc(proc_t _proc, void *_arg) {
  proc = _proc;
  arg  = _arg;
  stk  = initStk;
}

bool ThrH::lastRunOthers() {
  // Get ONE task from delayed stack WHEN needed
  if (last.lnk == &last) {
    ThrH *t;
    if (delayed.lnk != &delayed) {
      t = delayed.lnk;
      delayed.lnk = t->lnk;
      t->lnk = last.lnk;
      last.lnk = t;
      // Try sleeping task
    } else if (sleeping.lnk != &sleeping) {
      t = sleeping.lnk;
      if (t->readyTime <= wTime()) {
	sleeping.lnk = t->lnk;
	t->lnk = last.lnk;
	last.lnk = t;
      }
    }
  }
  bool result = last.lnk != &last;
  // Run ready tasks
  last.stkMin = 0;
  saveRestore(&last,last.lnk); 
  last.stk = 0; // Shutdown ThrH activity
  return result;
}

void ThrH::stopHereAndRunOthers() { 
  ThrH *nxt = last.lnk->lnk;
  last.lnk->lnk = lnk;
  lnk = last.lnk;
  last.lnk = nxt;
  saveRestore(lnk,nxt);
}

void ThrH::sleep(double sec) {
  // Leave execution flow
  ThrH *me = last.lnk;
  last.lnk = me->lnk;
  // Enter sleeping queue
  me->readyTime = wTime() + sec;
  ThrH *t = &sleeping;
  while (t->lnk != &sleeping && t->lnk->readyTime < me->readyTime) t = t->lnk;
  me->lnk = t->lnk;
  t->lnk = me;
  // Switch to ready thread
  saveRestore(me,last.lnk);
}

#endif//IMPL

// system page size 
// (it should not be exact page size, it may be for example 2 * page_size)
#define PAGE_SIZE 4096 //FIXME: use constant from system headers!
// size of stack for each PTh thread
#define STACK_SIZE 32*1024


#ifdef IMPL
#define DESIRED_STACK_SIZE (64*0x100000)
int myStackLimit() {
  static int result = -1;
  if (result < 0) {
    struct rlimit rm;
    assert(!getrlimit(RLIMIT_STACK,&rm));
    rm.rlim_cur =  rm.rlim_max > DESIRED_STACK_SIZE ? DESIRED_STACK_SIZE : rm.rlim_max;
    assert(!setrlimit(RLIMIT_STACK,&rm));
    assert(!getrlimit(RLIMIT_STACK,&rm));
    //uiprintf("... got %ld MB for stacks\n",rm.rlim_cur/0x100000); 
    result = rm.rlim_cur - 0x40000; // Reserve some space for bootstrap
  }
  return result;
}

long long myMemLimit() {
  static long long result = -1;
  if (result < 0) {
    long long desiredMemSize = (long long)sysconf(_SC_PHYS_PAGES)*(long long)sysconf(_SC_PAGESIZE);
    desiredMemSize -= 16*1024*1024; // Reserve some space for other processes
    assert(desiredMemSize > 0);
    struct rlimit rm;
    assert(!getrlimit(RLIMIT_DATA,&rm));
    rm.rlim_cur = rm.rlim_max > desiredMemSize ? desiredMemSize : rm.rlim_max;
    assert(!setrlimit(RLIMIT_DATA,&rm));
    assert(!getrlimit(RLIMIT_DATA,&rm));
    result = rm.rlim_cur;
    //uiprintf("%s: %d bytes\n", __FUNCTION__, result);
  }
  return result;
}

#else
int myStackLimit();
int myMemLimit();
#endif

class Task;
class ThreadCache {
public:
  int prio;//FIXME: use shared attr?
  int stack;
  ThrH *alloc_thr;
  ThrH *th,*wth;//FIXME: incapsulate ? (avoid ct!)
  Task *tsk;
  static int superthreads;

  void pre_ct() { 
    superthreads++;
    prio = thr_attrib->prio;
    stack= thr_attrib->stack * 1024;
    // we carefully set up manual stack for each thread
    if (stack < STACK_SIZE) stack = STACK_SIZE;
    // each new thread gets new stack
    stack_addr -= stack;
    // check that we have some space left
    if (stack_addr <= stack_begin) fatal ("Stack overflow\n");

    wth = new ThrH;
    th = new ThrH;
    th->lnk = wth;
    wth->lnk = th;

    tsk = NULL;
    alloc_thr = ThrH::self();
    th->initStk = (void *)(stack_addr + stack);
    th->stkMin = (void *)stack_addr;
   
    ThrH::spawn(th, cycle, this);
  }

  static void init_threads()
  {
#ifdef PTH_THREADS
	  ThrH::main_thread.tid = pth_self();
#endif
	  ThrH::active_thread = &ThrH::main_thread;
  }

  // FIXME!
  static void init_stack()
  {
	char a;

	// This is a hack to really allocate memory for global stack. If we just subtract
	// offset from %esp we will get a SIGSEGV signal, because Linux kernel treats page
	// fault below %esp as an bug and doesn't expand stack (see arch/i386/mm/fault.c
	// source file for details). If you know another solution please write me
       	_stack_touch();
	// get current stack & align it
	stack_addr = (unsigned long)&a;
	stack_addr = (stack_addr + PAGE_SIZE - 1) & ~(PAGE_SIZE-1);
	stack_begin = stack_addr - myStackLimit() + PAGE_SIZE;

	// initialize our own stack
	ThrH::main_thread.stk = (void *)stack_addr;
	stack_addr -= STACK_SIZE;
	ThrH::main_thread.stkMin = (void *)stack_addr;
  }

  static void cycle(void* _me) {
    ThreadCache* me = (ThreadCache*)_me; //v! "Superthread [%p]: initializing", me->th
    me->th->initProc((ThrH::proc_t)newTask, me);
    
    ThrH::yield(me->alloc_thr); 
    //assert(0 && "not reached");
    while (true) {
      ThrH::proc_t proc = me->th->proc;
      me->th->proc = NULL;
      assert(proc != NULL);
      proc(me->th->arg);
    }
  }
  static void newTask(ThreadCache*);

  static ThreadCache *thr_cache[TATTRIB_MAX_PRIO-TATTRIB_MIN_PRIO][TATTRIB_MAX_STACK];
  static inline ThreadCache **pcache(size_t) { 
    return &thr_cache[TATTRIB_MAX_PRIO-thr_attrib->prio][thr_attrib->stack];
  }

  static inline size_t& totalUsed() {
    static size_t total_threads_used = 0;
    return total_threads_used;
  }

private:
  static unsigned long stack_addr;
  static unsigned long stack_begin;
  
  // touch stack
  static void _stack_touch()
  {
	  char _tmp[myStackLimit()];
	  
	  // make page fault to expand stack
	  _tmp[0] = 0;
  }
};

#ifdef IMPL
unsigned long ThreadCache::stack_addr = 0;
unsigned long ThreadCache::stack_begin = 0;

int ThreadCache::superthreads = 0;
ThreadCache *ThreadCache::thr_cache[TATTRIB_MAX_PRIO-TATTRIB_MIN_PRIO][TATTRIB_MAX_STACK];

class THInit {
public:
  THInit() {
    ThreadCache::init_stack(); 
#ifdef PTH_THREADS
    if (!pth_init()) fatal("Threads initialization failed");
#endif
    ThreadCache::init_threads();
    sigset_t sigs;
    // Unblock SIGINT, SIGTERM
    sigemptyset(&sigs);
    sigaddset(&sigs, SIGINT);
    sigaddset(&sigs, SIGTERM);
#ifdef PTH_THREADS
    pth_sigmask(SIG_UNBLOCK, &sigs, NULL);
#else
    sigprocmask(SIG_UNBLOCK, &sigs, NULL);
#endif
  }
} thInit;

#endif//IMPL  

externNoImpl long long hallocated;

template <typename Cache>
class Header : public ACnt, public Cache {
public:
  Header(const Header<Cache>&)             { assert(0 && "Attempt to copy header"); }
  Header& operator= (const Header<Cache>&) { assert(0 && "Attempt to copy header"); return *this; }
  Header(){}

  size_t   size;
  Header  *next;

  virtual ~Header() {} 

  inline static Header<Cache>* halloc (size_t size) {
    Header<Cache> *h, **c = (Header<Cache>**)pcache(size);

    if (c && *c) {
      h = *c;
      *c = h->next;
    } else {
      h = (Header<Cache>*)malloc(size);
      assert(h && "malloc returns NULL");
      hallocated += size;
      assert(hallocated < myMemLimit());
      h->size = size;
      h->pre_ct();
    }
    totalUsed() += size;
    
    return h;
  }

  inline void* operator new (size_t size, size_t extra) {
    return halloc(size + extra);
  }
  
  inline static void hfree (Header<Cache>* h) {
    Header<Cache> **l = (Header<Cache>**)pcache(h->size);
    totalUsed() -= h->size;
    if (l) {
      h->next = *l;
      *l = h;
    } else {
      hallocated -= h->size;
      free(h);
    }
  }

  inline void* operator new (size_t size) {
    return halloc(size);
  }

  inline void operator delete (void* mem) {
    hfree((Header<Cache>*)mem);
  }

  virtual void onZero () { 
    delete this;  //v! "header = %p", this
  }
};


// Shortcuts for handy usage
#define SDATAINTRANS 32
#define SDATATRANS   64

class SData : public Header<MemCache>, public Viz {
public:
  int flags;
  static int ctFlags;
  inline SData() : flags (0) { ctFlags = 0; }

  virtual void onRecv() {
    //v! "%s: not overwritten", __PRETTY_FUNCTION__
  }

};

#ifdef IMPL
int SData::ctFlags;
#endif//IMPL

class SRef : public ARef {  //FIXME: may be improper ~ARef call !
public:
  inline SData* operator-> ()  { return (SData*) a; }
  inline SData& operator*  ()  { 
    //v! "SRef::operator* this=%p SData=%p\n", this, a
    return *(SData*)a; 
  }
  inline SData* operator= (SData *_a)      { 
    //v! "SRef::operator= this=%p SData=%p\n", this, _a
    return (SData*) (*(ARef*)this = _a); 
  }
  inline SData* operator= (const SRef& s)  { return (SData*) (*(ARef*)this = s.a);}
  inline operator SData* () { return (SData*)a; }
  bool operator == (ACnt *_a) { return a==_a; }
  inline void clearRef () { (*this) = (SData*)NULL; }
};

class SThread : public Header<ThreadCache> {
};

// MPI-related stuff

#define MPIOK(mpicall) { \
  int stat = mpicall;  \
  if (stat != MPI_SUCCESS) { \
    static char msg[0x1000]; \
    int len = sizeof(msg); \
    MPI_Error_string(stat,msg,&len); \
    fatal("MPI_Error[%d != %d]: %.*s",stat, MPI_SUCCESS,len,msg); \
  } \
}

class WTimer 
{
		double frequency;
		inline void getCpuInfo();
	public:
		inline WTimer();
		inline double wTime();
		inline unsigned long long tickTime();
		inline unsigned long long sec2Tick(double time);
		inline double tick2Sec(unsigned long long time);
};

inline WTimer::WTimer() {
	getCpuInfo();
}

inline void WTimer::getCpuInfo()
{
	const char* buf;

	buf=fileInfo("/proc/cpuinfo","cpu MHz",": ","\n");
	if (!buf)
		frequency=0;
	else
		frequency=1/(double)(atof(buf)*1000000);
	

#if !defined(IA32_ACCEL) && !defined(x86_64_ACCEL)
	frequency=1/(double)1000000;
#endif
	return;
}

inline double WTimer::wTime()
{
        return tickTime()*frequency;
}

inline unsigned long long WTimer::tickTime()
{
	unsigned long long tmp;
#ifdef IA32_ACCEL
	asm volatile ( "rdtsc\n\t"
	    : "=A" (tmp) );
#elif X86_64_ACCEL
	unsigned int low, high;
	asm volatile ( "rdtsc\n\t"
	    : "=a" (low), "=d" (high) );
	tmp = ((unsigned long long) high) << 32 | low;
#else
	tmp =(unsigned long long)(1000000* MPI_WTime());
#endif
        return tmp;
}

inline unsigned long long WTimer::sec2Tick(double time)
{
	return (unsigned long long)(time/frequency);
}

inline double WTimer::tick2Sec(unsigned long long time)
{
	return time*frequency;
}


extern WTimer localTimer;

#ifdef IMPL
WTimer localTimer;
#endif
double wTime();
unsigned long long tickTime();
double tick2Sec(unsigned long long);

#ifdef IMPL

unsigned long long tickTime() {
	return localTimer.tickTime();
}
double tick2Sec(unsigned long long tick) {
	return localTimer.tick2Sec(tick);
}
double wTime() {
	return localTimer.wTime(); 
}

#else
double wTime();
#endif

#define MAXMPITAG 4
#define MAXSENDRQ 0x4000 // FIXME!!!

externNoImpl Delay *mpiInitDelay;

externNoImpl bool mpiInitialized;

class MPITagHandler : public Service {
  static MPITagHandler* tagHandler[MAXMPITAG];
  static int nextTag;
  static int totalTags;

  int tag;
  int curRq;

  class Request {
  private:
    SRef data;
  public:
    MPI_Request mpiRq;
    
    bool isFree() {
      assert (!data  ||  (data->flags & SDATAINTRANS));
      return  !data  || !(data->flags & SDATAINTRANS);
    }

    void setFree() {
      assert (!isFree());
      if (data) { //FIXME: multiply send?
	data->flags &= ~SDATAINTRANS;
	data->flags |=  SDATATRANS;
	// Avoid manual destructor call - data is moved, not deleted!
	//FIXME: SData::hfree(&*data); why we change it to `data = (SData*)NULL' ?
	//FIXME: data.a = NULL;
	data.clearRef(); //FIXME ???
      }
    }
    
    void setBusy(SData *d) {
      assert (isFree());
      assert (d);
      assert (!(d->flags & SDATAINTRANS));
      data = d;
      data->flags |= SDATAINTRANS;
    }
    
    const SData* getData() {
      return data;
    }

    //FIXME: is it needed ?
    ~Request() {
      if (!isFree()) {
        data->flags &= ~SDATAINTRANS;
        data->flags |= SDATATRANS;
        data.clearRef();
      }
    }
     
  } rq[MAXSENDRQ]; //FIXME: add !MPI_Test window!
  
#define MAXTESTTRIES 1000000
  bool testRq(Request *rq) {
    MPI_Status status;
    int free;

    assert (rq);
    assert (!rq->isFree());
    MPIOK(MPI_Test (&rq->mpiRq, &free, &status));
    if (free) rq->setFree ();
    return free;
  }

  void sendRq(int dest, Request *rq, SData *d) {
    assert (rq);
    assert (d);
    assert (rq->isFree ());
    rq->setBusy (d);
    myStat->msgsSent++;
    if (tc.showMsgs) uiprintf ("rank%d,tag%d -> rank%d:  %s[%d]\n", myRank, tag, dest, d->viz(), d->size);
    {
      TaskStatistic::Holder __taskHolder((TaskStatistic*)&mpiTask);
      MPIOK(MPI_Isend (d, d->size, MPI_BYTE, dest, tag, MPI_COMM_WORLD, &rq->mpiRq));
      TaskStatistic::getCurrent()->msg_count++;
      TaskStatistic::getCurrent()->msg_size += d->size;
    }
  }
  
  Request* getFreeRq() {
    int i, tries;
    for (tries = 0; tries < MAXTESTTRIES; tries++,sched_yield()) { // Switch to communication process when needed
      // while (work ()) { fprintf (stderr, "[%d] getfree\n", myRank); }
      //v "try lock MPI request %d for data %p ",tries
      if (curRq >= MAXSENDRQ) curRq = 0;
      for (i = curRq; i < MAXSENDRQ; i++) 
	if (rq[i].isFree() || testRq(&rq[i])) 
	  { //v! "return MPI request i=%d rq=%p ", i, rq+i   
	    curRq = i+1;
            return &rq[i]; 
	  } 
    }
    fatal("No free MPI_Request (%d tries)", tries);
    return NULL;
  }
public:
  void forceTransmit (SData *d=NULL) {
    if (realsuperSize < 2) return;
    if (d && !( d->flags & SDATAINTRANS )) return; 
    int i, tries, free=0;
    for (i = 0; i < MAXSENDRQ; i++) { 
      if ((d && (rq[i].getData()!=d)) || rq[i].isFree()) continue ; 
  
      for (tries = 0; tries < MAXTESTTRIES; tries++,sched_yield()) { // Switch to communication process when needed
	//while (work ()) { fprintf (stderr, "[%d] force\n", myRank); }  //  FIXME!!!!!!!!!!!!!!!!!!!!!! hack !!!!!!!!!!!!!!!!
 	//v! "Forcing MPI request i=%d data=%p", i, rq[i].getData()
	if (testRq(&rq[i])) { free=1; break ; } 
	
      }
      	
      if (tries==MAXTESTTRIES) fatal("Unable to force transmit for %p (%d tries)", d, tries);
      else if (d) return ; 
    }

    if (d && !free) fatal("Unable to find Data %p in transmitter",d);
  }
   
  void asyncSend(SData* d, int to /* -1 => broadcast */) {
    //v! "asyncSend"
    if (!totalTags) return; // Ignore send if no handlers 

    int i = to>=0 ? to:0, end = to>=0 ? to+1:realsuperSize;
    for (; i < end; i++) {
      //FIXME: variable size
      Request *rq = getFreeRq();

      sendRq (i, rq, d);


      // FIXME: do not force transmitter HERE!
      testRq (rq);
      //MPI_Status status;
      //MPIOK(MPI_Wait(&rq->mpiRq, &status));

    }
  }

private:

  class MPI_Init_Delay : public Delay {
  public:
    ~MPI_Init_Delay() {
      MPIOK(MPI_Init(CmdLine::mainArgc,CmdLine::mainArgv));  
      MPIOK(MPI_Comm_rank(MPI_COMM_WORLD, &myRank));
      MPIOK(MPI_Comm_size(MPI_COMM_WORLD, &realsuperSize)); //FIXME

      mpiInitialized = true;
      
      assert(realsuperSize<=superSize);
      //v! "Initialized MPI: myRank=%d...", myRank
      Delay::force (mpiInitDelay);
    }
  };

public:
  MPITagHandler(const char *name,int servicePrio) : Service(name,servicePrio), tag(nextTag++), curRq(0) {
    if (!totalTags++) {
      (*new MPI_Init_Delay())(CmdLine::delay);
    }
    tagHandler[tag] = this; //v! "registering %s for MPI tag %d", name, tag
  }
  virtual ~MPITagHandler() {
    if (!mpiInitialized) {
	    tagHandler[tag] = NULL;
	    --totalTags;
	    return;
    }
    if (!Service::exitError) MPIOK(MPI_Barrier(MPI_COMM_WORLD));
    tagHandler[tag] = NULL;
    if (!--totalTags) { 
      MPIOK(MPI_Finalize()); //v! Finalizing MPI... ! done
      //uiprintf("rank %d: MPI finalized\n", myRank);
    }
  }


  virtual bool work() {
    int avail, count = 0;
    MPI_Status status;
    
    MPIOK(MPI_Iprobe(MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &avail, &status)); 
    if (!avail) return false;

    MPIOK(MPI_Get_count(&status, MPI_BYTE, &count));
    SData *d = (SData*)SData::halloc(count);
    {
      TaskStatistic::Holder __taskHolder((TaskStatistic*)&mpiTask);
      MPIOK(MPI_Recv(d, count, MPI_BYTE, status.MPI_SOURCE, tag, MPI_COMM_WORLD, &status)); 
      TaskStatistic::getCurrent()->msg_count++;
      TaskStatistic::getCurrent()->msg_size += d->size;
    }

    // Clear transmitter flags
    assert (d->flags & SDATAINTRANS);
    d->flags &= ~SDATAINTRANS;

    // Reinit data counter 
    assert (d->cnt > 0);
    d->cnt = 0;

    d->onRecv();  //v! Calling onRecv for incomming SData ! done

    return true; 
  }
};

#ifdef IMPL
MPITagHandler* MPITagHandler::tagHandler[MAXMPITAG];
int MPITagHandler::nextTag;
int MPITagHandler::totalTags;

void *dmpiEntry(int cmd, void *arg) { 
  assert(cmd==0); 
  static char status[0x1000];
  char hostname[0x100];
  memset(hostname, 0, sizeof(hostname));
  gethostname(hostname, sizeof(hostname));
  snprintf(status, sizeof(status)-1, "%s @ %s(build %s)", hostname, dmpi_so, mpi_portal->build_info);
  return status; 
}
Feature dmpiFeature("DMPI", "dynamic MPI selector", true, dmpiEntry);
#endif//IMPL

// Exported TransportHandler
class Cell {};
class TransportHandler : public MPITagHandler {
public:
  int cellRankSize;
  int cellStartOfs;
  int cellSize;
  int sharedSize;
// Hint for TSS: use it for lazy calls
  void (*onRead)(Cell*);
// Hint for pth: add waitq and do pth_suspend/pth_resume  
// Hint for TSS: use it for ALT implementation
  void (*onWrite)(Cell*);
  void (*inPlaceCtor)(Cell*);
  void (*clearCell)(Cell*);
  static void noop(Cell*) {}
  TransportHandler(const char *name, int prio) : MPITagHandler(name,prio),    // NULL for sigsegv
						 onRead(noop), onWrite(noop), inPlaceCtor(NULL), clearCell(noop) {}
};


// If it does not work on Linux - switch to another method
#define ZEROMAPSIZE 0x10000000
class VZeros {
public:  
  char *start,*end;
  char *nextChunk;
  static char* startMap, *endMap;
  static char* nextChunkMap;
  static void checkInit() {
    if (startMap) return;
    nextChunkMap = startMap = (char*)calloc(ZEROMAPSIZE,1); //mmap(NULL, ZEROMAPSIZE, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0); 
    assert(startMap != NULL);
    endMap = startMap + ZEROMAPSIZE;
  }
  VZeros(const char *name, size_t size) { 
    checkInit();
    start = nextChunk = nextChunkMap; 
    end = nextChunkMap = nextChunkMap + size; 
    assert(nextChunkMap < endMap); //v! "VZeroMap[%p-%p] for %s", start, end, name
  }
  char *allocZeros(size_t size) {
    char *result = nextChunk;
    nextChunk += size;
    assert(nextChunk <= end); //v! "result = %p", result
    return result;
  }
};

#ifdef IMPL
char *VZeros::startMap, *VZeros::endMap, *VZeros::nextChunkMap;
#endif//IMPL

//FIXME: optimize
inline int lg(int qty) {
  int l=0;
  while(qty >>= 1) l++;
  return l;
}

// Here we assume following functions in Chunk:
//  inline void chunkSetLg(int);
//  static void chunkBaseFree(Chunk*);
//  static void chunkInit(CachedChunks<Chunk>&);
#define ALIVECLNK ((void*)-1)
template <typename Chunk>
#define MAXCHUNKLG 32
class CachedChunks {
public:
  Chunk* cache[MAXCHUNKLG];
  int inUse;
  CachedChunks() : inUse(0) { memset(cache,0,sizeof(cache)); }
private:
  inline Chunk* pop(Chunk** p) {
    Chunk *c = *p;         
    if (!c) c = fillUp(p);
    *p = (Chunk*)c->clnk;  //v! "return = %p", c
    c->clnk = (Chunk*)ALIVECLNK; // Useful sanity for debugging
    return c;
  }
  inline void push(Chunk** p, Chunk *c) {
    c->clnk = *p;
    *p = c;
  }
  Chunk* fillUp(Chunk** p) {
    if ( p+1 >= cache+MAXCHUNKLG ) fatal ("Out of memory\n");
    Chunk *c = pop(p+1);
    int lg = p-cache; 
    Chunk *c1 = c + (1<<lg);
    c->chunkSetLg(lg); c1->chunkSetLg(lg);
    push(p,c1); push(p,c);   //v "return = %p", c
    return c;            
  }
public:
  inline Chunk* allocChunk(int qty) { //FIXME: BUG! lg -> size!!!
    //v! "qty = %d", qty
    inUse += qty;
    return pop(cache+lg(qty)); 
  } 
  inline void freeChunk(Chunk* c, int qty) { 
    inUse -= qty;
    push(cache+lg(qty),c); //v! "Free %d-chunk at %p; it pushed to %p (lg(qty)=%d), cache=%p", qty, c, cache+lg(qty), lg(qty), cache
  }
  void flushAllChunks() {
    Chunk** p;
    for (p=cache; p < cache+MAXCHUNKLG; p++) while (*p) Chunk::chunkBaseFree(pop(p));
  }
  void initCollect() {
    int _inUse = inUse;
    flushAllChunks();
    Chunk::chunkInit(this);
    inUse = _inUse;
  }
};


// Supercell (Supermatrix unit) 
template <TransportHandler& handler>
class SCell : public Cell, public SRef { //FIXME: add a method (ugly code)

  // copy constructor and operator= should NOT be used for any SCell, private to avoid compilation
  SCell(const SCell<handler>&)      { assert(0 && "Attempt to copy supercell"); }
  SCell<handler>& operator= (const SCell<handler>&) { assert(0 && "Attempt to copy supercell"); return *this; }

private:
  long long seqNo;
  
public:
  SCell<handler> *clnk;

  // FIXME: optimize for very large clusters (by means list)

  // Master-accessed
  unsigned mstRq:1;    
  unsigned mstDone:1;
  unsigned mstReady:1;

  // Slave-accessed
  unsigned slvRq:1;
  unsigned slvReady:1;

  unsigned ctDone:1;


  inline void dump() const { //????
    if(this)fprintf(stderr, "[%d:%x:%Ld|%c:%c]", (isShared()?-1:master()), offset(), seqNo,
		    (mstReady?'m':'-'), (slvReady?'s':'-'));
    else    fprintf(stderr,"[]");
  }

  inline SCell() { 
    if (isShared() || master() == myRank) {
      mstRq = mstDone = mstReady = 0;
    }
    slvRq = slvReady = 0;
    ctDone = 1;
    if (!isShared() && master() == myRank) seqNo++; 
    assert(cellIsOk()); 

    //FIXME: Reset Mst Flags for slaved mirrors   Is it needed  - finalizeMaster does it

  }  
  virtual ~SCell() { 
    if (master() == myRank) finalizeMaster(); 
    ctDone = 0; 
  }

  inline long long getSeqNo () const {return seqNo;}

  inline void setSlvSeqNo (long long _seqNo) {
    assert(cellIsOk());
    assert(!isShared());
    assert(_seqNo >= seqNo);
    if (master() == myRank) return;
    assert(_seqNo > seqNo);
    seqNo = _seqNo;
    assert(&data() == NULL);
    //if (slvRq != 0) uiprintf("[%d:%x:%Ld]\n", master(),oindex(),seqNo);
    //assert(slvRq == 0);
    assert(slvReady == 0);
  }

  /* Various cell ariphmetic */

  inline int offset() const { return (char*)this - VZeros::startMap; }

  inline int oindex() const { 
    //v "%s", __PRETTY_FUNCTION__
    //v "this=%p, offset()=%x, cellStartOfs=%x, cellRankSize=%d", this, offset(), handler.cellStartOfs, handler.cellRankSize 
    //v "oindex = %d", (offset()-handler.cellStartOfs) % handler.cellRankSize
    // FIXME:: currently assume that shared cells number is always less than cellRankSize
    return (offset()-handler.cellStartOfs) % handler.cellRankSize; 
  }

  inline int master() const { 
    assert(!isShared()); // if(isShared()) return -1;
    return (offset()-handler.cellStartOfs) / handler.cellRankSize;  
  }

  inline bool isShared() const {
    bool res = offset()-handler.cellStartOfs >= handler.cellRankSize*superSize;
    //v! "Cell %x ishared %d cellRankSize*superSize %d",offset(),res,handler.cellRankSize*superSize
    return res;
  }

  inline bool cellIsOk() const {
    if (!this) return false;
    bool ok = ctDone && ( offset() >= 0x1000 && offset() < handler.cellStartOfs + handler.cellRankSize*superSize+handler.sharedSize); // FIXME ALIVECLNK status 
    if (!ok) {
      fprintf (stderr," Wrong cell %p clnk %p offset %x bound %x \n", this, clnk,  offset(),  handler.cellStartOfs + handler.cellRankSize*superSize+handler.sharedSize);
    }
    return ok; 
  }

  static inline SCell<handler>& checkInit(SCell<handler>& c) { 
    if (!c.ctDone) handler.inPlaceCtor(&c);
    return c;
  }

  static inline SCell<handler>& cellAt(int offs) { 
    return checkInit(*(SCell<handler> *)(VZeros::startMap+offs));
  }

  /* Returns an index of "image" cell of supermatix on the node "rank" for current cell 
   */
  inline int mirror(int rank) const {
    assert(!isShared());
    //v! "mirror value %x", handler.cellRankSize + handler.cellStartOfs + oindex()
    return handler.cellStartOfs + rank*handler.cellRankSize + oindex();
  }
  /* Operator [] returns an "image" of cell with same oindex() that 
     is of current cell, but for a node which
     rank specified as a parameter.
     It is used to keep cell flags of supermemory 
   */ 
  inline SCell<handler>& operator[] (int rank) const {
    //v! "%s", __PRETTY_FUNCTION__
    //v! "this=%p, offset()=%x, cellStartOfs=%x, cellRankSize=%d", this, offset(), handler.cellStartOfs, handler.cellRankSize 
    //v! "oindex = %d", (offset()-handler.cellStartOfs) % handler.cellRankSize
    //v! " ---cell rank requested is %d",rank
    return cellAt(mirror(rank)); 
  }
  
  /* Moveable cell data */

  class Data : public SData {
  public:
    int cellOffset;
    long long seqNo;
    typedef void (SCell<handler>::*OnRecvMethod)(Data&);
    OnRecvMethod method;
    int srcRank;
    Data():cellOffset(0), seqNo(0), method(NULL), srcRank(myRank) {}; 
    Data(int _cellOffset, long long _seqNo, OnRecvMethod _method, const char* _vid) : // Control handler signals
      cellOffset(_cellOffset), seqNo(_seqNo), method(_method), srcRank(myRank) { vid = _vid; }
    
    virtual void onRecv() { 
      // Don't forget collect data (functions are collected manually)!
      SRef forCollect; 
      forCollect = this;

      SCell<handler> *c = &cellAt(cellOffset);

      assert((c->*method)!=NULL); 
      (c->*method)(*this); //v! Handling incomming data ! done
    }
  };

  inline Data& data() { return *(Data *)& * *(SRef*)this; } // returns a Data object associated with SCell object
  inline Data*  operator=(Data* d) { return (Data*)(*(SRef *)this = d); }
  inline operator Data*() { return &data(); }


  /* Read/Write cell */

  void read(long long _seqNo) {
    if (isShared() || master() == myRank) { assert(seqNo == _seqNo); handler.onRead(this); return; } // No remote 
    assert(!(slvRq && slvReady));
    if (slvRq) return; // Already reading...
    slvReady = 0;
    //v! "reaad at %x caused sending request w mirror %x seqNo %d to the master node  %d",offset(),mirror(myRank),_seqNo,master()
    handler.asyncSend(new Data(offset(),_seqNo,&SCell<handler>::onMstRead,"readRq"), master());
    slvRq = 1;
  }

  void onMstSetSlvUse(Data &d) {
    if (d.seqNo != seqNo) { 
      uiprintf("INFO: too old slave usage request: %lld != %lld\n", d.seqNo, seqNo);
      return;
    }
    SCell<handler>& slv = (*this)[d.srcRank];
    slv.mstDone = 1;
  }

  void onMstRead(Data &d) {
    //v! here
    if (d.seqNo != seqNo) { 
      uiprintf("INFO: too old master read request: %lld != %lld\n", d.seqNo, seqNo);
      return;
    }
    SCell<handler>& slv = (*this)[d.srcRank];
    if (!slv.mstReady) slv.mstRq = 1; 
    else sendToSlave(slv,d.srcRank); 
    handler.onRead(this);
  }

  void sendToSlave(SCell<handler>& slv, int slvRank) {
    assert(cellIsOk());  //v! "sendToSlave cell %p", this
    handler.forceTransmit(&data()); 

    data().method=&SCell<handler>::onSlvWrite;
    data().seqNo=seqNo; 
    data().cellOffset=offset();
    handler.asyncSend(&data(),slvRank);

    slv.mstRq = slv.mstReady = 0;
    slv.mstDone = 1;
  }

  void sendToMaster() {
    assert(cellIsOk()); //v! "sendToMaster cell %p\n", this
    handler.forceTransmit(&data()); // Wait for outcoming 

    data().method=&SCell<handler>::onMstWrite;
    data().seqNo=seqNo; 
    data().cellOffset=offset();
    handler.asyncSend(&data(),master());

    slvRq    = 0;
    slvReady = 1;
  }

  void onMstWrite(Data& d) { //FIXME: mark mstDone for incoming data rank 
    if (&data() != &d) *this = &d;
    //v! "Master node write superSize %d ", superSize
    //v! "This %p , oindex %x, offset %x ", this, oindex(), offset() 
    int i;
    for (i = 0; i < realsuperSize; i++) {
      SCell<handler> *c = &(*this)[i]; // get an image of remote cell on node i
      //v! "write to client for cell %p", c
      if  (c->mstRq && (i != d.srcRank)) sendToSlave(*c,i); 
      else c->mstReady = 1;
      if ((i != myRank) && (i == d.srcRank)) {
	      c->mstDone = 1;
      }
    }
    slvReady = 1;
    handler.onWrite(this);
  }

  void write(long long _seqNo) {
    //v! "write cell %p seqNo %d old seqNo %d ", this, (int)_seqNo, (int)seqNo    
    data().srcRank = myRank; //???????
    if (isShared()) { writeToAll(); return; }
    if (_seqNo < seqNo) { 
      uiprintf("INFO: attempt to overwrite with seqNo %lld less than current %lld\n",_seqNo,seqNo);
      return;
    }      
    if (master() == myRank) onMstWrite(data());  
    else  { sendToMaster(); handler.onWrite(this); }
  }

  void sendSlvUse() {
    if (master() != myRank) handler.asyncSend(new Data(offset(),seqNo,&SCell<handler>::onMstSetSlvUse,"setSlvUse"), master());
  }

  // Master-independed routine
  void writeToAll() {
    assert(isShared());
    data().method=&SCell<handler>::onSlvWrite;
    data().seqNo=seqNo; 
    data().cellOffset=offset();
    int i;
    for (i = 0; i < realsuperSize; i++) {
      if (i == myRank) continue; //FIXME: call onSharedWrite here ?
      handler.forceTransmit(&data()); // FIXME: optimize it!
      handler.asyncSend(&data(),i); 
    }
  }

  void onSlvWrite(Data &d) {
    //v! "onSlvWrite for cell %p\n", this
    if (d.seqNo < seqNo) { 
      uiprintf("INFO: too old master incoming data: %lld < %lld\n", d.seqNo, seqNo);
      return;
    }
    slvRq = 0; 
    *this=&d;         //v! Before data assign ! After data assign
    slvReady = 1;
    seqNo = d.seqNo;
    handler.onWrite(this);//v! Before onWrite() ! After onWrite
  }

  /* Finalization */

  void finalizeMaster() {
    int i; 
    for (i =0; i < realsuperSize; i++) {
      SCell *c = &(*this)[i];
      if (c->mstDone) handler.asyncSend(new Data(offset(),seqNo,&SCell<handler>::onSlvFinalize, "finalize"), i);
      // Reset slaved state
      c->mstDone = c->mstRq = c->mstReady = 0;
    }
    clearRef();
  }

  void onSlvFinalize(Data &d) {
    assert(cellIsOk() && seqNo == d.seqNo);
    clearRef();
    handler.clearCell(this);
    ctDone = 0;
  };


  /* Chunk-related code */

  static void chunkBaseFree(SCell<handler>*) {
    //v! FIXME: unimplemented
  }

  static void chunkInit(void *_cc) { // FIXME: ... parametrization
    CachedChunks< SCell<handler> >* cc = (CachedChunks< SCell<handler> >*) _cc;
    //v! "cellStartOfs=%p cellRankSize=%d cellSize=%d myRank=%d", handler.cellStartOfs, handler.cellRankSize, handler.cellSize, myRank
    cc->freeChunk((SCell<handler>*)(VZeros::startMap + handler.cellStartOfs + handler.cellRankSize * myRank), handler.cellRankSize/handler.cellSize);
  }

  void chunkSetLg(int) {/*FIXME*/}
};

template <typename Cell, TransportHandler& handler, int cellsPerNode, int sharedCells=0>
class SSegment : public CachedChunks<Cell>, public VZeros {
private:
  class SSegDelay : public Delay {
  private:
    CachedChunks<Cell> *data;
  public:
    SSegDelay (CachedChunks<Cell> *_data) : data(_data) {}
    ~SSegDelay() {data->initCollect();}
  };
 public:
  Cell *ns;
  SSegment() : VZeros(__PRETTY_FUNCTION__,sizeof(Cell)*(cellsPerNode*superSize+sharedCells) + 0x1000) { 
    ns = (Cell*) (allocZeros(sizeof(Cell)*(cellsPerNode*superSize+sharedCells) + 0x1000) + 0x1000); //FIXME: 0x1000 - to avoid zero offset
    handler.cellStartOfs = (char*)ns - VZeros::startMap;
    handler.cellSize = sizeof(Cell);
    handler.cellRankSize = sizeof(Cell)*cellsPerNode;
    handler.sharedSize = sizeof(Cell)*sharedCells;
    // Delayed part of initialization
    (*new SSegDelay(this))(mpiInitDelay);
   }
   Cell& operator[] (int i) { 
   //v! "Ssegop ns = %p, i = %d, cellsPerNode=%d", ns, i, cellsPerNode
    return Cell::checkInit(ns [i]); 
  }
  Cell& sharedCell (int i) {
    //v! "ns = %p, i = %d, cellsPerNode=%d", ns, i, cellsPerNode
    assert(i >= 0 && i < sharedCells);
    return Cell::checkInit(ns [cellsPerNode*superSize+i]);
  }
};

