// Built-in T++ benchmark/testing suite
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <setjmp.h>
#include "trt"
using namespace ts;

int errCnt = 0;
Prof prof;

static char status[0x10000];//FIXME: use String
#define success(t,iter)  (sprintf(status,"%.0f",(ts::tick2Sec(t.finish())*1000000000./(iter))),true) 
#define error(x...)      (errCnt++,sprintf(status,x),false)

// Test suite functions

#define NEW_ITERS 100000
#define NEW_SIZES 5
bool new_delete() {
  int i, j;
  Prof::Timer t(prof);
  for (i = 0; i < NEW_ITERS; i++) {
    char *m[NEW_SIZES] = { 
      new char[30], new char[60], new char[90], new char[120], new char[150]
    };
    for (j = 0; j<NEW_SIZES; j++) delete m[j];
  }
  return success(t,NEW_ITERS*NEW_SIZES);
}

class Char30  : public Header<MemCache> { char a[30]; };
class Char60  : public Header<MemCache> { char a[60]; };
class Char90  : public Header<MemCache> { char a[90]; };
class Char120 : public Header<MemCache> { char a[120]; };
class Char150 : public Header<MemCache> { char a[150]; };

bool super_new_delete() {
  int i, j;
  Prof::Timer t(prof);
  for (i = 0; i < NEW_ITERS; i++) {
    Header<MemCache> *m[NEW_SIZES] = { 
      new Char30, new Char60, new Char90, new Char120, new Char150
    };
    for (j = 0; j<NEW_SIZES; j++) delete m[j];
  }
  return success(t,NEW_ITERS*NEW_SIZES);
}

bool super_new_write_delete() {
  int i;
  Prof::Timer t(prof);
  for (i = 0; i < NEW_ITERS; i++) {
    TPtr<int> p = new TObj<int>;
    *p = 1;
  }
  return success(t,NEW_ITERS);
}

bool tobj_new_delete() {
  int i, j;
  Prof::Timer t(prof);

  //   return false;

  for (i = 0; i < NEW_ITERS; i++) {
    TObj<int> *m[NEW_SIZES] = { 
      new TObj<int>, new TObj<int>, new TObj<int>, new TObj<int>, new TObj<int>
    };
    for (j = 0; j<NEW_SIZES; j++) delete m[j];
  }
  return success(t,NEW_ITERS*NEW_SIZES);
}
#undef NEW_SIZES
#undef NEW_ITERS


#define C_CALLS 10000000
bool dummy_c_call() {
  int i, res;
  Prof::Timer t(prof);
  extern void dummy_c_fun(int *);
  for (i = 0; i < C_CALLS; i++) dummy_c_fun(&res);
  return success(t,C_CALLS);
}
#undef C_CALLS
void dummy_c_fun(int *res) { *res = 1; }

#define CXSWS 1000000
bool setjmp_longjmp() {
  int i;
  Prof::Timer t(prof);
  jmp_buf jmp;
  for (i = 0; i < CXSWS; i++) {
    if (!setjmp(jmp)) longjmp(jmp,1);
  }
  return success(t,CXSWS);
}


bool context_switch() {
  int i;
  Prof::Timer t(prof);
  for (i = 0; i < CXSWS; i++) {
  }
  return success(t,CXSWS);
}

//FIXME: BUG: save == restore!!!
bool super_context_switch() {
  int i;
  Prof::Timer t(prof);

  //printf("FIXME\n"); return false;

  for (i = 0; i < CXSWS; i++) {
    ThrH::lastRunOthers();
  }
  return success(t,CXSWS);
}

#undef CXSWS

#if 0
class FibTask : public TFun<int> {  
public:
  int n;
  inline FibTask(int _n) : n(_n) {
    //v! "new FibTask(%d)", n
  }
  static inline TObj<int>& result(int n)  {
    FibTask *tsk = new FibTask(n);
    TObj<int>* res = &tsk->retval().ownerCopy();
    tsk->stopAndContParent();
    return *res;
  }
  const char* viz() { VizStr v; v.printf("FibTask[%d]",n); return v.close(); } 
  virtual TVar<int> body() {
    if (n < 2) { //v! Leaf value
      return n;
    } else {
      TVar<int> n1 = result(n-1); //v! "n1 = %p", &n1()
      TVar<int> n2 = result(n-2); //v! "n2 = %p", &n2()
      return n1()()() + n2()()();  
      // fprintf(stderr,"fib(%d)=%d, n1=%d, n2=%d\n",n,r()()(),n1()()(),n2()()());
    }
  }
  // Autodrop in dt()
};

#else

TFUNDEF(int, fib, (int arg1), ,int arg1;, _tf->arg1=arg1;, , ,);
TFUNIMPL(int, fib, (int n), ,int n;, _tf->n=n;, , ,) {
  //v! "Starting fib(%d)", n
  if (n < 2) { 
    return n;
  } else {
    //ts::TVar<int> n1 = fib(n-1); 
    //ts::TVar<int> n2 = fib(n-2); 
    return fib(n-1) + fib(n-2);
  }
}

#endif

//#define FIB_N 28
#define FIB_N 8

bool fibt() {
  Prof::Timer t(prof);
  { // You will got failure without block here :-)
    mainResult = fib(FIB_N); //FibTask::result(FIB_N);


    //uiprintf("mainResult = %p, mainResult.owner = %d\n", &mainResult(), mainResult().owner);

    Service::doAllWork();
    uiprintf("\nResult of Fib(%d) = %d\n", FIB_N, (int)mainResult);
    uiprintf("Total %d superthreads created\n", ThreadCache::superthreads);
    mainResult.reset();
  }
  uiprintf("After test %d supercells are locked\n", tseg().inUse);
  return tseg().inUse ? error("supermemory leak") : success(t,1);
}

// from cilk
int fibc_(int n)
{
     if (n < 2)
          return (n);
     else {
          int x, y;
          x = fibc_(n - 1);
          y = fibc_(n - 2);
          return (x + y);
     }
}

bool fibc() {
  Prof::Timer t(prof);
  /*int r =*/ fibc_(FIB_N);
  //printf("Result of fibc = %d\n", r);
  return success(t,1);
}

#undef FIB_N

#define STREAMLEN 1000000

struct List {
  TPtr<List> next;
  int n;
  int sig;
  List() : next(NULL),sig(17) { fprintf(stderr,"\nCreated List %p\n", this); }
  List(const List &l) {
    fprintf(stderr,"\nCopying List %p -> %p\n", &l, this);
    n = l.n;
    assert(0);//sig = 0;
    next = l.next;
  }
  ~List() { fprintf(stderr,"\nDestroying List %p\n", this); assert(sig==17); }
};

#if 0
class Squarer : public MThread {  
public:
  TPtr<List> in;
  TPtr<List> out;
  inline Squarer(TObj<List>* _in) : in(_in) { out = new TObj<List>; }
  static inline TObj<List>* result(TObj<List>* in)  {
    Squarer *tsk = new Squarer(in);
    TObj<List> *res = &tsk->out().ownerCopy();
    tsk->stopAndContParent();
    return res;
  }
  const char* viz() { VizStr v; v.printf("Squarer"); return v.close(); } 
  virtual void body() {
    while (in) {
      out->next = new TObj<List>;
      out->n = in->n * in->n; fprintf(stderr,"Element Square = %d", out->n);
      in = in->next;     //v! GC here - producer
      out = out->next;   //v! "Autodrop here - producer, new out=%p", (List*)out
    }
  }
};

class SqSum : public MThread {  
public:
  int n;
  TVar<int> r;
  inline SqSum(int _n) : n(_n) {}
  static inline TObj<int>& result(int n)  {
    SqSum *tsk = new SqSum(n);
    TObj<int>* res = &tsk->r().ownerCopy();
    tsk->stopAndContParent();
    return *res;
  }
  const char* viz() { VizStr v; v.printf("SqSum[%d]",n); return v.close(); } 
  virtual void body() {
    TPtr<List> prod = new TObj<List>; 
    TPtr<List> res = Squarer::result(&prod().ownerCopy());
    r = 0;
    int i;
    for (i = 0; i < n; i++) {
      prod->n = i+7;  fprintf(stderr,"Element = %d", prod->n);
      prod->next = (i < n-1) ? new TObj<List> : NULL; //v! Fill prod->next
      prod = prod->next;  //v! "Autodrop here - consumer, new prod=%p", (List*)prod ! done
      r += res->n;     //v! "Accumulated value = %d", r()()()
      res = res->next; //v! GC here - consumer
    }
  }
};

#define SQLISTLEN 1
bool dlstream() {
  TVar<int> r = SqSum::result(SQLISTLEN);
  Prof::Timer t(prof);
  ThrH::lastRunOthers(); 
  uiprintf("**  DONE  **");
  uiprintf("Sum of list from squarer (7 .. %d) = %d\n", 7+SQLISTLEN-1, r()()());
  return success(t,STREAMLEN);
}
#undef STREAMLEN
#endif


#define THR_ITERS 100000
bool sthread_new_del() {
  int i;
  Prof::Timer t(prof);
  for (i = 0; i < THR_ITERS; i++) {
    SThread *p = new SThread;
    delete p;
  }
  return success(t,THR_ITERS);
}
#undef THR_ITERS

bool MPI2() {
  return error("Not implemented");
}

#define CATCH_ITER 1000
void mythrow() {
  throw "Help!";
}
bool try_catch() {
  int i = CATCH_ITER;
  Prof::Timer t(prof);
  while (i > 0) {
    try {
      mythrow(); 
    }
    catch(const char* p) {
      i--;
      continue;
    }
  }
  return i == 0 ? success(t,CATCH_ITER) : error("Something bad");
}
#undef CATCH_ITER

void test_result (const char *test, bool (*testfn)(void)) {
  VizStr s;
  const char *space= "                             ";
  const char* sppos= space + s.printf("%s",test);
  s.printf("%s", sppos); 
  ui->write(White, s.close());
  bool ok = (*testfn)();
  ui->write(ok? Green : bright(Red), ok? "OK\t\t\t\t":"Failed\t\t\t");
  ui->write(White,status);
  ui->write(White,"\n");
}

tfunmainTFunDef* tfunmainTFunDef::New() { return NULL; }//DELME

void performTests() {
  ui->write(White,"Built-in T++ benchmark/testing suite\n");
  ui->write(White,(const char*)ts::sysconf(_SC_SUMMARY));

  ui->write(White, "Benchmark                    Status                     NSec[/Error]\n");
  ui->write(Cyan,  "================             ======                     ============\n");
#define RUNTEST(fn) do { test_result(#fn,fn); } while(0)

  // FIXME: make header & groups
  //RUNTEST(dlstream); //FIXME-order

  //RUNTEST(new_delete);
  //RUNTEST(super_new_delete);
  //RUNTEST(super_new_write_delete);
  //  RUNTEST(tobj_new_delete);
  RUNTEST(dummy_c_call);
  RUNTEST(setjmp_longjmp);
  //RUNTEST(context_switch);
  //RUNTEST(super_context_switch);
  //  RUNTEST(sthread_new_del);
  //  RUNTEST(COW);
  //RUNTEST(MPI2);
  RUNTEST(fibc);
  RUNTEST(fibt); 

  //RUNTEST(try_catch);

#undef  RUNTEST
}

int main (int argc, char *argv[]) {
  TRT trt(&argc,&argv);
  if (myRank == mainRank) {
    MainTRT maintrt;
    performTests();
  } else {
    ts::Service::doAllWork();
  }
  return errCnt;
}


