// Open T-System Runtime 
// PSI RAS, Pereslavl-Zalessky, Russia.

#ifndef __TRT__
#define __TRT__

#include <unistd.h> 
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <malloc.h>
//#include <pth.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <regex.h>
#include <new>
#include <exception>
#include <typeinfo>
#ifdef __SMP__
#include <atomic.h>
#else //!__SMP__
typedef int atomic_t;
static inline int  atomic_dec_and_test(atomic_t *v) { return !--(*v); }
static inline void atomic_inc         (atomic_t *v) { (*v)++; }
#define atomic_read(v)    (*v)
#define atomic_set(v,i)  ((*v) = (i))
#endif
#include <dmpi.h>

#ifndef NOASSERT
#include <assert.h>
#else
#warning "runtime checks disabled!"
#define assert(x)
#endif

#ifdef IMPL
#define externNoImpl
#else //IMPL
#define externNoImpl extern
#endif //IMPL

#define TS_MODULE "trt"

extern "C" {
char *__cxa_demangle (const char *__mangled_name, char *__output_buffer, int *__length, int *__status);
}

namespace ts {

#define CLASSID(x) demangle(typeid(x).name())
char *demangle(const char* name);
#ifdef IMPL 
char *demangle(const char* name) {
  static char buf[0x1000];
  int length = sizeof(buf)-1;
  int err = 1;
  __cxa_demangle(name, buf, &length, &err);
  if (!err) 
    buf[length] = 0; 
  else      
    strncpy(buf,name,sizeof(buf));
  return buf;
}
#endif

const char *fileInfo(const char *file, const char *pat1, const char *pat2, const char *patTerm);

#ifdef IMPL
const char* fileInfo(const char *file, const char *pat1, const char *pat2, const char *patTerm) {
    static char buf[0x10000];
    int fd = open(file,O_RDONLY);
    assert(fd >= 0);
    assert(read(fd,buf,sizeof(buf)-1) > 0);
    close(fd);
    char *p = buf, *q = NULL; 
    (p=strstr(p,pat1)) && (p+=strlen(pat1)) && (p=strstr(p,pat2)) && (p+=strlen(pat2)) && (q=strstr(p,patTerm));
    return q ? (*q=0,p) : NULL;
}
#endif


class Atomic {
public:
  atomic_t cnt;
  Atomic(int _cnt)             { atomic_set(&cnt,_cnt); }
  inline operator int       () { return atomic_read(&cnt); }  
  // Only prefix operators
  inline void    operator++ () { atomic_inc(&cnt); } 
  inline bool    operator-- () { return !atomic_dec_and_test(&cnt); }
};

class ACnt : public Atomic {
public:
  ACnt() : Atomic(0) {}
  inline bool operator-- () {
    assert((int)(*this) > 0);
    return --(*(Atomic*)this) ? true : (onZero(),false);
  } 
  virtual void onZero() {}
};

class ARef {
private:
  inline void ctor(ACnt *_a)   { a = _a; if (a) ++(*a); }
  inline void dtor() { if (a)  { --(*a); a = NULL; } }
protected:
  ACnt *a;
public:

  inline ARef (ACnt* a = NULL) { ctor(a);    }
  inline ARef (const ARef &ar)   { ctor(ar.a); }
  inline ACnt* operator= (ACnt *_a) { 
    if (a !=   _a) { dtor(); ctor(_a);   } 
    return a; 
  } 
  inline ACnt* operator= (ARef& ar) { 
    if (a != ar.a) { dtor(); ctor(ar.a); } 
    return a; 
  } 
  inline ~ARef() { dtor(); }
};

class Spinlock : Atomic {
public:
  Spinlock() : Atomic(1)    {}
  inline void lock()        { while (--(*this)) { ++(*this); while((int)(*this)!=1)sched_yield(); } }
  inline bool try_lock()    { return (--(*this)) ? (++(*this),false) : true; }
  inline void unlock()      { ++(*this); }
  class Locker {
  public:
    Spinlock *sl;
    Locker(Spinlock *_sl) : sl(_sl) { sl->lock(); }
    ~Locker()                     { sl->unlock(); }
  };
};

#define callGraphPrintf(msg...) do {     \
  if (ts::tc.generateCallGraph) {        \
    static char buf[0x100000];           \
    char *p=buf, *e=buf+sizeof(buf)-1;   \
    p += snprintf(p,e-p,"\ncallGraph\t");\
    p += snprintf(p,e-p,msg);            \
    p += snprintf(p,e-p,"\n");           \
    fprintf(stderr, "%s\n", buf);        \
  }                                      \
} while (0)

// Superstring=fast temporary strings
//
// Usage: String s("hello %s", world); return s.close();
//
template <typename buftype, int levels>
class String {
  struct StrLevel {
    buftype buf; 
    char* p;
    void rewind() { p = buf; }
    int left() { return buf + sizeof(buf) - p; }
  };
  static StrLevel lvls[levels];
  static StrLevel *openl, *curl;
  char*  p;
public:
  static void rewind() { curl = NULL; }
  String() { open(); }
  String(const char *format, ...) {
    open();
    va_list args;
    va_start (args,format);
    p += vsnprintf(p,curl->left(),format,args); 
    va_end (args);
  }
  char *open() {
    if (curl == NULL) curl=openl=lvls-1; 
    if (++curl > openl) { 
      (++openl)->rewind(); 
      assert(openl < lvls + levels); 
    }
    return (p = curl->p); 
  }
  inline char operator* () { return *p; }
  inline operator char* () { return p; }
  inline char* operator ++() { return ++p; }
  inline char* operator ++(int) { return p++; }
  size_t printf (const char *format, ...) {
    va_list args;
    va_start (args,format);
    size_t n = vsnprintf(p,curl->left(),format,args); 
    p += n;
    va_end (args);
    return n;
  }
  char *close(int reserve = 0) { 
    char *ret = curl->p; 
    *p++ = 0; curl->p = p + reserve; 
    curl--; p = NULL;
    assert(curl->left() > 0); 
    return ret;
  } 
  ~String() { if (p) close(); }
};
  

#define VizStrType String<char[0x40000],8>
typedef VizStrType VizStr;
#ifdef IMPL
VizStrType::StrLevel* VizStrType::openl, *VizStr::curl;
VizStrType::StrLevel  VizStrType::lvls[8];
#endif //IMPL

// <Happy New Year>tree -style user interface

enum Color { Black, Red, Green, Yellow, Blue, Magenta, Cyan, White };
inline Color bright (Color c) { return (Color)(c | 8); }
#ifndef AUTOCOLOR
#define AUTOCOLOR ((Color)((myRank%4)+Green))
#endif
 
class UserInterface : Spinlock {
public:
  bool colorCapable;
  virtual int write (Color color, const char *what) {
    VizStr s;
    char colorFormat[] = "\033[0;30m%s\033[0;39m";
    colorFormat[2] += color >>3;
    colorFormat[5] += color & 7;
    int l = s.printf(colorCapable ? colorFormat : "%s", what);
    char *p = s.close();
    Locker x(this);
    return ::write(2,p,l);
  }
  UserInterface();
  virtual ~UserInterface() {}
#define MAXUIREAD 1024
  virtual const char* ask (const char* prompt) {
    VizStr s("\n\n * INPUT AWAITED * \n\n %s", prompt);
    Locker x(this);
    write(Yellow, s.close());
    char *p = s.open();
    int l = ::read(0, p, MAXUIREAD);
    assert (l >= 0);
    p[l] = 0;
    return s.close(MAXUIREAD);
  }
};

externNoImpl UserInterface* ui;
#ifdef IMPL
UserInterface defaultUI; 
UserInterface::UserInterface() { // Should work for any superclass
  colorCapable = isatty(1) || getenv("TTY_HAS_COLORS"); 
  ui = this;
}
#endif //IMPL

#define SUPERSIZE 128

#ifdef IMPL
int myRank = 0, superSize = SUPERSIZE, mainRank = 0,realsuperSize=1; // Default values 
#else
extern int myRank, superSize, mainRank,realsuperSize;
#endif //IMPL

  // task -> window
  // select what to print by command line (by regexp!)
		 

#ifndef VPP
#define VPP(lvl,msg...) do { {VizStr msgstr(msg); Viz::vpp(lvl,msgstr.close(),TS_MODULE,__PRETTY_FUNCTION__,__FUNCTION__);} VizStr::rewind(); } while (0)

#endif //VPP

#define uiprintf(msg...)  do { { VizStr msgstr(msg); ui->write(AUTOCOLOR,  msgstr.close()); } VizStr::rewind(); } while(0)
#define errprintf(msg...) do { { VizStr msgstr(msg); ui->write(bright(Red),msgstr.close()); } VizStr::rewind(); } while(0)
#define fatal(msg...)     do { errprintf("\n\n[%d]: ", myRank); errprintf(msg); errprintf("\n\n"); exit(1); } while(0)

class Viz { //FIXME: separate debug & viz ???
public:
  class Vid {
  public:
    char text[12];
    Vid() { text[0] = 0; }
    void operator= (const char *_text) { *this = *(Vid*)_text; }
  } vid;

  virtual const char* viz()  { 
    VizStr v("%s%s%.*s @%p", CLASSID(*this), vid.text[0] ? "/":"", sizeof(vid),vid.text, this);
    return v.close(); 
  }

  static regex_t rx; // regular expression from command line
  static bool rx_exist; // exist flag for rx
  static void vpp( int , const char * msg, const char *module, const char * pretty_function, const char * function)
  {
    VizStr hidstr("module=%s,method=%s",module,pretty_function);  
    
    if (!rx_exist ||
        (  
	 (0==regexec(&rx,hidstr.close(),0,NULL,0)) || 
	 (0==regexec(&rx,msg,0,NULL,0))
	) 
       )
    { 
      VizStr out ("[%d|%s|%s>   %s\n",myRank,module,function,msg);
      ui->write(AUTOCOLOR,out.close()); 
    };    
  };
};

#ifdef IMPL
bool Viz::rx_exist=false;
regex_t Viz::rx;
#endif //IMPL


  //#define RINGDEBUG 


class Ring {
  Ring *prev,*next;
public:
  inline void ok() {
#ifdef RINGDEBUG 
    Ring *n = next;
    do {
      assert(n->next->prev == n && n->prev->next == n);
      n = n->next;
    } while (n != this);
#endif //RINGDEBUG
  }
  inline void     ct()  { prev = next = this; }
  inline Ring() { ct(); }

  inline bool ringEmpty()  { return next == this; }
  inline Ring* first() { return ringEmpty() ? NULL : next; }
  inline Ring* last()  { return ringEmpty() ? NULL : prev; }

  inline Ring* popFront() { 
    Ring* e = ringEmpty() ? NULL : next->remove(); 
    ok();
    return e;
  }
  inline Ring* getFront()  { 
    Ring* e = ringEmpty() ? NULL : next; 
    ok();
    return e; 
  }    
  inline Ring* popBack()  { 
    Ring* e = ringEmpty() ? NULL : prev->remove(); 
    ok();
    return e; 
  }    
  inline Ring* getBack()  { 
    Ring* e = ringEmpty() ? NULL : prev; 
    ok();
    return e; 
  }    
  inline void pushFront(Ring& t) {
    if (!t.ringEmpty())
      t.remove();
#ifdef RINGDEBUG 
    assert(t.ringEmpty());
#endif //RINGDEBUG
    t.prev = this; t.next = next;
    next->prev = &t; next = &t; 
    ok();
  }
  inline void pushBack(Ring& t) {
    if (!t.ringEmpty())
      t.remove();
#ifdef RINGDEBUG 
    assert(t.ringEmpty());
#endif //RINGDEBUG
    t.next = this; t.prev = prev;
    prev->next = &t; prev = &t; 
    ok();
  }
  Ring* remove() { 
#ifdef RINGDEBUG 
    assert(!ringEmpty());
#endif //RINGDEBUG
    prev->next = next; 
    next->prev = prev; 
    next = prev = this;
    ok();
    return this;
  }
  ~Ring() { if (!ringEmpty()) remove(); }
};



// Construction = priority order

#define CTRL_SERVICE_PRIO 0
#define DATA_SERVICE_PRIO 1  
#define SCDL_SERVICE_PRIO 2

#define MAXSERVICE 3

class Service {
  int index;

public:
  static Service* service[MAXSERVICE];
  static bool exitFlag;
  static int exitError; 

  Service(const char *name,int prio) : index(prio) { 
    assert(service[index] == NULL);
    service[index] = this; //v! "registering service %s at index %d", name,index
  }
  virtual ~Service() { service[index] = NULL; }
  static void exitServices(int error) { 
    exitError=error;
    exitFlag =true; // no one should call exit(int) from the service code
  }; 
  static void doAllWork() { 
    while (!exitFlag) { 
      if (realsuperSize > 1) sched_yield();
      int i;
      for (i = 0; i < MAXSERVICE; i++) if (service[i] && service[i]->work()) break; 
    }
  }
  virtual bool work() = 0; // returns `true' when something done
};

#ifdef IMPL
Service* Service::service[MAXSERVICE]; // Zero-init
bool Service::exitFlag  = false; 
int  Service::exitError = 0;
#endif //IMPL


class Delay { 
public:
  Delay *next;
  Delay():next(NULL) {}
#define FORCED ((Delay*)-1)
  void operator()(Delay*& p) { 
    if (p == FORCED) delete this; 
    else { next = p; p = this; } 
  }
  static void force(Delay*& me) {
    if (me == NULL || me == FORCED) return;
    force(me->next); delete me; me = FORCED;
  }
#undef FORCED
  virtual ~Delay() {} // Do real work here
};


class FinalizeException {};

// Config Space constants
class SContext {
public:
  int nodePrio; 
  double heartbeatTimeout; // seconds (MPI_Wtime units)
  double heartbeatInterval;
  bool showMsgs;
  bool generateCallGraph;
  bool showFeatures;
  SContext(): nodePrio(0),heartbeatTimeout(300),heartbeatInterval(15),
	      showMsgs(false),generateCallGraph(false),showFeatures(false)  
  {};
};

class SchedContext {
public://FIXME
};

// Config Space attributes //FIXME: move to TFun
class TContext : public SContext, public SchedContext {
public:
  int prio;
  int mflops;
  int mbytes;
  int pflags;
  int fflags;
  int atRank;
  double cpuUsage;
  int extraSize;
  TContext() : prio(0), mflops(-10), mbytes(-10), pflags(0), fflags(0), atRank(-1), cpuUsage(1.), extraSize(0) {}
};
externNoImpl TContext tc;

// TCT operands

#define __GLUE(a,b)      a##b 
#define _GLUE(a,b) __GLUE(a,b)
#define _TRIGGER _GLUE(_trigger,__LINE__)

#define _tc_cpuUsage(n) ts::cpuUsageTrigger _TRIGGER (n)
class cpuUsageTrigger {
  double savedCpuUsage;
 public:
  cpuUsageTrigger(double n) : savedCpuUsage(tc.cpuUsage) { tc.cpuUsage = n; }
  ~cpuUsageTrigger()                              { tc.cpuUsage = savedCpuUsage; }
};

#define _tc_atRank(n) ts::atRankTrigger _TRIGGER (n)
class atRankTrigger {
  int savedAtRank;
 public:
  atRankTrigger(int n) : savedAtRank(tc.atRank) { tc.atRank = n; }
  ~atRankTrigger()                              { tc.atRank = savedAtRank; }
};

#define _tc_extraSize(n) ts::extraSizeTrigger _TRIGGER (n)
class extraSizeTrigger {
  int savedExtraSize;
 public:
  extraSizeTrigger(int n) : savedExtraSize(tc.extraSize) { tc.extraSize = n; }
  ~extraSizeTrigger()                                    { tc.extraSize = savedExtraSize; }
};

#define _tc_priority(n) ts::priorityTrigger _TRIGGER (n)
class priorityTrigger {
  int savedPriority;
 public:
  priorityTrigger(int n) : savedPriority(tc.prio) { tc.prio = n; }
  ~priorityTrigger()                              { tc.prio = savedPriority; }
};

#define _tc_setLabel(l...) do { if (ts::Task::getCurrent()) ts::Task::getCurrent()->setLabel(l); } while(0)

double wTime();
unsigned long long tickTime();
double tick2Sec(unsigned long long);
class Prof {
public:
  long long total;
  long long startAt;

  void reset() {
    total = 0;
  }
  inline Prof() : total(0),startAt(0) {
    //v! "In Prof constructor"
  }
  void start()   { startAt = tickTime(); }
  long long stop()  { 
    long long timeLeft; 
    if (!startAt) {
      timeLeft = 0;
    } else {
      timeLeft = tickTime() - startAt; 
      total += timeLeft;
    }
    startAt = 0;
    return timeLeft; 
  }
  
  virtual const char *codeId() { return "<unnamed>"; }

  class Timer {
  public:
    Prof *p;
    Timer(Prof& _p) : p(&_p) { p->start(); }
    long finish() { return p->stop(); }
    ~Timer() { finish(); } 
  };
};
class Task;
class TaskStatistic : public Prof { //FIXME: make it private
public:
  long msg_count; // messages sent/received by task
  long msg_size; // total amount of data in aforesaid messages
  long event_count; 
  static long long total_time;
private:
  static TaskStatistic* current;
public:
  static inline TaskStatistic* getCurrent() {return current;}
  static inline void setCurrent(TaskStatistic* stat) {
    //uiprintf("Switch: %p -> %p\n", current, stat);
    if (current) current->stop();
    current = stat;
    if (current) current->start();
  }

  TaskStatistic() : msg_count(0),msg_size(0),event_count(0) {}

  virtual ~TaskStatistic() {
    stop();
    total_time += total;
  }

  class Holder {
    TaskStatistic* saved_current;
  public:
    Holder(TaskStatistic *t) : saved_current(getCurrent()) {
      setCurrent(t);
    }
    ~Holder() {
      setCurrent(saved_current);
    }
  };
};
class Task;
extern Task mpiTask;
#ifdef IMPL
long long TaskStatistic::total_time = 0;
TaskStatistic* TaskStatistic::current = NULL;
#endif

// Local statistic  

struct Statistic {
public:
  bool alive;
  long long tasksActivated;
  long long tasksExported;
  long long msgsSent;
  double total_time;
  double tasks_time;
  double idle_time;
  double sched_time;
  double mpi_time;
  int visit_count;
  
  Statistic() : alive(true), tasksActivated(0), tasksExported(0), msgsSent(0), mpi_time(0.) {}
};
externNoImpl Statistic *myStat; // Allocated in MoTOR
#define TCT_OPTION "-tct" 


class CmdLine {
public:
  static int    *mainArgc;
  static char** *mainArgv;
  static Delay  *delay;
  CmdLine(int* argc, char** *argv) {
    mainArgc = argc; 
    mainArgv = argv;
    Delay::force(delay);
    ProcessCmdLine();
  }
  void ProcessCmdLine()
  {
    int i;
    int tct_count=0; 
    bool tct_prev=false;

    //    fprintf (stderr,"TCT Command line processing\n");

    for  (i=1; i<*mainArgc ; i++)
    {
      //fprintf (stderr,"TCT processing loop i %d, tct_count %d, tct_prev %d, %s \n", i, tct_count, tct_prev,(*mainArgv)[i] );
      if (tct_prev) {
	ProcessTCTOption((*mainArgv)[i]);
	tct_prev=false;
	tct_count++;
      } else if ( strcmp(TCT_OPTION, (*mainArgv)[i] )==0)   { // process cmd line option
       tct_count++;
       tct_prev=true;
      } else  {
	(*mainArgv)[i-tct_count]=(*mainArgv)[i];
      }
    }
    *mainArgc -= tct_count;
  }
  void ProcessTCTOption(char *option) {  // FIXME Processing only DP(" regexp ") option
    int opt_len;

    opt_len=strlen(option);
    assert (opt_len<512);
    if ((strncmp("DP(\"",option,4)==0) && (strncmp("\")",option+opt_len-2,2)==0)) {
      char regex[1024]={0,};     
      memcpy(regex,option+4,opt_len-6);
      //      fprintf (stderr,"Regex=%s\n",regex);
      if (regcomp (&Viz::rx, regex, REG_EXTENDED)) {
              fatal("Regexp `%s' compilation failed!\n", regex);
      }
      fprintf(stderr, "Debug Pattern was set to %s\n", regex);
      Viz::rx_exist=true;
    } else if (strcmp(option,"callGraph")==0) {
      tc.generateCallGraph = true;
    } else if (strcmp(option,"showMsgs")==0) {
      tc.showMsgs = true;
    } else if (strcmp(option,"showFeatures")==0) {
      tc.showFeatures = true;
    } else { 
      fatal ("Unknown TCT option %s !\n", option);
    }
  }
};
 

#ifdef IMPL
int    *CmdLine::mainArgc = NULL;
char** *CmdLine::mainArgv = NULL;
Delay  *CmdLine::delay    = NULL;
#endif


// Features API
typedef void* (*entry_t)(int cmd, void *arg); // initialize: entry(0,char*) -> char* status
extern "C" {
void registerFeature(const char *name, const char *comment, entry_t entry); // `entry' can be NULL
entry_t featureEntry(const char *name);
}

// Features IMPL
class Feature {
  Feature *next;
  entry_t entry;
  const char *name;
  const char *comment;
  const char *status;
  bool internal;
  static Feature *all;
public:
  Feature(const char *_name, const char *_comment, bool _internal = true, entry_t _entry = NULL) : 
    next(all), entry(_entry), name(_name), comment(_comment), status(NULL), internal(_internal) { 
    all = this; 
  }
  inline const char* error() { return (status && !strncmp(status,"ERROR",5)) ? status+5 : NULL; }
  static entry_t getEntry(const char *name) {
    Feature *f;
    for (f = all; f; f = f->next) if (!f->error() && !strcmp(f->name,name)) return f->entry;
    return NULL;
  }
  void doInitialize() {
    if (verbose()) uiprintf("  %s (%s)", name, comment); 
    if (entry)   status = (const char*)(*entry)(0, NULL); // Initialize feature  FIXME: pass cmdline argument
    // Print out status
    if (status && verbose()) {
      if (error()) errprintf(": %s [disabled]", error()); 
      else          uiprintf(": %s", status);
    }
    if (verbose()) uiprintf("\n");
    if (internal && error()) fatal("Internal feature failed\n");
  }
  static void initialize(const char *msg, bool internal) {
    if (verbose()) uiprintf("%s\n", msg);
    Feature *f;
    for (f = all; f; f = f->next) if (f->internal == internal) f->doInitialize();
  }
  static void initialize() {
    initialize("Internal features: ", true);
    initialize("External features: ", false);
  }
  static bool verbose() { return tc.showFeatures; }
};

#ifdef IMPL
Feature* Feature::all = NULL;
extern "C" {
void registerFeature(const char *name, const char *comment, entry_t entry) {
  new Feature(name, comment, false, entry);
}
entry_t featureEntry(const char *name) {
  return Feature::getEntry(name);
}
}
#endif


  // Include superstructure levels
#undef  TS_MODULE
#define TS_MODULE "super"
#include "super.v.hh"
#undef  TS_MODULE
#define TS_MODULE "motor"
#include "motor.v.hh"
#undef  TS_MODULE
#define TS_MODULE "tss"
#include "tss.v.hh"
#undef  TS_MODULE
#define TS_MODULE "trt"
#include "txx"

#ifndef TCT 
#define TCT "stack(64)"
#endif  //TCT

#ifdef IMPL
void sleep(double sec) { //ts::sleep 
  assert(Task::getRealCurrent());
  Task *saved_current = Task::getRealCurrent();
  Task::setCurrent(NULL); // sanity
  ThrH::sleep(sec);
  Task::setCurrent(saved_current); 
}
#else
void sleep(double sec=0.);
#endif

class DefaultExceptionHandler {
public:
  static void unexpectedException() {
    fprintf(stderr, "%s: may be some 'try {} catch()' missed?\n", __FUNCTION__);
    fprintf(stderr, "%s: terminating program execution\n", __FUNCTION__);
    exit(1); //FIXME?
  }
  void (*savedHandler)();
  DefaultExceptionHandler() {
    savedHandler = std::set_unexpected(unexpectedException);
  }
  ~DefaultExceptionHandler() {
    std::set_unexpected(savedHandler);
  }
};

class TRT : public CmdLine, DefaultExceptionHandler { 
public:
  static TRT *trt;
  MacroScheduler *ms;

  TRT(int* argc, char *** argv) : CmdLine(argc,argv),
    ms(MacroScheduler::New())
  {
    assert (!trt);   
    trt = this ; 
    atexit(atExit);
    Task::setCurrent(&rootTask);
    Feature::initialize(); //v! TRT initialized
    trt->ms->exchangeByResources();
  }
  static void atExit() {
    //v! atExit
    if (!Service::exitFlag) { // If unexpected exit only!
      trt->ms->tb.relaxData() = 1;
      trt->ms->tb.relaxCell().write(0);
      trt->ms->setStatistic(); // FIXME: to late!
      //v! "%s: in hope to terminate TRT on all nodes", __PRETTY_FUNCTION__
    }
  }
  ~TRT() {
    if (tseg().inUse) uiprintf("rank%d: %d supercells are locked\n", myRank, tseg().inUse);
    delete ms;
    trt = NULL;
  }
};
#ifdef IMPL
TRT *TRT::trt = NULL;
#endif

class MainTRT : Prof {
public:
public:
  Timer *timer;
  MainTRT() {
    uiprintf("Open T-System Runtime v1.9, 2003, PSI RAS, Russia\n");
    uiprintf("Running under %s MPI on %d-rank %scluster:\n  ", 
	     mpi_portal->name, realsuperSize, dmpi_meta ? "meta":"");
    TRT::trt->ms->printClusterResources();

    //v! Compiled at %time @ node

    //v! "Supercell size is %d bytes", sizeof(TObj<int>)
    timer = new Timer(*this);
    uiprintf("Starting tfun main, good luck!\n\n");
    //v! 
  }
  virtual ~MainTRT() {
    //v!
    TRT::trt->ms->printStatistic();

    //v! "dd supercells used"

    //v! Top of CPUsage records:
    //v! ...

    //v! KPD = ...
  }
};

  // FIXME: inheritance!

// System configuration

enum _sc_name { 
  _SC_START = 0x1000,
  // Cluster configuration
  _SC_CL_RUN_DATE,      // Program start date
  _SC_CL_CPU_NAME,      // CPU achitecture
  _SC_CL_CPU_FLOPS,     // CPU performance
  _SC_NPROCESSORS_CONF, // CPUs per this node
  _SC_CL_SIZE,          // Processes at cluster
  _SC_CL_CLUSTER_FLOPS, // Cluster peak power
  _SC_CL_MPI_RANK,      // My MPI rank
  _SC_CL_MPI_IMPL,      // MPI system kind
  _SC_CL_MPI_OPTIONS,   // MPI options
  _SC_CL_EMULATION,     // Cluster emulation
  _SC_CL_NODE_NAME,     // Node name asciiz
  _SC_CL_OS_VERSION,    // OS name & version
  _SC_CL_L1_CACHE_SIZE, // L1 cache size
  _SC_CL_L2_CACHE_SIZE, // L2 cache size
  // Compile-time configuration
  _SC_CT_GCC_VERSION,   // GCC compiler version
  _SC_CT_OPTIONS,       // Used options
  _SC_CT_DEBUG,         // Debug mode
  _SC_CT_CHECK,         // Do checks
  _SC_CT_THREAD_IMPL,   // Used thread package   
  // Run-time configuration
  _SC_RT_TASK_BOARD,    // My task board rank
  _SC_RT_CELLS,         // Cell matrix dimension
  _SC_RT_FEATURES,      // Features asciiz
  _SC_RT_THREAD_POOL,   // Thread pool size
  _SC_RT_MEMORY_POOL,   // Memory pool size     
  _SC_RT_PARALLEL,      // Parallel mode 
  _SC_RT_PROFILE,       // Profiling mode 
  _SC_RT_TRACE,         // Tracing mode 
  _SC_RT_DEBUG,         // Debug mode 
  // Environment attributes
  _SC_ENV_STACK_SIZE,    // Default stack size 
  _SC_ENV_THREAD_PRIO,   // Default thread priority 
  _SC_ENV_PTR_MODE,      // Pointer creation mode
  _SC_ENV_MEASURE,       // Time measure mode
  _SC_ENV_VERBOSE,       // Verbose execution
  // Summary information 
  _SC_SUMMARY           // All above asciiz
};

  //  conf_chain;

long sysconf(int name);

#ifdef IMPL
long _sc[_SC_SUMMARY-_SC_START-1]; 

long sysconf(int name) {  
  if (name >= _SC_START && name < _SC_SUMMARY) return _sc[name];
  if (name <  _SC_START || name > _SC_SUMMARY) return -1;
  // Produce sysconf summary
  static char summary[0x4000]; 
  char *p = summary;//FIXME: use String
  p += sprintf(p,"FIXME!\n");
  return (long)summary;
}


// Static TVars as shared array
TFrz<int> mainResult;
bool mainResultReady() {
  return mainResult().isAbsolutelyReady();
}
#else
extern 
TFrz<int> mainResult;
#endif //IMPL
}

TFUNDEF(int, tfunmain, (int argc, char *argv[]), /*static*/, int argc; char **argv;, _tf->argc=argc; _tf->argv=argv ;, , , 0)

int main(int argc, char *argv[]) __attribute__ ((weak));
#ifdef IMPL
int main(int argc, char *argv[]) {
  if (getenv("TSLEEP"))
    ::sleep(10);
  int result = 0;
  {
    ts::TRT trt(&argc,&argv);
    if (ts::myRank == ts::mainRank) {
      callGraphPrintf("digraph callgraph {\n");
      ts::MainTRT maintrt;
      {
	      tct(atRank(0));
	      ts::mainResult = tfunmain(argc,argv);
      }
      ts::Service::doAllWork();
      callGraphPrintf("};\n");
      result = ts::mainResult;
      ts::mainResult.reset();
    } else {
      ts::Service::doAllWork();
    }
    // Destroy trt before MPI_Finalize
  }
  MPI_Finalize();
  ts::mpiInitialized = false;
  return result;
}
#endif //IMPL

#endif // __TRT__
