//+
//  Open T-Superstructure for C++
//
//  See article: "New words for old song"
//  S.M.Abramov,A.I.Adamovich,V.A.Roganov
//-

// all static - to tmain!

// TDsc (TObj descriptor)
// Hot/Cold segment (to aviod twice TObj)

template <typename type> 
class TObj;

externNoImpl int freezeMode; // initially zero
class Freezer {
public:
  inline Freezer()  { 
    assert(freezeMode >= 0);
    freezeMode++;   //v! Freezing ++
  }
  inline ~Freezer() { 
    freezeMode--;   //v! Freezing --
    assert(freezeMode >= 0);
  }
};

#define HASDESCR 1
template <typename type>
class TDsc : public MRef<type> {
public:
  inline TDsc() { SData::ctFlags |= HASDESCR; }

  inline TDsc(const TDsc<type>& d) {
    SData::ctFlags |= HASDESCR;
    (*this) = &((TDsc<type>&) d)();
  }

  inline TDsc<type>& operator= (TDsc<type> d) {
    (*this) = &d();
    return *this;
  }

  inline TDsc(TObj<type>*);
  inline TObj<type>* operator= (TObj<type>*);
  inline TObj<type>& operator()() { 
    //v! "=%p", &(TObj<type>&)**this
    return (TObj<type>&)**this; 
  }
  inline TObj<type>* getTObjPtr() { return (TObj<type> *)getMObjPtr(); }

  //????inline operator bool() { return cellOffset != 0; }
  inline bool notNull() { return cellOffset != 0; }

  inline void drop();
  virtual ~TDsc() { } // Do not drop here!
};

typedef int AnyType;

class MDataNonReady;

template <typename type>
class TObj : public MObj<type>, public Ring {
public:
  //Not MRef - no RC!
  TObj<type>* dropTo; // FIXME: dsize - for MPI2 ///FIXME: translate dropTo to remote nodes (use invariant cellOffset)
  int dropSeqNo;
  TDsc<type> asnFrom; // != NULL only for awaiting values; always cold
  bool dropAccepted;

  inline TObj() : dropTo(NULL),asnFrom(NULL),dropAccepted(false) {}
  inline TObj(const TObj<type>& t) : dropTo(NULL),asnFrom(NULL) { *this = (TObj<type>&)t; }

  // Semistrong binary operators
  //inline type operator+ (type t) { return (*this)()() + t; }
  //inline type operator- (type t) { return (*this)()() - t; }
  //inline type operator* (type t) { return (*this)()() * t; }
  //inline type operator/ (type t) { return (*this)()() / t; }

  inline bool isBound() { return (&data()!=NULL) || asnFrom.notNull(); }
  inline void makeUnbound() { *this = (MData<type>*)NULL; assignFrom(NULL); }

  inline void assignFrom(TObj<type>* t) {  // should be called only by owner for hot data ONLY!
    assert(cellIsOk() && (!t || t->cellIsOk()) && t != this);
    if (&asnFrom() == t) return; 
    if (asnFrom.notNull()) { assert(!ringEmpty()); remove(); asnFrom = NULL; } 
    if (t) { 
      *this = (MData<type>*)NULL; // Reset ready state
      assert(t->owner && (!Task::getRealCurrent() || t->owner != Task::getCurrent()->taskId));
      asnFrom = t; t->pushFront(*this); //v! "assignFrom %p = %p", this, t 
    }
  }

  // Assignment to MData 
  inline MData<type>* operator= (MData<type> *data) {
    *(SCell<dataHandler>*)this = data; //v! "assign TObj to MData" ! "assigning TObj to MData done"
    return data;
  }

  inline operator type& () { return (*this)()(); }

      //FIXME: onRead() -> asnFrom to force lazy function to run

  inline MData<type>& operator()() { // It does COW!
    while (asnFrom.notNull()) waiting();
    return (*(MObj<type>*)this)(); 
  }

  inline bool isAbsolutelyReady() { //FIMXE?
    bool result = !asnFrom.notNull() && !owner; //v! "result for %p = %d\n", this, result
    return result;
  }

  inline TObj<type>& getMyDrop() { // should be called only by owner for hot data ONLY!
    assert(cellIsOk());
    assert(Task::getRealCurrent() && owner == Task::getRealCurrent()->taskId);
    if (!dropTo) { 
      dropTo = new TObj<type>; //v! "dropTo is TObj %p (for %p)", dropTo, this
      dropTo->noDRCDelete = 1;
      dropSeqNo = dropTo->getSeqNo();
      // FIXME: vita is unimplemented
      // dropTo->vita = ((MThread*)Task::current)->vita; // Current thread vita condition
    }
    return *dropTo;
  }

  inline TDsc<type> ownerCopy() {  // Should be called only by owner!
    if (this == NULL) return NULL; // ??????
    assert(cellIsOk());
    assert(Task::getRealCurrent());
    if (asnFrom.notNull()) return &asnFrom();
    if (!owner || owner != Task::getRealCurrent()->taskId) return this;
    if (&data()!=NULL) { 
      TObj<type>* t = new TObj<type>; 
      *t = &ownerCopyData(); 

      //t->unlock(); 
      // ->
      TDsc<type> d = t; // GC!
      t->onWrite(); //FIXME?

      return d; 
    } else {
      return &getMyDrop(); 
    }
  }


  // Assignment to value
  inline TObj<type>& operator= (const type& v) { //FIXME: use reference to avoid copying?
    assert(cellIsOk());
    assignFrom(NULL);
    (*this)()() = v; //v! "Assignment to ready value for %p", this
    ///onWrite();      //FIXME: call onWrite for cold values only 
    //v  onWrite
    return *this;
  }

  inline TObj<type>& operator= (TObj<type>& t) {
    assert(cellIsOk());
    assert(!Task::getRealCurrent() || owner == Task::getCurrent()->taskId);
    if (this != &t) {
      if (t.asnFrom.notNull()) {
	assignFrom(&t.asnFrom());
      } else {        // We allow assignment to locked w/o current, it means assignFrom!
	if (!t.owner || (Task::getRealCurrent() && t.owner == Task::getCurrent()->taskId)) {      //v! "%p = to ready value %p", this, &t
	  //COW
	  ///FIXME read and wait for a proper seqNo if slave; slave  cell should be like non-ready value,
          /// freezing accessor until data arrived from the master node. seqNo is instead of owner for a slave. 
   
	  *this = (MData<type> *)&t.data();//v! "assign TObj data by COW" ! "assigning TObj data by COW done"
	  assignFrom(NULL);
	  //onWrite();
	} else {             //v! "%p = to non-ready value %p", this, &t
	  assignFrom(&t);
	}
      }
    }
    return *this;
  }

  // Drop the value //FIXME: should stop lazy function
  void drop(bool doUnbound = true) {
    assert(cellIsOk()); //v! "drop cell %p, dropTo = %p", this, dropTo
    if (Task::getRealCurrent())
    if (dropTo && dropTo->ctDone && dropTo->getSeqNo() == dropSeqNo) {

      //uiprintf("%s:%d for %p\n", __FUNCTION__, __LINE__, this);

      if (!asnFrom.notNull()) {
	*dropTo = &ownerCopyData();
      } else {
        MData<AnyType> *nonReady = new (MData<AnyType>::halloc(sizeof(MDataNonReady))) MDataNonReady((TObj<AnyType>*)&asnFrom());
	*(TObj<AnyType>*)dropTo = nonReady;
      }
      dropTo->dropAccepted = true;
      dropTo->onWrite(); //v! dropTo->onWrite()
    }
    dropTo = NULL; 
    if (doUnbound) makeUnbound();
  }

  virtual void onSlvRetrieve() { // On slave (local) cell finish we don't reset data!
    drop(false);
  }

  inline MData<type>& ownerCopyData() {
    assert(cellIsOk());
    assert((Task::getRealCurrent() && owner == Task::getRealCurrent()->taskId)|| ( *(int*)0=0 ) );
    assert(!asnFrom.notNull());
    MData<type> *my = &(*this)();
    if (!(my->flags & HASDESCR)) 
      return *my;
    else {
      Freezer frz;
      MData<type>* d = (MData<type>*)(my->clone()); //FIXME:  Why we need do clone here ?
      d->flags &= ~HASDESCR;  // Already frozen value - no problem for the further
      return *d;
    }
  }

  void clearDropTo()
  {
    if (myRank == master()) {
      MRef<AnyType> cleaner(this);
    }
  }

  virtual ~TObj() { 
    assert(cellIsOk());
    drop(); //v! "destructor for %p", this ! done
  }

  // Allocation service
#define TCELLSPERNODE ((ZEROMAPSIZE/(sizeof(TObj<AnyType>)*SUPERSIZE)) - 0x10) // Reserve some space for ControlSegment
  void* operator new (size_t size)   { 
    extern SSegment< TObj<AnyType>, dataHandler,  TCELLSPERNODE >& tseg();
    return tseg().allocChunk(size/sizeof(TObj<AnyType>)); 
  }
  void* operator new[] (size_t size) { 
    extern SSegment< TObj<AnyType>, dataHandler, TCELLSPERNODE >& tseg();
    return tseg().allocChunk(size/sizeof(TObj<AnyType>)); 
  }    
  void operator delete (void* mem, size_t size)   { 
    extern SSegment< TObj<AnyType>, dataHandler, TCELLSPERNODE >& tseg();
    TObj<AnyType>*m = (TObj<AnyType>*)mem;
    assert(m->master() == myRank);
    tseg().freeChunk(m,size/sizeof(TObj<AnyType>));
  }
  void operator delete[] (void* mem, size_t size) { 
    extern SSegment< TObj<AnyType>, dataHandler,  TCELLSPERNODE >& tseg();
    TObj<AnyType>*m = (TObj<AnyType>*)mem;
    assert(m->master() == myRank);
    tseg().freeChunk(m,size/sizeof(TObj<AnyType>)); 
  }  
  // for inPlaceCtor
  void* operator new (size_t size, void *mem) { return mem; }
  void clearCell() {
    dropTo = 0;
    dropSeqNo = 0;
    dropAccepted = false;
    assignFrom(0);
    owner = 0;
    extraSize = 0;
    ((DRCObj *)this)->~DRCObj();
  }
};

// Used to pass non-Ready values across network
class MDataNonReady : public MData<AnyType> {
public:
  MRef<AnyType> nonReady;
  long long owner; 
  MDataNonReady(TObj<AnyType> *to) { 
    nonReady = to; 
    assert(to); 
    owner = to->owner;
    assert(owner); 
    //isReady();//FIXME
  }
  virtual bool isReady () { 
    return false; 
  }
};

#ifdef IMPL
//FIXME: add onReadMethod
//FIXME: add non-ready value passing throw network


void assignOnWrite(TObj<AnyType>* src, TObj<AnyType>* dst) {
  MData<AnyType> *d = (MData<AnyType> *) &src->data();

  //  uiprintf("assignOnWrite src = %p(data = %p), dst = %p\n", src, d, dst);

  MDataNonReady* nrd = d->isReady() ? NULL : (MDataNonReady *)d;
  if (nrd) {
    TObj<AnyType>* nonReady = (TObj<AnyType>*) nrd->nonReady.getMObjPtr();
    assert(nonReady != src && nonReady != dst);
    nonReady->owner = nrd->owner;
    *dst = *nonReady;
  } else {
    src->owner = 0;

    // uiprintf("src = %p(data = %p), dst = %p(data = %p)\n", src, &src->data(), dst, &dst->data());
    *dst = *src;
    //  uiprintf("src = %p(data = %p), dst = %p(data = %p)\n", src, &src->data(), dst, &dst->data());

    if (dst->dropAccepted) dst->owner = 0; 

  }
  dst->contTasks();
}

void tssOnWrite (Cell *c) {
  TObj<AnyType> *t = (TObj<AnyType>*)c;
  //v! "enter tssOnWrite by write access for cell %p", t

  assert(!t->asnFrom.notNull()); // Must have real on NotReady MData bound!
  {
    Task *savedCurrent = Task::getCurrent(); 
    Task::setCurrent(NULL); 
  
    if (!t->ringEmpty()) { // FIXME: To avoid MRef takeWE...!!!
      MRef<AnyType> hold(t); // dropTo can be autocleaned!
      t->noDRCDelete = 0;
      while (!t->ringEmpty()) assignOnWrite(t, (TObj<AnyType> *)t->getFront());
      Task::setCurrent(savedCurrent);
    } else {
      t->noDRCDelete = 0;
      assignOnWrite(t,t);
      Task::setCurrent(savedCurrent);
      t->clearDropTo();
    }

  }

  //v! "done"
}
class InitDataHandler {
public:
  static void inPlaceCtor(Cell *c) { new (c) TObj<AnyType>(); }
  static void clearCell(Cell *c) { ((TObj<AnyType> *)c)->clearCell(); }
  InitDataHandler() { 
    dataHandler.onWrite     = tssOnWrite; 
    dataHandler.inPlaceCtor = inPlaceCtor;
    dataHandler.clearCell   = clearCell;
  }
} initDataHandler;
#endif

#define tObjAt(offs) (*(TObj<type>*)&TObj<type>::cellAt(offs))

template <typename type>
inline TDsc<type>::TDsc<type>(TObj<type>* t) { 
  *(MRef<type>*)this = t; //v "new TDsc -> %p", t
}

template <typename type>
inline TObj<type>* TDsc<type>::operator= (TObj<type>* t) { *(MRef<type>*)this = t; return t; }

template <typename type>
inline void TDsc<type>::drop() { if (cellOffset) tObjAt(cellOffset).drop(); }

externNoImpl SSegment< TObj<AnyType>,dataHandler,TCELLSPERNODE > _tseg;
#ifdef IMPL
SSegment< TObj<AnyType>,dataHandler,TCELLSPERNODE >& tseg() { return _tseg; }
#else
extern SSegment< TObj<AnyType>,dataHandler,TCELLSPERNODE >& tseg();
#endif

// FIXME: vita is unimplemented
/*
inline MObj<int>* newVita() { 
  TObj<AnyType>* vita = new TObj<AnyType>; //v! "new Vita TObj %p", vita
  return vita; 
} */

//FIXME: add check for null dsc dereference (and use another way than &* to take address :-)

// Intermediate gate between inner and outher TFunction's worlds
template <typename type>
class TFrz {
  TDsc<type> d;
public:
  TObj<type>&  operator()() { return d(); }
  operator TObj<type>& ()   { return d(); }
  operator type&       ()   { return d()()(); } 
  inline TFrz() {};

  //inline TFrz(const type v) : d(new TObj<type>) { Freezer frz; d()()() = v; d().onWrite(); } // For TFrz = CVal 

  // FIXME: Uses stack!
  inline TFrz(type v) : d(new TObj<type>) { Freezer frz; d()()() = v; d().onWrite(); } // For TFrz = CVal, no stack pass 

  inline TFrz(TObj<type>* t) : d(t->ownerCopy())                      {} // For TFrz = &TVar
  inline TFrz(TObj<type>& t) : d(t.ownerCopy())                       {} // For TFrz = TObj
  inline TFrz(TDsc<type>& t) : d(t().ownerCopy())                     {} // For TFrz = TVar,TPtr
  inline TFrz(const TFrz<type>& t) : d(&((TFrz<type>&)t)()) {}
  inline TFrz<type>& operator = (TFrz<type> t) { d = &t(); return *this; }


  inline void reset() { d = NULL; } //FIXME!

  // Semistrong binary operators
  //inline type operator+ (TObj<type>& t) { TDsc<type> md(to), td(&t); uiprintf(" + "); return (*to)()() + t()(); }
  //inline type operator- (TObj<type>& t) { TDsc<type> md(to), td(&t); return (*to)()() - t()(); }
  //inline type operator* (TObj<type>& t) { TDsc<type> md(to), td(&t); return (*to)()() * t()(); }
  //inline type operator/ (TObj<type>& t) { TDsc<type> md(to), td(&t); return (*to)()() / t()(); }

  // Strong binary operators
  //inline type operator+ (type& t) { TDsc<type> md(to); return (*to)()() + t; }
  //inline type operator- (type& t) { TDsc<type> md(to); return (*to)()() - t; }
  //inline type operator* (type& t) { TDsc<type> md(to); return (*to)()() * t; }
  //inline type operator/ (type& t) { TDsc<type> md(to); return (*to)()() / t; }
};

//FIXME: constant operator

template <typename type>
class TPtr : public TDsc<type> {  
public:
  inline TPtr() {}
  inline TPtr(const TObj<type>* t) {
    //v! "new TPtr %p, TObj=%p", this, &(*this)()
    *(TDsc<type>*)this = (TObj<type>*)t;
  }

  inline TObj<type>* operator= (TObj<type>* t) {
    *(TDsc<type>*)this = (TObj<type>*)t; //v! "TPtr %p = TObj %p (count = %d)", this, t, t?t->lcnt:0
    return (TObj<type>*)t;
  }

  inline TObj<type>* operator= (TFrz<type> frz) {
    *(TDsc<type>*)this = &(TObj<type>&)frz; // "TPtr %p = TObj %p (count = %d)", this, t, t?t->lcnt:0
    return &(TObj<type>&)frz;
  }

  inline TPtr(const TPtr<type>& t) { *this = t; } // Memory looks initialized enough //FIXME: check it!
  inline TPtr<type>& operator= (const TPtr<type>& t) {
    //v! Copying TPtr
    if (!t.cellOffset) {
      *(TDsc<type>*)this = NULL;  //v! =NULL ! done
    } else { 
      //assert(((TDsc<type>&)t)().cellIsOk());
      if (!freezeMode) {
	// *(TDsc<type>*)this = &((TPtr<type>&)t)();           //v! TDsc COPY (w/o freeze) ! done
	*(TDsc<type>*)this = ((TPtr<type>&)t).getTObjPtr();           //v! TDsc COPY (w/o freeze) ! done
	seqNo = t.seqNo;
      } else {
	*(TDsc<type>*)this = ((TPtr<type>&)t)().ownerCopy(); //v! TDsc COPY (w/  freeze) ! done
      }
    }
    return *this;
  }

  inline operator type* () { return cellOffset ? &(*this)()()() : NULL; }
  inline operator bool () const { return cellOffset != 0; }

  inline type* operator->() { 
    type *p = (type*)*this;
    if (p == NULL) uiprintf("\nNULL TPtr dereference!\n");
    return p;
  }


  inline TObj<type>& operator[] (int n)  { return tObjAt(cellOffset+n*sizeof(TObj<type>)); }
  inline TObj<type>& operator*  ()       { return (*this)(); }
  inline TObj<type>* operator + (int n)  { return &(*this)[+n]; }
  inline TObj<type>* operator - (int n)  { return &(*this)[-n]; } 
  // Constant 
  inline const TObj<type>& operator[] (int n) const { return tObjAt(cellOffset+n*sizeof(TObj<type>)); }
  inline const TObj<type>& operator*  ()      const { return (*this)(); }
  inline const TObj<type>* operator + (int n) const { return &(*this)[+n]; }
  inline const TObj<type>* operator - (int n) const { return &(*this)[-n]; } 
  // Shifts
  inline TObj<type>* operator += (int n) { return *this = *this + n; }
  inline TObj<type>* operator -= (int n) { return *this = *this - n; }
  inline TObj<type>* operator ++ ()      { return (*this)+=1; }
  inline TObj<type>* operator -- ()      { return (*this)-=1; }
  inline TObj<type>* operator ++ (int)   { return (++(*this))-1; }
  inline TObj<type>* operator -- (int)   { return (--(*this))+1; }
  // Comparision
  inline const bool operator == (TPtr<type>& t) const { return cellOffset == t.cellOffset; }
  inline const bool operator != (TPtr<type>& t) const { return cellOffset != t.cellOffset; }
};

// Fixed-size array
template <typename type, int dim>
class TPtrA : private TPtr<type> {
public:
  inline TObj<type>& operator[] (int n)  { return tObjAt(cellOffset+n*sizeof(TObj<type>)); }
  inline TObj<type>& operator*  ()       { return (*this)[0]; }
  inline TObj<type>* operator+  (int n)  { return &(*this)[+n]; }
  inline TObj<type>* operator-  (int n)  { return &(*this)[-n]; } 
  // Constant version
  inline const TObj<type>& operator[] (int n) const { return tObjAt(cellOffset+n*sizeof(TObj<type>)); }
  inline const TObj<type>& operator*  ()      const { return (*this)[0]; }
  inline const TObj<type>* operator+  (int n) const { return &(*this)[+n]; }
  inline const TObj<type>* operator-  (int n) const { return &(*this)[-n]; } 
  // Shifts
  TPtrA() : TPtr<type>(new TObj<type>[dim]) {} //FIXME: set dropTo !
};

template <typename type>
class TVar : public TDsc<type> {
public:
  inline TVar() : TDsc<type>(new TObj<type>) {
    //v! "new TVar %p, TObj=%p", this, &(*this)()
  }

  inline TVar(type v) : TDsc<type>(new TObj<type>) {
    //v! "new TVar %p from value", this
    (*this)()()() = v;
  }

  inline TVar<type>& operator= (const type& v)  {
    (*this)() = v;
    return *this;
  }

  inline TVar(TObj<type>& t) : TDsc<type>(new TObj<type>) {
    //v! "new TVar %p, TObj=%p", this, &(*this)()
    (*this)() = t;
  }

  inline TVar(const TVar<type>& v) : TDsc<type>(new TObj<type>) {
    //v! "new TVar %p, TObj=%p", this, &(*this)()
    *this = v;
  }
  inline TVar<type>& operator= (TObj<type>& to) {
    //v! "Copying TVar %p from TObj %p", this, &to
    if (!freezeMode) {
      (*this)() = to;   //v! TDsc COPY (w/o freeze)
    } else {
      TDsc<type>  copy(to.ownerCopy()); // + GC
      (*this)() = copy();               //v! TDsc COPY (w/  freeze)

      // Only when data ready
      if (!(*this)().asnFrom.notNull()) (*this)().onWrite();//

    }
    assert((*this)().cellIsOk());
    return *this;
  }

  inline TVar<type>& operator= (const TVar<type>& v) {
    assert((*v.ampersand())().cellIsOk());
    return (*this) = ((TDsc<type>&)v)();
  }

  inline TVar(const TFrz<type>& frz) : TDsc<type>(new TObj<type>) {
    *this = ((TFrz<type>&)frz)();
  }

  inline TVar<type>& operator= (TFrz<type> frz) {
    //???    assert((*this)().ringEmpty());

    (*this) = frz();

      // Only when data ready
    if (!(*this)().asnFrom.notNull()) (*this)().onWrite(); ///FIXME:!!!!! ????????????????????????????????????????????????????????????????????

    return *this;
  }

  inline TVar<type>* ampersand()  const { return (TVar<type>*)this; } 
  inline TObj<type>* operator& () { return &(*this)(); }
  inline operator type& () { return (*this)()()(); }

  // inline TObj<type>& operator= (TObj<type>& t) { return (*this)() = t; }
  //inline TObj<type>& operator= (type v) { (*this)()()() = v; return (*this)(); }

  // Shifts
  inline type& operator += (type n) { return (*this)()()() += n; }
  inline type& operator -= (type n) { return (*this)()()() -= n; }
  inline type& operator ++ ()       { return ++(*this)()()(); }
  inline type& operator -- ()       { return --(*this)()()(); }
  inline type& operator ++ (int)    { return (*this)()()()++; }
  inline type& operator -- (int)    { return (*this)()()()--; }

  // Semistrong binary operators
  inline type operator+ (TObj<type>& t) { return (*this)()()() + t()(); }
  inline type operator- (TObj<type>& t) { return (*this)()()() - t()(); }
  inline type operator* (TObj<type>& t) { return (*this)()()() * t()(); }
  inline type operator/ (TObj<type>& t) { return (*this)()()() / t()(); }

  // Strong binary operators
  inline type operator+ (type t) { return (*this)()()() + t; }
  inline type operator- (type t) { return (*this)()()() - t; }
  inline type operator* (type t) { return (*this)()()() * t; }
  inline type operator/ (type t) { return (*this)()()() / t; }
  
  //To pass TVar as tin/tout
  inline operator TObj<type>& () { return (*this)(); }
};

// Holds retval during TFun creation 
externNoImpl long toutChain;
class toutChainReset {
public:
  toutChainReset() { toutChain = 0; }
};

template <typename type>
class TOut : public TVar<type> {
  // Used to chain all TOut's together by mobile way
  long nextOffset;
  char sign;
  // Used to pack/unpack TFun 
  long dropToOffset;
  long long toutSeqNo, dropToSeqNo;
  
public:
  inline TOut<type>* next()  { 
    TOut<type>* res = (TOut<type>*) (nextOffset+(long)this); 
    assert(res->sign == 'o');
    return res;
  }
  TOut() : nextOffset(0), sign('o') { 
    if (!toutChain) 
      toutChain = (long)this; 
    else {
      nextOffset = (long)((TOut<type>*)toutChain)->next() - (long)this;
      ((TOut<type>*)toutChain)->nextOffset = (long)this - toutChain;
    }
  }

  void pack() {
    TObj<type> *to = &(*this)();

    assert(to->cellIsOk());
    if (to->dropTo && to->dropTo->ctDone && to->dropTo->getSeqNo() == to->dropSeqNo) { // Valid TOUT
      assert(to->dropTo->cellIsOk());
      dropToOffset = to->dropTo->offset(); //v! "dropTo = %p", to->dropTo
      dropToSeqNo = to->dropTo->getSeqNo();
    } else {   // Invalid OUT!
      dropToOffset = 0;
    }
    toutSeqNo = to->getSeqNo(); 
  }
  void unpack(long long taskId) {
    // Unpack seqNo first to avoid blocking!
    getTObjPtr()->setSlvSeqNo(toutSeqNo); 
    getTObjPtr()->sendSlvUse();

    TObj<type> *to = &(*this)();
    if (dropToOffset) { // Valid TOUT
      to->dropTo = (TObj<type>*)&TObj<type>::cellAt(dropToOffset); //v! "tout %p, dropTo = %p", to, to->dropTo
      to->dropSeqNo = dropToSeqNo;
      to->dropTo->setSlvSeqNo(dropToSeqNo);
      // Now set ownership
      to->dropTo->owner = taskId;
      to->dropTo->sendSlvUse();
      assert(to->dropTo->cellIsOk());
    } else { // Ignored OUT!
      to->dropTo = NULL;
    }
    to->owner = taskId;
    assert(to->cellIsOk());
  }

  TObj<type>& operator= (TObj<type>& to) {
    (*this)() = to;
    return (*this)();
  }
  TObj<type>& operator= (const type& v) {
    (*this)()()() = v;
    return (*this)();
  }
  TObj<type>& operator= (TFrz<type> frz) {
    (*this)() = frz();
    return (*this)();
  }
};
class TFunCtxt : public TaskCtxt {
public:
  TFunCtxt(char *name) : TaskCtxt(name) {}
};
// TFunction itself
#define TFUNLAZY 1
template <typename type, TFunCtxt& ctxt>
class TFun : public MThread, toutChainReset {
private:
  long long parentTaskId; // Why not in Task ???
public:
  int tfunId;
  int fflags;
  TOut<type> retval;

  TFun() : parentTaskId(parent?parent->taskId:-1), fflags(tc.fflags), memoLnk(NULL) {
    if (parentTaskId >= 0) {
      callGraphPrintf("\tf%Lx -> f%Lx;\n", parent->taskId, taskId);
    }
    if (!(TFUNLAZY & fflags)) run(); 
  }
  TFunCtxt& context() {
    return ctxt;
  }
  virtual void packState() {
    Task::packState();
    TOut<type>* t = (TOut<type>*)retval.ampersand();
    do t->pack(); while ((t = t->next()) != retval.ampersand()); 
  }
  virtual void unpackState() {
    TOut<type>* t = (TOut<type>*)retval.ampersand();
    do t->unpack(taskId); while ((t = t->next()) != retval.ampersand());
  }

  const char *name() {
    char *p = CLASSID(*this), *q;
    if ((q=strstr(p,"TFunImpl"))) *q = 0;
    if (!strcmp(p,"tfunmain"))    p += 4;
    return p;
  }
  
  virtual void work() {
    static char * clr[] = {"yellow", "red", "green", "blue", "magenta", "cyan"};
    const int clrQty = sizeof(clr)/sizeof(clr[0]);
    //FIXME: collector can erase body() result ?
    try { 
      retval = body();
      callGraphPrintf("\tf%Lx [label=\"%s %s\",style=\"filled\",fillcolor=\"%s\"];\n",
		      taskId, name(), label, clr[myRank%clrQty]); 
    } catch (FinalizeException e) {
      //v! "MThread %p got finalize exception", this
    }
  }

  virtual TFrz<type> body() = 0;

  virtual ~TFun() {
    // uiprintf("Destroying function label=%s\n", label);
  }

  //Memoization


  TFun<type,ctxt>* memoLnk; 
  
  class MemoTable {
  public:
    TFun<type,ctxt>** table;
    int mtSize;

    MemoTable(int _mtSize) : mtSize(_mtSize) { 
      assert( mtSize > 0 );
      table = new TFun<type,ctxt>* [mtSize];
      memset(table,0,sizeof(table[0])*mtSize);
    }

    static inline unsigned hash(const char *pos, size_t size) {
      unsigned res = 0;
#define stepsFor(type) \
      while(size >= sizeof(type)) { \
        res  ^= *(type*)pos;  \
        pos  += sizeof(type); \
        size -= sizeof(type); \
      }
      stepsFor(unsigned int); 
      stepsFor(unsigned short); 
      stepsFor(unsigned char);
#undef stepsFor    
      return res;
    }
  };

  static inline TFun<type,ctxt>** intern(void *key, size_t size) {
    TFun<type,ctxt>** p = table + hash((const char*)key,size) % mtSize; 
    while (*p && memcmp((*p)->key,key,size)) p = &(*p)->memoLnk;
    return p;
  }

};


