//     pi.
//    :
//        1
// pi = 4*Sdx/(1+x*x)
//        0
//  S -    0  1.
//    .   
//  10^-8     10^6 . 

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>

// nstripes -    .
int 
ccalcPi(int ns0, int ns1, int nstripes, double *sum)
{
  int i;
  double w, v;
 
  w = 1.0 / nstripes;
  *sum = 0.0;
  //   
  for (i = ns0; i < ns1; i++) {
       v = (i - 0.5 ) * w;
       v = 4.0 / (1.0 + v * v);
       *sum = *sum + v;
  }

  return(0);
}

tfun int 
tcalcPi(int ns0, int ns1, int nstripes, tout double sum)
//tcalcPi(int ns0, int ns1, int nstripes, ts::TObj<double> sum)
//tcalcPi(int ns0, int ns1, int nstripes, ts::TVar<double> sum)
{
  int i;
  double w, v, vsum;
 
  w = 1.0 / nstripes;
  vsum = 0.0;
  //   
  for (i = ns0; i < ns1; i++) {
       v = (i - 0.5 ) * w;
       v = 4.0 / (1.0 + v * v);
       vsum = vsum + v;
  }
  sum = (double)vsum;

  return(0);
}

tfun int 
main(int argc, char *argv[])
{
  int nstripes, nprocessors, ns, np;
  double w, totalsumm, mflops, calctime;
  struct timeval time0, time1, subtime;

  if (argc < 3) {
      printf("Program calculates the 'pi' number.\n");
      printf("Usage: calc_pi <n_stripes> <n_processors>\n"
             "where:\n"
             "       n_stripes - number of the stripes\n"
             "       n_processors - number of the processors\n");
      exit(1);          
  }
  nstripes = (int)strtod(argv[1], NULL);
  if (nstripes <= 0) {
      printf("Bad number of the stripes.\n");
      exit(1); 
  }
  nprocessors = (int)strtod(argv[2], NULL);
  if (nprocessors <= 0) {
      printf("Bad number of the processors.\n");
      exit(1); 
  }

  //  
  gettimeofday(&time0, NULL);
  //double tptr summ = new tval double[nprocessors];
  //double tptr summ = new ts::TObj<double> [nprocessors];
  ts::TPtr<double> summ = new ts::TObj<double> [nprocessors];
  double csumm;
  tval double s[nprocessors];

  totalsumm = 0.0;
  if (nstripes < 100000) {
      nprocessors = 1; // one processor
      ccalcPi(1, 1 + nstripes, nstripes, &csumm);
      totalsumm += csumm;
  } else {
          for (np = 0, ns = 1; np < nprocessors; 
                               np++, ns = np * (nstripes / nprocessors) + 1) {
               tcalcPi(ns, ns + (nstripes / nprocessors), nstripes, 
                       summ[np]);
          }

          for (np = 0; np < nprocessors; np++) {
               s[np] = summ[np];
               totalsumm += (double)s[np];
               //totalsumm += (double)summ[np];
          }
    }

  //  ,   
  gettimeofday(&time1, NULL);
  timersub(&time1, &time0, &subtime);

  calctime = (double)subtime.tv_sec + (double)(subtime.tv_usec * 1.0e-6);
  if (calctime == 0.0) calctime = 1.0e-6;

  //     Mflops
  mflops = 9 * nstripes / (1000000.0 * calctime);
  w = 1.0 / nstripes;
  printf("number of the stripes = %d\n", nstripes);
  printf("pi is approximated with %.20f\n", totalsumm * w);
  printf("calc time = %.6f seconds\n", calctime);
  printf("mflops = %f Mflops on %d processor(s)\n", mflops, nprocessors);
  printf("mflops = %f Mflops for one processor\n", mflops/nprocessors);
  
  //delete summ;
  summ = NULL;
  
  return(0);
}
