
topmost:: build_test

# Modules/components layout
include $(TOP)/wizard/layout.mk

CXX=g++ -Wall -g 
LDFLAGS=
#-Wl,-rpath -Wl,$(shell cd $(PTHLIBDIR) && pwd)

# To make recursively only when goal is present or for "wizard" Makefile
MAKEREC= $(shell for dir in *; do \
  if [ -f "$$dir/Makefile" ]; then \
     if grep -q '^$@:\|include.*wizard/rules.mk' $$dir/Makefile; then \
        echo -n "make -C $$dir TOP=$(TOP) MODULE=$$dir $@ && "; \
     else \
        if [ "$@" = "build" ]; then echo -n "make -C $$dir && "; fi; \
     fi; \
  fi; \
done; echo "true")

build_test::
	make build test_status 2>&1 
test::  
	@[ "$$TESTALL" != "yes" ] || ($(MAKEREC))
	@echo "Perform quick tests for $(MODULE):"
	@[ "$$TESTALL" != "yes" ] || make test_long
test_status:: test
	@echo ""
	@echo "SUCCESS!  Quick tests for module $(MODULE) perfectly completed"
	@echo ""

test_all:: build 
	export TESTALL; TESTALL=yes; make test_all_status 2>&1 
test_all_status:: test
	@echo ""
	@echo "SUCCESS!  All tests for module $(MODULE) perfectly completed"
	@echo ""
test_long:: 
	@echo "Perform long tests for $(MODULE):"

build::
	@[ "$$BUILDRECURSIVE" = "no" ] || ($(MAKEREC))

manual::
	@echo "Building manual for $(MODULE):"

install:: manual $(SKIFROOT) $(DEVROOT) 
	@BUILDRECURSIVE=no; export BUILDRECURSIVE; make build 
	$(MAKEREC)

# Visualization rules

.SUFFIXES: .v.hh .v.cc

%.v.hh : %.hh 
	$(VPP) $< > $@

%.v.cc : %.cc 
	$(VPP) $< > $@


clean:: 
	@[ "$$CLEANRECURSIVE" = "no" ] || ($(MAKEREC))
	rm -f *.[oa] *.so *.v.cc *.v.hh *.log a.out core .rpmmacros

distclean:: 
	@CLEANRECURSIVE=no; export CLEANRECURSIVE; make clean
	$(MAKEREC)
	rm -f *~

snapshot:: $(DEVROOT)
	make clean test_all 
	make distclean
	find -name '*~' -exec rm -rfv {} \;
	(LANG=C; export LANG; date ; hostname ; echo $$USER $$PWD) > .id
	cd .. && TAR=$(DEVROOT)/SOURCE/$(MODULE).tar.bz2; \
	tar cjf $$TAR $(MODULE) &&  ls -l $$TAR

### RPM generation ###

RPMNAME=$(MODULE)
SPEC=$(SKIFROOT)/usr/src/SPECS/$(MODULE).spec
_SPEC=$(SKIFROOT)/usr/src/SPECS/_.spec
INSTDIR=$(SKIFROOT)/usr/src/BUILD/install-$(MODULE)

rpm:    snapshot
	./configure --prefix=/usr && make
	rm -f $(_SPEC)
	rm -fr $(INSTDIR)
	make install DESTDIR=$(INSTDIR)/usr
	rm -f .rpmmacros $(SPEC)
	make .rpmmacros $(SPEC)
	HOME=`pwd` rpmbuild -ba $(SPEC)
	rm -f .rpmmacros $(_SPEC)
	rm -fr $(INSTDIR)


.rpmmacros: 
	echo >> $@ "%_fixperms chmod -R a+rX,u+w,g-w,o-w" 
	echo >> $@ "%_topdir $(INSTDIR)" 
	echo >> $@ "%_rpmdir $(DEVROOT)/RPMS" 
	echo >> $@ "%_srcrpmdir $(DEVROOT)/SRPMS" 
	echo >> $@ "%_builddir $(INSTDIR)" 
	echo >> $@ "%_sourcedir $(DEVROOT)/SOURCE"
	echo >> $@ "%_specdir $(DEVROOT)/SPECS" 
	echo >> $@ "%_make_smp -j 2" 
	echo >> $@ "%_rpmfilename %%{NAME}-%%{VERSION}-%%{RELEASE}.%%{ARCH}.rpm"; 

$(SPEC): 
	echo >> $@ "Name: $(RPMNAME)"
	echo >> $@ "Version: $(RPMVERSION)"
	echo >> $@ "Release: $(RPMRELEASE)"
	echo >> $@ "Serial:  $(RPMSERIAL)"
	echo >> $@ "Vendor:  PSI RAS, Pereslavl-Zalessky, Russia"
	echo >> $@ "BuildRoot: $(INSTDIR)"
	echo >> $@ "Source: $(RPMNAME).tar.bz2"
	echo >> $@ "License: Currently all rights reserved by PSI RAS"
	echo >> $@ "Summary: $(RPMSUMMARY)"
	echo >> $@ "Group: Development/Libraries"
	echo >> $@ "%description"
	echo >> $@ "$(RPMDESCRIPTION)"
	echo >> $@ "%prep"
	echo >> $@ "%build"
	echo >> $@ "%files"
	cat  >> $@ $(_SPEC)

